/*
 * AddNSController.java
 *
 * Copyright (c) 2005 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

package net.emulab.netlab.client;

import net.emulab.ns.NSObject;

/**
 * Controller for the "Add NS" dialog.
 */
public class AddNSController extends ControllerBase {
	/**
	 * The component representing the NS object.
	 */
	protected Object nsComponent;

	/**
	 * The original prepend value, used to restore the object if the user
	 * cancelled the operation.
	 */
	protected String originalPrepend;

	/**
	 * The original append value, used to restore the object if the user
	 * cancelled the operation.
	 */
	protected String originalAppend;

	/**
	 * Thinlet outlet for the append textarea, used to force focus to the
	 * textarea in the append pane.
	 */
	public Object nsAppend;

	/**
	 * Construct an AddNSController with the given value.
	 * 
	 * @param netlab
	 *            The NetlabClient this controller is attached to.
	 */
	public AddNSController(NetlabClient netlab) {
		super(netlab, null);
	}

	/**
	 * @see net.emulab.netlab.client.Controller#triggeredBy(java.lang.Object)
	 */
	public Controller triggeredBy(Object sender) throws Exception {
		NSObject nso;

		/* Get the button for the NSObject, */
		this.nsComponent = this.netlab.getParent(this.netlab.getParent(sender));

		/* ... the model, */
		nso = (NSObject) this.netlab.getProperty(this.nsComponent, "model");

		/* ... the original values for "Prepend/Append", and then */
		this.originalPrepend = nso.getNSPrepend();
		this.originalAppend = nso.getNSAppend();
		/* ... set the content for this controller. */
		this.tkv.setKeyValue(this, "content", nso);

		/* XXX Manually update the dialog window's title. */
		this.netlab.setString(this.view, "text", "Add NS to '" + nso.getName()
				+ "'");

		super.triggeredBy(sender);

		return this;
	}

	/**
	 * Method called when the dialog is focusable, we forcefully switch to the
	 * "Append" pane and focus on the textarea so the user can start typing
	 * immediately.
	 */
	public void focusable() {
		this.thinlet.requestFocus(this.nsAppend);
	}

	/**
	 * Clear internal state setup in triggeredBy().
	 */
	private void clearState() {
		this.nsComponent = null;
		this.originalPrepend = null;
		this.originalAppend = null;
		this.nsAppend = null;
	}

	/**
	 * Method called when the "Ok" button is clicked, just closes the window and
	 * clears some internal state.
	 * 
	 * @param prepend
	 *            The text from the prepend textarea.
	 * @param append
	 *            The text from the append textarea.
	 */
	public void ok() {
		this.netlab.remove(this.view);

		this.clearState();
	}

	/**
	 * Method called when the "Cancel" button is clicked, just closes the window
	 * and restores "Prepend/Append" to their original values.
	 */
	public void cancel() {
		this.netlab.remove(this.view);

		this.tkv.setKeyValue(this, "content.NSPrepend", this.originalPrepend);
		this.tkv.setKeyValue(this, "content.NSAppend", this.originalAppend);

		this.clearState();
	}

	/**
	 * @see java.lang.Object#toString()
	 */
	public String toString() {
		return "AddNSController[nsComponent=" + this.nsComponent
				+ "; originalPrepend=" + this.originalPrepend
				+ "; originalAppend=" + this.originalAppend + "; nsAppend="
				+ this.nsAppend + "]";
	}
}
