/*
 * BrowserController.java
 *
 * Copyright (c) 2005 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

package net.emulab.netlab.client;

import java.io.IOException;

/**
 * Helper class used to open a URL in a web browser.
 */
public final class BrowserController
{
    /**
     * Indicates whether or not the Browser properties have been loaded
     * successfully.
     */
    private static boolean loaded = false;

    /**
     * Flush the Browser properties so they will be reloaded on the next view.
     */
    public static synchronized void flush()
    {
	try
	{
	    Browser.exec = null;
	    loaded = false;
	}
	catch(NoClassDefFoundError e)
	{
	}
    }

    /**
     * Display a URL in a web browser.
     * 
     * @param netlab The NetlabClient instance this controller is being used
     * by.
     * @param url The URL to open.
     * @throws IOException if there was a problem opening the browser window.
     */
    public static synchronized void displayURL(NetlabClient netlab, String url)
	throws IOException
    {
	try
	{
	    if( !loaded )
	    {
		Browser.load(netlab.getPreferences());
		loaded = true;
	    }
	    Browser.displayURL(url);
	}
	catch(NoClassDefFoundError e)
	{
	    netlab.showErrorDialog("Unable to open browser window: "
				   + e.getMessage());
	}
    }

    /**
     * Private constructor for a static-only class.
     */
    private BrowserController()
    {
    }
}
