/*
 * ControllerBase.java
 *
 * Copyright (c) 2005 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

package net.emulab.netlab.client;

import thinlet.ObjectController;
import thinlet.ThinletKeyValueObserver;

/**
 * Common base class for "normal" NetlabClient controllers.
 */
public class ControllerBase extends ObjectController implements Controller {
	/**
	 * The NetlabClient instance this controller is attached to.
	 */
	public final NetlabClient netlab;

	/**
	 * The view that is being controlled.
	 */
	protected Object view;

	/**
	 * Construct a ControllerBase with the given values.
	 * 
	 * @param netlab
	 *            The NetlabClient instance this controller is attached to.
	 * @param content
	 *            The initial value for this controller's content.
	 */
	public ControllerBase(NetlabClient netlab, Object content) {
		super(netlab, content);

		this.netlab = netlab;

		this.tkv.observeKeyValue(netlab, "", new ThinletKeyValueObserver() {
			public void observedKeyValue(Object obj, String keyPath, int flags,
					Object oldValue, Object newValue) {
				ControllerBase.this.tkv.notifyKeyValue(ControllerBase.this,
						"netlab." + keyPath, flags, oldValue, newValue);
			}
		});
	}

	/**
	 * Construct a ControllerBase with the given argument.
	 * 
	 * @param netlab
	 *            The NetlabClient instance this controller is attached to.
	 */
	public ControllerBase(NetlabClient netlab) {
		super(netlab);

		this.netlab = netlab;

		this.tkv.observeKeyValue(netlab, "", new ThinletKeyValueObserver() {
			public void observedKeyValue(Object obj, String keyPath, int flags,
					Object oldValue, Object newValue) {
				ControllerBase.this.tkv.notifyKeyValue(ControllerBase.this,
						"netlab." + keyPath, flags, oldValue, newValue);
			}
		});
	}

	public NetlabClient getNetlabClient() {
		return this.netlab;
	}

	/**
	 * @see net.emulab.netlab.client.Controller#showView(java.lang.Object)
	 */
	public void showView(Object component) throws Exception {
		this.netlab.showView(component);
	}

	/**
	 * @see net.emulab.netlab.client.Controller#setView(java.lang.Object)
	 */
	public Controller setView(Object view) {
		this.view = view;
		return this;
	}
	
	public Object getView() {
		return this.view;
	}

	/**
	 * @see net.emulab.netlab.client.Controller#modelChanged(java.lang.Object)
	 */
	public Controller modelChanged(Object component) {
		return this;
	}

	/**
	 * @see net.emulab.netlab.client.Controller#triggeredBy(java.lang.Object)
	 */
	public Controller triggeredBy(Object component) throws Exception {
		if (this.netlab.getParent(this.view) == null)
			this.netlab.add(this.view);
		
		return this;
	}

	/**
	 * @see java.lang.Object#toString()
	 */
	public String toString() {
		return "ControllerBase[net.emulab.netlab.client=" + this.netlab
				+ "; view=" + this.view + "]";
	}
}
