/*
 * DelayConfigChange.java
 *
 * Copyright (c) 2005 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */
package net.emulab.netlab.client;

import java.io.IOException;
import java.util.Hashtable;

import net.emulab.netlab.client.emulab.EmulabExperimentProxy;
import net.emulab.ns.NSLanLink;

import org.apache.xmlrpc.XmlRpcException;

public class DelayConfigChange
    extends Hashtable
    implements Change
{
    private final NSLanLink nsll;
    
    public DelayConfigChange(NSLanLink nsll, String aspect, Object value)
    {
	this.nsll = nsll;
	super.put(aspect, value);
    }

    public NSLanLink getLanLink()
    {
	return this.nsll;
    }

    public String getDescription()
    {
	String aspect, unit = "", retval = "";
	Object value;

	aspect = (String)this.keys().nextElement();
	value = this.get(aspect);
	if( "bandwidth".equals(aspect) )
	{
	    unit = "kbps";
	}
	else if( "delay".equals(aspect) )
	{
	    aspect = "latency";
	    unit = "ms";
	}
	retval += this.nsll.getName().toShortString(8)
	    + "/"
	    + aspect
	    + " -> "
	    + value.toString()
	    + unit;
	return retval;
    }

    public boolean sameChange(Change ch)
    {
	boolean retval = false;
	
	if( ch instanceof DelayConfigChange )
	{
	    DelayConfigChange dcc = (DelayConfigChange)ch;

	    if( (this.nsll == dcc.nsll) &&
		this.keys().nextElement().equals(dcc.keys().nextElement()) )
	    {
		retval = true;
	    }
	}
	return retval;
    }

    public void changeExperiment(ExperimentParameters ep,
				 EmulabExperimentProxy eep)
	throws XmlRpcException, IOException
    {
	eep.delay_config(ep.getProject(),
			 ep.getName(),
			 this.nsll.getName().toString(),
			 this,
			 true);
    }
    
    public String toString()
    {
	return "DelayConfigChange["
	    + super.toString()
	    + "; nsll="
	    + this.nsll
	    + "]";
    }
}
