package net.emulab.netlab.client;

import java.awt.Font;

import net.emulab.ns.NSObject;
import thinlet.ObjectController;
import thinlet.ThinletKeyValue;
import thinlet.ThinletKeyValueObserver;

public abstract class ElementControllerBase extends ObjectController implements
		ElementController, Controller {
	public final NetlabClient netlab;

	protected final NetlabConfiguration nc;

	public NSObject nsobject;

	public Object mapping;

	protected Object selection;

	protected Object element;

	/**
	 * The font to use for the tooltip. An italic font is used when the link has
	 * all the default settings and the tooltip is " (default config) ",
	 * otherwise a plain font is used.
	 */
	public Font tooltipFont = null;

	public ElementControllerBase(NetlabClient netlab) {
		super(netlab);

		this.netlab = netlab;
		this.nc = this.netlab.getConfiguration();

		this.tkv.addKeyDependency(this, "shortName", "nsobject.name");
		this.tkv.observeKeyValue(netlab, "", new ThinletKeyValueObserver() {
			public void observedKeyValue(Object obj, String keyPath, int flags,
					Object oldValue, Object newValue) {
				ElementControllerBase.this.tkv.notifyKeyValue(
						ElementControllerBase.this, "netlab." + keyPath, flags,
						oldValue, newValue);
			}
		});
	}

	/**
	 * @return The abbreviated version of the link's name.
	 */
	public String shortName() {
		String retval = "";

		if (this.nsobject != null) {
			retval = this.nsobject.getName().toShortString(
					this.nc.getShortNameLength());
		}
		return retval;
	}

	public void showView(Object component) throws Exception {
		this.netlab.showView(component);
	}

	public Controller setView(Object view) {
		return this;
	}
	
	public Object getView() {
		return null;
	}

	public Controller triggeredBy(Object component) throws Exception {
		return this;
	}

	public Object getElement() {
		return this.element;
	}

	public void setElement(Object component) {
		// System.out.println("set element!");
		// new Throwable().printStackTrace();
		this.element = component;
		this.tkv.setKeyValue(this, "nsobject", this.netlab.getProperty(
				component, "model"));
		this.tkv.setKeyValue(this, "mapping", this.netlab.getProperty(
				component, "mapping"));
		if (this.mapping != null) {
			this.selection = this.mapping;
		} else {
			this.selection = this.nsobject;
		}
		this.tkv.notifyKeyValue(this, "selection",
				ThinletKeyValue.CHANGE_SETTING, null, this.selection);
		this.tkv.observeKeyValue(this.nsobject, "",
				new ThinletKeyValueObserver() {
					public void observedKeyValue(Object obj, String keyPath,
							int flags, Object oldValue, Object newValue) {
						ElementControllerBase.this.tkv.notifyKeyValue(
								ElementControllerBase.this, "nsobject."
										+ keyPath, flags, oldValue, newValue);
					}
				});
	}

	public boolean hasMapping() {
		return this.mapping != null;
	}

	public String toString() {
		return "ElementControllerBase[" + "]";
	}
}
