/*
 * OSListController.java
 *
 * Copyright (c) 2005 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

package net.emulab.netlab.client;

import java.util.Arrays;
import java.util.List;
import java.util.Vector;

import net.emulab.ns.FirewallRule;
import net.emulab.ns.Pair;
import thinlet.ArrayController;
import thinlet.AutoFillController;
import thinlet.NumberFormatter;
import thinlet.ThinletKeyValue;

/**
 * Controller for the operating system ID dialog box.
 */
public class ExpPropertiesController extends PropertiesController {

	public static final NumberFormatter RULE_NUMBER_FORMATTER = new NumberFormatter()
			.setMinimum(0.0).setEmpty(new Integer(-1)).setMaximum(65535.0);

	private static final List ROUTING_AUTO_FILL = Arrays.asList(new String[] {
			"Static", "Session", "Manual" });

	private static final List FIREWALL_TYPE_AUTO_FILL = Arrays.asList(new String[] {
			"ipfw2-vlan" });

	private static final List FIREWALL_STYLE_AUTO_FILL = Arrays.asList(new String[] {
			"basic", "open", "closed" });

	public final AutoFillController rc;

	public final AutoFillController ssc;

	public final AutoFillController ftc;
	public final AutoFillController fsc;
	public final ArrayController fc;

	public final ArrayController cpc;
	
	public final TopologyController tc;

	public final AutoFillController josc;
	public final AutoFillController dosc;
	
	/**
	 * Construct a controller with the given parameter.
	 * 
	 * @param netlab
	 *            The NetlabClient this controller is working under.
	 */
	public ExpPropertiesController(NetlabClient netlab) {
		super(netlab);

		this.tc = netlab.getTopologyController();
		
		this.vis.put("net", Boolean.TRUE);
		this.vis.put("firewall", Boolean.TRUE);
		this.vis.put("param", Boolean.TRUE);
		this.vis.put("os", Boolean.FALSE);

		this.rc = new AutoFillController(netlab, new Vector(ROUTING_AUTO_FILL));
		
		this.ssc = new AutoFillController(netlab);
		this.tkv.observeKeyValue(this.tc, "content.nodes.name",
				ThinletKeyValue.CHANGE_ANY, this.ssc, "content");

		this.ftc = new AutoFillController(netlab, new Vector(FIREWALL_TYPE_AUTO_FILL));
		this.fsc = new AutoFillController(netlab, new Vector(FIREWALL_STYLE_AUTO_FILL));
		
		this.fc = new ArrayController(netlab);
		this.fc.setObjectClass(FirewallRule.class);
		this.tkv.observeKeyValue(this.tc, "content.firewallRules",
				ThinletKeyValue.CHANGE_ANY, this.fc, "content");
		
		this.cpc = new ArrayController(netlab);
		this.cpc.setObjectClass(Pair.class);
		this.tkv.observeKeyValue(this.tc, "content.capturedParameters",
				ThinletKeyValue.CHANGE_ANY, this.cpc, "content");
		
		this.josc = new AutoFillController(netlab);
		this.dosc = new AutoFillController(netlab);
	}
}
