/*
 * FileListController.java
 *
 * Copyright (c) 2005 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */
package net.emulab.netlab.client;

import java.util.Date;
import java.util.Vector;

import java.text.SimpleDateFormat;

import java.awt.Color;

public class FileListController
    implements Controller
{
    protected static final SimpleDateFormat DATE_FORMAT =
	new SimpleDateFormat("MMM dd yyyy");
    
    protected final NetlabClient netbuild;
    protected Object view;
    protected String path;
    protected Vector sparePaths;
    protected DirectoryEntry de[];
    
    public FileListController(NetlabClient netbuild)
    {
	this.netbuild = netbuild;
    }

    public void showView(Object component)
	throws Exception
    {
	this.netbuild.showView(component);
    }

    public Controller setView(Object view)
    {
	this.view = view;
	return this;
    }

    public Controller triggeredBy(Object component)
	throws Exception
    {
	this.netbuild.add(this.view);
	
	return this;
    }
    
    public Controller modelChanged(Object component)
    {
	return this;
    }

    protected void setCellValue(Object row, String cellName, String value)
    {
	Object cell;

	cell = this.netbuild.find(row, cellName);
	this.netbuild.setString(cell, "text", value);
    }
    
    protected String getCellValue(Object row, String cellName)
    {
	String retval;
	Object cell;

	cell = this.netbuild.find(row, cellName);
	retval = this.netbuild.getString(cell, "text");
	return retval;
    }

    protected Object addCell(Object row, String cellName, String cellValue)
    {
	Object retval;
	
	retval = this.netbuild.create("cell");
	this.netbuild.setString(retval, "name", cellName);
	this.netbuild.setString(retval, "text", cellValue);
	this.netbuild.add(row, retval);
	return retval;
    }

    protected Object addRow(Object filelist, DirectoryEntry de)
    {
	Object cell, retval;
	Date date;
	
	retval = this.netbuild.create("row");
	cell = this.addCell(retval, "name", de.name);
	switch( de.type )
	{
	case 'd':
	    this.netbuild.setColor(cell, "foreground", Color.blue);
	    cell = this.addCell(retval, "size", "(dir)");
	    break;
	default:
	    cell = this.addCell(retval, "size", String.valueOf(de.size));
	    break;
	}
	this.netbuild.setChoice(cell, "alignment", "right");
	cell = this.addCell(retval,
			    "last-modified",
			    DATE_FORMAT.format(new Date(de.mtime)));
	this.netbuild.setChoice(cell, "alignment", "right");
	this.netbuild.add(filelist, retval);
	return retval;
    }

    protected Object addRow(Object filelist, String spath)
    {
	Object cell, retval;

	retval = this.netbuild.create("row");
	cell = this.addCell(retval, "path", spath);
	this.netbuild.setColor(cell, "foreground", Color.blue);
	this.netbuild.add(filelist, retval);
	return retval;
    }

    protected void updateFileList(Object filelist,
				  Object dir,
				  Object file,
				  Object parent)
    {
	if( this.path != null )
	{
	    int lpc;

	    this.netbuild.setString(dir, "text", this.path);
	    this.netbuild.setString(file, "text", "");
	    if( "/".equals(this.path) ) // XXX Fix me
		this.netbuild.setBoolean(parent, "enabled", false);
	    else
		this.netbuild.setBoolean(parent, "enabled", true);
	    this.netbuild.removeAll(filelist);
	    for( lpc = 0; lpc < this.de.length; lpc++ )
	    {
		this.addRow(filelist, this.de[lpc]);
	    }
	}
	else
	{
	    int lpc, size;

	    this.netbuild.setString(dir, "text", "");
	    this.netbuild.setString(file, "text", "");
	    this.netbuild.setBoolean(parent, "enabled", false);
	    this.netbuild.removeAll(filelist);
	    size = this.sparePaths.size();
	    for( lpc = 0; lpc < size; lpc++ )
	    {
		String spath = (String)this.sparePaths.elementAt(lpc);

		this.addRow(filelist, spath);
	    }
	}
    }

    public void headerChange()
    {
    }

    public void selectFile(final Object filelist,
			   Object item,
			   Object dir,
			   Object file)
    {
	if( this.path == null )
	{
	    this.netbuild.setString(dir,
				    "text",
				    this.getCellValue(item, "path"));
	}
	else
	{
	    this.netbuild.setString(file,
				    "text",
				    this.getCellValue(item, "name"));
	}
    }

    public void cancel()
    {
	this.netbuild.remove(this.view);
    }

    public String toString()
    {
	return "FileListController[net.emulab.netlab.client="
	    + this.netbuild
	    + "; view="
	    + this.view
	    + "; path="
	    + this.path
	    + "; sparePaths="
	    + this.sparePaths
	    + "; de="
	    + this.de
	    + "]";
    }
}
