/*
 * IPv4AddressFormatter.java
 *
 * Copyright (c) 2005 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

package net.emulab.netlab.client;

import java.text.FieldPosition;

import net.emulab.ns.NSInterface;
import thinlet.ThinletAPI;
import thinlet.ThinletEntryException;
import thinlet.ThinletFormatter;

/**
 * A ThinletFormatter for IPv4 addresses. Note that this formatter does not
 * create java.net.InetAddresses, it just returns the String object. It does
 * support partial validation of the string though. The behavior of the
 * validator is to restrict each component to three digits (i.e. any extra
 * characters are dropped) followed by a single dot, and there can only be four
 * components.
 */
public class IPv4AddressFormatter implements ThinletFormatter {
	/**
	 * Construct a default IPv4AddressFormatter.
	 */
	public IPv4AddressFormatter() {
	}

	public Object getObjectValue(String string) throws Exception {
		
		if (!"".equals(string)) {
			try {
				NSInterface.checkIPv4Address(string);
			}
			catch (IllegalArgumentException e) {
				throw new ThinletEntryException(e.getMessage(),
				"IPv4 Addresses should look like '192.168.0.1'.");
			}
		}
		
		return string;
	}

	public String stringForObjectValue(ThinletAPI thinlet, Object component,
			Object value) {
		return (String) value;
	}

	public boolean isPartialStringValid(ThinletAPI thinlet, Object component,
			StringBuffer partialString, FieldPosition fp) {
		int byteStart, dots = 0, lpc;
		boolean retval = true;

		for (lpc = 0, byteStart = 0; (lpc < partialString.length()) && retval; lpc++) {
			if (((lpc - byteStart) == 0) && (partialString.charAt(lpc) == '.')) {
				partialString.deleteCharAt(lpc);
			} else if ((lpc - byteStart) < 3) {
				if (partialString.charAt(lpc) == '.') {
					byteStart = lpc + 1;
					dots += 1;
				} else if (!Character.isDigit(partialString.charAt(lpc))) {
					retval = false;
				}
			} else if (partialString.charAt(lpc) == '.') {
				byteStart = lpc + 1;
				dots += 1;
			} else {
				retval = false;
			}
		}
		if (dots > 3) {
			retval = false;
		}
		return retval;
	}

	public String toString() {
		return "IPv4AddressFormatter[" + "]";
	}
}
