/*
 * InterfaceController.java
 *
 * Copyright (c) 2005 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

package net.emulab.netlab.client;

import java.awt.Font;

import net.emulab.ns.NSInterface;
import net.emulab.ns.NSLan;
import net.emulab.ns.NSNode;
import net.emulab.ns.NSObject;
import thinlet.ThinletKeyValueObserver;

public class InterfaceController extends ElementControllerBase {
	
	public NSInterface nsi;
	
	public InterfaceController(NetlabClient netlab) {
		super(netlab);

		this.tkv.addKeyDependency(this, "nsi", "nsobject");
		this.tkv.addKeyDependency(this, "tooltip", "nsi");
	}

	public String shortName() {
		return "";
	}

	public void showView(Object component) throws Exception {
		this.netlab.showView(component);
	}

	public Controller setView(Object view) {
		return this;
	}

	public Controller triggeredBy(Object component) throws Exception {
		return this;
	}

	public Controller modelChanged(Object component) {
		return this;
	}

	public void delete(Object component) {
		NSLan nsl;
		
		nsl = (NSLan) this.netlab.getProperty(component, "lan");
		nsl.removeMember((NSNode) nsi.getNSObject());
		this.netlab.remove(component);
	}

	public Object copyElement(NSObject component) {
		return null;
	}

	public boolean verifyElement(ProgressController pc, Object component) {
		boolean retval = true;

		return retval;
	}

	public void menuProperties(Object popup) throws Exception {
		this.showView(this.netlab.getParent(popup));
	}

	public void setElement(Object component) {
		if (this.nsi != null)
			throw new IllegalStateException();

		super.setElement(component);

		this.tkv.setKeyValue(this, "nsi", this.nsobject);
		this.tkv.observeKeyValue(this.nsi, "", new ThinletKeyValueObserver() {
			public void observedKeyValue(Object obj, String keyPath, int flags,
					Object oldValue, Object newValue) {
				InterfaceController.this.tkv.notifyKeyValue(InterfaceController.this,
						"nsi." + keyPath, flags, oldValue, newValue);
			}
		});
	}
	
	public String tooltip() {
		String retval = "";
		
		if (this.nsi != null) {
			Font newFont;

			if (nsi.getBandwidth() > 0) {
				retval += ("  BW: " + InterfacePropertiesController.BW_FORMATTER.
						stringForObjectValue(this.netlab, null, new Long(nsi.getBandwidth())));
			}
			if (nsi.getLatency() > 0.0) {
				retval += "  Lat: " + nsi.getLatency() + "ms";
			}
			if (nsi.getLossRate() > 0.0) {
				retval += "  PLR: " + (nsi.getLossRate() * 100.0) + "%";
			}

			if (retval.length() == 0) {
				retval = "  (default config)  ";
				newFont = this.nc.getTooltipFont();
			} else {
				retval += "  ";
				newFont = this.nc.getPlainFont();
			}
			this.tkv.setKeyValue(this, "tooltipFont", newFont);
		}
		
		return retval;
	}

	public String toString() {
		return "InterfaceController[" + "]";
	}
}
