/*
 * LanController.java
 *
 * Copyright (c) 2005 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

package net.emulab.netlab.client;

import java.awt.Color;
import java.awt.Font;
import java.awt.Image;

import net.emulab.ns.NSInterface;
import net.emulab.ns.NSLan;
import net.emulab.ns.NSNode;
import net.emulab.ns.NSObject;
import thinlet.Thinlet;
import thinlet.ThinletDragSource;
import thinlet.ThinletDragSourceBase;
import thinlet.ThinletKeyValueObserver;

public class LanController extends LanLinkController implements
		ElementController {
	private static final String DEFAULT_LAN_NAME = "Modem";

	private static final ThinletDragSource DRAG_SOURCE = new ThinletDragSourceBase() {
		public Object dragStart(Thinlet thinlet, Object sender) {
			thinlet.setBoolean(sender, "selected", false);

			return sender;
		}
	};
	
	public NSLan nsl;

	public LanController(NetlabClient netlab) {
		super(netlab);

		this.tkv.addKeyDependency(this, "nsl", "nsobject");
		this.tkv.addKeyDependency(this, "icon", "netlab");
		this.tkv.addKeyDependency(this, "tooltip", "nsl");
	}

	/**
	 * @return The appropriate link icon that corresponds to its current state.
	 */
	public Image getIcon() {
		Image retval;

		retval = this.netlab.getImageManager().getObjectImage("node-m");
		return retval;
	}

	public void showView(Object component) throws Exception {
		this.netlab.showView(component);
	}

	public Controller modelChanged(Object component) {
		return this;
	}

	private void primeComponent(Object component) throws Exception {
		TopologyController tc = this.netlab.getTopologyController();
		String name;
		NSLan nsl;

		nsl = (NSLan) this.netlab.getProperty(component, "model");
		name = this.netlab.getString(component, "name");
		if ((name == null) || "nslan".equals(name)) {
			this.netlab.setString(component, "name", "nslan-"
					+ tc.incrementObjectCount());
		}
		this.netlab.setDelegate(component, "dragsrc", DRAG_SOURCE);
		this.netlab.setColor(component, "background", NetlabClient.LAN_COLOR);
		this.netlab.setColor(component, "foreground", Color.white);
		this.netlab.setChoice(component, "style", "round");
		this.netlab.setBoolean(component, "usegradient", true);
		this.netlab.bindAttribute(component, "text", "@shortName", this);
		this.netlab.bindAttribute(component, "icon", "@icon", this);
		this.netlab.bindAttribute(component, "tooltip", "@tooltip", this);
		this.netlab.bindAttribute(component, "tooltipfont", "@tooltipFont",
				this);
		this.netlab.setString(component, "proximitytooltip",
				ElementController.PROXIMITY_TOOLTIP);
		this.netlab.setMethod(component, "action", "elementAction(this)",
				component, this.netlab);
		this.netlab.setMethod(component, "focusgained", "elementAction(this)",
				component, this.netlab);
		this.netlab.setMethod(component, "connect",
				"gotConnect(connector, connectee)", component, this);
		this.netlab.setMethod(component, "delete", "deleteNSObject(this)", component,
				this.netlab);
		this.netlab.putProperty(component, "nsocontroller", this);
		this.netlab.putProperty(component, "view", "lanproperties.xml");
		this.netlab.putProperty(component, "controller",
				"net.emulab.netlab.client.LanPropertiesController");
		this.netlab.setInteger(component, "proximity", NetlabClient.PROXIMITY);

		this.netlab.add(component, this.netlab.parse("lanmenu.xml", this));

		tc.addNSObject(nsl);
	}

	public Controller triggeredBy(Object component) throws Exception {
		TopologyController tc = this.netlab.getTopologyController();
		NSLan nsl;

		if ((nsl = (NSLan) this.netlab.getProperty(component, "model")) == null) {
			nsl = new NSLan();
			nsl.setName(tc.getNamer().suggest(DEFAULT_LAN_NAME));
			this.netlab.putProperty(component, "model", nsl);
		} else {
			tc.getNamer().put(nsl.getName());
		}

		this.primeComponent(component);
		this.tkv.setKeyValue(this, "element", component);
		this.showView(component);
		this.netlab.requestFocus(component);

		return this;
	}

	public void gotConnect(Object left, Object right) throws Exception {
		this.netlab.gotConnect(left, right);
	}

	public void delete(Object component) {
		TopologyController tc = this.netlab.getTopologyController();
		Object parent, items[];
		NSLan nsl;
		int lpc;

		nsl = (NSLan) this.netlab.getProperty(component, "model");
		parent = this.netlab.getParent(component);
		tc.getNamer().remove(nsl.getName());
		this.netlab.remove(component);
		items = this.netlab.getItems(parent);
		for (lpc = 0; lpc < items.length; lpc++) {
			NSObject nso;

			nso = (NSObject) this.netlab.getProperty(items[lpc], "model");
			if (nso instanceof NSInterface) {
				NSInterface nsi = (NSInterface) nso;

				if (nsl.removeMember((NSNode) nsi.getNSObject()))
					this.netlab.remove(items[lpc]);
			}
		}
		tc.removeNSObject(nsl);
	}

	public Object copyElement(NSObject model) throws Exception {
		Object retval;

		retval = this.netlab.create("togglebutton");
		this.netlab.putProperty(retval, "model", model);
		this.primeComponent(retval);
		this.tkv.setKeyValue(this, "element", retval);
		
		return retval;
	}

	public boolean verifyElement(ProgressController pc, Object component) {
		boolean retval = true;

		return retval;
	}

	public void menuProperties(Object sender) throws Exception {
		this.showView(this.netlab.getParent(sender));
	}

	public void setElement(Object component) {
		if (this.nsl != null)
			throw new IllegalStateException();

		super.setElement(component);

		this.tkv.setKeyValue(this, "nsl", this.nsobject);
		this.tkv.observeKeyValue(this.nsl, "", new ThinletKeyValueObserver() {
			public void observedKeyValue(Object obj, String keyPath, int flags,
					Object oldValue, Object newValue) {
				LanController.this.tkv.notifyKeyValue(LanController.this,
						"nsl." + keyPath, flags, oldValue, newValue);
			}
		});
	}
	
	public String tooltip() {
		String retval = "";
		
		if (this.nsl != null) {
			String fullName, shortName;
			Font newFont;

			fullName = this.nsl.getName().toString();
			shortName = this.nsl.getName().toShortString(this.nc.getShortNameLength());

			if (fullName.length() > shortName.length())
				retval += "  " + fullName;
			if (nsl.getTemplateInterface().getBandwidth() > 0) {
				retval += ("  BW: " + InterfacePropertiesController.BW_FORMATTER.
						stringForObjectValue(this.netlab, null, new Long(nsl.getTemplateInterface().getBandwidth())));
			}
			if (nsl.getTemplateInterface().getLatency() > 0) {
				retval += "  Lat: " + nsl.getTemplateInterface().getLatency()
						+ "ms";
			}
			if (nsl.getTemplateInterface().getLossRate() > 0.0) {
				retval += "  PLR: "
						+ (nsl.getTemplateInterface().getLossRate() * 100.0) + "%";
			}

			if (retval.length() == 0) {
				retval = "  (default config)  ";
				newFont = this.nc.getTooltipFont();
			} else {
				retval += "  ";
				newFont = this.nc.getPlainFont();
			}
			this.tkv.setKeyValue(this, "tooltipFont", newFont);
		}
		
		return retval;
	}

	public String toString() {
		return "LanController[" + "]";
	}
}
