/*
 * LanLinkPropertiesController.java
 *
 * Copyright (c) 2005 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

package net.emulab.netlab.client;

import java.util.Arrays;
import java.util.List;
import java.util.Vector;

import net.emulab.ns.NSInterface;
import net.emulab.ns.NSObject;
import thinlet.AutoFillController;
import thinlet.NumberFormatter;
import thinlet.SizeFormatter;

public class LanLinkPropertiesController extends PropertiesController {
	public static final IPv4AddressFormatter IPV4_FORMATTER = new IPv4AddressFormatter();

	public static final NumberFormatter BW_FORMATTER = new SizeFormatter("",
			"bps").setMinimum(0.1).setMaximum(NSInterface.MAX_BANDWIDTH)
			.setEmpty(new Long(-1));

	public static final NumberFormatter LATENCY_FORMATTER = new NumberFormatter()
			.setMinimum(0.0).setEmpty(new Double(-1.0));

	public static final NumberFormatter LOSS_RATE_FORMATTER = new NumberFormatter()
			.setMinimum(0.0).setMaximum(1.0).setEmpty(new Double(-1.0));

	private static final List QUEUEING_AUTO_FILL = Arrays.asList(new String[] {
			"DropTail", "RED", "GRED" });

	public static final NumberFormatter SNAPLEN_FORMATTER = new NumberFormatter()
			.setMinimum(0.0).setEmpty(new Integer(-1));

	private static final List TRACING_AUTO_FILL = Arrays.asList(new String[] {
			"header", "packet", "monitor" });

	public static final NumberFormatter RATE_FORMATTER = new SizeFormatter(
			null, "bps").setMinimum(0.0);

	public static final NumberFormatter PACKET_SIZE_FORMATTER = new NumberFormatter()
			.setMinimum(0.0);

	public static final NumberFormatter INTERVAL_FORMATTER = new NumberFormatter()
			.setMinimum(0.0);


	public static final NumberFormatter COST_FORMATTER = new NumberFormatter()
                        .setMinimum(0.0);

	public final AutoFillController qc;

	public final AutoFillController ltc;
	
	public LanLinkPropertiesController(NetlabClient netlab) {
		super(netlab);
		
		this.vis.put("trace", Boolean.TRUE);

		this.qc = new AutoFillController(netlab, new Vector(QUEUEING_AUTO_FILL));
		this.ltc = new AutoFillController(netlab, new Vector(TRACING_AUTO_FILL));
	}

	/**
	 * @see net.emulab.netlab.client.Controller#setView(java.lang.Object)
	 */
	public Controller setView(Object view) {
		super.setView(view);

		return this;
	}

	public void setNSObject(NSObject nso) {
		super.setNSObject(nso);

		/*
		 * if( nso instanceof NSLan ) { this.tkv.setKeyValue(this, "nslink",
		 * null); this.tkv.setKeyValue(this, "nslan", nso);
		 * this.tkv.setKeyValue( this, "nsi",
		 * ((NSLan)nso).getTemplateInterface());
		 * 
		 * this.tkv.setKeyValue(this.tc, "content", Collections.EMPTY_LIST); }
		 */
	}

	public String toString() {
		return "LANLinkPropertiesController[" + super.toString() + "]";
	}
}
