package net.emulab.netlab.client;

import java.util.Arrays;
import java.util.List;
import java.util.Vector;

import net.emulab.ns.NSLan;
import net.emulab.ns.NSObject;
import net.emulab.ns.Pair;
import thinlet.ArrayController;
import thinlet.AutoFillController;
import thinlet.ThinletKeyValue;

public class LanPropertiesController extends LanLinkPropertiesController {

	private static final List PROTOCOL_AUTO_FILL = Arrays.asList(new String[] {
			"80211a", "80211b", "80211g", "flex900" });

	public final TopologyController tc;
	public final AutoFillController wpc;
	public final AutoFillController apc;
	public final ArrayController lsc;
	
	public LanPropertiesController(NetlabClient netlab) {
		super(netlab);
		
		this.setObjectClass(NSLan.class);

		this.vis.put("pipe", Boolean.TRUE);
		this.vis.put("wifi", Boolean.FALSE);
		
		this.tc = this.netlab.getTopologyController();
		this.wpc = new AutoFillController(netlab, new Vector(PROTOCOL_AUTO_FILL));

		this.apc = new AutoFillController(netlab);
		this.tkv.observeKeyValue(this, "selection.members.NSObject.name",
				ThinletKeyValue.CHANGE_ANY, this.apc, "content");

		this.lsc = new ArrayController(netlab);
		this.lsc.setObjectClass(Pair.class);
	}
	
	public void setTopologySelection(Object selection) {
		super.setTopologySelection(selection);
		
		Object newContent;
		
		if (this.nsobject == null) {
			newContent = null;
		}
		else {
			newContent = this.tkv.getKeyValue(this.nsobject, "settings");
			if (!"".equals(this.tkv.getKeyValueAsString(this.nsobject, "protocol"))) {
				this.tkv.setKeyValue(this, "vis.wifi", true);
			}
		}
		
		this.tkv.setKeyValue(this.lsc, "content", newContent);
	}


	/**
	 * @see net.emulab.netlab.client.Controller#setView(java.lang.Object)
	 */
	public Controller setView(Object view) {
		super.setView(view);

		return this;
	}

	public void setNSObject(NSObject nso) {
		super.setNSObject(nso);
	}

	public String toString() {
		return "LinkPropertiesController[" + super.toString() + "]";
	}
}
