package net.emulab.netlab.client;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import net.emulab.ns.LinkTraffic;
import net.emulab.ns.NSLink;
import net.emulab.ns.NSNode;
import net.emulab.ns.NSObject;
import net.emulab.ns.NSTopology;

public class LinkPropertiesController extends LanLinkPropertiesController {

	public static final String TRAFFIC_NAME_BASE = "tg";
	
	public final UniqueNameFormatter trafficNameFormatter;

	public Object trafficDialog;

	public boolean trafficDialogOpened;

	public final TrafficArrayController tc;

	public LinkPropertiesController(NetlabClient netlab) {
		super(netlab);
		
		this.trafficNameFormatter = new UniqueNameFormatter(netlab.getTopologyController());
		
		this.setObjectClass(NSLink.class);

		this.vis.put("ips", Boolean.FALSE);
		this.vis.put("pipe", Boolean.TRUE);
		this.vis.put("traf", Boolean.TRUE);

		this.tc = new TrafficArrayController(netlab, this.trafficNameFormatter) {
			public Object newObject() {
				TopologyController tc = this.netlab.getTopologyController();
				NSLink nsl = (NSLink)LinkPropertiesController.this.selection;
				LinkTraffic lt;
				Object retval;

				lt = new LinkTraffic();
				lt.setName(tc.getNamer().suggest(LinkPropertiesController.TRAFFIC_NAME_BASE));
				lt.setLink(nsl);
				lt.setLeft((NSNode) nsl.getLeftMember().getNSObject());
				lt.setRight((NSNode) nsl.getRightMember().getNSObject());
				lt.setType(TrafficArrayController.DEFAULT_TYPE);
				lt.getAttributes().putAll(TrafficArrayController.DEFAULT_ATTRIBUTES);
				retval = lt;

				this.netlab.getTopologyController().addAggregate(lt);

				return retval;
			}
		};

		this.tkv.observeKeyValue(this.tc, "selection.attributes", new TrafficAttrObserver(this.tc, this.tkv));
	}

	/**
	 * @see net.emulab.netlab.client.Controller#setView(java.lang.Object)
	 */
	public Controller setView(Object view) {
		super.setView(view);

		try {
			this.trafficDialog = this.thinlet.parse("trafficdialog.xml", this);
		} catch (IOException e) {
			throw new InternalError();
		}
		return this;
	}

	public void setNSObject(NSObject nso) {
		super.setNSObject(nso);

		if (nso == null)
			this.tkv.setKeyValue(this, "vis.ips", false);
	}
	
	public void setTopologySelection(Object selection) {
		super.setTopologySelection(selection);

		NSTopology nst = this.netlab.getTopologyController().getNSTopology();
		List tccontents = new ArrayList();
		
		if (this.selection instanceof List) {
			List list = (List)this.selection;
			Iterator iter;
			
			for (iter = list.iterator(); iter.hasNext(); ) {
				NSLink nsl = (NSLink)iter.next();
				List aggs;
				
				aggs = nst.findAggregates(nsl, LinkTraffic.class);
				aggs.removeAll(tccontents);
				tccontents.addAll(aggs);
			}
		}
		else if (this.selection instanceof NSLink) {
			tccontents.addAll(nst.findAggregates((NSObject)this.selection, LinkTraffic.class));
		}
		this.tkv.setKeyValue(this.tc, "content", tccontents);
	}

	public void openTrafficDialog(Object sender) {
		if (!this.trafficDialogOpened) {
			this.thinlet.add(this.trafficDialog);
			this.trafficDialogOpened = true;
		}
	}

	public void closeTrafficDialog() {
		this.thinlet.remove(this.trafficDialog);
		this.trafficDialogOpened = false;
	}
	
	public String toString() {
		return "LinkPropertiesController[" + super.toString() + "]";
	}
}
