/*
 * LocalFileListController.java
 *
 * Copyright (c) 2005 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */
package net.emulab.netlab.client;

import java.util.Date;
import java.util.Vector;
import java.util.SortedSet;

import java.io.File;

public class LocalFileListController
    extends FileListController
{
    private static final int RECENT_DIR_COUNT = 10;
    private static final String DEFAULT_RECENT_DIRS[] = {
	System.getProperty("user.home"),
	System.getProperty("user.dir"),
    };
    
    protected String direction;
    protected File jPath;
    protected File jFile;
    
    public LocalFileListController(NetlabClient netbuild)
    {
	super(netbuild);

	NetlabPreferences nbp;
	int lpc;
	
	nbp = this.netbuild.getPreferences();
	this.sparePaths = nbp.getVector(NetlabPreferences.RECENT_DIR_PREFIX,
					RECENT_DIR_COUNT);
	if( this.sparePaths.size() == 0 )
	{
	    for( lpc = 0; lpc < DEFAULT_RECENT_DIRS.length; lpc++ )
	    {
		this.sparePaths.addElement(DEFAULT_RECENT_DIRS[lpc]);
	    }
	}
    }

    public Controller triggeredBy(Object component)
	throws Exception
    {
	this.direction = (String)
	    this.netbuild.getProperty(component, "direction");

	return super.triggeredBy(component);
    }

    private void updateDirectoryEntries(final Object filelist,
					final Object dir,
					final Object file,
					final Object parent)
    {
	File files[];
	int lpc;
	
	this.path = this.jPath.getAbsolutePath();
	files = this.jPath.listFiles();
	this.de = new DirectoryEntry[files.length];
	for( lpc = 0; lpc < this.de.length; lpc++ )
	{
	    this.de[lpc] = new DirectoryEntry(files[lpc]); // XXX Stupid
	}
	this.updateFileList(filelist, dir, file, parent);
    }
    
    public void showPanel(final Object panel,
			  final Object filelist,
			  final Object dir,
			  final Object file,
			  final Object spare,
			  final Object parent)
    {
	this.netbuild.setString(spare, "text", "Recent");
	this.netbuild.setString(spare,
				"tooltip",
				"Show the list of recently used directories");
	if( this.path != null )
	{
	    this.updateDirectoryEntries(filelist, dir, file, parent);
	}
	else
	{
	    this.updateFileList(filelist, dir, file, parent);
	}
    }
    
    public void acceptFile(final Object filelist,
			   Object item,
			   Object dir,
			   Object file,
			   Object parent)
	throws Exception
    {
	if( this.path == null )
	{
	    Object firstItem;
	    
	    this.jPath = new File(this.getCellValue(item, "path"));
	    this.updateDirectoryEntries(filelist, dir, file, parent);
	    this.netbuild.requestFocus(filelist);
	}
	else
	{
	    String name;

	    name = this.getCellValue(item, "name");
	    if( "(dir)".equals(this.getCellValue(item, "size")) )
	    {
		this.jPath = new File(this.jPath, name);
		this.updateDirectoryEntries(filelist, dir, file, parent);
		this.netbuild.requestFocus(filelist);
	    }
	    else
	    {
		this.netbuild.setString(file, "text", name);
		this.ok(dir, file);
	    }
	}
    }
    
    public void setPath(Object filelist,
			Object dir,
			Object file,
			Object parent)
    {
	String dirText;

	dirText = this.netbuild.getString(dir, "text");
	if( "".equals(dirText) )
	{
	    this.spare(filelist, dir, file, parent);
	}
	else
	{
	    File newPath;

	    newPath = new File(dirText);
	    if( newPath.exists() )
	    {
		this.jPath = newPath;
		this.updateDirectoryEntries(filelist, dir, file, parent);
		this.netbuild.requestFocus(file);
	    }
	    else
	    {
		this.netbuild.getToolkit().beep();
		this.netbuild.requestFocus(dir);
	    }
	}
    }

    public void setFile(Object ok)
    {
	this.netbuild.requestFocus(ok);
    }
    
    public void spare(Object filelist, Object dir, Object file, Object parent)
    {
	this.jPath = null;
	this.path = null;
	this.updateFileList(filelist, dir, file, parent);
    }
    
    public void parent(Object filelist, Object dir, Object file, Object parent)
    {
	if( this.jPath.getParentFile() != null )
	{
	    this.jPath = this.jPath.getParentFile();
	    this.updateDirectoryEntries(filelist, dir, file, parent);
	}
    }

    public void ok(Object dir, Object file)
	throws Exception
    {
	String dirText, fileText;
	NetlabPreferences nbp;
	int lpc;
	
	this.netbuild.remove(this.view);

	dirText = this.netbuild.getString(dir, "text");
	fileText = this.netbuild.getString(file, "text");

	try
	{
	    this.jFile = new File(dirText, fileText);
	    
	    if( "open".equals(this.direction) )
	    {
		this.netbuild.openFile(this.jFile);
	    }
	    else
	    {
		if( this.jFile.exists() )
		{
		    this.netbuild.showChoiceDialog("Overwrite existing file?",
						   this,
						   "overwrite(choice, 1)",
						   "overwrite(choice, 0)");
		}
		else
		{
		    this.writeFile();
		}
	    }
	}
	catch(Throwable th)
	{
	    th.printStackTrace();
	    this.netbuild.showErrorDialog(th.getMessage(), null);
	}

	if( !this.sparePaths.contains(dirText) )
	{
	    if( this.sparePaths.size() == RECENT_DIR_COUNT )
		this.sparePaths.remove(DEFAULT_RECENT_DIRS.length);
	    this.sparePaths.addElement(dirText);
	    nbp = this.netbuild.getPreferences();
	    nbp.setVector(NetlabPreferences.RECENT_DIR_PREFIX,
			  this.sparePaths);
	    nbp.store();
	}
    }

    public void overwrite(Object choiceDialog, int value)
	throws Exception
    {
	this.netbuild.remove(choiceDialog);
	if( value != 0 )
	{
	    this.writeFile();
	}
	else
	{
	    this.netbuild.add(this.view);
	}
    }

    protected void writeFile()
	throws Exception
    {
	if( "save".equals(this.direction) )
	    this.netbuild.saveAs(this.jFile);
	else if( "export".equals(this.direction) )
	    this.netbuild.exportAs(this.jFile);
	else
	    throw new InternalError();
    }
    
    public String toString()
    {
	return "LocalFileListController["
	    + super.toString()
	    + "]";
    }
}
