/*
 * NetlabArrayController.java
 *
 * Copyright (c) 2005 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

package net.emulab.netlab.client;

import thinlet.AutoFillController;

/**
 * ArrayController that is also implements the netlab Controller interface.
 */
public class NetlabArrayController extends AutoFillController implements
		Controller {
	/**
	 * The NetlabClient instance this controller is attached to.
	 */
	public final NetlabClient netlab;

	/**
	 * The view that is being controlled.
	 */
	protected Object view;

	/**
	 * Construct a NetlabArrayController with the given value.
	 * 
	 * @param netlab
	 *            The NetlabClient this controller is working with.
	 */
	public NetlabArrayController(NetlabClient netlab) {
		super(netlab);

		this.netlab = netlab;
	}

	/**
	 * @see net.emulab.netlab.client.Controller#showView(java.lang.Object)
	 */
	public void showView(Object component) throws Exception {
		this.netlab.showView(component);
	}

	/**
	 * @see net.emulab.netlab.client.Controller#setView(java.lang.Object)
	 */
	public Controller setView(Object view) {
		this.view = view;
		return this;
	}
	
	public Object getView() {
		return this.view;
	}

	/**
	 * @see net.emulab.netlab.client.Controller#modelChanged(java.lang.Object)
	 */
	public Controller modelChanged(Object component) {
		return this;
	}

	/**
	 * @see net.emulab.netlab.client.Controller#triggeredBy(java.lang.Object)
	 */
	public Controller triggeredBy(Object component) throws Exception {
		this.netlab.add(this.view);
		
		return this;
	}

	public String toString() {
		return "NetlabArrayController[" + super.toString() + "; view="
				+ this.view + "]";
	}
}
