/*
 * NetlabPreferencesController.java
 *
 * Copyright (c) 2005 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

package net.emulab.netlab.client;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.Vector;

import thinlet.ArrayController;
import thinlet.AutoFillController;
import thinlet.PropertiesController;

public class NetlabPreferencesController
    extends PropertiesController
    implements Controller
{
    protected final NetlabClient netlab;
    protected NetlabPreferences nbp;

    protected Object view;
    
    public ArrayController ac;
    public AutoFillController sc;
    
    /* Thinlet outlets... */
    
    public Object browser;
    public Object sshTemplate;
    
    public Object testTable;
    
    public NetlabPreferencesController(NetlabClient netlab)
    {
	super(netlab, netlab.getPreferences());

	this.netlab = netlab;
	this.nbp = this.netlab.getPreferences();
	this.setAppliesImmediately(false);

	this.ac = new ArrayController(netlab);

	Vector vector = new Vector();
	vector.add(new DirectoryEntry(new File("/")));
	vector.add(new DirectoryEntry(new File("/z")));
	vector.add(new DirectoryEntry(new File("/z/stack")));

	this.ac.getFilterKeys().add("name");
	
	this.netlab.getNamespace().
	    setKeyValue(this.ac, "content", vector);
	this.ac.setObjectClass(DirectoryEntry.class);

	this.sc = new AutoFillController(netlab);
    }

    public void showView(Object component)
	throws Exception
    {
	this.netlab.showView(component);
    }
    
    public Controller setView(Object view)
    {
	this.view = view;
	return this;
    }
    
    public Controller triggeredBy(Object component)
	throws Exception
    {
	this.netlab.add(this.view);
	return this;
    }
    
    public Controller modelChanged(Object component)
    {
	return this;
    }
    
    public void showPanel()
    {
	if( this.nbp != null )
	{
	    this.netlab.setString(this.browser,
				  "text",
				  this.nbp.getBrowserCommand());
	    this.netlab.setString(this.sshTemplate,
				  "text",
				  this.nbp.getSshTemplate());
	}
    }

    public void resetSshTemplate()
    {
	this.netlab.setString(this.sshTemplate,
			      "text",
			      NetlabPreferences.DEFAULT_SSH_CONFIG_TEMPLATE);
    }

    public void ok(String browser, String sshTemplate)
    {
	this.save();
	
	this.nbp.clearSshConfiguration();
	
	if( !"".equals(browser) )
	{
	    this.nbp.setProperty(NetlabPreferences.BROWSER_COMMAND, browser);
	}
	else
	{
	    this.nbp.remove(NetlabPreferences.BROWSER_COMMAND);
	}
	if( sshTemplate.equals(NetlabPreferences.DEFAULT_SSH_CONFIG_TEMPLATE) )
	{
	    NetlabPreferences.NETLAB_SSH_CONFIG_TEMPLATE.delete();
	}
	else
	{
	    try
	    {
		FileOutputStream fos;

		fos = new FileOutputStream(NetlabPreferences.
					   NETLAB_SSH_CONFIG_TEMPLATE);
		fos.write(sshTemplate.getBytes());
		fos.close();
	    }
	    catch(IOException e)
	    {
		e.printStackTrace(System.err);
	    }
	}
	BrowserController.flush();
	try
	{
	    this.nbp.store();
	}
	catch(IOException e)
	{
	    this.netlab.showErrorDialog("Error while saving preferences: "
					+ e.getMessage(),
					null);
	}
	this.netlab.updateExperimentMenu();
	this.netlab.remove(this.view);
    }

    public void cancel()
    {
	this.revert();
	this.netlab.remove(this.view);
    }
    
    public String toString()
    {
	return "NetlabPreferencesController[netlab="
	    + this.netlab
	    + "; nbp="
	    + this.nbp
	    + "]";
    }
}
