package net.emulab.netlab.client;

import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.Image;
import java.awt.Rectangle;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.PrintStream;
import java.lang.reflect.Constructor;
import java.util.Arrays;
import java.util.Hashtable;
import java.util.List;
import java.util.Vector;

import net.emulab.netlab.client.emulab.EmulabUserInfo;
import net.emulab.netlab.client.emulab.EmulabXmlRpcException;
import net.emulab.ns.NSFile;
import net.emulab.ns.NSInterface;
import net.emulab.ns.NSLan;
import net.emulab.ns.NSLanLink;
import net.emulab.ns.NSLink;
import net.emulab.ns.NSNode;
import net.emulab.ns.NSObject;
import net.emulab.ns.NSTopology;
import net.emulab.ns.Namer;
import thinlet.FrameLauncher;
import thinlet.Thinlet;
import thinlet.ThinletDragSourceBase;
import thinlet.ThinletDropDestinationBase;
import thinlet.ThinletEntryException;

public class OldNetlabClient extends Thinlet {
	private static final String BASE_UI_FILE = "netlab.xml";

	private static final String PALETTE_UI_FILE = "palette.xml";

	private static final String ERROR_UI_FILE = "errordialog.xml";

	private static final String CHOICE_UI_FILE = "choicedialog.xml";

	private static final String PROGRESS_UI_FILE = "progressdialog.xml";

	private static final String PROPERTIES_UI_FILE = "properties.xml";

	private static final String EMPTY_PROPERTIES_UI_FILE = "emptyproperties.xml";

	private static final String DEFAULT_LINK_NAME = "link";

	public static final ThinletDragSourceBase NODE_DRAG_SOURCE = new ThinletDragSourceBase() {
		public Object dragStart(Thinlet thinlet, Object sender) {
			Object retval;

			retval = thinlet.create("togglebutton");

			thinlet.setString(retval, "text", "Node");
			thinlet.setIcon(retval, "icon", thinlet.getIcon(sender, "icon"));
			thinlet.putProperty(retval, "controller",
					"net.emulab.netlab.client.NodeController");

			return retval;
		}
	};

	public static final ThinletDropDestinationBase DROP_DST = new ThinletDropDestinationBase() {
		public boolean draggingEntered(Thinlet thinlet, Object component,
				Object dropzone, int x, int y) {
			System.out.println("drag enter " + x + " " + y);
			thinlet.setBoolean(dropzone, "border", true);
			return true;
		}

		public void draggingExited(Thinlet thinlet, Object component,
				Object dropzone) {
			thinlet.setBoolean(dropzone, "border", false);
		}

	};

	public static final int SHORT_NAME_LENGTH = 8;

	public Object version;

	private static final Hashtable CONTROLLERS = new Hashtable();

	private boolean changed = false;

	private final Hashtable images = new Hashtable();

	private final Object palette;

	private boolean paletteVisible;

	public Object emulabMenu;

	private final Namer namer;

	private int objectCount;

	private final EmulabUserInfo eui = new EmulabUserInfo();

	private final Object errorDialog;

	public Object errorWhat;

	public Object errorFix;

	public Object errorDetails;

	public Object errorOk;

	private final Object choiceDialog;

	public Object choiceMsg;

	public Object choiceYes;

	public Object choiceNo;

	private final ProgressController pc;

	private final Object progressDialog;

	private final Object propertiesDialog;

	private final Object emptyPropertiesPanel;

	private ExperimentParameters ep;

	private String experimentState;

	private final NSFile nsf;

	private final NSTopology nst;

	private Object currentPropertiesPanel;

	private Object nextFocus;

	private ElementController clipController;

	private NSObject clipObject;

	private String clipBaseName;

	private Rectangle pasteBounds;

	public boolean errorDetailsVisible;

	public NetlabClient(String args[]) throws Exception {
		this.add(this.parse(BASE_UI_FILE));

		this.palette = this.parse(PALETTE_UI_FILE);

		this.errorDialog = this.parse(ERROR_UI_FILE);

		this.choiceDialog = this.parse(CHOICE_UI_FILE);

		this.pc = new ProgressController(this);
		this.progressDialog = this.parse(PROGRESS_UI_FILE, this.pc);

		this.propertiesDialog = this.parse(PROPERTIES_UI_FILE);
		this.add(this.find("deck"), this.propertiesDialog, 0);

		this.emptyPropertiesPanel = this.parse(EMPTY_PROPERTIES_UI_FILE);
		this.setPropertiesPanel(this.emptyPropertiesPanel);

		this.ep = new ExperimentParameters();

		this.nsf = new NSFile();
		this.nst = new NSTopology();
		this.namer = this.nst.getNamer();

		this.images.put("incomplete", this.getIcon("/icon/incomplete_tsk.gif"));
		this.images.put("complete", this.getIcon("/icon/complete_tsk.gif"));
		this.images.put("error", this.getIcon("/icon/error_tsk.gif"));
		this.images.put("warn", this.getIcon("/icon/warn_tsk.gif"));

		this.images.put("favicon", this.getIcon("/icon/favicon.gif"));

		this.images.put("folder", this.getIcon("/icon/folder.gif"));
		this.images.put("loading", this.getIcon("/icon/loading.gif"));

		this.images.put("blank", this.getIcon("/icon/blankball.gif"));
		this.images.put("black", this.getIcon("/icon/blackball.gif"));
		this.images.put("white", this.getIcon("/icon/whiteball.gif"));
		this.images.put("blue", this.getIcon("/icon/blueball.gif"));
		this.images.put("yellow", this.getIcon("/icon/yellowball.gif"));
		this.images.put("green", this.getIcon("/icon/greenball.gif"));
		this.images.put("pink", this.getIcon("/icon/pinkball.gif"));
		this.images.put("red", this.getIcon("/icon/redball.gif"));

		this.images.put("node-none", this.getIcon("/icon/node-none.gif"));
		this.images.put("node-up", this.getIcon("/icon/node-up.gif"));
		this.images.put("node-unknown", this.getIcon("/icon/node-unknown.gif"));
		this.images.put("node-down", this.getIcon("/icon/node-down.gif"));

		this.images.put("link-none", this.getIcon("/icon/link-none.gif"));
		this.images.put("link-up", this.getIcon("/icon/link-up.gif"));

		this.images.put("switch", this.getIcon("/icon/switch.gif"));

		this.images.put("coming-soon", this
				.getIcon("/icon/coming-soon-thumb.gif"));

		if (args.length > 0) {
			try {
				this.openFile(new File(args[0]));
			} catch (Throwable th) {
				this.showErrorDialog(th.getMessage());
			}
		}
	}

	private Dimension minSize = new Dimension(300, 250);

	public Dimension getMinimumSize() {
		return minSize;
	}

	public Image getObjectImage(String type) {
		Image retval;

		retval = (Image) this.images.get(type);
		return retval;
	}

	public Namer getNamer() {
		return this.namer;
	}

	public EmulabUserInfo getUserInfo() {
		return this.eui;
	}

	public NSTopology getNSTopology() {
		return this.nst;
	}

	public synchronized int incrementObjectCount() {
		int retval;

		retval = this.objectCount;
		this.objectCount += 1;
		return retval;
	}

	/**
	 * Open a link in a browser window.
	 * 
	 * @param sender
	 *            The Thinlet component that triggered this call and,
	 *            presumably, has the URL in an "aref" property.
	 * @throws Exception
	 *             if there was a problem opening the URL.
	 */
	public void openLinkInObject(Object sender) throws Exception {
		this.openLink((String) this.getProperty(sender, "aref"));
	}

	public void openLink(String aref) {
		setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
		if (!aref.startsWith("http://")) {
			// XXX aref = this.nbp.getEmulabWeb() + aref;
		}
		/* XXX showDocument */
		setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
	}

	public synchronized void addNSObject(NSObject nso) {
		this.nst.add(nso);
		this.updateExperimentMenu();
	}

	public synchronized void removeNSObject(NSObject nso) {
		this.nst.remove(nso);
		this.requestFocus(this.find("field"));
		this.updateExperimentMenu();
	}

	public synchronized void replaceNSObject(NSObject nsoOld, NSObject nsoNew) {
		this.nst.remove(nsoOld);
		this.nst.add(nsoNew);
		this.updateExperimentMenu();
	}

	public synchronized void updateNSObject(NSObject nso) {
		this.updateExperimentMenu();
	}

	public synchronized void setPaletteVisible(boolean value) {
		if (this.paletteVisible != value) {
			if (value)
				this.add(this.find("deck"), this.palette, 0);
			else
				this.remove(this.palette);
			this.paletteVisible = value;
		}
	}

	public void updateExperimentMenu() {
	}

	public Controller getController(String name) throws Exception {
		Controller retval;

		if ((retval = (Controller) CONTROLLERS.get(name)) == null) {
			Constructor constructor;
			String viewFileName;
			Class cc;

			cc = Class.forName(name);
			constructor = cc.getConstructor(new Class[] { NetlabClient.class });
			retval = (Controller) constructor
					.newInstance(new Object[] { this });

			if (!(retval instanceof ElementController))
				CONTROLLERS.put(name, retval);
		}
		return retval;
	}

	public void showView(Object sender) throws Exception {
		Object controllerObject, component;
		Controller controller;
		Object ui;

		if (sender instanceof List)
			component = ((List) sender).get(0);
		else
			component = sender;
		controllerObject = this.getProperty(component, "controller");
		if (controllerObject instanceof String) {
			if ((controller = (Controller) CONTROLLERS.get(controllerObject)) == null) {
				Constructor constructor;
				String viewFileName;
				Class cc;

				cc = Class.forName((String) controllerObject);
				constructor = cc
						.getConstructor(new Class[] { NetlabClient.class });
				controller = (Controller) constructor
						.newInstance(new Object[] { this });

				if ((viewFileName = (String) this
						.getProperty(component, "view")) != null) {
					Object view;

					view = this.parse(viewFileName, controller);
					controller.setView(view);
				}

				if (!(controller instanceof ElementController))
					CONTROLLERS.put(controllerObject, controller);
			}
			this.putProperty(component, "controller", controller);
		} else {
			controller = (Controller) controllerObject;
		}

		controller.triggeredBy(sender);
	}

	public void elementAction(Object sender) throws Exception {
		Object selected[];

		selected = this.getSelectedItems(this.find("field"));
		if (selected.length == 0) {
		} else if (selected.length == 1) {
			this.showView(sender);
		} else {
			ElementController first;
			boolean same = true;
			int lpc;

			first = (ElementController) this.getProperty(selected[0],
					"nsocontroller");
			for (lpc = 1; (lpc < selected.length) && same; lpc++) {
				ElementController ec = (ElementController) this.getProperty(
						selected[lpc], "nsocontroller");

				if (ec.getClass() != first.getClass())
					same = false;
			}
			if (same) {
				this.showView(Arrays.asList(selected));
			} else {
				// XXX this.setPropertiesPanel(this.multiplePropertiesPanel);
			}
		}
	}

	public void dropView(Object sender) throws Exception {
		this.showView(sender);
	}

	public synchronized void setPropertiesPanel(Object panel) {
		Object oldPanel;

		oldPanel = this.currentPropertiesPanel;
		this.remove(this.currentPropertiesPanel);
		/*
		 * XXX The remove can change focus and cause this method to be called
		 * again...
		 */
		if (oldPanel == this.currentPropertiesPanel) {
			if (panel != null)
				this.currentPropertiesPanel = panel;
			else
				this.currentPropertiesPanel = this.emptyPropertiesPanel;
			this.add(this.propertiesDialog, this.currentPropertiesPanel);
		}
	}

	public void debugPrint() {
		Object field = this.find("field");
		Object items[];
		int lpc;

		// System.out.println("net.emulab.netlab.client = " + this);
		items = this.getItems(field);
		for (lpc = 0; lpc < items.length; lpc++) {
			Hashtable ht;

			// System.out.println("model = "
			// + this.getProperty(items[lpc], "model");
			ht = new Hashtable();
			ht.put("OS", "RHL-STD");
			ht.put("hardware", "pc850");
			ht.put("startup", "foo");
			// this.putProperty(items[lpc], "mapping", ht);
		}
		this.nsf.toStream(System.out, this.nst);

		System.out.println("ag " + this.nst.getAgents());
		System.out.println("agg " + this.nst.getAggregates());
		// System.out.println("cl = " + this.changeLog);
		// System.out.println("cl = " + this.nst.getHardwareUsage());
		// this.tkv.setKeyValue(this, "experimentState", "active");
	}

	public void openFile(Object action) throws Exception {
		try {
			Object flow = this.findFlow(action);

			this.openFile((File) this.getProperty(flow, "file"));
		} catch (FileNotFoundException e) {
			this.showLongNoticeDialog("Could not find file named '"
					+ e.getMessage() + "'.",
					"Make sure the file exists and is " + "readable.", e
							.toString(), "closeErrorDialog", this);
		}
	}

	public void newFile() {
		Object field = this.find("field");

		this.tkv.setKeyValue(this, "experimentState", null);
		this.ep.clear();
		this.nst.clear();
		this.namer.clear();
		this.removeAll(field);
		this.setPropertiesPanel(null);
	}

	public void exportAs(File file) throws Exception {
		Object field = this.find("field");
		File tempFile, parent;
		PrintStream ps;
		Object items[];
		int lpc;

		if (file.exists() && !file.isFile()) {
			throw new IOException("'" + file + "' is not a valid file name");
		}

		parent = file.getParentFile();
		if (!parent.canWrite()) {
			throw new IOException("Cannot write to directory: " + parent);
		}

		tempFile = File.createTempFile("netbuildns", null, parent);
		tempFile.deleteOnExit();
		ps = new PrintStream(new FileOutputStream(tempFile));
		this.nsf.toStream(ps, this.nst);
		ps.close();

		if (!tempFile.renameTo(file)) {
			tempFile.delete();
			throw new IOException("Unable to move file " + tempFile + " to "
					+ file);
		}
	}

	private static final Hashtable EMPTY_MAPPING = new Hashtable();

	public void debugCheck() throws Exception {
		final ProgressController pc;

		pc = this.showProgressDialog("Checking your topology...", false);
		new Downloader(this, pc) {
			private Object currentItem;

			public void download() throws Throwable {
				Object field = super.netlab.find("field");
				boolean goodToGo = true;
				Object items[];
				int lpc;

				items = super.netlab.getItems(field);
				for (lpc = 0; (lpc < items.length) && goodToGo; lpc++) {
					PropertiesController ec;
					Object controller;

					currentItem = items[lpc];
					super.netlab.showView(items[lpc]);
					controller = super.netlab.getProperty(items[lpc],
							"controller");
					if (controller instanceof PropertiesController) {
						ec = (PropertiesController) controller;

						goodToGo = ec.verifyElement(super.pc, items[lpc]);
					} else {
						System.out.println("not a cont " + controller);
					}
				}
			}

			public void downloadError(EmulabXmlRpcException e) {
				super.downloadError(e);
				super.netlab.setBoolean(this.currentItem, "selected", true);
				try {
					super.netlab.showView(this.currentItem);
				} catch (Exception e2) {
				}
			}

			public void downloadCancelled() {
			}
		}.start();
	}

	public void showLongNoticeDialog(String what, String fix, String details,
			String action, Object handler) {
		this.setString(this.errorWhat, "text", what);
		this.setInteger(this.errorWhat, "rows", 1);
		this.setString(this.errorFix, "text", fix);
		this.setInteger(this.errorFix, "rows", 2);
		this.setString(this.errorDetails, "text", details);
		this.setMethod(this.errorOk, "action", action, this.errorDialog,
				handler);
		this.add(this.errorDialog);
	}

	public void showNoticeDialog(String msg, String title, String action,
			Object handler) {
		int len;

		if (msg == null)
			msg = "Unknown error";
		len = msg.length();
		this.setString(this.errorDialog, "text", title);
		this.setString(this.errorDetails, "text", "\n" + msg);
		this.setInteger(this.errorDetails, "columns", len < 40 ? len : 40);
		this.setInteger(this.errorDetails, "rows", (len + 39) / 40 + 1);
		this.setMethod(this.errorOk, "action", action, this.errorDialog,
				handler);
		this.add(this.errorDialog);
	}

	public void showErrorDialog(String msg, Object nextFocus) {
		// this.nextFocus = nextFocus;
		this.showNoticeDialog(msg, "Error", "closeErrorDialog", this);
	}

	public void showErrorDialog(String msg) {
		this.showErrorDialog(msg, null);
	}

	public void showChoiceDialog(String msg, Object handler, String methodYes,
			String methodNo, String lButton, String rButton) {
		this.setString(this.choiceMsg, "text", msg);
		this.setMethod(this.choiceYes, "action", methodYes, this.choiceDialog,
				handler);
		this.setMethod(this.choiceNo, "action", methodNo, this.choiceDialog,
				handler);
		this.setString(this.choiceYes, "text", lButton);
		this.setString(this.choiceNo, "text", rButton);
		this.add(this.choiceDialog);
	}

	public void showChoiceDialog(String msg, Object handler, String methodYes,
			String methodNo) {
		this.showChoiceDialog(msg, handler, methodYes, methodNo, "Yes", "No");
	}

	public ProgressController showProgressDialog(String title, boolean oneShot) {
		this.pc.setDialog(this.progressDialog);
		this.pc.setTitle(title);
		this.pc.setOneShot(oneShot);
		this.add(this.progressDialog);
		return this.pc;
	}

	public ProgressController showProgressDialog(String title) {
		return this.showProgressDialog(title, true);
	}

	public void closeErrorDialog() {
		this.remove(this.errorDialog);
		if (this.nextFocus != null) {
			this.requestFocus(this.nextFocus);
			this.nextFocus = null;
		}
	}

	public void closeChoiceDialog() {
		this.remove(this.choiceDialog);
	}

	public Object gotConnect(Object left, Object right) throws Exception {
		String cname, lname, rname;
		NSObject lobj, robj;
		Object parent;

		parent = this.getParent(left);
		lname = this.getString(left, "name");
		rname = this.getString(right, "name");
		if (lname.compareTo(rname) < 0)
			cname = lname + "|" + rname;
		else
			cname = rname + "|" + lname;
		lobj = (NSObject) this.getProperty(left, "model");
		robj = (NSObject) this.getProperty(right, "model");
		if (this.find(parent, cname) != null) {
			this
					.showErrorDialog("Multiple links between nodes is not supported.");
		} else if ((lobj instanceof NSLan) && (robj instanceof NSLan)) {
			this
					.showErrorDialog("Cannot link LANs together without a router node.");
		} else {
			ElementController controller;
			boolean islink = false;
			Object connector;
			NSInterface nsi;
			NSLanLink nsll;

			if ((lobj instanceof NSNode) && (robj instanceof NSNode)) {
				Vector members;

				islink = true;
				nsll = new NSLink();
				nsll.setName(this.namer.suggest(DEFAULT_LINK_NAME));
				members = nsll.getMembers();
				members.add(new NSInterface().setNSObject(lobj));
				members.add(new NSInterface().setNSObject(robj));
				controller = (ElementController) this
						.getController("net.emulab.netlab.client.LinkController");
				connector = this.parse("nslink.xml", controller);
				this.setIcon(connector, "icon", this
						.getObjectImage("link-none"));
				this.setString(connector, "text", nsll.getName().toShortString(
						SHORT_NAME_LENGTH));
				this.putProperty(connector, "model", nsll);
				this.nst.add(nsll);
				this.add(connector, this.parse("linkmenu.xml", controller));
				this.putProperty(connector, "view", "linkproperties.xml");
				this.putProperty(connector, "controller",
						"net.emulab.netlab.client.LinkPropertiesController");
			} else {
				NSNode nsn;

				if (lobj instanceof NSNode) {
					nsn = (NSNode) lobj;
					nsi = new NSInterface().setNSObject(lobj);
					nsll = (NSLan) robj;
				} else {
					nsn = (NSNode) robj;
					nsll = (NSLan) lobj;
					nsi = new NSInterface().setNSObject(robj);
				}
				nsi.setTemplate(((NSLan) nsll).getTemplateInterface());
				if (!nsll.isMember(nsn)) // This may be a bind
					nsll.addMember(nsi);
				controller = (ElementController) this
						.getController("net.emulab.netlab.client.InterfaceController");
				connector = this.parse("nslanlink.xml", controller);
				this.putProperty(connector, "model", nsi);
				this.putProperty(connector, "lan", nsll);
				this.add(connector, this.parse("intmenu.xml", controller));
				this.putProperty(connector, "view", "intproperties.xml");
				this
						.putProperty(connector, "controller",
								"net.emulab.netlab.client.InterfacePropertiesController");
			}
			this.setString(connector, "left", lname);
			this.setString(connector, "right", rname);
			this.setString(connector, "name", cname);
			this.setMethod(connector, "action", "showView(this)", connector,
					controller);
			this.setMethod(connector, "delete", "delete(this)", connector,
					controller);
			this.putProperty(connector, "nsocontroller", controller);

			this.add(parent, connector, 0);

			this.tkv.setKeyValue(controller, "element", connector);
			controller.modelChanged(connector);

			return connector;
		}
		return null;
	}

	private int nodeIndex() {
		boolean done = false;
		int lpc, retval = 0;
		Object items[];

		items = this.getItems(this.find("field"));
		for (lpc = 0; (lpc < items.length) && !done; lpc++) {
			NSObject nso;

			nso = (NSObject) this.getProperty(items[lpc], "model");
			if (nso instanceof NSNode) {
				retval = lpc;
				done = true;
			}
		}
		return retval;
	}

	public void doCutCopy(int copy) throws Exception {
		Object obj;

		if ((this.experimentState == null)
				&& (obj = this.getSelectedItem(this.find("field"))) != null) {
			NSObject nso;

			nso = (NSObject) this.getProperty(obj, "model");
			if ((nso instanceof NSNode) || (nso instanceof NSLan)) {
				this.clipController = (ElementController) this.getProperty(obj,
						"nsocontroller");
				if (nso instanceof NSLan)
					((NSLan) nso).getMembers().clear();
				this.clipObject = (NSObject) nso.clone();
				this.clipBaseName = nso.getName().getBase();
				this.pasteBounds = (Rectangle) (this
						.getRectangle(obj, "bounds").clone());
				if (copy == 0) {
					this.clipObject.setName(nso.getName());
					this.clipController.delete(obj);
				}
			} else {
				this.clipController = null;
				this.clipObject = null;
				this.clipBaseName = null;
				this.pasteBounds = null;
			}
			this.updateExperimentMenu();
		}
	}

	public void doPaste() throws Exception {
		if ((this.experimentState == null) && (this.clipObject != null)) {
			Rectangle bounds;
			NSObject nso;
			Object obj;

			nso = (NSObject) this.clipObject.clone();
			if ((nso.getName() != null) && !this.namer.contains(nso.getName())) {
				this.namer.put(nso.getName());
			} else {
				nso.setName(this.namer.suggest(this.clipBaseName));
			}
			obj = this.clipController.copyElement(nso);
			this.pasteBounds.x += 40;
			this.pasteBounds.y += 40;
			this.setRectangle(obj, "bounds", this.pasteBounds.x,
					this.pasteBounds.y, this.pasteBounds.width,
					this.pasteBounds.height);
			this.add(this.find("field"), obj, this.nodeIndex());
		}
	}

	public void deselect(Object field) {
		Object selected[];

		if ((selected = this.getSelectedItems(field)) != null) {
			int lpc;

			for (lpc = 0; lpc < selected.length; lpc++) {
				this.setBoolean(selected[lpc], "selected", false);
			}
			this.updateExperimentMenu();
		}
		this.setPropertiesPanel(null);
	}

	public void quit() {
		if (this.changed) {
			this.showChoiceDialog("Exit without saving changes?", this,
					"quitStage2()", "closeChoiceDialog()");
		} else {
			this.quitStage2();
		}
	}

	public void quitStage2() {
		this.closeChoiceDialog();
		if ("active".equals(this.experimentState)) {
			this.showChoiceDialog("Your Emulab experiment is still active:",
					this, "swapExperiment(this, 1)", "forceQuit()", "Swap Out",
					"Leave Running");
		} else {
			this.forceQuit();
		}
	}

	public void forceQuit() {
		System.exit(0);
	}

	protected String traceToString(Throwable th) {
		ByteArrayOutputStream baos = new ByteArrayOutputStream(1024);
		PrintStream ps = new PrintStream(baos);
		String retval;

		th.printStackTrace(ps);
		retval = baos.toString().replaceAll("\t", "        ");
		return retval;
	}

	protected void handleException(Object component, Throwable th) {
		th.printStackTrace();
		if (th instanceof ThinletEntryException) {
			this.showLongNoticeDialog(th.getMessage(),
					((ThinletEntryException) th).getExplanation(), this
							.traceToString(th), "closeErrorDialog", this);
		} else {
			showErrorDialog(th.getMessage(), component);
		}
	}

	public String toString() {
		return "NetlabClient[namer=" + this.namer + "]";
	}

	public static void main(String args[]) throws Throwable {
		int width = 700, height = 600;
		NetlabClient nb;

		System.setProperty("java.util.prefs.PreferencesFactory",
				"net.emulab.netlab.client.DisabledPreferencesFactory");

		nb = new NetlabClient(args);

		if (nb.getWidth() != 0)
			width = nb.getWidth();
		if (nb.getHeight() != 0)
			height = nb.getHeight();

		new FrameLauncher("NetlabClient", nb, width, height);
	}
}
