/*
 * PropertiesController.java
 *
 * Copyright (c) 2005 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

package net.emulab.netlab.client;

import java.util.Hashtable;
import java.util.List;
import java.util.Vector;

import net.emulab.ns.NSObject;
import thinlet.ThinletController;
import thinlet.ThinletKeyValue;

/**
 * Base class for controllers of the properties dialog.
 */
public class PropertiesController extends ControllerBase {
	public final UniqueNameFormatter nameFormatter;

	/**
	 * The component being manipulated.
	 */
	protected Object component;

	/**
	 * The model whose properties are being shown/edited.
	 */
	protected NSObject nsobject;

	protected Object topologySelection;
	
	protected Object selection;

	public boolean locked;

	public final Hashtable vis = new Hashtable();

	/**
	 * Construct a PropertiesController with the given value.
	 * 
	 * @param netlab
	 *            The NetlabClient instance this controller is attached to.
	 */
	protected PropertiesController(NetlabClient netlab) {
		super(netlab);

		this.nameFormatter = new UniqueNameFormatter(netlab.
				getTopologyController());

		this.tkv.addKeyDependency(this, "isSingleSelection", "selection");
		this.tkv.addKeyDependency(this, "isMultipleNSObjectSelection", "selection");
		this.tkv.observeKeyValue(this.netlab.getTopologyController(), "selection",
				ThinletKeyValue.CHANGE_SETTING, this, "topologySelection");
	}

	public void setSelection(Object selection) {
		this.selection = selection;
	}
	
	public Object getTopologySelection() {
		return this.topologySelection;
	}
	
	public void setTopologySelection(Object selection) {
		Object newSelection;
		
		this.topologySelection = selection;
		if (selection instanceof List) {
			List newSelectionList = new Vector();
			List list = (List)selection;
			NSObject nnso = null;
			int lpc, size;

			size = list.size();
			for (lpc = 0; lpc < size; lpc++) {
				Object model, item = list.get(lpc);

				model = this.thinlet.getProperty(item, "model");
				if (model.getClass() == this.objectClass())
					newSelectionList.add(model);
			}

			switch (newSelectionList.size()) {
			case 0:
				newSelection = null;
				break;
			case 1:
				newSelection = newSelectionList.get(0);
				nnso = (NSObject)newSelection;
				break;
			default:
				newSelection = newSelectionList;
			}
			this.nameFormatter.setNSObject(nnso);
			this.tkv.setKeyValue(this, "NSObject", nnso);
		} else {
			NSObject nso;

			nso = (NSObject) this.netlab.getProperty(selection, "model");
			if (nso != null && nso.getClass() == this.objectClass()) {
				newSelection = nso;
				this.tkv.setKeyValue(this, "NSObject", nso);
				this.nameFormatter.setNSObject(nso);
			}
			else {
				this.tkv.setKeyValue(this, "NSObject", null);
				newSelection = null;
			}
		}
		
		this.tkv.setKeyValue(this, "selection", newSelection);
	}

	public boolean isSingleSelection() {
		return (this.selection != null) && !(this.selection instanceof List);
	}

	/**
	 * @see net.emulab.netlab.client.Controller#triggeredBy(java.lang.Object)
	 */
	public Controller triggeredBy(Object sender) {
		this.setComponent(sender);
		this.tkv.setKeyValue(this.netlab, "propertiesIndex", this.netlab.propertyPanelIndex(this.view));
		
		return this;
	}

	/**
	 * @param component
	 *            The component being manipulated.
	 */
	public void setComponent(Object component) {
		this.component = component;
	}

	/**
	 * @return The component being manipulated.
	 */
	public Object getComponent() {
		return this.component;
	}

	/**
	 * @param nsobject
	 *            The model whose properties are being shown/edited.
	 */
	public void setNSObject(NSObject nsobject) {
		this.nsobject = nsobject;
	}

	/**
	 * @return The model whose properties are being shown/edited.
	 */
	public NSObject getNSObject() {
		return this.nsobject;
	}

	public boolean verifyElement(ProgressController pc, Object component)
			throws Exception {
		boolean retval = true;

		return retval;
	}
	
	public boolean isMultipleNSObjectSelection() {
		return (this.selection instanceof List);
	}

	public Object selection() {
		Object retval;
		
		if (this.selection != null)
			retval = this.selection;
		else
			retval = ThinletController.NO_SELECTION;
		
		return retval;
	}

	/**
	 * @see java.lang.Object#toString()
	 */
	public String toString() {
		return "PropertiesController[net.emulab.netlab.client=" + this.netlab
				+ "; component=" + this.component + "; nsobject="
				+ this.nsobject + "]";
	}
}
