/*
 * NodeController.java
 *
 * Copyright (c) 2005 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */
package net.emulab.netlab.client;

import java.awt.Font;
import java.awt.Image;
import java.text.MessageFormat;
import java.util.Iterator;

import net.emulab.ns.NSInterface;
import net.emulab.ns.NSLanLink;
import net.emulab.ns.NSLink;
import net.emulab.ns.NSNode;
import net.emulab.ns.NSObject;
import net.emulab.ns.NSProgram;
import net.emulab.ns.NSTopology;
import net.emulab.ns.Namer;
import thinlet.Thinlet;
import thinlet.ThinletDragSource;
import thinlet.ThinletDragSourceBase;
import thinlet.ThinletKeyValueObserver;

public class RAPController extends ElementControllerBase {
	private static final String TOOLTIP_FORMAT = "{0,choice,0#|1#  {1}}"
			+ "{2,choice,0#|1#  HW: {3}}" + "{4,choice,0#|1#  OS: {5}}";

	private static final String DEFAULT_NODE_NAME = "RAP";

	private static final ThinletDragSource DRAG_SOURCE = new ThinletDragSourceBase() {
		public Object dragStart(Thinlet thinlet, Object sender) {
			thinlet.setBoolean(sender, "selected", false);

			return sender;
		}
	};

	public NSNode nsn;

	public RAPController(NetlabClient netlab) {
		super(netlab);

		this.tkv.addKeyDependency(this, "nsn", "nsobject");
		this.tkv.addKeyDependency(this, "icon", "netlab");
		this.tkv.addKeyDependency(this, "tooltip", "nsn");
	}

	/**
	 * @return The appropriate link icon that corresponds to its current state.
	 */
	public Image getIcon() {
		Image retval;

		retval = this.netlab.getImageManager().getObjectImage("node-rap");
		return retval;
	}

	public void showView(Object component) throws Exception {
		this.netlab.showView(component);
	}

	public Controller setView(Object view) {
		return this;
	}

	public Controller modelChanged(Object component) {
		return this;
	}

	private void primeComponent(Object component) throws Exception {
		TopologyController tc = this.netlab.getTopologyController();
		String name;
		NSNode nsnode;

		nsnode = (NSNode) this.netlab.getProperty(component, "model");
		name = this.netlab.getString(component, "name");
		if ((name == null) || "".equals(name) || "nsnode".equals(name)) {
			this.netlab.setString(component, "name", "nsnode-"
					+ tc.incrementObjectCount());
		}
		nsnode.setOS("ncSCADA-RAP");
		this.netlab.setDelegate(component, "dragsrc", DRAG_SOURCE);
		this.netlab.bindAttribute(component, "text", "@shortName", this);
		this.netlab.bindAttribute(component, "icon", "@icon", this);
		this.netlab.bindAttribute(component, "tooltip", "@tooltip", this);
		this.netlab.bindAttribute(component, "tooltipfont", "@tooltipFont",
				this);
		this.netlab.setMethod(component, "action", "elementAction(this)",
				component, this.netlab);
		this.netlab.setMethod(component, "focusgained", "elementAction(this)",
				component, this.netlab);
		this.netlab.setMethod(component, "connect",
				"gotConnect(connector, connectee)", component, this);
		this.netlab.setMethod(component, "delete", "deleteNSObject(this)", component,
				this.netlab);
		this.netlab.setString(component, "proximitytooltip",
				ElementController.PROXIMITY_TOOLTIP);
		this.netlab.putProperty(component, "nsocontroller", this);
		this.netlab.putProperty(component, "view", "RAPproperties.xml");
		this.netlab.putProperty(component, "controller",
				"net.emulab.netlab.client.RAPPropertiesController");
		this.netlab.setInteger(component, "proximity", NetlabClient.PROXIMITY);

		this.netlab.add(component, this.netlab.parse("RAPmenu.xml", this));

		tc.addNSObject(nsnode);
	}

	public Controller triggeredBy(Object component) throws Exception {
		TopologyController tc = this.netlab.getTopologyController();
		NSNode nsnode;

		if ((nsnode = (NSNode) this.netlab.getProperty(component, "model")) == null) {
			nsnode = new NSNode();
			nsnode.setName(tc.getNamer().suggest(DEFAULT_NODE_NAME));
			this.netlab.putProperty(component, "model", nsnode);
		} else {
			tc.getNamer().put(nsnode.getName());
		}

		this.primeComponent(component);

		this.tkv.setKeyValue(this, "element", component);

		this.showView(component);
		this.netlab.requestFocus(component);

		return this;
	}

	public void gotConnect(Object left, Object right) throws Exception {
		this.netlab.gotConnect(left, right);
	}

	public void delete(Object component) {
		TopologyController tc = this.netlab.getTopologyController();
		Namer namer = tc.getNamer();
		Object parent, items[];
		Iterator iter;
		NSNode nsnode;
		int lpc;

		nsnode = (NSNode) this.netlab.getProperty(component, "model");
		parent = this.netlab.getParent(component);
		namer.remove(nsnode.getName());
		for (iter = nsnode.getPrograms().iterator(); iter.hasNext(); ) {
			NSProgram nsp = (NSProgram)iter.next();
			
			namer.remove(nsp.getName());
		}
		this.netlab.remove(component);
		// XXX this.netlab.setPropertiesPanel(null);
		items = this.netlab.getItems(parent);
		for (lpc = 0; lpc < items.length; lpc++) {
			ElementController ec;
			NSObject nso;

			ec = (ElementController) this.netlab.getProperty(items[lpc],
					"nsocontroller");
			nso = (NSObject) this.netlab.getProperty(items[lpc], "model");
			if (nso instanceof NSLanLink) {
				NSLanLink nsll = (NSLanLink) nso;

				if (nsll.removeMember(nsnode) && (nsll instanceof NSLink))
					ec.delete(items[lpc]);
			} else if (nso instanceof NSInterface) {
				NSInterface nsi = (NSInterface) nso;

				if (nsi.getNSObject().equals(nsnode)) {
					this.netlab.remove(items[lpc]);
				}
			}
		}
		tc.removeNSObject(nsnode);
	}

	public Object copyElement(NSObject model) throws Exception {
		TopologyController tc = this.netlab.getTopologyController();
		NSTopology nst = tc.getNSTopology();
		Namer namer = nst.getNamer();
		Iterator iter;
		Object retval;

		retval = this.netlab.create("togglebutton");
		this.netlab.putProperty(retval, "model", model);
		this.primeComponent(retval);
		this.tkv.setKeyValue(this, "element", retval);
		
		for (iter = this.nsn.getPrograms().iterator(); iter.hasNext(); ) {
			NSProgram nsp = (NSProgram) iter.next();
			
			nsp.setName(namer.suggest(nsp.getName().toString()));
			nst.add(nsp);
		}
		
		return retval;
	}

	public boolean verifyElement(ProgressController pc, Object component)
			throws Exception {
		boolean retval = true;

		return retval;
	}

	public void menuProperties(Object popup) throws Exception {
		this.showView(this.netlab.getParent(popup));
	}

	public void setElement(Object component) {
		if (this.nsn != null)
			throw new IllegalStateException();

		super.setElement(component);

		this.tkv.setKeyValue(this, "nsn", this.nsobject);
		this.tkv.observeKeyValue(this.nsn, "", new ThinletKeyValueObserver() {
			public void observedKeyValue(Object obj, String keyPath, int flags,
					Object oldValue, Object newValue) {
				RAPController.this.tkv.notifyKeyValue(RAPController.this,
						"nsn." + keyPath, flags, oldValue, newValue);
			}
		});
	}

	public String tooltip() {
		String retval = "";

		if (this.nsn != null) {
			String fullName, shortName;
			Font newFont;

			fullName = this.nsn.getName().toString();
			shortName = this.nsn.getName().toShortString(
					this.nc.getShortNameLength());

			retval = MessageFormat.format(TOOLTIP_FORMAT, new Object[] {
					new Integer((fullName.length() > shortName.length()) ? 1
							: 0), fullName,

					new Integer("".equals(this.nsn.getHardware()) ? 0 : 1),
					this.nsn.getHardware(),

					new Integer("".equals(this.nsn.getOS()) ? 0 : 1),
					this.nsn.getOS(), });

			if (retval.length() == 0) {
				retval = "  (default config)  ";
				newFont = this.nc.getTooltipFont();
			} else {
				retval += "  ";
				newFont = this.nc.getPlainFont();
			}
			this.tkv.setKeyValue(this, "tooltipFont", newFont);
		}
		return retval;
	}

	public String toString() {
		return "RAPController[" + super.toString() + "; nsn=" + this.nsn + "]";
	}
}
