/*
 * UniqueNameFormatter.java
 *
 * Copyright (c) 2005 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

package net.emulab.netlab.client;

import java.text.FieldPosition;

import net.emulab.ns.NSObject;
import net.emulab.ns.Namer;
import net.emulab.ns.UniqueName;
import thinlet.ThinletAPI;
import thinlet.ThinletEntryException;
import thinlet.ThinletFormatter;

public class UniqueNameFormatter implements ThinletFormatter {
	private static final String STD_EXPLANATION = "Names must start with a letter and be unique.";

	private final TopologyController tc;

	private NSObject nso;

	public UniqueNameFormatter(TopologyController tc) {
		this.tc = tc;
	}

	public void setNSObject(NSObject nso) {
		this.nso = nso;
	}

	public NSObject getNSObject() {
		return this.nso;
	}

	public Object getObjectValue(String string) throws Exception {
		Object retval;

		if ("".equals(string)) {
			throw new ThinletEntryException("Empty names are not allowed.",
					STD_EXPLANATION);
		} else if (Character.isDigit(string.charAt(0))) {
			throw new ThinletEntryException("Name cannot start with a digit: "
					+ string, STD_EXPLANATION);
		} else if (this.nso.getName().toString().equals(string)) {
			retval = this.nso.getName();
		} else {
			UniqueName un = new UniqueName(string);

			try {
				Namer namer = this.tc.getNamer();
				
				namer.put(un);
				namer.remove(this.nso.getName());
				retval = un;
			} catch (IllegalArgumentException e) {
				e.printStackTrace();
				
				/* TODO: Make sure this is why the exception was thrown. */
				throw new ThinletEntryException("Name is already in use: "
						+ string, STD_EXPLANATION);
			}
		}

		return retval;
	}

	public String stringForObjectValue(ThinletAPI thinlet, Object component,
			Object value) {
		String retval = "";

		if (value != null)
			retval = value.toString();
		return retval;
	}

	public boolean isPartialStringValid(ThinletAPI thinlet, Object component,
			StringBuffer partialString, FieldPosition fp) {
		return true;
	}

	public String toString() {
		return "UniqueNameFormatter[tc=" + this.tc + "]";
	}
}
