/*
 * EmulabEmulabProxy.java
 *
 * Copyright (c) 2005 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */
package net.emulab.netlab.client.emulab;

import java.io.IOException;
import java.net.MalformedURLException;
import java.util.Date;
import java.util.Hashtable;
import java.util.Vector;

import org.apache.xmlrpc.XmlRpcException;

import thinlet.ObjectPool;
import thinlet.ThinletKeyValue;

/**
 * Client-side XML-RPC proxy for accessing the Emulab "emulab" module.
 */
public class EmulabEmulabProxy extends EmulabProxy {
	/**
	 * The "news" method name.
	 */
	private static final String METHOD_NEWS = "emulab.news";

	private static final String METHOD_MESSAGE = "emulab.message";

	/**
	 * Construct an EmulabUserProxy with the given values.
	 * 
	 * @param nbp
	 *            The NetlabPreferences to use when constructing the URL for the
	 *            RPC server.
	 * @param po
	 *            The ProgressObserver that is interested in this proxy.
	 * @throws MalformedURLException
	 *             if the preferences specified by the user form a bad URL.
	 */
	public EmulabEmulabProxy(ThinletKeyValue tkv, ObjectPool op)
			throws MalformedURLException {
		super(tkv, op);
	}

	/**
	 * Get the news item headers for a period of time.
	 * 
	 * @param starting
	 *            The date to start searching for news items or null if it
	 *            should start from the beginning of time.
	 * @param ending
	 *            The date to stop searching for news items or null if it should
	 *            stop at the end of time.
	 * @return A vector of tables that describe each news item.
	 * @throws XmlRpcException
	 *             if there was an error while trying to execute the method.
	 * @throws IOException
	 *             if there was an error while communicating with the server.
	 */
	public synchronized Vector news(Date starting, Date ending)
			throws XmlRpcException, IOException {
		Hashtable args;
		Vector retval;
		Object rc;

		args = new Hashtable();
		if (starting != null)
			args.put("starting", starting);
		if (ending != null)
			args.put("ending", ending);

		rc = this._invoke(METHOD_NEWS, args);

		if (!(rc instanceof Vector))
			throw new XmlRpcException(0, "Bad response: " + rc);

		retval = (Vector) rc;

		return retval;
	}

	public synchronized String message() throws XmlRpcException, IOException {
		String retval;
		Object rc;

		rc = this._invoke(METHOD_MESSAGE);

		if (!(rc instanceof String))
			throw new XmlRpcException(0, "Bad response: " + rc);

		retval = (String) rc;

		return retval;
	}

	/**
	 * Get all of the news item headers.
	 * 
	 * @return A vector of tables that describe each news item.
	 * @throws XmlRpcException
	 *             if there was an error while trying to execute the method.
	 * @throws IOException
	 *             if there was an error while communicating with the server.
	 */
	public synchronized Vector news() throws XmlRpcException, IOException {
		return this.news(null, null);
	}

	/**
	 * @see java.lang.Object#toString()
	 */
	public String toString() {
		return "EmulabEmulabProxy[" + super.toString() + "]";
	}
}
