/*
 * EmulabOSIDProxy.java
 *
 * Copyright (c) 2005 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */
package net.emulab.netlab.client.emulab;

import java.io.IOException;
import java.net.MalformedURLException;
import java.util.Hashtable;

import org.apache.xmlrpc.XmlRpcException;

import thinlet.ObjectPool;
import thinlet.ThinletKeyValue;

/**
 * Client-side XML-RPC proxy for accessing the Emulab "osid" module.
 */
public class EmulabOSIDProxy extends EmulabProxy {
	/**
	 * The "getlist" method name.
	 */
	private static final String METHOD_GETLIST = "osid.getlist";

	/**
	 * Construct an EmulabOSIDProxy with the given values.
	 * 
	 * @param nbp
	 *            The NetlabPreferences to use when constructing the URL for the
	 *            RPC server.
	 * @param po
	 *            The ProgressObserver that is interested in this proxy.
	 * @throws MalformedURLException
	 *             if the preferences specified by the user form a bad URL.
	 */
	public EmulabOSIDProxy(ThinletKeyValue tkv, ObjectPool op)
			throws MalformedURLException {
		super(tkv, op);
	}

	/**
	 * Get the list of OS IDs visible to this user.
	 * 
	 * @return A hashtable indexed by OS ID where each element is another table
	 *         containing attributes.
	 * @throws XmlRpcException
	 *             if there was an error while trying to execute the method.
	 * @throws IOException
	 *             if there was an error while communicating with the server.
	 */
	public synchronized Hashtable getlist() throws XmlRpcException, IOException {
		Hashtable retval;
		Object rc;

		rc = this._invoke(METHOD_GETLIST);
		if (!(rc instanceof Hashtable)) {
			throw new XmlRpcException(0, "Bad response: " + rc);
		}

		retval = (Hashtable) rc;

		return retval;
	}

	/**
	 * @see java.lang.Object#toString()
	 */
	public String toString() {
		return "EmulabOSIDProxy[" + super.toString() + "]";
	}
}
