/*
 * EmulabXmlRpcException.java
 *
 * Copyright (c) 2005 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

package net.emulab.netlab.client.emulab;

import org.apache.xmlrpc.XmlRpcException;

/**
 * An XmlRpcException .
 */
public class EmulabXmlRpcException extends XmlRpcException {
	/**
	 * The integer code of the response.
	 * 
	 * @see net.emulab.netlab.client.emulab.EmulabResponse
	 */
	private final int emulabCode;

	/**
	 * Construct an EmulabXmlRpcException with the given values.
	 * 
	 * @see net.emulab.netlab.client.emulab.EmulabResponse
	 * @param emulabCode
	 *            The integer code of the response.
	 * @param output
	 *            The error message from the server.
	 */
	public EmulabXmlRpcException(int emulabCode, Object output) {
		super(0, EmulabResponse.codeToString(emulabCode) + ": "
				+ output.toString());

		this.emulabCode = emulabCode;
	}

	/**
	 * @return The integer code of the response.
	 * @see net.emulab.netlab.client.emulab.EmulabResponse
	 */
	public int getCode() {
		return this.emulabCode;
	}
}
