package net.emulab.netlab.oper;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.StringTokenizer;

import thinlet.ArrayController;
import thinlet.AutoFillController;
import thinlet.FrameLauncher;
import thinlet.Thinlet;

public class Operator extends Thinlet {
	/**
	 * The keys to scan when filtering out entries.
	 */
	private static final String FILTER_KEYS[] = { "action", "pid", "eid",
			"user", };

	public final ArrayController ac = new ArrayController(this);

	public final AutoFillController sc = new AutoFillController(this);

	public static class Record {
		public final String action;

		public final String pid;

		public final String eid;

		public final String user;

		public Record(StringTokenizer st) {
			this.action = st.nextToken();
			this.pid = st.nextToken();
			this.eid = st.nextToken();
			this.user = st.nextToken();
		}

		public String toString() {
			return "Record[action=" + this.action + "; pid=" + this.pid
					+ "; eid=" + this.eid + "; user=" + this.user + "]";
		}
	}

	public Operator(String args[]) throws Exception {
		this.add(this.parse("oper.xml"));

		this.ac.addFilterKeys(FILTER_KEYS);
	}

	public static void main(String args[]) throws Throwable {
		int width = 600, height = 500;
		final Operator oper;

		System.setProperty("java.util.prefs.PreferencesFactory",
				"net.emulab.netlab.client.DisabledPreferencesFactory");

		oper = new Operator(args);
		if (oper.getWidth() != 0)
			width = oper.getWidth();
		if (oper.getHeight() != 0)
			height = oper.getHeight();

		new FrameLauncher("NetlabOperator", oper, width, height);

		new Thread() {
			public void run() {
				try {
					InputStreamReader isr = new InputStreamReader(System.in);
					BufferedReader br = new BufferedReader(isr);
					String line;

					while ((line = br.readLine()) != null) {
						StringTokenizer st = new StringTokenizer(line);
						Record record = new Record(st);

						oper.ac.addObject(record);
					}
				} catch (IOException e) {
					e.printStackTrace();
					System.exit(1);
				}
			}
		}.start();
	}
}
