package net.emulab.ns;

import java.util.Hashtable;
import java.util.List;
import java.util.SortedSet;
import java.util.Vector;

public class LinkTraffic extends NSBase implements NSAggregate {
	public static final int BIDIRECTIONAL = 0;

	public static final int LEFT_TO_RIGHT = 1;

	public static final int RIGHT_TO_LEFT = 2;

	public static final Hashtable TRAFFIC_TYPES = new Hashtable();

	static {
		TRAFFIC_TYPES.put("UDP", new String[] { "Agent/UDP",
				"Application/Traffic/CBR", "Agent/Null", });
		TRAFFIC_TYPES.put("TCP", new String[] { "Agent/TCP",
				"Application/Traffic/CBR", "Agent/TCPSink", });
	}

	static List group(NSTopology nst) {
		NSConnectableObject nsco[];
		List retval = new Vector();
		Namer namer = nst.getNamer();
		List list;
		int lpc;

		list = nst.getConnectors();
		nsco = (NSConnectableObject[]) list.toArray(new NSConnectableObject[list
				.size()]);
		for (lpc = 0; lpc < nsco.length; lpc++) {
			if ((nsco[lpc] instanceof NSAgent)
					&& nst.getAgents().contains(nsco[lpc])) {
				NSAgent nsa = (NSAgent) nsco[lpc];

				if (!(nsa.getAttachPoint() instanceof NSAgent)
						|| !(nsa.getConnectPoint() instanceof NSAgent)) {
				} else {
					NSAgent attach, connect;
					NSNode left, right;

					attach = (NSAgent) nsa.getAttachPoint();
					connect = (NSAgent) nsa.getConnectPoint();
					if (("Application/Traffic/CBR".equals(nsa.getType())
							&& "Agent/UDP".equals(attach.getType())
							&& "Agent/Null".equals(connect.getType())) ||
							("Application/Traffic/CBR".equals(nsa.getType())
									&& "Agent/TCP".equals(attach.getType())
									&& "Agent/TCPSink".equals(connect.getType()))) {
						SortedSet membership = nst.findEventGroup(nsa);
						LinkTraffic lt;
						NSLink nsl;

						left = (NSNode) attach.getAttachPoint();
						right = (NSNode) connect.getAttachPoint();
						
						if ((nsl = nst.findLink(left, right)) == null)
							continue;
						
						lt = new LinkTraffic();
						if (membership.size() == 1) {
							NSEventGroup nseg;
							NSAgent other;

							nseg = (NSEventGroup) membership.first();
							lt.setName(nseg.getName());
							lt.setDirection(BIDIRECTIONAL);
							nst.remove(nseg);

							nseg.getMembers().remove(nsa);
							other = (NSAgent) nseg.getMembers().first();
							nst.remove(other);
							namer.remove(other.getName());
							namer.remove(nsa.getName());
							nst.remove(other.getAttachPoint());
							namer.remove(other.getAttachPoint().getName());
							nst.remove(other.getConnectPoint());
							namer.remove(other.getConnectPoint().getName());
						} else {
							lt.setName(nsa.getName());
							lt.setDirection(LEFT_TO_RIGHT);
						}
						lt.setLink(nsl);
						lt.setType(attach.getType().substring(6));
						lt.setLeft(left);
						lt.setRight(right);
						lt.getAttributes().putAll(nsa.getAttributes());
						lt.setStartAtSwapin(nsa.getStartAtSwapin());
						nst.remove(nsa);
						nst.remove(attach);
						namer.remove(attach.getName());
						nst.remove(connect);
						namer.remove(connect.getName());
						
						retval.add(lt);
					}
				}
			}
		}
		return retval;
	}
	
	protected NSLink nsl;

	protected NSNode left;

	protected NSNode right;

	protected String type = "UDP";

	protected int direction = BIDIRECTIONAL;

	protected final Hashtable attributes = new Hashtable();

	protected boolean startAtSwapin = true;
	
	public LinkTraffic() {
	}

	public boolean contains(NSObject nso) {
		return this.nsl.equals(nso);
	}
	
	public LinkTraffic setLink(NSLink nsl) {
		this.nsl = nsl;
		return this;
	}

	public LinkTraffic setLeft(NSNode left) {
		this.left = left;
		return this;
	}

	public NSNode getLeft() {
		return this.left;
	}

	public LinkTraffic setRight(NSNode right) {
		this.right = right;
		return this;
	}

	public NSNode getRight() {
		return this.right;
	}

	public LinkTraffic setType(String type) {
		this.type = type;
		return this;
	}

	public String getType() {
		return this.type;
	}

	public LinkTraffic setDirection(int direction) {
		this.direction = direction;
		return this;
	}

	public int getDirection() {
		return this.direction;
	}

	public Hashtable getAttributes() {
		return this.attributes;
	}

	public LinkTraffic setStartAtSwapin(boolean sas) {
		this.startAtSwapin = sas;
		return this;
	}
	
	public boolean getStartAtSwapin() {
		return this.startAtSwapin;
	}

	private NSAgent addAgents(NSTopology nst, NSNode src, NSNode dst) {
		NSAgent transport, sink, application;
		String agentTypes[];
		Namer namer;

		namer = nst.getNamer();
		
		agentTypes = (String[]) TRAFFIC_TYPES.get(this.type);

		transport = new NSAgent();
		transport.setType(agentTypes[0]);
		transport.setAttachPoint(src);
		transport.setName(namer.suggest(this.name + "src"));
		
		sink = new NSAgent();
		sink.setType(agentTypes[2]);
		sink.setAttachPoint(dst);
		sink.setName(namer.suggest(this.name + "sink"));

		application = new NSAgent();
		application.setType(agentTypes[1]);
		application.setAttachPoint(transport);
		application.setConnectPoint(sink);
		if (this.direction == BIDIRECTIONAL)
			application.setName(namer.suggest(this.name + "app"));
		else
			application.setName(this.name);
		application.getAttributes().putAll(this.attributes);
		application.setStartAtSwapin(this.startAtSwapin);

		nst.add(transport);
		nst.add(sink);
		nst.add(application);

		return application;
	}
	
	public void ungroup(NSTopology nst) {
		NSAgent ltor = null, rtol = null;

		if ((this.direction == BIDIRECTIONAL)
				|| (this.direction == LEFT_TO_RIGHT)) {
			ltor = this.addAgents(nst, this.left, this.right);
		}
		if ((this.direction == BIDIRECTIONAL)
				|| (this.direction == RIGHT_TO_LEFT)) {
			rtol = this.addAgents(nst, this.right, this.left);
		}
		if (this.direction == BIDIRECTIONAL) {
			NSEventGroup nseg;

			nseg = new NSEventGroup();
			nseg.setName(this.name);
			nseg.addMember(ltor);
			nseg.addMember(rtol);
			nst.add(nseg);
		}
	}

	public String toString() {
		return "LinkTraffic[key=" + this.nsl + "; name=" + this.name
				+ "; left=" + this.left + "; right=" + this.right + "; type="
				+ this.type + "; direction=" + this.direction + "; attributes"
				+ this.attributes + "]";
	}
}
