package net.emulab.ns;

import java.util.Enumeration;
import java.util.Hashtable;
import java.util.List;

public class NSAgent extends NSBase implements NSAttachableObject,
		NSConnectableObject {
	
	public static final Hashtable DEFAULT_CBR_ATTRIBUTES = new Hashtable();

	public static final double DEFAULT_RATE = 100 * 1000 * 1000;
	public static final double DEFAULT_PACKET_SIZE = 210;
	public static final double DEFAULT_INTERVAL =
		DEFAULT_PACKET_SIZE * 8 / DEFAULT_RATE;
	
	static {
		DEFAULT_CBR_ATTRIBUTES.put("rate_", new Double(DEFAULT_RATE));
		DEFAULT_CBR_ATTRIBUTES.put("packetSize_", new Double(DEFAULT_PACKET_SIZE));
		DEFAULT_CBR_ATTRIBUTES.put("interval_", new Double(DEFAULT_INTERVAL));
	}
	
	private String type;

	private NSObject attachPoint;

	private NSObject connectPoint;

	private final Hashtable attributes;
	
	private boolean startAtSwapin;


	/*** ADDED BY BELA ***/
	private String accessSecZone;
	private String egressSecZone;
	private String securityConduit;


	public NSAgent(Hashtable attributes) {
		this.attributes = attributes;
	}

	public NSAgent() {
		this(new Hashtable());
	}

	public NSAgent setType(String type) {
		this.type = type;
		return this;
	}

	public String getType() {
		return this.type;
	}


	/**** ADDED BY BELA ***/
	public NSAgent setAccessSecZone(String accessSecZone) {
                this.accessSecZone = accessSecZone;
                return this;
        }
        public String getAccessSecZone() {
                return this.accessSecZone;
        }
	public NSAgent setEgressSecZone(String egressSecZone) {
                this.egressSecZone = egressSecZone;
                return this;
        }
        public String getEgressSecZone() {
                return this.egressSecZone;
        }
	public NSAgent setSecurityConduit(String securityConduit) {
                this.securityConduit = securityConduit;
                return this;
        }
        public String getSecurityConduit() {
                return this.securityConduit;
        }





	public NSAttachableObject setAttachPoint(NSObject attachPoint) {
		this.attachPoint = attachPoint;
		return this;
	}

	public NSObject getAttachPoint() {
		return this.attachPoint;
	}

	public NSConnectableObject setConnectPoint(NSObject connectPoint) {
		this.connectPoint = connectPoint;
		return this;
	}

	public NSObject getConnectPoint() {
		return this.connectPoint;
	}

	public NSAgent setStartAtSwapin(boolean sas) {
		this.startAtSwapin = sas;
		return this;
	}
	
	public boolean getStartAtSwapin() {
		return this.startAtSwapin;
	}

	public Hashtable getAttributes() {
		return this.attributes;
	}
	
	public boolean initFromEvent(List warnings, NSEvent nse) {
		boolean retval = false;
		
		switch (nse.getEventType()) {
		case NSEvent.EVENT_TYPE_START:
			if (nse.getTime() == 0.0) {
				this.startAtSwapin = true;
				retval = true;
			}
			break;
		case NSEvent.EVENT_TYPE_MODIFY:
			double packetSize = -1, rate = -1, interval = -1;
			Hashtable ht = nse.getArguments();
			Enumeration en;
			
			en = ht.keys();
			while (en.hasMoreElements()) {
				String key = (String)en.nextElement();
				String value = (String)ht.get(key);
				
				if ("PACKETSIZE".equals(key)) {
					packetSize = Double.parseDouble(value);
				}
				else if ("RATE".equals(key)) {
					rate = Double.parseDouble(value) * 1000;
				}
				else if ("INTERVAL".equals(key)) {
					interval = Double.parseDouble(value);
				}
				else if ("IPTOS".equals(key)) {
					int iptos = Integer.parseInt(value);
					
					if (iptos != -1)
						this.attributes.put("iptos_", new Integer(iptos));
				}
				else {
					warnings.add("Unknown event argument: " + key);
				}
			}
			
			if (interval == -1) {
				interval = packetSize * 8.0 / rate;
			}
			
			this.attributes.put("rate_", new Double(rate));
			this.attributes.put("packetSize_", new Double(packetSize));
			this.attributes.put("interval_", new Double(interval));
			break;
		}
		
		return retval;
	}

	public String toString() {
		return "NSAgent[" + super.toString() + "; type=" + this.type
				+ "; attachPoint=" + this.attachPoint + "; connectPoint="
				+ this.connectPoint + "; attributes=" + this.attributes + "]";
	}
}
