/*
 * NSFile.java
 *
 * Copyright (c) 2005 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

package net.emulab.ns;

import java.io.ByteArrayOutputStream;
import java.io.PrintStream;
import java.text.MessageFormat;
import java.text.NumberFormat;
import java.io.File;
import java.io.FileWriter;
import java.io.BufferedWriter;
import java.io.IOException;
import java.util.Arrays;
import java.util.Collections;
import java.util.Date;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.ResourceBundle;
import java.util.SortedSet;
import java.util.TreeSet;



import thinlet.ThinletEntryException;

/**
 * Class used to convert between an NSTopology and an NS file.
 */
public class NSFile1 {
	private static final NumberFormat NUMBER_FORMAT;
	File file;
	
	//if (file.exists()){
			//file.createNewFile();
	//}
	FileWriter fileWriter;
	BufferedWriter bufferWritter;
	static int writeToFile;
	static String location;
	
	
	static {
		NUMBER_FORMAT = NumberFormat.getNumberInstance();
		NUMBER_FORMAT.setGroupingUsed(false);
		NUMBER_FORMAT.setMaximumFractionDigits(10);
	}

	/**
	 * The ResourceBundle that has the strings used to generate NS syntax.
	 */
	protected final MutableResourceBundle fileFormat;
	
	/**
	 * Construct an empty NSFile.
	 */
	public NSFile1() {
		ResourceBundle base;

		base = ResourceBundle.getBundle(this.getClass().getName());
		this.fileFormat = new MutableResourceBundle(base);
	}

	/**
	 * @return The ResourceBundle that has the strings used to generate NS
	 *         syntax.
	 */
	public MutableResourceBundle getFileFormat() {
		return this.fileFormat;
	}

	protected String getNSResourceAsString(Object object, String property,
			Object formatArguments[]) {
		String retval;

		if (object == null)
			throw new IllegalArgumentException("object is null");
		if (property == null)
			throw new IllegalArgumentException("property is null");

		retval = this.fileFormat.getString(object.getClass().getName() + "."
				+ property);
		if (formatArguments != null) {
			retval = MessageFormat.format(retval, formatArguments);
		}
		return retval;
	}

	protected String getNSResourceAsString(Object object, String property) {
		return this.getNSResourceAsString(object, property, null);
	}

	/**
	 * Print virtual hardware type information.
	 * 
	 * @param ps
	 *            The stream to write NS syntax to.
	 */
	protected void printVType(PrintStream ps, NSVType nsvt) throws IOException{
		String kind, weight, memberString = "";
		Iterator iter;

		if (nsvt.getMembers().size() == 0) {
			throw new IllegalArgumentException("NSVType has no members: "
					+ nsvt);
		}

		iter = nsvt.getMembers().iterator();
		while (iter.hasNext()) {
			memberString += " " + iter.next();
		}

		if (nsvt instanceof NSHardVType) {
			kind = "hard";
			weight = "";
		} else if (nsvt instanceof NSSoftVType) {
			kind = "soft";
		} else {
			kind = "weighted";
		}
		weight = String.valueOf(((NSSoftVType) nsvt).getWeight());

		ps.println(this.getNSResourceAsString(nsvt, kind, new Object[] {
				nsvt.getName(), weight, memberString }));
		if (writeToFile==1)
			bufferWritter.write(this.getNSResourceAsString(nsvt, kind, new Object[] {
				nsvt.getName(), weight, memberString })+"\n");
	}

	/**
	 * Print user-specified NS to the output stream.
	 * 
	 * @param ps
	 *            The stream to write NS syntax to.
	 * @param ns
	 *            The NS syntax to write, in a form acceptable to
	 *            java.text.MessageFormat.
	 * @param parameters
	 *            The parameters to pass when performing formatting.
	 */
	protected void printExtraNS(PrintStream ps, String ns, Object parameters[]) throws IOException{
		if ((ns != null) && !"".equals(ns)) {
			ps.println(MessageFormat.format(ns, parameters));
			if (writeToFile==1)
				bufferWritter.write(MessageFormat.format(ns, parameters)+"\n");
		}
	}

	/**
	 * Print an NSNode object to the output stream.
	 * 
	 * @param ps
	 *            The stream to write NS syntax to.
	 * @param nsn
	 *            The NSNode to write.
	 */
	protected void printNode(PrintStream ps, NSNode nsn) throws IOException{
		Object formatArguments[];
		SortedSet ss;
		String name;
		List list;

		name = nsn.getName().toString();

		formatArguments = new Object[] { name, nsn.getOS(), nsn.getHardware(),
				nsn.getStartup(), nsn.getFixedNode(), nsn.getFailureAction(), nsn.getNodeCost(), 
				nsn.getSecurityZone1Cost(), nsn.getSecurityZone2Cost(), nsn.getSecurityZone3Cost(), };

		/* ... create the object, */
		ps.println(this.getNSResourceAsString(nsn, "node", formatArguments));
		if (writeToFile==1)
			bufferWritter.write(this.getNSResourceAsString(nsn, "node", formatArguments)+"\n");

		/* ... set any non-default attributes, and */
		if (!"".equals(nsn.getOS())){
			ps.println(this.getNSResourceAsString(nsn, "os", formatArguments));
		if (writeToFile==1)
				bufferWritter.write(this.getNSResourceAsString(nsn, "os", formatArguments)+"\n");
		}


		/*** ADDED by Bela ***/
		if (nsn.getOS().contains("CoS") && !"".equals(nsn.getNodeCost())){
                        ps.println(this.getNSResourceAsString(nsn, "nodeCost", formatArguments));
                if (writeToFile==1 && nsn.getOS().contains("CoS"))
                                bufferWritter.write(this.getNSResourceAsString(nsn, "nodeCost", formatArguments)+"\n");
                }
		if (nsn.getOS().contains("CoS") && !"".equals(nsn.getSecurityZone1Cost())){
                        ps.println(this.getNSResourceAsString(nsn, "securityZone1Cost", formatArguments));
                if (writeToFile==1 && nsn.getOS().contains("CoS"))
                                bufferWritter.write(this.getNSResourceAsString(nsn, "securityZone1Cost", formatArguments)+"\n");
                }
		if (nsn.getOS().contains("CoS") && !"".equals(nsn.getSecurityZone2Cost())){
                        ps.println(this.getNSResourceAsString(nsn, "securityZone2Cost", formatArguments));
                if (writeToFile==1 && nsn.getOS().contains("CoS"))
                                bufferWritter.write(this.getNSResourceAsString(nsn, "securityZone2Cost", formatArguments)+"\n");
                }
 		if (nsn.getOS().contains("CoS") && !"".equals(nsn.getSecurityZone3Cost())){
                        ps.println(this.getNSResourceAsString(nsn, "securityZone3Cost", formatArguments));
                if (writeToFile==1 && nsn.getOS().contains("CoS"))
                                bufferWritter.write(this.getNSResourceAsString(nsn, "securityZone3Cost", formatArguments)+"\n");
                }




		if (!"".equals(nsn.getHardware())) {
			ps.println(this.getNSResourceAsString(nsn, "hardware",
					formatArguments));
		if (writeToFile==1)
				bufferWritter.write(this.getNSResourceAsString(nsn, "hardware",formatArguments)+"\n");
		}
		if (!"".equals(nsn.getStartup())) {
			ps.println(this.getNSResourceAsString(nsn, "startup",
					formatArguments));
		if (writeToFile==1)
				bufferWritter.write(this.getNSResourceAsString(nsn, "startup",formatArguments)+"\n");
		}
		if (!"".equals(nsn.getFixedNode())) {
			ps.println(this.getNSResourceAsString(nsn, "fix", formatArguments));
		if (writeToFile==1)
				bufferWritter.write(this.getNSResourceAsString(nsn, "fix", formatArguments)+"\n");
		}
		if (!"".equals(nsn.getFailureAction()) && !"fatal".equals(nsn.getFailureAction())) {
			ps.println(this.getNSResourceAsString(nsn, "fail",
					formatArguments));
		if (writeToFile==1)
				bufferWritter.write(this.getNSResourceAsString(nsn, "fail",	formatArguments)+"\n");
		}
		if (((ss = nsn.getTarFiles()) != null) && (ss.size() > 0)) {
			NSTarFile nstf[];
			int lpc;

			nstf = (NSTarFile[]) ss.toArray(new NSTarFile[ss.size()]);
			ps.print(this.getNSResourceAsString(nsn, "tarfiles",
					formatArguments));
		if (writeToFile==1)
				bufferWritter.write(this.getNSResourceAsString(nsn, "tarfiles",	formatArguments));
			for (lpc = 0; lpc < nstf.length; lpc++) {
				ps.print(" " + nstf[lpc].getDestination() + " "
						+ nstf[lpc].getTarFile());
		if (writeToFile==1)
					bufferWritter.write(" " + nstf[lpc].getDestination() + " "
						+ nstf[lpc].getTarFile());
			}
			ps.println();
		if (writeToFile==1)
				bufferWritter.write("\n");
		}
		if (((ss = nsn.getRPMFiles()) != null) && (ss.size() > 0)) {
			String rpms[];
			int lpc;

			rpms = (String[]) ss.toArray(new String[ss.size()]);
			ps.print(this.getNSResourceAsString(nsn, "rpms", formatArguments));
		if (writeToFile==1)
				bufferWritter.write(this.getNSResourceAsString(nsn, "rpms", formatArguments));
			for (lpc = 0; lpc < rpms.length; lpc++) {
				ps.print(" " + rpms[lpc]);
		if (writeToFile==1)
					bufferWritter.write(" " + rpms[lpc]);
			}
			ps.println();
		if (writeToFile==1)
				bufferWritter.write("\n");
		}
		if (((list = nsn.getDesires()) != null) && (list.size() > 0)) {
			int lpc;

			for (lpc = 0; lpc < list.size(); lpc++) {
				Pair pair = (Pair)list.get(lpc);
				
				ps.println(this.getNSResourceAsString(nsn, "desire", new Object[] {
						name, pair.getKey(), pair.getValue(),
				}));
		if (writeToFile==1)
					bufferWritter.write(this.getNSResourceAsString(nsn, "desire", new Object[] {
						name, pair.getKey(), pair.getValue(),
				})+"\n");
			}
		}
	}

	/**
	 * Print the IP address for a member of a LAN/link.
	 * 
	 * @param ps
	 *            The stream to write NS syntax to.
	 * @param command
	 *            The Emulab-specific command to set the IP.
	 * @param nsi
	 *            The LAN/link member whose address should be written.
	 * @param name
	 *            The name of the LAN/link the member is connected to.
	 */
	protected void printLanLinkIP(PrintStream ps, NSObject nso,
			NSInterface nsi, String name) throws IOException{
		if (!"".equals(nsi.getIPv4Address())) {
			ps.println(this.getNSResourceAsString(nso, "ip", new Object[] {
					nsi.getNSObject().getName().toString(), name,
					nsi.getIPv4Address() }));
		if (writeToFile==1)
				bufferWritter.write(this.getNSResourceAsString(nso, "ip", new Object[] {
					nsi.getNSObject().getName().toString(), name,
					nsi.getIPv4Address() })+"\n");
		}
	}

	protected void printLanLinkTracing(PrintStream ps, NSLanLink ll) throws IOException{
		if (!"".equals(ll.getTracingType())) {
			String filter = ll.getTracingFilter();
			
			if (!"".equals(filter))
				filter = "\"" + filter + "\"";
			ps.println(this.getNSResourceAsString(ll, "trace", new Object[] {
					ll.getName().toString(),
					ll.getTracingType(),
					filter }));
					if (writeToFile==1)
				bufferWritter.write(this.getNSResourceAsString(ll, "trace", new Object[] {
					ll.getName().toString(),
					ll.getTracingType(),
					filter })+"\n");
			if (ll.getSnapLen() > 0) {
				ps.println(this.getNSResourceAsString(ll, "snaplen", new Object[] {
						ll.getName().toString(),
						new Integer(ll.getSnapLen()) }));
				if (writeToFile==1)
				bufferWritter.write(this.getNSResourceAsString(ll, "snaplen", new Object[] {
						ll.getName().toString(),
						new Integer(ll.getSnapLen()) })+"\n");
			}
		}
	}
	
	/**
	 * Print an NSLink object to the output stream.
	 * 
	 * @param ps
	 *            The stream to write NS syntax to.
	 * @param nsl
	 *            The NSLink to write.
	 */
	protected void printLink(PrintStream ps, NSLink nsl) throws IOException{
		UniqueName names[] = new UniqueName[2];
		String name;

		name = nsl.getName().toString();

		names[0] = nsl.getLeftMember().getNSObject().getName();
		names[1] = nsl.getRightMember().getNSObject().getName();
		
		Arrays.sort(names);
		
		/* ... create the link, */
		ps.println(this.getNSResourceAsString(nsl, "link", new Object[] { name,
				names[0].toString(),
				names[1].toString(),
				String.valueOf(nsl.getLeftMember().getActualBandwidth() / 1000.0),
				String.valueOf(nsl.getLeftMember().getActualLatency()),
				nsl.getQueueing() }));
		if (writeToFile==1)
			bufferWritter.write(this.getNSResourceAsString(nsl, "link", new Object[] { name,
				names[0].toString(),
				names[1].toString(),
				String.valueOf(nsl.getLeftMember().getActualBandwidth() / 1000.0),
				String.valueOf(nsl.getLeftMember().getActualLatency()),
				nsl.getQueueing() })+"\n");

		/** ADDED BY Bela  ***/
                ps.println(this.getNSResourceAsString(nsl, "bandwidthCost", new Object[] { name,
                                String.valueOf(nsl.getLeftMember().getBandwidthCost()) }));
                        bufferWritter.write(this.getNSResourceAsString(nsl, "bandwidthCost", new Object[] { name,
                                String.valueOf(nsl.getLeftMember().getBandwidthCost()) }) + "\n");

                ps.println(this.getNSResourceAsString(nsl, "conduit1Cost", new Object[] { name,
                                String.valueOf(nsl.getLeftMember().getConduit1Cost()) }));
                        bufferWritter.write(this.getNSResourceAsString(nsl, "conduit1Cost", new Object[] { name,
                                String.valueOf(nsl.getLeftMember().getConduit1Cost()) }) + "\n");

                ps.println(this.getNSResourceAsString(nsl, "conduit2Cost", new Object[] { name,
                                String.valueOf(nsl.getLeftMember().getConduit2Cost()) }));
                        bufferWritter.write(this.getNSResourceAsString(nsl, "conduit2Cost", new Object[] { name,
                                String.valueOf(nsl.getLeftMember().getConduit2Cost()) }) + "\n");

                ps.println(this.getNSResourceAsString(nsl, "conduit3Cost", new Object[] { name,
                                String.valueOf(nsl.getLeftMember().getConduit3Cost()) }));
                        bufferWritter.write(this.getNSResourceAsString(nsl, "conduit3Cost", new Object[] { name,
                                String.valueOf(nsl.getLeftMember().getConduit3Cost()) }) + "\n");



		/* ... set any non-default attributes, and */
		if (!"".equals(nsl.getNetMask())) {
			ps.println(this.getNSResourceAsString(nsl, "netmask", new Object[] {
					name, nsl.getNetMask(), }));
		if (writeToFile==1)
				bufferWritter.write(this.getNSResourceAsString(nsl, "netmask", new Object[] {
					name, nsl.getNetMask(), })+"\n");
		}
		if (nsl.getMultiplexed()) {
			ps.println(this
					.getNSResourceAsString(nsl, "multiplexed", new Object[] {
							name, Boolean.valueOf(nsl.getMultiplexed()), }));
		if (writeToFile==1)
				bufferWritter.write(this
					.getNSResourceAsString(nsl, "multiplexed", new Object[] {
							name, Boolean.valueOf(nsl.getMultiplexed()), })+"\n");
		}
		if (nsl.getLeftMember().getActualLossRate() > 0.0) {
			ps.println(this.getNSResourceAsString(nsl, "loss",
					new Object[] {
					name,
					String.valueOf(nsl.getLeftMember().getActualLossRate()), }));
			if (writeToFile==1)
				bufferWritter.write(this.getNSResourceAsString(nsl, "loss",
					new Object[] {
					name,
					String.valueOf(nsl.getLeftMember().getActualLossRate()), })+"\n");
		}
		this.printLanLinkIP(ps, nsl, nsl.getLeftMember(), name);
		this.printLanLinkIP(ps, nsl, nsl.getRightMember(), name);
		/* XXX no support for simplex. */
		this.printLanLinkTracing(ps, nsl);
	}

	/**
	 * Convert an array of LAN members to a NS list.
	 * 
	 * @param members
	 *            The array of members to convert.
	 * @return NS syntax that contains the names of all of the members.
	 */
	protected String lanMembersAsString(NSInterface members[]) {
		UniqueName names[] = new UniqueName[members.length];
		String retval = "";
		int lpc;
		
		for (lpc = 0; lpc < members.length; lpc++) {
			names[lpc] = members[lpc].getNSObject().getName();
		}
		
		Arrays.sort(names);

		for (lpc = 0; lpc < members.length; lpc++) {
			if (lpc > 0)
				retval += " ";
			retval += "$" + members[lpc].getNSObject().getName();
		}
		return retval;
	}

	/**
	 * Print an NSLan object to the output stream.
	 * 
	 * @param ps
	 *            The stream to write NS syntax to.
	 * @param nsl
	 *            The NSLan to write.
	 */
	protected void printLan(PrintStream ps, NSLan nsl) throws IOException{
		NSInterface nsi = nsl.getTemplateInterface();
		NSInterface members[];
		SortedSet ss;
		String name;
		int lpc;

		name = nsl.getName().toString();

		/* ... check for members, and */
		ss = new TreeSet(nsl.getMembers());
		members = (NSInterface[]) ss.toArray(new NSInterface[ss.size()]);
		if (members.length < 2) {
			throw new ThinletEntryException(nsl.getName().toString() + " is invalid",
					"LANs must have atleast two member nodes.");
		}

		/* There are members, create the object, */
		ps.println(this.getNSResourceAsString(nsl, "lan", new Object[] { name,
				this.lanMembersAsString(members),
				String.valueOf(nsi.getActualBandwidth() / 1000.0),
				String.valueOf(nsi.getActualLatency()) }));
		if (writeToFile==1)
			bufferWritter.write(this.getNSResourceAsString(nsl, "lan", new Object[] { name,
				this.lanMembersAsString(members),
				String.valueOf(nsi.getActualBandwidth() / 1000.0),
				String.valueOf(nsi.getActualLatency()) })+"\n");
		/* ... set the template parameters, and */
		if (!"".equals(nsl.getNetMask())) {
			ps.println(this.getNSResourceAsString(nsl, "netmask", new Object[] {
					name, nsl.getNetMask(), }));
		if (writeToFile==1)
				bufferWritter.write(this.getNSResourceAsString(nsl, "netmask", new Object[] {
					name, nsl.getNetMask(), })+"\n");
		}
		if (!"".equals(nsl.getProtocol()) && !"ethernet".equals(nsl.getProtocol())) {
			ps.println(this.getNSResourceAsString(nsl, "protocol",
					new Object[] { name, nsl.getProtocol(), }));
		if (writeToFile==1)
				bufferWritter.write(this.getNSResourceAsString(nsl, "protocol",
					new Object[] { name, nsl.getProtocol(), })+"\n");
		}
		if (!"".equals(nsl.getAccessPoint())) {
			ps.println(this.getNSResourceAsString(nsl, "accesspoint",
					new Object[] { name,	nsl.getAccessPoint(), }));
		if (writeToFile==1)
				bufferWritter.write(this.getNSResourceAsString(nsl, "accesspoint",
					new Object[] { name,	nsl.getAccessPoint(), })+"\n");
		}
		if (nsl.getSettings().size() > 0) {
			List settings;
			Iterator iter;
			
			settings = nsl.getSettings();
			iter = settings.iterator();
			while (iter.hasNext()) {
				Pair pair = (Pair) iter.next();

				ps.println(this.getNSResourceAsString(nsl, "setting",
						new Object[] { name, pair.getKey(), pair.getValue() }));
		if (writeToFile==1)
					bufferWritter.write(this.getNSResourceAsString(nsl, "setting",
						new Object[] { name, pair.getKey(), pair.getValue() })+"\n");
			}
		}
		if (nsi.getActualLossRate() > 0.0) {
			ps.println(this.getNSResourceAsString(nsl, "loss", new Object[] {
					name, String.valueOf(nsi.getActualLossRate()) }));
		if (writeToFile==1)
				bufferWritter.write(this.getNSResourceAsString(nsl, "loss", new Object[] {
					name, String.valueOf(nsi.getActualLossRate()) })+"\n");
		}
		/* ... the node-specific parameters. */
		for (lpc = 0; lpc < members.length; lpc++) {
			NSNode nsnode;

			nsnode = (NSNode) members[lpc].getNSObject();
			this.printLanLinkIP(ps, nsl, members[lpc], name);
			if (members[lpc].getLatency() >= 0) {
				ps.println(this.getNSResourceAsString(nsl, "ndelay",
						new Object[] { nsnode.getName().toString(), name,
								String.valueOf(members[lpc].getLatency()) }));
		if (writeToFile==1)
					bufferWritter.write(this.getNSResourceAsString(nsl, "ndelay",
						new Object[] { nsnode.getName().toString(), name,
								String.valueOf(members[lpc].getLatency()) })+"\n");
			}
			if (members[lpc].getBandwidth() >= 0) {
				ps.println(this.getNSResourceAsString(nsl, "nbandwidth",
						new Object[] { nsnode.getName().toString(), name,
								String.valueOf(members[lpc].getBandwidth() / 1000.0) }));
		if (writeToFile==1)
					bufferWritter.write(this.getNSResourceAsString(nsl, "nbandwidth",
						new Object[] { nsnode.getName().toString(), name,
								String.valueOf(members[lpc].getBandwidth() / 1000.0) })+"\n");
			}
			if (members[lpc].getLossRate() >= 0) {
				ps.println(this.getNSResourceAsString(nsl, "nloss",
						new Object[] { nsnode.getName().toString(), name,
								String.valueOf(members[lpc].getLossRate()), }));
		if (writeToFile==1)
					bufferWritter.write(this.getNSResourceAsString(nsl, "nloss",
						new Object[] { nsnode.getName().toString(), name,
								String.valueOf(members[lpc].getLossRate()), })+"\n");
			}
		}
		this.printLanLinkTracing(ps, nsl);
	}

	protected void printProgram(PrintStream ps, NSProgram nsp) throws IOException{
		String name;

		name = nsp.getName().toString();

		ps.println(this.getNSResourceAsString(nsp, "program", new Object[] {
				name, nsp.getNode().getName().toString(),

				new Integer(nsp.getDirectory().length() > 0 ? 1 : 0),
				nsp.getDirectory(),

				new Integer(nsp.getCommand().length() > 0 ? 1 : 0),
				"{" + nsp.getCommand() + "}",

				new Integer(nsp.getTimeout() > 0 ? 1 : 0),
				new Integer(nsp.getTimeout()),
				
				new Integer(nsp.getExpectedExitCode() != 0 ? 1 : 0),
				new Integer(nsp.getExpectedExitCode()) }));
		if (writeToFile==1)
			bufferWritter.write(this.getNSResourceAsString(nsp, "program", new Object[] {
				name, nsp.getNode().getName().toString(),

				new Integer(nsp.getDirectory().length() > 0 ? 1 : 0),
				nsp.getDirectory(),

				new Integer(nsp.getCommand().length() > 0 ? 1 : 0),
				"{" + nsp.getCommand() + "}",

				new Integer(nsp.getTimeout() > 0 ? 1 : 0),
				new Integer(nsp.getTimeout()),
				
				new Integer(nsp.getExpectedExitCode() != 0 ? 1 : 0),
				new Integer(nsp.getExpectedExitCode()) })+"\n");
	}

	protected void printAgent(PrintStream ps, NSAgent nsa) throws IOException{
		Hashtable attributes;
		Iterator iter;
		String name;

		name = nsa.getName().toString();

		ps.println(this.getNSResourceAsString(nsa, "agent", new Object[] {
				name, nsa.getType(), }));
		if (writeToFile==1)
			bufferWritter.write(this.getNSResourceAsString(nsa, "agent", new Object[] {
				name, nsa.getType(), })+"\n");

		attributes = nsa.getAttributes();
		iter = new TreeSet(attributes.keySet()).iterator();
		while (iter.hasNext()) {
			String key = (String) iter.next();
			Object value;

			value = attributes.get(key);
			if (value instanceof NSObject) {
				NSObject nso = (NSObject) value;

				value = "$" + nso.getName().toString();
			}
			else if (value instanceof Number) {
				value = NUMBER_FORMAT.format(value);
			}
			ps.println(this.getNSResourceAsString(nsa, "attribute",
					new Object[] { name, key, value.toString() }));
			if (writeToFile==1)
				bufferWritter.write(this.getNSResourceAsString(nsa, "attribute",
					new Object[] { name, key, value.toString() })+"\n");
		}

		/** ADDED BY BELA **/
                /** Given that the extra design attributes are not specific to CBR, I did not add them to the hash vector **/
                /** Therefore, these need to be written outside the while loop **/
                /** We only print the NS in the case that there is something configured **/

                if ( nsa.getAccessSecZone() != null ) {
                        ps.println(this.getNSResourceAsString(nsa, "agentAccessSecZone", new Object[] {
                                name, nsa.getAccessSecZone(), }));
			if (writeToFile==1)
                        	bufferWritter.write(this.getNSResourceAsString(nsa, "agentAccessSecZone", new Object[] {
                                	name, nsa.getAccessSecZone(), })+"\n");
                }

                if ( nsa.getEgressSecZone() != null ) {
                        ps.println(this.getNSResourceAsString(nsa, "agentEgressSecZone", new Object[] {
                                name, nsa.getEgressSecZone(), }));
			if (writeToFile==1)
                        	bufferWritter.write(this.getNSResourceAsString(nsa, "agentEgressSecZone", new Object[] {
                                	name, nsa.getEgressSecZone(), })+"\n");
                }

                if ( nsa.getSecurityConduit() != null ) {
                        ps.println(this.getNSResourceAsString(nsa, "agentSecConduit", new Object[] {
                                name, nsa.getSecurityConduit(), }));
			if (writeToFile==1)
                        	bufferWritter.write(this.getNSResourceAsString(nsa, "agentSecConduit", new Object[] {
                                	name, nsa.getSecurityConduit(), })+"\n");
                }
	}

	protected void printEventGroup(PrintStream ps, NSEventGroup nseg) throws IOException{
		String name, memberList = "";
		Iterator iterator;
		SortedSet ss;

		name = nseg.getName().toString();

		ps.println(this.getNSResourceAsString(nseg, "eg",
				new Object[] { name, }));
		if (writeToFile==1)
			bufferWritter.write(this.getNSResourceAsString(nseg, "eg",
				new Object[] { name, })+"\n");

		ss = nseg.getMembers();
		iterator = ss.iterator();
		while (iterator.hasNext()) {
			NSObject nso = (NSObject) iterator.next();

			memberList += " $" + nso.getName();
		}

		ps.println(this.getNSResourceAsString(nseg, "add", new Object[] { name,
				memberList, }));
				if (writeToFile==1)
			bufferWritter.write(this.getNSResourceAsString(nseg, "add", new Object[] { name,
				memberList, })+"\n");
	}

	protected void printNSObject(PrintStream ps, NSTopology nst, NSObject nso) throws IOException{
		if (nso.getName() == null)
			throw new IllegalArgumentException("nso has no name: " + nso);
		
		/* Write any prepended user-specified NS, */
		this.printExtraNS(ps, nso.getNSPrepend(), new Object[] { "$"
				+ nso.getName().toString() });

		if (nso instanceof NSVType)
			this.printVType(ps, (NSVType) nso);
		else if (nso instanceof NSNode)
			this.printNode(ps, (NSNode) nso);
		else if (nso instanceof NSProgram)
			this.printProgram(ps, (NSProgram) nso);
		else if (nso instanceof NSLink)
			this.printLink(ps, (NSLink) nso);
		else if (nso instanceof NSLan)
			this.printLan(ps, (NSLan) nso);
		else if (nso instanceof NSAgent)
			this.printAgent(ps, (NSAgent) nso);
		else if (nso instanceof NSEventGroup)
			this.printEventGroup(ps, (NSEventGroup) nso);
		else
			throw new InternalError();

		/* ... write any appended user-specified NS. */
		this.printExtraNS(ps, nso.getNSAppend(), new Object[] { "$"
				+ nso.getName().toString() });
	}

	protected void printNSObjects(PrintStream ps, NSTopology nst, List list, String desc) throws IOException{
		NSObject nso[];
		
		Collections.sort(list);

		nso = (NSObject[]) list.toArray(new NSObject[list.size()]);
		if (nso.length > 0) {
			int lpc;

			ps.println("# " + desc);
		if (writeToFile==1)
				bufferWritter.write("# "+desc+"\n");
			for (lpc = 0; lpc < nso.length; lpc++) {
				this.printNSObject(ps, nst, nso[lpc]);
			}
			ps.println();
		if (writeToFile==1)
				bufferWritter.write("\n");
		}
	}

	protected void printNSAttachableObjects(PrintStream ps, NSTopology nst,
			List list) throws IOException{
		NSAttachableObject nsao[];

		Collections.sort(list);
		
		nsao = (NSAttachableObject[]) list.toArray(new NSAttachableObject[list
				.size()]);
		if (nsao.length > 0) {
			int lpc;

			for (lpc = 0; lpc < nsao.length; lpc++) {
				if (nsao[lpc].getAttachPoint() != null) {
					String attachType, ns;

					if (nsao[lpc].getAttachPoint() instanceof NSNode)
						attachType = "attach-node";
					else
						attachType = "attach";
					ns = MessageFormat.format(this.fileFormat
							.getString(NSAttachableObject.class.getName() + "."
									+ attachType), new Object[] {
							nsao[lpc].getName().toString(),
							nsao[lpc].getAttachPoint().getName().toString(), });
					ps.println(ns);
					if (writeToFile==1)
						bufferWritter.write(ns+"\n");

				}
			}
			ps.println();
				if (writeToFile==1)
				bufferWritter.write("\n");
		}
	}

	protected void printNSConnectableObjects(PrintStream ps, NSTopology nst,
			List list) throws IOException{
		NSConnectableObject nsco[];

		nsco = (NSConnectableObject[]) list.toArray(new NSConnectableObject[list
				.size()]);
		if (nsco.length > 0) {
			int lpc;

			for (lpc = 0; lpc < nsco.length; lpc++) {
				if (nsco[lpc].getConnectPoint() != null) {
					String aname, ns;

					if (nsco[lpc] instanceof NSAttachableObject) {
						aname = ((NSAttachableObject)nsco[lpc]).getAttachPoint().getName().toString();
					}
					else {
						aname = nsco[lpc].getName().toString();
					}
					ns = MessageFormat.format(this.fileFormat
							.getString(NSConnectableObject.class.getName()
									+ ".connect"), new Object[] {
						aname,
						nsco[lpc].getConnectPoint().getName().toString(), });
					ps.println(ns);
				if (writeToFile==1)
						bufferWritter.write(ns+"\n");
					
					if (nsco[lpc] instanceof NSAgent) {
						NSAgent nsa = (NSAgent)nsco[lpc];
						
						if (nsa.getStartAtSwapin()) {
							ps.println(this.getNSResourceAsString(nsa, "start", new Object[] {
									nsa.getName().toString() }));
					if (writeToFile==1)
								bufferWritter.write(this.getNSResourceAsString(nsa, "start", new Object[] {
									nsa.getName().toString() })+"\n");
						}
					}
				}
			}
			ps.println();
				if (writeToFile==1)
				bufferWritter.write("\n");
		}
	}
	
	protected void printNSFirewall(PrintStream ps, NSTopology nst) throws IOException{
		if (!"".equals(nst.getFirewallType())) {
			List rules = nst.getFirewallRules();
			int lpc, size;
			
			ps.println(MessageFormat.format(this.getNSResourceAsString(this, "firewall"),
					new Object[] { nst.getFirewallName(), nst.getFirewallType(), nst.getFirewallStyle() }));
		if (writeToFile==1)
				bufferWritter.write(MessageFormat.format(this.getNSResourceAsString(this, "firewall"),
					new Object[] { nst.getFirewallName(), nst.getFirewallType(), nst.getFirewallStyle() })+"\n");
			size = rules.size();
			for (lpc = 0; lpc < size; lpc++) {
				FirewallRule fr = (FirewallRule)rules.get(lpc);
				
				if (!"".equals(fr.getRule())) {
					if (fr.getNumber() != -1) {
						ps.println(MessageFormat.format(this.getNSResourceAsString(fr, "numberedrule"),
								new Object[] { nst.getFirewallName(), new Integer(fr.getNumber()), fr.getRule() }));
		if (writeToFile==1)
							bufferWritter.write(MessageFormat.format(this.getNSResourceAsString(fr, "numberedrule"),
								new Object[] { nst.getFirewallName(), new Integer(fr.getNumber()), fr.getRule() })+"\n");
					}
					else {
						ps.println(MessageFormat.format(this.getNSResourceAsString(fr, "rule"),
								new Object[] { nst.getFirewallName(), fr.getRule() }));
					if (writeToFile==1)
							bufferWritter.write(MessageFormat.format(this.getNSResourceAsString(fr, "rule"),
								new Object[] { nst.getFirewallName(), fr.getRule() })+"\n");
					}
				}
			}
			if (size > 0) {
				ps.println();
				if (writeToFile==1)
					bufferWritter.write("\n");
			}
		}
	}
	
	protected void printCapturedParameters(PrintStream ps, NSTopology nst) throws IOException{
		SortedSet ss = new TreeSet(nst.getCapturedParameters());
		Iterator iter = ss.iterator();

		while (iter.hasNext()) {
			Pair cp = (Pair)iter.next();
			
			if ("".equals(cp.getKey())) {
				// Skip blanks
			}
			else {
				String value = cp.getValue();
				
				if ("".equals(value))
					value = "\"\"";
				ps.println(MessageFormat.format(this.getNSResourceAsString(this, "param"),
						new Object[] { cp.getKey(), value }));
					if (writeToFile==1)
					bufferWritter.write(MessageFormat.format(this.getNSResourceAsString(this, "param"),
						new Object[] { cp.getKey(), value })+"\n");
			}
		}
	}

	/**
	 * Print an NSTopology to a stream.
	 * 
	 * @param ps
	 *            The stream to write the NS file to.
	 * @param nst
	 *            The NSTopology to write to the stream.
	 */
	protected void printTopology(PrintStream ps, NSTopology nst) throws IOException{
		boolean lf = false;
		

		
		fileWriter = new FileWriter (file,false);
		if (writeToFile==1)
			bufferWritter = new BufferedWriter (fileWriter);
		if (writeToFile==1)
			bufferWritter.write("# Generated by NetlabClient\n");
		if (writeToFile==1)
			bufferWritter.write("\nset ns [new Simulator]\n");
		if (writeToFile==1)
			bufferWritter.write("source tb_compat.tcl\n\n");
		
		
		
		
		if (nst.getUseEndNodeShaping()) {
			ps.println(MessageFormat.format(this.getNSResourceAsString(this, "useendnodeshaping"),
					new Object[] { "1" }));
		if (writeToFile==1)
				bufferWritter.write(MessageFormat.format(this.getNSResourceAsString(this, "useendnodeshaping"),
					new Object[] { "1" })+"\n");
			lf = true;
		}
		if (nst.getForceEndNodeShaping()) {
			ps.println(MessageFormat.format(this.getNSResourceAsString(this, "forceendnodeshaping"),
					new Object[] { "1" }));
		if (writeToFile==1)
				bufferWritter.write(MessageFormat.format(this.getNSResourceAsString(this, "forceendnodeshaping"),
					new Object[] { "1" })+"\n");
			lf = true;
		}
		if (!"".equals(nst.getDelayOS())) {
			ps.println(MessageFormat.format(this.getNSResourceAsString(this, "delayos"),
					new Object[] { nst.getDelayOS() }));
		if (writeToFile==1)
				bufferWritter.write(MessageFormat.format(this.getNSResourceAsString(this, "delayos"),
					new Object[] { nst.getDelayOS() })+"\n");
			lf = true;
		}
		if (!"".equals(nst.getJailOS())) {
			ps.println(MessageFormat.format(this.getNSResourceAsString(this, "jailos"),
					new Object[] { nst.getJailOS() }));
		if (writeToFile==1)
				bufferWritter.write(MessageFormat.format(this.getNSResourceAsString(this, "jailos"),
					new Object[] { nst.getJailOS() })+"\n");
			lf = true;
		}
		if (lf){
			ps.println();
		if (writeToFile==1)
				bufferWritter.write("\n");
		}
		this.printNSObjects(ps, nst, nst.getVTypes(), "Virtual Types");
		this.printNSObjects(ps, nst, nst.getNodes(), "Nodes");
		this.printNSObjects(ps, nst, nst.getPrograms(), "Program Agents");
		this.printNSObjects(ps, nst, nst.getLinks(), "Links");
		this.printNSObjects(ps, nst, nst.getLans(), "Lans");
		this.printNSObjects(ps, nst, nst.getAgents(), "Event Agents");
		this.printNSObjects(ps, nst, nst.getEventGroups(), "Event Groups");
		this.printNSAttachableObjects(ps, nst, nst.getAttachments());
		this.printNSConnectableObjects(ps, nst, nst.getConnectors());
		this.printNSFirewall(ps, nst);
		if (!"".equals(nst.getSyncServer())) {
			ps.println(MessageFormat.format(this.getNSResourceAsString(this, "syncserver"),
					new Object[] { nst.getSyncServer() }));
				if (writeToFile==1)
				bufferWritter.write(MessageFormat.format(this.getNSResourceAsString(this, "syncserver"),
					new Object[] { nst.getSyncServer() })+"\n");
		}
		if (!"".equals(nst.getRoutingProtocol())) {
			ps.println(this.getNSResourceAsString(nst, "rtproto",
					new Object[] { nst.getRoutingProtocol() }));
				if (writeToFile==1)
				bufferWritter.write(this.getNSResourceAsString(nst, "rtproto",
					new Object[] { nst.getRoutingProtocol() })+"\n");
		}
		//bufferWritter.close();
	}

	/**
	 * Write the contents of this object, in NS syntax, to the given stream.
	 * 
	 * @param ps
	 *            The stream to write the NS file to.
	 * @param nst
	 *            The NSTopology to write to the stream.
	 * @return this
	 */
	public NSFile1 toStream(PrintStream ps, NSTopology nst, int wToFile, String loc) throws IOException{

		writeToFile = wToFile;
		location = loc;

		if (writeToFile==1)
			file = new File(location);
		if (!file.exists()){
			file.createNewFile();
		}
		if (ps == null)
			throw new IllegalArgumentException("ps is null");
		if (nst == null)
			throw new IllegalArgumentException("nst is null");


		try {
			nst.flatten();
			this.printCapturedParameters(ps, nst);
			ps.print(this.getNSResourceAsString(this, "header"));
			//bufferWritter.write(this.getNSResourceAsString(this, "header")+"\n");
			this.printTopology(ps, nst);
			ps.print(this.getNSResourceAsString(this, "footer",
					new Object[] { new Date(), }));
						if (writeToFile==1)
				bufferWritter.write(this.getNSResourceAsString(this, "footer",new Object[] { new Date(), })+"\n");
		} finally {
			ps.flush();
			nst.unflatten();
		}
					if (writeToFile==1)
			bufferWritter.close();
		return this;
	}

	/**
	 * Convenience function that calls #toStream and returns the result as a
	 * string.
	 * 
	 * @param nst
	 *            The NSTopology to convert to an NS formatted string.
	 * @return The contents of this object, in NS.
	 */
	public String toNSString(NSTopology nst) throws IOException{
		ByteArrayOutputStream baos;
		String retval;

		if (nst == null)
			throw new IllegalArgumentException("nst is null");

		baos = new ByteArrayOutputStream(1024);
		this.toStream(new PrintStream(baos), nst,writeToFile,location);
		retval = new String(baos.toByteArray());
		return retval;
	}

	/**
	 * @see java.lang.Object#toString
	 */
	public String toString() {
		return "NSFile[fileFormat=" + this.fileFormat + "]";
	}

	public static void main(String args[]) throws Throwable {
		NSTopology nst = new NSTopology();
		NSVType nsvt = new NSSoftVType();
		NSProgram nsp = new NSProgram();
		NSFile1 nsf = new NSFile1();
		NSLink nsl = new NSLink();
		NSAgent nsaUDP, nsaNull, nsa;
		NSNode nsn1, nsn2, nsn;

		nsl.setName(new UniqueName("link0"));

		nsn = new NSNode();
		nsn.setName(new UniqueName("foobar"));
		nsn.setHardware("end");
		nst.add(nsn);

		nsp = new NSProgram();
		nsp.setName(new UniqueName("worker"));
		nsp.setNode(nsn);
		nst.add(nsp);

		nsp = new NSProgram();
		nsp.setName(new UniqueName("worker2"));
		nsp.setNode(nsn);
		nsp.setDirectory("/tmp");
		nst.add(nsp);

		nsp = new NSProgram();
		nsp.setName(new UniqueName("worker3"));
		nsp.setNode(nsn);
		nsp.setDirectory("/tmp");
		nsp.setCommand("echo 'Hello, World'");
		nst.add(nsp);

		nsp = new NSProgram();
		nsp.setName(new UniqueName("worker4"));
		nsp.setNode(nsn);
		nsp.setDirectory("/tmp");
		nsp.setCommand("echo 'Hello, World'");
		nsp.setTimeout(100);
		nst.add(nsp);

		nsl.addMember(new NSInterface().setNSObject(nsn).setIPv4Address(
				"10.1.1.2").setLossRate(0.1));

		nsn1 = nsn;

		nsn = new NSNode();
		nsn.setName(new UniqueName("barfoo"));
		nsn.setHardware("end");
		nst.add(nsn);
		nsl.addMember(new NSInterface().setNSObject(nsn).setIPv4Address(
				"10.1.1.3"));

		nsn2 = nsn;

		nst.add(nsl);

		nsvt.setName(new UniqueName("end"));
		nsvt.addMember("pc850");
		nsvt.addMember("pc600");
		nst.add(nsvt);

		if (false) {
			nsa = new NSAgent();
			nsa.setType("Agent/Null");
			nsa.setAttachPoint(nsn);
			nst.add(nsa);

			nsaNull = nsa;

			nsa = new NSAgent();
			nsa.setType("Agent/UDP");
			nsa.setAttachPoint(nsn);
			nst.add(nsa);

			nsaUDP = nsa;

			nsa = new NSAgent();
			nsa.setName(new UniqueName("cbr0"));
			nsa.setType("Application/Traffic/CBR");
			nsa.getAttributes().put("packetSize_", new Integer(500));
			nsa.getAttributes().put("interval_", new Double(0.005));
			nsa.setAttachPoint(nsaUDP);
			nsa.setConnectPoint(nsaNull);
			nst.add(nsa);
		} else {
			LinkTraffic lt;

			lt = new LinkTraffic();
			lt.setName(new UniqueName("cbr0"));
			lt.setLeft(nsn1);
			lt.setRight(nsn2);
			lt.setType("UDP");
			lt.getAttributes().put("packetSize_", new Integer(500));
			lt.getAttributes().put("interval_", new Double(0.005));
			nst.addAggregate(lt);
		}

		nsf.toStream(System.out, nst,writeToFile,location);

		System.out.println("====");

		nsf.toStream(System.out, nst,writeToFile,location);
	}
}
