/*
 * NSLan.java
 *
 * Copyright (c) 2005 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

package net.emulab.ns;

import java.util.ArrayList;
import java.util.List;

/**
 * Class that represents a "lan" in NS, with Emulab extensions.
 */
public class NSLan extends NSLanLink {
	/**
	 * Unique identifier for serialization.
	 */
	static final long serialVersionUID = -1499132398753865339L;

	/**
	 * The template interface that should be applied to all interfaces that are
	 * a member of this LAN.
	 */
	protected final NSInterface templateInterface = new NSInterface();

	/**
	 * The wire protocol for this LAN.
	 */
	protected String protocol = "";

	/**
	 * The access point for a wireless network.
	 */
	protected String accessPoint = "";

	/**
	 * Additional key/value settings for this LAN.
	 */
	protected List settings = new ArrayList();

	/**
	 * Construct an NSLan with the default values.
	 */
	public NSLan() {
	}

	/**
	 * Extend to set the template interface of 'nsi' to this LAN's template.
	 * 
	 * @see net.emulab.ns.NSLanLink#addMember(net.emulab.ns.NSInterface)
	 */
	public NSLanLink addMember(NSInterface nsi) {
		super.addMember(nsi);
		if (nsi != null)
			nsi.setTemplate(this.templateInterface);
		return this;
	}

	/**
	 * @return The template interface that should be applied to all interfaces
	 *         that are a member of this LAN.
	 */
	public NSInterface getTemplateInterface() {
		return this.templateInterface;
	}

	/**
	 * @param protocol
	 *            The wire protocol for this LAN.
	 * @return this
	 */
	public NSLan setProtocol(String protocol) {
		this.protocol = protocol;
		return this;
	}

	/**
	 * @return The wire protocol for this LAN.
	 */
	public String getProtocol() {
		return this.protocol;
	}

	/**
	 * @param accessPoint
	 *            The access point for a wireless network.
	 * @return this
	 */
	public NSLan setAccessPoint(String accessPoint) {
		this.accessPoint = accessPoint;
		return this;
	}

	/**
	 * @return The access point for a wireless network.
	 */
	public String getAccessPoint() {
		return this.accessPoint;
	}

	/**
	 * @return Additional key/value settings for this LAN.
	 */
	public List getSettings() {
		return this.settings;
	}
	
	public void setSettings(List list) {
		this.settings = list;
	}

	/**
	 * @see java.lang.Object#toString
	 */
	public String toString() {
		return "NSLan[" + super.toString() + "; templateInterface="
				+ this.templateInterface + "; protocol=" + this.protocol
				+ "; accessPoint=" + this.accessPoint + "; settings="
				+ this.settings + "]";
	}
}
