/*
 * NSObject.java
 *
 * Copyright (c) 2005 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

package net.emulab.ns;

/**
 * Interface that represents "objects" in NS.
 */
public interface NSObject extends Cloneable, Comparable {
	/**
	 * @param un
	 *            The unique name for this object.
	 * @return this
	 */
	NSObject setName(UniqueName un);

	/**
	 * @return The unique name for this object.
	 */
	UniqueName getName();

	/**
	 * @param prepend
	 *            NS syntax to prepend to this definition.
	 * @return this
	 */
	NSObject setNSPrepend(String prepend);

	/**
	 * @return NS syntax to prepend to this definition.
	 */
	String getNSPrepend();

	/**
	 * @param append
	 *            NS syntax to append to this definition.
	 * @return this
	 */
	NSObject setNSAppend(String append);

	/**
	 * @return NS syntax to append to this definition.
	 */
	String getNSAppend();

	/**
	 * @see java.lang.Object#toString
	 */
	Object clone();
}
