/*
 * NSVType.java
 *
 * Copyright (c) 2005 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

package net.emulab.ns;

import java.util.SortedSet;
import java.util.TreeSet;

/**
 * This class represents a "virtual type" class in Emulab extended NS.
 */
public class NSVType extends NSBase {
	/**
	 * Unique identifier for serialization.
	 */
	static final long serialVersionUID = -761869584379214248L;

	/**
	 * Sorted set of hardware types that are represented by this virtual type.
	 */
	protected TreeSet members = new TreeSet();

	/**
	 * The weight for this soft type.
	 */
	protected double weight;

	/**
	 * Construct an empty NSVType.
	 */
	public NSVType() {
	}

	/**
	 * @return The hardware members that are represented by this virtual type.
	 */
	public SortedSet getMembers() {
		return this.members;
	}

	/**
	 * @param members
	 *            The hardware members that are represented by this virtual
	 *            type.
	 * @return this
	 */
	public NSVType setMembers(TreeSet members) {
		this.members = members;
		return this;
	}

	/**
	 * Add a hardware member to this virtual type.
	 * 
	 * @param hardware
	 *            The hardware type to add.
	 * @return this
	 */
	public NSVType addMember(String hardware) {
		this.members.add(hardware);
		return this;
	}

	/**
	 * @return The weight for this soft type.
	 */
	public double getWeight() {
		return this.weight;
	}

	/**
	 * Set the weight for this soft type.
	 * 
	 * @param weight
	 *            The weight value as a floating point number between 0.0 and
	 *            1.0, where the higher the value the more consistent the actual
	 *            mapping will be.
	 */
	public NSVType setWeight(double weight) {
		if (weight < 0.0) {
			throw new IllegalArgumentException("weight is negative: " + weight);
		}
		if (weight > 1.0)
			throw new IllegalArgumentException("weight is > 1.0: " + weight);

		this.weight = weight;
		return this;
	}

	/**
	 * @see java.lang.Object#toString()
	 */
	public String toString() {
		return "NSVType[" + super.toString() + "; members=" + this.members
				+ "]";
	}
}
