package net.emulab.ns;

import java.util.ArrayList;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.SortedSet;
import java.util.Vector;

public class NodeTraffic extends NSBase implements NSAggregate {
	
	public static final int SOURCE = 0;
	public static final int TARGET = 1;
	
	public static final Hashtable TRAFFIC_TYPES = new Hashtable();

	static {
		TRAFFIC_TYPES.put("UDP", new String[] { "Agent/UDP",
				"Application/Traffic/CBR", "Agent/Null", });
		TRAFFIC_TYPES.put("TCP", new String[] { "Agent/TCP",
				"Application/Traffic/CBR", "Agent/TCPSink", });
	}

	static List group(NSTopology nst) {
		Namer namer = nst.getNamer();
		NSConnectableObject nsco[];
		List retval = new Vector();
		List list;
		int lpc;

		list = nst.getConnectors();
		nsco = (NSConnectableObject[]) list.toArray(new NSConnectableObject[list.size()]);
		for (lpc = 0; lpc < nsco.length; lpc++) {
			if ((nsco[lpc] instanceof NSAgent)
					&& nst.getAgents().contains(nsco[lpc])) {
				NSAgent nsa = (NSAgent) nsco[lpc];

				if (!(nsa.getAttachPoint() instanceof NSAgent)
						|| !(nsa.getConnectPoint() instanceof NSAgent)) {
				} else {
					NSAgent attach, connect;
					NSNode src, dst;

					attach = (NSAgent) nsa.getAttachPoint();
					connect = (NSAgent) nsa.getConnectPoint();
					if (("Application/Traffic/CBR".equals(nsa.getType())
							&& "Agent/UDP".equals(attach.getType())
							&& "Agent/Null".equals(connect.getType())) ||
							("Application/Traffic/CBR".equals(nsa.getType())
									&& "Agent/TCP".equals(attach.getType())
									&& "Agent/TCPSink".equals(connect.getType()))) {
						SortedSet membership = nst.findEventGroup(nsa);
						NodeTraffic nt;

						src = (NSNode) attach.getAttachPoint();
						dst = (NSNode) connect.getAttachPoint();
						
						if (nst.findLink(src, dst) != null)
							continue;
						
						nt = new NodeTraffic();
						if (membership.size() == 1) {
							boolean sameSrc = true;
							NSEventGroup nseg;
							Iterator iter;

							nseg = (NSEventGroup) membership.first();
							for (iter = nseg.getMembers().iterator(); iter.hasNext() && sameSrc; ) {
								NSAgent mem = (NSAgent)iter.next();
								NSNode msrc, mdst;
								
								msrc = (NSNode)((NSAgent)mem.getAttachPoint()).getAttachPoint();
								mdst = (NSNode)((NSAgent)mem.getConnectPoint()).getAttachPoint();
								sameSrc = sameSrc && msrc.equals(src);
								if (!nt.members.contains(msrc))
									nt.members.add(msrc);
								if (!nt.members.contains(mdst))
									nt.members.add(mdst);
								
								iter.remove();

								nst.remove(mem);
								namer.remove(mem.getName());
								nst.remove(mem.getAttachPoint());
								namer.remove(mem.getAttachPoint().getName());
								nst.remove(mem.getConnectPoint());
								namer.remove(mem.getConnectPoint().getName());
							}
							nt.setDirection(sameSrc ? SOURCE : TARGET);
							if (sameSrc)
								nt.setNode(src);
							else
								nt.setNode(dst);
							nt.setName(nseg.getName());
							nst.remove(nseg);
						} else {
							nt.setName(nsa.getName());
							nt.setDirection(TARGET);
							nt.members.add(src);
							nt.members.add(dst);
							nt.node = dst;
							nst.remove(nsa);
							nst.remove(attach);
							namer.remove(attach.getName());
							nst.remove(connect);
							namer.remove(connect.getName());
						}
						nt.setType(attach.getType().substring(6));
						nt.getAttributes().putAll(nsa.getAttributes());
						
						retval.add(nt);
					}
				}
			}
		}
		return retval;
	}

	protected List members = new ArrayList();
	
	protected NSNode node;

	protected String type = "UDP";

	protected int direction = TARGET;
	
	protected final Hashtable attributes = new Hashtable();
	
	protected boolean startAtSwapin = true;


	/** ADDED BY BELA **/
	protected String accessSecZone = "Zone1";
	protected String egressSecZone = "Zone1";
	protected String securityConduit = "Conduit1";

	public NodeTraffic() {
	}
	
	public boolean contains(NSObject nso) {
		return this.members.contains(nso);
	}

	public NodeTraffic setMembers(List members) {
		this.members = members;
		return this;
	}

	public List getMembers() {
		return this.members;
	}

	public NodeTraffic setNode(NSNode node) {
		this.node = node;
		return this;
	}
	
	public NSNode getNode() {
		return this.node;
	}

	public NodeTraffic setType(String type) {
		this.type = type;
		return this;
	}

	public String getType() {
		return this.type;
	}

	/** ADDED by BELA **/
	public NodeTraffic setAccessSecZone(String accessSecZone) {
		this.accessSecZone = accessSecZone;
		return this;
	}
	public String getAccessSecZone() {
		return this.accessSecZone;
	}
	public NodeTraffic setEgressSecZone(String egressSecZone) {
                this.egressSecZone = egressSecZone;
                return this;
        }
        public String getEgressSecZone() {
                return this.egressSecZone;
        }
	public NodeTraffic setSecurityConduit(String securityConduit) {
                this.securityConduit = securityConduit;
                return this;
        }
        public String getSecurityConduit() {
                return this.securityConduit;
        }



	public NodeTraffic setDirection(int direction) {
		this.direction = direction;
		return this;
	}

	public int getDirection() {
		return this.direction;
	}

	public Hashtable getAttributes() {
		return this.attributes;
	}
	
	public NodeTraffic setStartAtSwapin(boolean sas) {
		this.startAtSwapin = sas;
		return this;
	}
	
	public boolean getStartAtSwapin() {
		return this.startAtSwapin;
	}

	private NSAgent addAgents(NSTopology nst, NSNode src, NSNode dst) {
		NSAgent transport, sink, application;
		String agentTypes[];
		Namer namer;

		namer = nst.getNamer();
		
		agentTypes = (String[]) TRAFFIC_TYPES.get(this.type);

		transport = new NSAgent();
		transport.setType(agentTypes[0]);
		transport.setAttachPoint(src);
		transport.setName(namer.suggest(this.name + "src"));
		transport.setAccessSecZone(this.accessSecZone);

		sink = new NSAgent();
		sink.setType(agentTypes[2]);
		sink.setAttachPoint(dst);
		sink.setName(namer.suggest(this.name + "sink"));
		sink.setEgressSecZone(this.egressSecZone);

		application = new NSAgent();
		application.setType(agentTypes[1]);
		application.setAttachPoint(transport);
		application.setConnectPoint(sink);
		application.setSecurityConduit(this.securityConduit);
		if (this.members.size() > 2)
			application.setName(namer.suggest(this.name + "app"));
		else
			application.setName(this.name);
		application.getAttributes().putAll(this.attributes);
		
		application.setStartAtSwapin(this.startAtSwapin);

		nst.add(transport);
		nst.add(sink);
		nst.add(application);

		return application;
	}
	
	public void ungroup(NSTopology nst) {
		NSEventGroup nseg = null;
		int lpc, size;
				
		size = this.members.size();
		if (size > 2) {
			nseg = new NSEventGroup();
			nseg.setName(this.name);
			nst.add(nseg);
		}
		for (lpc = 0; lpc < size; lpc++) {
			NSNode nsn = (NSNode)this.members.get(lpc);
			NSAgent nsa = null;
			
			if (nsn == this.node) {
			}
			else if (this.direction == TARGET) {
				nsa = this.addAgents(nst, nsn, this.node);
			}
			else {
				nsa = this.addAgents(nst, this.node, nsn);
			}
			
			if (size > 2 && nsa != null) {
				nseg.addMember(nsa);
			}
		}
	}

	public String toString() {
		return "NodeTraffic[name=" + this.name
				+ "; members=" + this.members + "; node=" + this.node + "; type="
				+ this.type + "; direction=" + this.direction + "; attributes"
				+ this.attributes + "]";
	}
}
