package net.emulab.util;

import java.applet.Applet;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLEncoder;
import java.net.URLStreamHandler;
import java.util.List;

import net.emulab.xmlrpc.SshHandler;
import net.emulab.xmlrpc.SshXmlRpcTransportFactory;

import org.apache.xmlrpc.XmlRpcTransportFactory;

import thinlet.ThinletKeyValue;

public class ServerProperties {
	public int port = 0;

	public String protocol = "https";
	
	public String user = "stack"; // System.getProperty("user.name");

	public String host = "boss.emulab.net";

	public String path = "";
	
	public String auth = "";
	
	public ProgressObserver observer;

	public XmlRpcTransportFactory transportFactory;

	private boolean haveSSH = false;

	private boolean haveHTTPS = false;

	public ServerProperties(ThinletKeyValue tkv, Object pool) {
		Applet applet;
		
		tkv.observeKeyValue(pool, "environ.SSH_AUTH_SOCK",
				ThinletKeyValue.CHANGE_SETTING, this, "sshAuthSock");

		if ((applet = (Applet)tkv.getKeyValue(pool, "applet")) != null) {
			this.user = applet.getParameter("uid");
			this.auth = applet.getParameter("auth");
			if (this.user != null && this.auth != null) {
				this.host = applet.getCodeBase().getHost();
				this.path = (applet.getCodeBase().getPath() + "/xmlrpcpipe.php3?"
						+ "nocookieuid=" + URLEncoder.encode(this.user)
						+ "&nocookieauth=" + URLEncoder.encode(this.auth));
			}
		}
		else {
			// this.path = "xmlrpc";
			// this.path = "/usr/testbed/devel/stack/sbin/xmlrpc/server";
		}
		
		tkv.addKeyDependency(this, "URL", "observer");
		tkv.addKeyDependency(this, "URL", "protocol");
		tkv.addKeyDependency(this, "URL", "user");
		tkv.addKeyDependency(this, "URL", "host");
		tkv.addKeyDependency(this, "URL", "port");
		tkv.addKeyDependency(this, "URL", "path");

		tkv.addKeyDependency(this, "transportFactory", "URL");
	}

	public void setSshAuthSock(String value) {
		this.haveSSH = true;
		this.protocol = "ssh";
		this.path = "/usr/testbed/devel/stack/sbin/xmlrpc/server";
	}
	
	public URL getURL(String path, String args[]) {
		URL retval = null;

		try {
			URLStreamHandler handler = null;
			String userPart = "";
			int lpc;
			
			if ("ssh".equals(this.protocol)) {
				handler = new SshHandler(this.observer);
				userPart = this.user + "@";
			}
			if (args != null) {
				for (lpc = 0; lpc < args.length; lpc += 2) {
					if (lpc == 0)
						path += "?";
					else
						path += "&";
					path += args[lpc] + "=" + URLEncoder.encode(args[lpc + 1]);
				}
				if (!"".equals(this.auth)) {
					path += "&nocookieuid=" + URLEncoder.encode(this.user);
					path += "&nocookieauth=" + URLEncoder.encode(this.auth);
				}
			}
			retval = new URL(null, this.protocol + "://"
					+ userPart
					+ this.host
					+ (this.port != 0 ? ":" + this.port : "")
					+ "/"
					+ path,
					handler);
		} catch (MalformedURLException e) {
			e.printStackTrace();
		}

		return retval;
	}

	public URL getURL() {
		return this.getURL(this.path, null);
	}
	
	public XmlRpcTransportFactory getTransportFactory() {
		XmlRpcTransportFactory retval = null;
		URL url;

		url = this.getURL();
		if (url != null) {
			retval = new SshXmlRpcTransportFactory(url, this.observer);
		}

		return retval;
	}

	public String toString() {
		return "ServerProperties[port=" + this.port + "; protocol="
				+ this.protocol + "; host=" + this.host + "; path=" + this.path
				+ "; observer=" + this.observer + "; transportFactory="
				+ this.transportFactory + "]";
	}
}
