/*
 * SshXmlRpcTransportFactory.java
 *
 * Copyright (c) 2005 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

package net.emulab.xmlrpc;

import java.net.URL;

import net.emulab.util.ProgressObserver;

import org.apache.xmlrpc.DefaultXmlRpcTransportFactory;
import org.apache.xmlrpc.XmlRpcClientException;
import org.apache.xmlrpc.XmlRpcTransport;

/**
 * Extension to the DefaultXmlRpcTransportFactory that works with SSH URLs.
 */
public class SshXmlRpcTransportFactory extends DefaultXmlRpcTransportFactory {
	/**
	 * The ProgressObserver that is interested in connections created by this
	 * factory.
	 */
	private final ProgressObserver po;

	/**
	 * Construct an SshXmlRpcTransportFactory with the given values.
	 * 
	 * @param url
	 *            The URL of the RPC peer.
	 * @param po
	 *            The ProgressObserver that is interested in connections created
	 *            by this factory.
	 */
	public SshXmlRpcTransportFactory(URL url, ProgressObserver po) {
		super(url);

		this.po = po;
	}

	/**
	 * @see org.apache.xmlrpc.XmlRpcTransportFactory#createTransport()
	 */
	public XmlRpcTransport createTransport() throws XmlRpcClientException {
		XmlRpcTransport retval;

		retval = new SshXmlRpcTransport(this.url, this.po);
		
		return retval;
	}
}
