package thinlet;

import java.util.List;
import java.util.Arrays;
import java.util.Vector;

public class MatchController extends ArrayController {
	protected String fillKey = "";

	protected final boolean underline[] = new boolean[80];

	protected final boolean tmpUnderline[] = new boolean[80];

	protected final StringBuffer input = new StringBuffer();

	public MatchController(Thinlet thinlet, Vector content) {
		super(thinlet, content);
	}

	public MatchController(Thinlet thinlet) {
		super(thinlet);
	}

	public String fillKey() {
		return this.fillKey;
	}

	public void setFillKey(String fillKey) {
		this.fillKey = fillKey;
	}

	public void initComponent(Object sender) {
		this.thinlet.setObject(sender, "underline", this.underline);
	}

	protected void inputChanged(Object sender) {
		String text;

		this.rearrangeObjects();
		text = this.tkv.getKeyValueAsString(this.selection, this.fillKey);
		this.thinlet.setString(sender, "text", text);
		this.thinlet.setInteger(sender, "start", 0);
		this.thinlet.setInteger(sender, "end", text.length());
	}

	public void insertText(Object sender, String text) {
		this.input.append(text);
		this.inputChanged(sender);
	}

	public void removeText(Object sender) {
		if (this.input.length() > 0) {
			this.input.setLength(this.input.length() - 1);
			this.inputChanged(sender);
		}
	}

	protected boolean objectMatches(String value) {
		boolean retval = false;
		int start;

		if (Character.toLowerCase(value.charAt(0)) == Character
				.toLowerCase(this.input.charAt(0))) {
			int lpc = 1, vindex = 1, ilength, vlength;

			this.tmpUnderline[0] = true;
			ilength = this.input.length();
			vlength = value.length();
			Arrays.fill(this.tmpUnderline, 1, vlength, false);
			while ((lpc < ilength) && (vindex < vlength)) {
				if (Character.toLowerCase(value.charAt(vindex)) != Character
						.toLowerCase(this.input.charAt(lpc))) {
					this.tmpUnderline[vindex] = false;
					for (vindex++; (vindex < vlength)
									&& Character.isLetter(value.charAt(vindex)); vindex++) {
						this.tmpUnderline[vindex] = false;
					}
					for (; (vindex < vlength)
							&& Character.isWhitespace(value.charAt(vindex)); vindex++) {
						this.tmpUnderline[vindex] = false;
					}
				} else {
					this.tmpUnderline[vindex] = true;
					lpc += 1;
					vindex += 1;
				}
			}
			if (lpc == ilength) {
				retval = true;
			}
		}
		return retval;
	}

	protected List arrangeObjects(List list) {
		Object newSelection = null;
		List retval;

		retval = super.arrangeObjects(list);
		if ((this.input != null) && (this.input.length() > 0)) {
			int lpc, length;

			length = retval.size();
			for (lpc = 0; lpc < length;) {
				String value;

				value = this.tkv.getKeyValueAsString(retval.get(lpc),
						this.fillKey);
				if (this.objectMatches(value)) {
					System.arraycopy(this.tmpUnderline, 0, this.underline, 0,
							this.underline.length);
					newSelection = retval.get(lpc);
					lpc += 1;
				} else {
					retval.remove(lpc);
					length -= 1;
				}
			}
		}
		this.tkv.setKeyValue(this, "selection", newSelection);
		return retval;
	}

	public String toString() {
		return "MatchController[" + super.toString() + "]";
	}
}
