package thinlet;

import java.lang.reflect.InvocationTargetException;

import java.util.Stack;

import java.lang.reflect.Constructor;

public class ObjectFactory {
	public static final int DEFAULT_STACK_LENGTH = 5;

	private final ThinletKeyValue tkv;

	private final Object pool;

	private final Constructor constructor;

	private final Object constructorArgs[];

	private final Stack stack = new Stack();

	private final int maxStackLength;

	public ObjectFactory(ThinletKeyValue tkv, Object pool, Class cl,
			int maxStackLength) throws NoSuchMethodException, SecurityException {
		this.tkv = tkv;
		this.pool = pool;
		this.constructor = cl.getConstructor(new Class[] { this.getClass() });
		this.constructorArgs = new Object[] { this };
		this.maxStackLength = maxStackLength;
	}

	public ObjectFactory(ThinletKeyValue tkv, Object pool, Class cl)
			throws NoSuchMethodException, SecurityException {
		this(tkv, pool, cl, DEFAULT_STACK_LENGTH);
	}

	public ThinletKeyValue getNamespace() {
		return this.tkv;
	}

	public Object getPool() {
		return this.pool;
	}

	public void setKeyValue(String keyPath, Object value) {
		this.tkv.setKeyValue(this.pool, keyPath, value);
	}

	public Object getKeyValue(String keyPath) {
		return this.tkv.getKeyValue(this.pool, keyPath);
	}

	public void wirePathToObjectKey(String keyPath, Object obj, String objKey) {
		this.tkv.observeKeyValue(this.pool, keyPath,
				ThinletKeyValue.CHANGE_SETTING, obj, objKey);
	}

	public synchronized void blackhole(Object obj) {
		if (this.stack.size() < this.maxStackLength)
			this.stack.push(obj);
	}

	public synchronized Object whitehole() throws InvocationTargetException,
			InstantiationException, IllegalAccessException {
		Object retval;

		if (this.stack.size() == 0) {
			retval = this.constructor.newInstance(this.constructorArgs);
		} else {
			retval = this.stack.pop();
		}
		return retval;
	}

	public String toString() {
		return "ObjectFactory[]";
	}
}
