package thinlet;

import java.util.Hashtable;

public class ObjectPool extends Hashtable {
	
	private final ObjectPool parents[];
	
	public ObjectPool(ObjectPool parents[]) {
		if (parents == null)
			throw new IllegalArgumentException();
		
		this.parents = parents;
	}
	
	public ObjectPool(ObjectPool parent) {
		this(new ObjectPool[] { parent });
	}
	
	public ObjectPool() {
		this(new ObjectPool[0]);
	}
	
	public Object get(Object key) {
		Object retval;
		
		if ((retval = super.get(key)) == null) {
			int lpc;
			
			for (lpc = 0; lpc < this.parents.length && retval == null; lpc++) {
				retval = this.parents[lpc].get(key);
			}
		}
		
		return retval;
	}

	public boolean containsKey(Object key) {
		boolean retval;
		
		if (!(retval = super.containsKey(key))) {
			int lpc;
			
			for (lpc = 0; lpc < this.parents.length && !retval; lpc++) {
				retval = this.parents[lpc].containsKey(key);
			}
		}
		
		return retval;
	}
	
	public int hashCode() {
		return System.identityHashCode(this);
	}

	public boolean equals(Object obj) {
		return (obj == this);
	}
}
