package thinlet;

import java.awt.AWTEvent;
import java.awt.Component;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.StringSelection;
import java.awt.datatransfer.Transferable;
import java.awt.event.ComponentEvent;
import java.awt.event.FocusEvent;
import java.awt.event.InputEvent;
import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;
import java.io.Reader;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

public abstract class ThinletEvent extends ThinletLayout {
	transient Thread timer;

	transient long watchdelay;

	transient long watch;

	transient String clipboard;

	// enter the starting characters of a list item text within a short time to
	// select the first matching item
	transient String findprefix = "";

	transient long findtime;

	transient Object mouseinside;

	transient Object insidepart;

	transient Object mousepressed;

	transient Object pressedpart;

	transient int referencex, referencey;

	transient int mousex, mousey;

	transient int mousebutton;

	transient boolean mousewasdragged;

	transient Object popupowner;

	transient Object proxinside;

	transient Object tooltipowner;

	transient Object draggerParent;

	transient int dragx, dragy, dragProximity;

	transient Image dragImage;

	transient final Rectangle dragBounds = new Rectangle();
	
	transient int dragSelect = 0;
	transient boolean dragPaint = false;
	transient final Point dragSelectStart = new Point();
	transient final Point dragSelectCurrent = new Point();
	transient final Rectangle dragSelectBounds = new Rectangle();
	transient final Rectangle dragSelectLimit = new Rectangle();
	transient final Rectangle dragSelectInt = new Rectangle();
	transient final Rectangle dragSelectRepaint = new Rectangle();
	transient final List dragSelectInit = new ArrayList();
	
	transient ThinletDragSource tds = null;
	transient ThinletDropDestination tdd;

	transient ThinletDragInfo tdi = new ThinletDragInfo();

	transient AWTEvent event;

	// -----------------------------------------------------------------------
	private static final int DRAG_ENTERED = AWTEvent.RESERVED_ID_MAX + 1;

	private static final int DRAG_EXITED = AWTEvent.RESERVED_ID_MAX + 2;

	private static long WHEEL_MASK = 0;

	private static int MOUSE_WHEEL = 0;

	private static Method wheelrotation = null;

	static {
		try {
			WHEEL_MASK = AWTEvent.class.getField("MOUSE_WHEEL_EVENT_MASK")
					.getLong(null);
			MOUSE_WHEEL = MouseEvent.class.getField("MOUSE_WHEEL").getInt(null);
		} catch (Exception exc) { /* not 1.4 */
		}
	}
	{
		// disable global focus-manager for this component in 1.4
		if (MOUSE_WHEEL != 0) {
			try {
				getClass().getMethod("setFocusTraversalKeysEnabled",
						new Class[] { Boolean.TYPE }).invoke(this,
						new Object[] { Boolean.FALSE });
			} catch (Exception exc) { /* never */
			}
		}
		// set listeners flags
		enableEvents(AWTEvent.COMPONENT_EVENT_MASK | AWTEvent.FOCUS_EVENT_MASK
						| AWTEvent.KEY_EVENT_MASK | AWTEvent.MOUSE_EVENT_MASK
						| AWTEvent.MOUSE_MOTION_EVENT_MASK | WHEEL_MASK);
	}

	/**
	 * A second thread is used to repeat value change events for scrollbar or
	 * spinbox during the mouse is pressed, or to pop up tooltip
	 */
	public synchronized void run() {
		while (timer == Thread.currentThread()) {
			try {
				if (watch == 0) {
					wait(0);
				} else {
					long current = System.currentTimeMillis();
					if (watch > current) {
						wait(watch - current);
					} else {
						watch = 0;
						if ((watchdelay == 300L) || (watchdelay == 60L)) {
							if (processScroll(mousepressed, pressedpart)) {
								setTimer(60L);
							}
						} else if ((watchdelay == 301L) || (watchdelay == 61L)) {
						} else if ((watchdelay == 375L) || (watchdelay == 75L)) {
							if (processSpin(mousepressed, pressedpart)) {
								setTimer(75L);
							}
						} else if (watchdelay == 750L) {
							showTip();
						}
					}
				}
			} catch (InterruptedException ie) {
			} // ie.printStackTrace();
		}
	}

	/**
	 * 
	 */
	void setTimer(long delay) {
		watchdelay = delay;
		if (delay == 0) {
			watch = 0;
		} else {
			long prev = watch;
			watch = System.currentTimeMillis() + delay;
			if (timer == null) {
				timer = new Thread(this);
				timer.setPriority(Thread.MIN_PRIORITY);
				timer.setDaemon(true);
				timer.start();
			}
			if ((prev == 0) || (watch < prev)) {
				synchronized (this) {
					notify();
				}
				// synchronized (this) { try { notify(); }catch
				// (IllegalMonitorStateException imse) {} }
			}
		}
	}

	/**
	 * Dispatches mouse, key, focus, and component events occurring on the
	 * <i>Thinlet</i> component internally
	 */
	protected void processEvent(AWTEvent e) {
		// evm (touchscreen) events: entered/moved/pressed -> dragged ->
		// dragged/released/exited
		int id = e.getID();
		event = e;
		if ((id == MouseEvent.MOUSE_ENTERED) || (id == MouseEvent.MOUSE_MOVED)
			|| (id == MouseEvent.MOUSE_EXITED)
			|| (id == MouseEvent.MOUSE_PRESSED)
			|| (id == MouseEvent.MOUSE_DRAGGED)
			|| (id == MouseEvent.MOUSE_RELEASED)) {
			MouseEvent me = (MouseEvent) e;
			int x = me.getX();
			int y = me.getY();
			int clickcount = me.getClickCount();
			boolean shiftdown = me.isShiftDown();
			boolean controldown = me.isControlDown();
			boolean popuptrigger = me.isPopupTrigger();
			if (connector != null) {
				if (id == MouseEvent.MOUSE_MOVED) {
					id = MouseEvent.MOUSE_DRAGGED;
					mousepressed = connector;
				} else if (id == MouseEvent.MOUSE_PRESSED) {
					id = MouseEvent.MOUSE_RELEASED;
				}
			}
			if (id == MouseEvent.MOUSE_ENTERED) {
				if (mousepressed == null) {
					findComponent(content, x, y);
					handleMouseEvent(x, y, clickcount, shiftdown, controldown,
							popuptrigger, MouseEvent.MOUSE_ENTERED,
							mouseinside, insidepart);
				}
			} else if (id == MouseEvent.MOUSE_MOVED) {
				Object previnside = mouseinside;
				Object prevpart = insidepart;
				findComponent(content, x, y);
				if ((previnside == mouseinside) && (prevpart == insidepart)) {
					handleMouseEvent(x, y, clickcount, shiftdown, controldown,
							popuptrigger, MouseEvent.MOUSE_MOVED, mouseinside,
							insidepart);
				} else {
					handleMouseEvent(x, y, clickcount, shiftdown, controldown,
							popuptrigger, MouseEvent.MOUSE_EXITED, previnside,
							prevpart);
					handleMouseEvent(x, y, clickcount, shiftdown, controldown,
							popuptrigger, MouseEvent.MOUSE_ENTERED,
							mouseinside, insidepart);
				}
			} else if (id == MouseEvent.MOUSE_EXITED) {
				if (mousepressed == null) {
					Object mouseexit = mouseinside;
					Object exitpart = insidepart;
					mouseinside = insidepart = null;
					handleMouseEvent(x, y, clickcount, shiftdown, controldown,
							popuptrigger, MouseEvent.MOUSE_EXITED, mouseexit,
							exitpart);
				}
			} else if (id == MouseEvent.MOUSE_PRESSED) {
				if (mousepressed != null)
					return;
				mousebutton = me.getButton();
				mousewasdragged = false;
				if (popupowner != null) { // remove popup
					String classname = getClass(mouseinside);
					if ((popupowner != mouseinside) && (classname != ":popup")
						&& (classname != ":combolist")) {
						closeup();
					}
				}
				hideTip(); // remove tooltip
				mousepressed = mouseinside;
				pressedpart = insidepart;
				handleMouseEvent(x, y, clickcount, shiftdown, controldown,
						popuptrigger, MouseEvent.MOUSE_PRESSED, mousepressed,
						pressedpart);
			} else if (id == MouseEvent.MOUSE_DRAGGED) {
				hideTip(); // remove tooltip
				mousewasdragged = true;
				Object previnside = mouseinside;
				Object prevpart = insidepart;
				findComponent(content, x, y);
				boolean same = (previnside == mouseinside)
								&& (prevpart == insidepart);
				boolean isin = (mousepressed == mouseinside)
								&& (pressedpart == insidepart);
				boolean wasin = (mousepressed == previnside)
								&& (pressedpart == prevpart);

				if (wasin && !isin) {
					handleMouseEvent(x, y, clickcount, shiftdown, controldown,
							popuptrigger, MouseEvent.MOUSE_EXITED,
							mousepressed, pressedpart);
				} else if (!same && (popupowner != null) && !wasin) {
					handleMouseEvent(x, y, clickcount, shiftdown, controldown,
							popuptrigger, DRAG_EXITED, previnside, prevpart);
				}
				if (isin && !wasin) {
					handleMouseEvent(x, y, clickcount, shiftdown, controldown,
							popuptrigger, MouseEvent.MOUSE_ENTERED,
							mousepressed, pressedpart);
				} else if (!same && (popupowner != null) && !isin) {
					handleMouseEvent(x, y, clickcount, shiftdown, controldown,
							popuptrigger, DRAG_ENTERED, mouseinside, insidepart);
				}
				if (isin == wasin) {
					handleMouseEvent(x, y, clickcount, shiftdown, controldown,
							popuptrigger, MouseEvent.MOUSE_DRAGGED,
							mousepressed, pressedpart);
				}
				if (!same && getClass(previnside) == "togglebutton") {
					setPoint(previnside, ":mouse", -1, -1);
				}
			} else if (id == MouseEvent.MOUSE_RELEASED) {
				if (mousebutton != me.getButton()) {
					if (connector != null) {
						connector = null;
						connectorList = null;
						mousepressed = pressedpart = null;
						findComponent(content, x, y);
						repaint(content);
						return;
					}
					return;
				}
				mousebutton = me.getButton();
				hideTip(); // remove tooltip
				Object mouserelease = mousepressed;
				Object releasepart = pressedpart;
				mousepressed = pressedpart = null;
				handleMouseEvent(x, y, clickcount, shiftdown, controldown,
						popuptrigger, MouseEvent.MOUSE_RELEASED, mouserelease,
						releasepart);
				if ((mouseinside != null)
					&& ((mouserelease != mouseinside) || (releasepart != insidepart))) {
					handleMouseEvent(x, y, clickcount, shiftdown, controldown,
							popuptrigger, MouseEvent.MOUSE_ENTERED,
							mouseinside, insidepart);
				}
			}
		} else if (id == MOUSE_WHEEL) {
			Rectangle port = getRectangle(mouseinside, ":port");
			if (port != null) { // is scrollable
				// TODO hide tooltip?
				Rectangle bounds = getRectangle(mouseinside, "bounds");
				try { // mouse wheel is supported since 1.4 thus it use
						// reflection
					if (wheelrotation == null) {
						wheelrotation = e.getClass().getMethod(
								"getWheelRotation", null);
					}
					int rotation = ((Integer) wheelrotation.invoke(e, null))
							.intValue();

					if (port.x + port.width < bounds.width) { // has vertical
																// scrollbar
						processScroll(mouseinside, (rotation > 0) ? "down"
								: "up"); // TODO scroll panels too
					} else if (port.y + port.height < bounds.height) { // has
																		// horizontal
																		// scrollbar
						processScroll(mouseinside, (rotation > 0) ? "right"
								: "left");
					}
				} catch (Exception exc) { /* never */
				}
			}
		} else if ((id == KeyEvent.KEY_PRESSED) || (id == KeyEvent.KEY_TYPED)) {
			if (focusinside && ((popupowner != null) || (focusowner != null))) {
				hideTip(); // remove tooltip
				KeyEvent ke = (KeyEvent) e;
				int keychar = ke.getKeyChar();
				boolean control = (keychar <= 0x1f)
									|| ((keychar >= 0x7f) && (keychar <= 0x9f))
									|| (keychar >= 0xffff)
									|| ke.isControlDown();
				int keycode = control ? ke.getKeyCode() : 0;

				if ((control == (id == KeyEvent.KEY_PRESSED))
					&& processKeyPress((popupowner != null) ? popupowner
							: focusowner, ke.isShiftDown(), ke.isControlDown(),
							ke.getModifiers(), control ? 0 : keychar, keycode)) {
					ke.consume();
				} else if ((keycode == KeyEvent.VK_TAB)
							|| ((keycode == KeyEvent.VK_F6) && (ke.isAltDown() || ke
									.isControlDown()))) {
					boolean outgo = (keycode == KeyEvent.VK_F6);
					if (!ke.isShiftDown() ? setNextFocusable(focusowner, outgo)
							: setPreviousFocusable(focusowner, outgo)) {
						ke.consume();
					} else if (MOUSE_WHEEL != 0) { // 1.4
						if (!ke.isShiftDown()) {
							transferFocus();
						} else {
							try {
								getClass().getMethod("transferFocusBackward",
										null).invoke(this, null);
							} catch (Exception exc) { /* never */
							}
						}
					}
					repaint(focusowner);
					closeup();
				} else if (keycode == KeyEvent.VK_F8) {
					for (Object splitpane = focusowner; splitpane != null; splitpane = getParent(splitpane)) {
						if (getClass(splitpane) == "splitpane") {
							setFocus(splitpane);
							repaint(splitpane);
							ke.consume();
							break; // middle
						}
					}
				} else if ((id == KeyEvent.KEY_PRESSED)
							&& ((keychar != 0) || ke.isActionKey())
							&& checkMnemonic(focusowner, true, null, ke
									.getKeyCode(), ke.getModifiers())) {
					ke.consume();
				}
			}
		} else if (id == FocusEvent.FOCUS_LOST) {
			focusinside = false;
			if (focusowner != null) {
				repaint(focusowner);
			}
			closeup();
		} else if (id == FocusEvent.FOCUS_GAINED) {
			focusinside = true;
			if (focusowner == null) {
				setFocus(content);
			} else {
				repaint(focusowner);
			}
		} else if ((id == ComponentEvent.COMPONENT_RESIZED)
					|| (id == ComponentEvent.COMPONENT_SHOWN)) {
			Dimension d = getSize();
			setRectangle(content, "bounds", 0, 0, d.width, d.height);
			validate(content);
			closeup();
			if (!focusinside) {
				requestFocus();
			}
		}
	}

	/**
	 * Check the previous mouse location again because of a possible layout
	 * change
	 */
	void checkLocation(Object component) {
		if (mouseinside == component) {
			findComponent(content, mousex, mousey);
			handleMouseEvent(mousex, mousex, 1, false, false, false,
					MouseEvent.MOUSE_ENTERED, mouseinside, insidepart);
		}
	}

	/**
	 * 
	 */
	boolean processKeyPress(Object component, boolean shiftdown,
			boolean controldown, int modifiers, int keychar, int keycode) {
		String classname = getClass(component);
		if ("button" == classname) {
			if (keychar == KeyEvent.VK_SPACE
				|| ((keycode == KeyEvent.VK_ENTER) && (get(component, "type") == "default"))
				|| ((keycode == KeyEvent.VK_ESCAPE) && // ...
				(get(component, "type") == "cancel"))) {
				// pressedkey = keychar;
				invoke(component, null, "action");
				repaint(component);
				return true;
			} else if (keycode == KeyEvent.VK_BACK_SPACE
						|| keycode == KeyEvent.VK_DELETE) {
				invoke(component, null, "delete");
				return true;
			}
		} else if ("connector" == classname) {
			if (keycode == KeyEvent.VK_BACK_SPACE
				|| keycode == KeyEvent.VK_DELETE) {
				invoke(component, null, "delete");
				return true;
			}
		} else if ("dropzone" == classname) {
			 if (keycode == KeyEvent.VK_BACK_SPACE
						|| keycode == KeyEvent.VK_DELETE) {
				invoke(component, null, "delete");
				return true;
			}
		} else if (("checkbox" == classname) || ("togglebutton" == classname)) {
			if (keychar == KeyEvent.VK_SPACE) {
				changeCheck(component, true);
				repaint(component);
				return true;
			} else if (keycode == KeyEvent.VK_BACK_SPACE
						|| keycode == KeyEvent.VK_DELETE) {
				invoke(component, null, "delete");
				return true;
			}
		} else if ("combobox" == classname) {
			Object combolist = get(component, ":combolist");
			if (combolist == null) { // the drop down list is not visible
				boolean editable = getBoolean(component, "editable", true)
									|| getBoolean(component, "selectable", true);
				if (editable
					&& processField(component, shiftdown, controldown,
							modifiers, keychar, keycode, false, false, false)) {
					int sel;

					if ((sel = getInteger(component, "selected")) != -1) {
						Object choice = getItem(component, sel);

						if (true || getString(choice, "text").equals(
								getString(component, "text"))) {
							/* The textfield matches a choice, pop the combobox. */
							popupCombo(component);
						} else {
							/*
							 * The textfield does not match a choice, clear the
							 * selection.
							 */
							setInteger(component, "selected", -1, -1);
						}
					}
					return true;
				}
				if (getBoolean(component, "pickable")
					&& ((keychar == KeyEvent.VK_SPACE) || (keycode == KeyEvent.VK_DOWN))) {
					popupCombo(component);
				}
				// +findText
				else {
					return false;
				}
			} else {
				if ((keycode == KeyEvent.VK_UP)
					|| (keycode == KeyEvent.VK_DOWN)
					|| (keycode == KeyEvent.VK_PAGE_UP)
					|| (keycode == KeyEvent.VK_PAGE_DOWN)
					|| (keycode == KeyEvent.VK_HOME)
					|| (keycode == KeyEvent.VK_END)) {
					Object next = getListItem(component, combolist, keycode,
							get(combolist, ":lead"), false);
					if (next != null) {
						String text = getString(next, "text");

						setInside(combolist, next, true);
						/*
						 * The user is navigating manually, update the text in
						 * the field,
						 */
						setString(component, "text", text);
						/* ... select everything, */
						setInteger(component, "start", 0);
						setInteger(component, "end", text.length());
						/* ... and update the "selected" value. */
						setInteger(component, "selected", getIndex(component,
								next));
					}
				} else if ((keycode == KeyEvent.VK_ENTER)
							|| (keychar == KeyEvent.VK_SPACE)) {
					boolean editable = getBoolean(component, "editable");
					Object selection;

					if (editable)
						selection = null;
					else
						selection = get(combolist, ":lead");
					closeCombo(component, combolist, selection); // Alt+Up
					invoke(component, null, "perform");
				} else if (keycode == KeyEvent.VK_ESCAPE) {
					closeCombo(component, combolist, null);
				} else if (!processField(component, shiftdown, controldown,
						modifiers, keychar, keycode, false, false, false)) {
					Object item = findText((char) keychar, component,
							combolist, false);
					if (item != null) {
						setInside(combolist, item, true);
					} else
						return false;
				}
			}
			return true;
		} else if (("textfield" == classname) || ("passwordfield" == classname)) {
			boolean retval = processField(component, shiftdown, controldown, modifiers,
					keychar, keycode, false, ("passwordfield" == classname),
					false);
			return retval;
		} else if ("textarea" == classname) {
			char[] chars = (char[]) get(component, ":text");
			int start = getInteger(component, "start", 0);
			int end = getInteger(component, "end", 0);

			int istart = start;
			int iend = end;
			String insert = null;
			if ((keycode == KeyEvent.VK_HOME) && !controldown) {
				while ((iend > 0) && (chars[iend - 1] != '\n')) {
					iend--;
				}
				if (!shiftdown) {
					istart = iend;
				}
			} else if ((keycode == KeyEvent.VK_END) && !controldown) {
				while ((iend < chars.length) && (chars[iend] != '\n')) {
					iend++;
				}
				if (!shiftdown) {
					istart = iend;
				}
			} else if ((keycode == KeyEvent.VK_UP)
						|| (keycode == KeyEvent.VK_PAGE_UP)
						|| (keycode == KeyEvent.VK_DOWN)
						|| (keycode == KeyEvent.VK_PAGE_DOWN)) {
				Font currentfont = (Font) get(component, "font");
				FontMetrics fm = getFontMetrics((currentfont != null) ? currentfont
						: font);
				int fh = fm.getHeight();
				int y = 0;
				int linestart = 0;
				for (int i = 0; i < iend; i++) {
					if ((chars[i] == '\n') || (chars[i] == 0)) { /* wrap */
						linestart = i + 1;
						y += fh;
					}
				}
				if (keycode == KeyEvent.VK_UP) {
					y -= fh;
				} else if (keycode == KeyEvent.VK_DOWN) {
					y += fh;
				} else {
					int dy = getRectangle(component, ":port").height;
					y += (keycode == KeyEvent.VK_PAGE_UP) ? -dy : dy; // VK_PAGE_DOWN
				}
				int x = fm.charsWidth(chars, linestart, iend - linestart);
				iend = getCaretLocation(component, x, y, true, false);
				if (!shiftdown) {
					istart = iend;
				}
			} else
				return processField(component, shiftdown, controldown,
						modifiers, keychar, keycode, true, false, false);
			return changeField(component, getString(component, "text", ""),
					insert, istart, iend, start, end);
		} else if ("tabbedpane" == classname) {
			if ((keycode == KeyEvent.VK_RIGHT) || (keycode == KeyEvent.VK_DOWN)
				|| (keycode == KeyEvent.VK_LEFT) || (keycode == KeyEvent.VK_UP)) {
				int selected = getInteger(component, "selected", 0);
				boolean increase = (keycode == KeyEvent.VK_RIGHT)
									|| (keycode == KeyEvent.VK_DOWN);
				int newvalue = selected;
				int n = increase ? getItemCountImpl(component, ":comp") : 0;
				int d = (increase ? 1 : -1);
				for (int i = selected + d; increase ? (i < n) : (i >= 0); i += d) {
					if (getBoolean(getItem(component, i), "enabled", true)) {
						newvalue = i;
						break;
					}
				}
				if (newvalue != selected) {
					setInteger(component, "selected", newvalue, 0);
					checkOffset(component);
					repaint(component);
					invoke(component, getItem(component, newvalue), "action");
				}
			}
		} else if ("spinbox" == classname) {
			if ((keycode == KeyEvent.VK_UP) || (keycode == KeyEvent.VK_DOWN)) {
				processSpin(component, (keycode == KeyEvent.VK_UP) ? "up"
						: "down");
				return true;
			}
			return processField(component, shiftdown, controldown, modifiers,
					keychar, keycode, false, false, true);
		} else if ("slider" == classname) {
			int value = getInteger(component, "value", 0);
			int d = 0;
			if ((keycode == KeyEvent.VK_HOME) || (keycode == KeyEvent.VK_LEFT)
				|| (keycode == KeyEvent.VK_UP)
				|| (keycode == KeyEvent.VK_PAGE_UP)) {
				d = getInteger(component, "minimum", 0) - value;
				if ((keycode == KeyEvent.VK_LEFT)
					|| (keycode == KeyEvent.VK_UP)) {
					d = Math.max(d, -getInteger(component, "unit", 5));
				} else if (keycode == KeyEvent.VK_PAGE_UP) {
					d = Math.max(d, -getInteger(component, "block", 25));
				}
			} else if ((keycode == KeyEvent.VK_END)
						|| (keycode == KeyEvent.VK_RIGHT)
						|| (keycode == KeyEvent.VK_DOWN)
						|| (keycode == KeyEvent.VK_PAGE_DOWN)) {
				d = getInteger(component, "maximum", 100) - value;
				if ((keycode == KeyEvent.VK_RIGHT)
					|| (keycode == KeyEvent.VK_DOWN)) {
					d = Math.min(d, getInteger(component, "unit", 5));
				} else if (keycode == KeyEvent.VK_PAGE_DOWN) {
					d = Math.min(d, getInteger(component, "block", 25));
				}
			}
			if (d != 0) {
				setInteger(component, "value", value + d, 0);
				repaint(component);
				invoke(component, null, "action");
			}
		} else if ("splitpane" == classname) {
			int divider = getInteger(component, "divider", -1);
			int d = 0;
			if (keycode == KeyEvent.VK_HOME) {
				d = -divider;
			} else if ((keycode == KeyEvent.VK_LEFT)
						|| (keycode == KeyEvent.VK_UP)) {
				d = Math.max(-10, -divider);
			} else if ((keycode == KeyEvent.VK_END)
						|| (keycode == KeyEvent.VK_RIGHT)
						|| (keycode == KeyEvent.VK_DOWN)) {
				boolean horizontal = ("vertical" != get(component,
						"orientation"));
				Rectangle bounds = getRectangle(component, "bounds");
				int max = (horizontal ? bounds.width : bounds.height) - 5;
				d = max - divider;
				if (keycode != KeyEvent.VK_END) {
					d = Math.min(d, 10);
				}
			}
			if (d != 0) {
				setInteger(component, "divider", divider + d, -1);
				validate(component);
			}
		} else if (("list" == classname) || ("table" == classname)) {
			return processList(component, shiftdown, controldown, keychar,
					keycode, false);
		} else if ("tree" == classname) {
			// ? clear childs' selection, select this is its subnode was
			// selected
			if (keycode == KeyEvent.VK_LEFT) {
				Object lead = get(component, ":lead");
				if ((get(lead, ":comp") != null)
					&& getBoolean(lead, "expanded", true)) { // collapse
					setBoolean(lead, "expanded", false, true);
					selectItem(component, lead, true);
					validate(component);
					invoke(component, lead, "collapse"); // lead
					return true;
				} else { // select parent
					Object parent = getParent(lead);
					if (parent != component) {
						selectItem(component, parent, true);
						setLead(component, lead, parent);
						return true;
					}
				}
			}
			// ? for interval mode select its all subnode or deselect all after
			else if (keycode == KeyEvent.VK_RIGHT) {
				Object lead = get(component, ":lead");
				Object node = get(lead, ":comp");
				if (node != null) {
					if (getBoolean(lead, "expanded", true)) { // select its
																// first subnode
						selectItem(component, node, true);
						setLead(component, lead, node);
					} else { // expand
						setBoolean(lead, "expanded", true, true);
						selectItem(component, lead, true);
						validate(component);
						invoke(component, lead, "expand"); // lead
					}
					return true;
				}
			}
			return processList(component, shiftdown, controldown, keychar,
					keycode, true);
		} else if (("menubar" == classname) || ("popupmenu" == classname)) {
			// find the last open :popup and the previous one
			Object previous = null;
			Object last = null;
			for (Object i = get(component, ":popup"); i != null; i = get(i,
					":popup")) {
				previous = last;
				last = i;
			}
			// selected is the current item of the last, or the previous :popup,
			// or null
			Object selected = get(last, "selected");
			Object hotpopup = ((selected != null) || (previous == null)) ? last
					: previous;
			if ((selected == null) && (previous != null)) {
				selected = get(previous, "selected");
			}

			if ((keycode == KeyEvent.VK_UP) || (keycode == KeyEvent.VK_DOWN)) {
				Object next = getMenu(hotpopup, selected,
						keycode == KeyEvent.VK_DOWN, true);
				if (next != null) {
					set(hotpopup, "selected", null);
					popupMenu(hotpopup);
					set(hotpopup, "selected", next);
					repaint(hotpopup);
				}
			} else if (keycode == KeyEvent.VK_LEFT) {
				if (previous != null) { // close the last :popup
					selected = get(previous, "selected");
					set(previous, "selected", null);
					popupMenu(previous);
					set(previous, "selected", selected);
					repaint(previous); // , selected
				} else if ("menubar" == classname) { // select the previous
														// menubar menu
					Object next = getMenu(component,
							get(component, "selected"), false, false);
					if (next != null) {
						set(component, "selected", next);
						Object popup = popupMenu(component);
						set(popup, "selected", getMenu(popup, null, true, true));
						repaint(component); // , selected
					}
				}
			} else if (keycode == KeyEvent.VK_RIGHT) {
				if ((previous != null) && (selected == null)) { // ?
					set(last, "selected", get(get(last, "menu"), ":comp"));
					repaint(last); // , selected
				} else if ((selected != null) && (getClass(selected) == "menu")) { // expand
																					// menu
					Object popup = popupMenu(last);
					set(popup, "selected", getMenu(popup, null, true, true));
				} else if ("menubar" == classname) { // select the next
														// menubar menu
					Object next = getMenu(component,
							get(component, "selected"), true, false);
					if (next != null) {
						set(component, "selected", next);
						Object popup = popupMenu(component);
						set(popup, "selected", getMenu(popup, null, true, true));
						repaint(component); // , selected
					}
				}
			} else if ((keycode == KeyEvent.VK_ENTER)
						|| (keychar == KeyEvent.VK_SPACE)
						|| (keycode == KeyEvent.VK_ESCAPE)) {
				if ((keycode != KeyEvent.VK_ESCAPE)
					&& getBoolean(selected, "enabled", true)) {
					if ((selected != null)
						&& (getClass(selected) == "checkboxmenuitem")) {
						changeCheck(selected, false);
					} else
						invoke(selected, null, "action");
				}
				closeup();
			} else
				return false;
			return true;
		}
		return false;
	}

	/**
	 * 
	 */
	boolean changeCheck(Object component, boolean box) {
		String group = getString(component, "group", null);
		if (group != null) {
			if (getBoolean(component, "selected", false)) {
				return false;
			}
			for (Object comp = get(getParent(component), ":comp"); comp != null; comp = get(
					comp, ":next")) {
				if (comp == component) {
					setBoolean(component, "selected", true);
				} else if (group.equals(get(comp, "group"))
							&& getBoolean(comp, "selected", false)) {
					setBoolean(comp, "selected", false);
					if (box) {
						repaint(comp);
					} // checkbox only
				}
			}
		} else {
			setBoolean(component, "selected", !getBoolean(component,
					"selected", false), false);
			if (box) {
				repaint(component);
			}
		}
		invoke(component, null, "action");
		return true;
	}

	/**
	 * @param component
	 *            a :popup or a menubar
	 * @param part
	 *            the currently selected item, return the first/last if null
	 * @param forward
	 *            find the next item if true, the previous otherwise
	 * @param popup
	 *            the given component is :popup if true, menubar otherwise
	 * @return the next/previous item relative to the current one excluding
	 *         separators, or null
	 */
	Object getMenu(Object component, Object part, boolean forward, boolean popup) {
		Object previous = null;
		for (int i = 0; i < 2; i++) { // 0: next to last, 1: first to previous
			for (Object item = (i == 0) ? get(part, ":next") : get(popup ? get(
					component, "menu") : component, ":comp"); (i == 0) ? (item != null)
					: (item != part); item = get(item, ":next")) {
				if ((getClass(item) != "separator")
					&& getBoolean(item, "enabled", true)) {
					if (forward) {
						return item;
					}
					previous = item;
				}
			}
		}
		return previous;
	}

	/**
	 * Process keyboard events for textfield, passwordfield, textarea, combobox,
	 * and spinbox
	 * 
	 * @param multiline
	 *            true for textarea, otherwise false
	 * @param hidden
	 *            true for passwordfield, otherwise false
	 * @param filter
	 *            true for spinbox, otherwise false
	 */
	boolean processField(Object component, boolean shiftdown,
			boolean controldown, int modifiers, int keychar, int keycode,
			boolean multiline, boolean hidden, boolean filter) {
		String text = getString(component, "text", "");
		int start = getInteger(component, "start", 0);
		int end = getInteger(component, "end", 0);
		boolean editable = getBoolean(component, "editable", true);

		int istart = start;
		int iend = end;
		String insert = null;
		if (editable && (keychar != 0) &&
		// ((modifiers == 0) || (modifiers == InputEvent.SHIFT_MASK))) {
			(modifiers != InputEvent.ALT_MASK)) {
			insert = String.valueOf((char) keychar);
		} else if (editable && (keycode == KeyEvent.VK_ENTER)) {
			if (multiline) {
				insert = "\n";
			} else {
				try {
					if (focusformatter != null) {
						Object value;

						/*
						 * There is a formatter, try to translate the text into
						 * an object.
						 */
						value = focusformatter.getObjectValue(text);
						setReal(component, "text", value);
						text = focusformatter.stringForObjectValue(
								(Thinlet) this, component, value);
						tkv.notifyKeyValue(component, "text",
								ThinletKeyValue.CHANGE_SETTING, null, value);
					}
					invoke(component, null, "perform");
				} catch (Exception e) {
					handleException(component, e);
				}
				setInteger(component, "start", 0, 0);
				setInteger(component, "end", text != null ? text.length() : 0,
						0);
				validate(component);
				return false;
			}
		} else if (editable && (keycode == KeyEvent.VK_BACK_SPACE)) {
			insert = "";
			if (start == end) {
				istart -= 1;
			}
		} else if (keycode == KeyEvent.VK_END) {
			iend = text.length();
			if (!shiftdown) {
				istart = iend;
			}
		} else if (keycode == KeyEvent.VK_HOME) {
			iend = 0;
			if (!shiftdown) {
				istart = iend;
			}
		} else if (keycode == KeyEvent.VK_LEFT) {
			if (controldown) {
				for (int i = 0; i < 2; i++) {
					while ((iend > 0)
							&& ((i != 0) == Character.isLetterOrDigit(text
									.charAt(iend - 1)))) {
						iend--;
					}
				}
			} else {
				iend -= 1;
			}
			if (!shiftdown) {
				istart = iend;
			}
		} else if (keycode == KeyEvent.VK_RIGHT) {
			if (controldown) {
				for (int i = 0; i < 2; i++) {
					while ((iend < text.length())
							&& ((i == 0) == Character.isLetterOrDigit(text
									.charAt(iend)))) {
						iend++;
					}
				}
			} else {
				iend += 1;
			}
			if (!shiftdown) {
				istart = iend;
			}
		} else if (editable && (keycode == KeyEvent.VK_DELETE)) {
			insert = "";
			if (start == end) {
				iend += 1;
			}
		} else if (controldown
					&& ((keycode == KeyEvent.VK_A) || (keycode == 0xBF))) {
			istart = 0; // KeyEvent.VK_SLASH
			iend = text.length();
		} else if (controldown && (keycode == 0xDC)) {
			istart = iend = text.length(); // KeyEvent.VK_BACK_SLASH
		} else if ((editable && !hidden && controldown && (keycode == KeyEvent.VK_X))
					|| (!hidden && controldown && (keycode == KeyEvent.VK_C))) {
			if (start != end) {
				clipboard = text.substring(Math.min(start, end), Math.max(
						start, end));
				try {
					getToolkit().getSystemClipboard().setContents(
							new StringSelection(clipboard), null);
				} catch (Exception exc) {
				}
				try {
					getToolkit().getSystemSelection().setContents(
							new StringSelection(clipboard), null);
				} catch (Exception exc) {
				}
				if (keycode == KeyEvent.VK_X) {
					insert = "";
				} else {
					return true;
				}
			} else {
				invoke(component, null, "sigint");
			}
		} else if (editable && controldown && (keycode == KeyEvent.VK_V)) {
			try {
				insert = (String) getToolkit().getSystemClipboard()
						.getContents(this).getTransferData(
								DataFlavor.stringFlavor);
			} catch (Exception exc) {
				try {
					insert = (String) getToolkit().getSystemSelection()
							.getContents(this).getTransferData(
									DataFlavor.stringFlavor);
				} catch (Exception exc2) {
					insert = clipboard;
				}
			}
			// new Throwable().printStackTrace();
			System.out.println("ins " + insert);
			if (insert != null) { // no text on system clipboard nor internal
									// clipboard text
				insert = filter(insert, multiline);
			}
		}
		if (filter && (insert != null)) { // contributed by Michael Nascimento
			for (int i = insert.length() - 1; i >= 0; i--) {
				if (!Character.isDigit(insert.charAt(i))) {
					getToolkit().beep();
					return false;
				}
			}
		}
		if (insert != null) {
			String allowed, notallowed;

			allowed = (String) get(component, "allowed");
			notallowed = (String) get(component, "notallowed");
			if ((allowed != null) || (notallowed != null)) {
				for (int i = insert.length() - 1; i >= 0; i--) {
					if (allowed != null
						&& allowed.indexOf(insert.charAt(i)) == -1) {
						getToolkit().beep();
						return false;
					}
					if (notallowed != null
						&& notallowed.indexOf(insert.charAt(i)) != -1) {
						getToolkit().beep();
						return false;
					}
				}
			}
		}

		if ((istart > 0) && (istart < text.length())
			&& (text.charAt(istart) == '\b')) {
			if (istart > start)
				istart += 2;
			else
				istart -= 2;
		}
		if ((iend > 0) && (iend < text.length()) && (text.charAt(iend) == '\b')) {
			if (iend > end)
				iend += 2;
			else
				iend -= 2;
		}
		return changeField(component, text, insert, istart, iend, start, end);
	}

	/**
	 * @param text
	 * @param multiline
	 * @return
	 */
	static String filter(String text, boolean multiline) {
		StringBuffer filtered = new StringBuffer(text.length());
		for (int i = 0; i < text.length(); i++) {
			char ckey = text.charAt(i);
			if (((ckey > 0x1f) && (ckey < 0x7f))
				|| ((ckey > 0x9f) && (ckey < 0xffff))
				|| (multiline && (ckey == '\n'))) {
				filtered.append(ckey);
			}
		}
		return (filtered.length() != text.length()) ? filtered.toString()
				: text;
	}

	/**
	 * @param component
	 *            a textfield, passwordfield, textarea, combobox, or spinbox
	 * @param text
	 *            current text
	 * @param insert
	 *            a string to replace thr current selection
	 * @param movestart
	 *            new selection start position
	 * @param moveend
	 *            new caret (selection end) position
	 * @param start
	 *            current selection start position
	 * @param end
	 *            current caret position
	 * @return true if selection, caret location, or text content changed
	 */
	boolean changeField(Object component, String text, String insert,
			int movestart, int moveend, int start, int end) {
		movestart = Math.max(0, Math.min(movestart, text.length()));
		moveend = Math.max(0, Math.min(moveend, text.length()));
		if ((insert == null) && (start == movestart) && (end == moveend)) {
			return false;
		}
		if (insert != null) {
			try {
				StringBuffer newText = new StringBuffer();
				int min = Math.min(movestart, moveend);

				newText.append(text.substring(0, min));
				newText.append(insert);
				newText.append(text.substring(Math.max(movestart, moveend)));
				movestart = moveend = min + insert.length();
				if (focusformatter != null) {
					fp.setBeginIndex(Math.min(movestart, moveend));
					fp.setEndIndex(Math.max(movestart, moveend));
					if (focusformatter.isPartialStringValid((Thinlet) this,
							component, newText, fp)) {
						movestart = fp.getBeginIndex();
						moveend = fp.getEndIndex();
						setReal(component, "text", newText.toString());
					} else {
						movestart = start;
						moveend = end;
					}
				} else {
					set(component, "text", newText.toString());
				}
				setReal(component, keyForCache("text"), null);
				invoke(component, null, "action"); // deprecated
			} catch (IllegalArgumentException e) {
				handleException(component, e);
				movestart = start;
				moveend = end;
			}
		}
		if (start != movestart) {
			setInteger(component, "start", movestart, 0);
		}
		if (end != moveend) {
			setInteger(component, "end", moveend, 0);
		}
		validate(component);
		invoke(
				component,
				null,
				(insert != null) ? ((insert.length() > 0) ? "insert" : "remove")
						: "caret");
		return true;
	}

	/**
	 * 
	 */
	boolean processList(Object component, boolean shiftdown,
			boolean controldown, int keychar, int keycode, boolean recursive) {
		if ((keycode == KeyEvent.VK_UP) || // select previous/next/first/...
											// item
			(keycode == KeyEvent.VK_DOWN) || (keycode == KeyEvent.VK_PAGE_UP)
			|| (keycode == KeyEvent.VK_PAGE_DOWN)
			|| (keycode == KeyEvent.VK_HOME) || (keycode == KeyEvent.VK_END)) {
			Object lead = get(component, ":lead");
			Object row = getListItem(component, component, keycode, lead,
					recursive);
			if (row != null) {
				String selection = getString(component, "selection", "single");
				if (shiftdown && (selection != "single") && (lead != null)) {
					extend(component, lead, row, recursive);
				} else if (!controldown) {
					selectItem(component, row, recursive);
				}
				setLead(component, lead, row);
				return true;
			}
		} else if (keycode == KeyEvent.VK_LEFT) {
			return processScroll(component, "left");
		} else if (keycode == KeyEvent.VK_RIGHT) {
			return processScroll(component, "right");
		} else if (keychar == KeyEvent.VK_SPACE) { // select the current item
			select(component, get(component, ":lead"), recursive, shiftdown,
					controldown); // ...
			return true;
		} else if (keycode == KeyEvent.VK_ENTER) {
			Object item = getSelectedItem(component);

			if (item != null) {
				focusowner = null;
				return invoke(component, getSelectedItem(component), "perform");
			}
			return false;
		} else if (controldown) {
			if (((keycode == KeyEvent.VK_A) || (keycode == 0xBF)) && // KeyEvent.VK_SLASH
				(getString(component, "selection", "single") != "single")) { // select
																				// all
				selectAll(component, true, recursive);
				return true;
			} else if (keycode == 0xDC) { // KeyEvent.VK_BACK_SLASH //
											// deselect all
				selectAll(component, false, recursive);
				return true;
			}
		} else {
			Object item = findText((char) keychar, component, component,
					recursive);
			if (item != null) {
				select(component, item, recursive, false, false);
				return true;
			}
		}
		return false;
	}

	/**
	 * Search for the next/first appropriate item starting with the collected
	 * string or the given single character
	 * 
	 * @param keychar
	 *            the last typed character
	 * @param component
	 *            a list, tree, table, or combobox
	 * @param leadowner
	 *            the list, tree, table, or the combobox's drop down list
	 * @param recursive
	 *            if the component is a tree
	 * @return the appropriate item or null
	 */
	Object findText(char keychar, Object component, Object leadowner,
			boolean recursive) {
		if (keychar != 0) {
			long current = System.currentTimeMillis();
			int i = (current > findtime + 1000) ? 1 : 0; // clear the
															// starting string
															// after a second
			findtime = current;
			Object lead = get(leadowner, ":lead");
			for (; i < 2; i++) { // 0: find the long text, 1: the stating
									// character only
				findprefix = (i == 0) ? (findprefix + keychar) : String
						.valueOf(keychar);
				for (int j = 0; j < 2; j++) { // 0: lead to last, 1: first to
												// lead
					for (Object item = (j == 0) ? ((i == 0) ? lead
							: getNextItem(component, lead, recursive)) : get(
							component, ":comp"); (j == 0) ? (item != null)
							: (item != lead); item = getNextItem(component,
							item, recursive)) {
						if (getString(item, "text", "").regionMatches(true, 0,
								findprefix, 0, findprefix.length())) { // table
																		// first
																		// column...
							return item;
						}
					}
				}
			}
		}
		return null;
	}

	/**
	 * 
	 */
	Object getListItem(Object component, Object scrollpane, int keycode,
			Object lead, boolean recursive) {
		Object row = null;
		if (keycode == KeyEvent.VK_UP) {
			for (Object prev = get(component, ":comp"); prev != lead; prev = getNextItem(
					component, prev, recursive)) {
				row = prev; // component -> getParent(lead)
			}
		} else if (keycode == KeyEvent.VK_DOWN) {
			row = (lead == null) ? get(component, ":comp") : getNextItem(
					component, lead, recursive);
		} else if ((keycode == KeyEvent.VK_PAGE_UP)
					|| (keycode == KeyEvent.VK_PAGE_DOWN)) {
			Rectangle view = getRectangle(scrollpane, ":view");
			Rectangle port = getRectangle(scrollpane, ":port");
			Rectangle rl = (lead != null) ? getRectangle(lead, "bounds") : null;
			int vy = (keycode == KeyEvent.VK_PAGE_UP) ? view.y
					: (view.y + port.height);
			if ((keycode == KeyEvent.VK_PAGE_UP) && (rl != null)
				&& (rl.y <= view.y)) {
				vy -= port.height;
			}
			if ((keycode == KeyEvent.VK_PAGE_DOWN) && (rl != null)
				&& (rl.y + rl.height >= view.y + port.height)) {
				vy += port.height;
			}
			for (Object item = get(component, ":comp"); item != null; item = getNextItem(
					component, item, recursive)) {
				Rectangle r = getRectangle(item, "bounds");
				if (keycode == KeyEvent.VK_PAGE_UP) {
					row = item;
					if (r.y + r.height > vy) {
						break;
					}
				} else {
					if (r.y > vy) {
						break;
					}
					row = item;
				}
			}
		} else if (keycode == KeyEvent.VK_HOME) {
			row = get(component, ":comp");
		} else if (keycode == KeyEvent.VK_END) {
			for (Object last = lead; last != null; last = getNextItem(
					component, last, recursive)) {
				row = last;
			}
		}
		return row;
	}

	/**
	 * Select all the items
	 * 
	 * @param component
	 *            a list/tree/table
	 * @param selected
	 *            selects or deselects items
	 * @param recursive
	 *            true for tree
	 */
	void selectAll(Object component, boolean selected, boolean recursive) {
		boolean changed = false;
		for (Object item = get(component, ":comp"); item != null; item = getNextItem(
				component, item, recursive)) {
			if (setBoolean(item, "selected", selected, false)) {
				repaint(component, null, item);
				changed = true;
			}
		}
		set(component, ":anchor", null);
		if (changed) {
			invoke(component, null, "action");
		}
	}

	/**
	 * Select a single given item, deselect others
	 * 
	 * @param component
	 *            a list/tree/table
	 * @param row
	 *            the item/node/row to select
	 * @param recursive
	 *            true for tree
	 */
	void selectItem(Object component, Object row, boolean recursive) {
		boolean changed = false;
		int lpc = 0, selectedIndex = -1;
		for (Object item = get(component, ":comp"); item != null; item = getNextItem(
				component, item, recursive)) {
			if (item == row)
				selectedIndex = lpc;
			if (setBoolean(item, "selected", (item == row), false)) {
				repaint(component, null, item);
				changed = true;
			}
			lpc += 1;
		}
		set(component, ":anchor", null);
		if (changed) {
			tkv.notifyKeyValue(component, "selectedIndex",
					ThinletKeyValue.CHANGE_SETTING, null, new Integer(
							selectedIndex));
			invoke(component, row, "action");
		}
	}

	/**
	 * 
	 */
	void extend(Object component, Object lead, Object row, boolean recursive) {
		Object anchor = get(component, ":anchor");
		if (anchor == null) {
			set(component, ":anchor", anchor = lead);
		}
		char select = 'n';
		boolean changed = false;
		for (Object item = get(component, ":comp"); // anchor - row
		item != null; item = getNextItem(component, item, recursive)) {
			if (item == anchor)
				select = (select == 'n') ? 'y' : 'r';
			if (item == row)
				select = (select == 'n') ? 'y' : 'r';
			if (setBoolean(item, "selected", (select != 'n'), false)) {
				repaint(component, null, item);
				changed = true;
			}
			if (select == 'r')
				select = 'n';
		}
		if (changed) {
			invoke(component, row, "action");
		}
	}

	/**
	 * Update the lead item of a list/tree/table, repaint, and scroll
	 * 
	 * @param component
	 *            a list, tree, or table
	 * @param oldlead
	 *            the current lead item
	 * @param lead
	 *            the new lead item
	 */
	protected void setLead(Object component, Object oldlead, Object lead) {
		if (oldlead != lead) { // ?
			if (oldlead != null) {
				repaint(component, null, oldlead);
			}
			set(component, ":lead", lead);
			repaint(component, null, lead);

			Rectangle r = getRectangle(lead, "bounds");
			if (r != null)
				scrollToVisible(component, r.x, r.y, 0, r.height);
		}
	}

	/**
	 * Update the lead item of a combolist, repaint, and scroll
	 * 
	 * @param component
	 *            a combobox drop down list
	 * @param part
	 *            the current hotspot item
	 * @param scroll
	 *            scroll to the part if true
	 */
	void setInside(Object component, Object part, boolean scroll) {
		Object previous = get(component, ":lead");
		if (previous != null) {
			repaint(component, ":combolist", previous);
		}
		set(component, ":lead", part);
		if (part != null) {
			repaint(component, ":combolist", part);
			if (scroll) {
				Rectangle r = getRectangle(part, "bounds");
				if (r != null)
					scrollToVisible(component, r.x, r.y, 0, r.height);
			}
		}
	}

	/**
	 * @param x
	 *            mouse x position relative to thinlet component
	 * @param y
	 *            mouse y position relative to the main desktop
	 */
	void handleMouseEvent(int x, int y, int clickcount, boolean shiftdown,
			boolean controldown, boolean popuptrigger, int id,
			Object component, Object part) {
		if (false) {
			if (id == MouseEvent.MOUSE_PRESSED) {
				System.out.println("press! " + System.currentTimeMillis());
			}
			if (id == MouseEvent.MOUSE_RELEASED) {
				System.out.println("released! " + System.currentTimeMillis());
			}
			if (id == MouseEvent.MOUSE_ENTERED) {
				System.out.println("entered! " + System.currentTimeMillis());
			}
			if (id == MouseEvent.MOUSE_EXITED) {
				System.out.println("exited! " + System.currentTimeMillis());
			}
		}
		if (id == MouseEvent.MOUSE_ENTERED) {
			setTimer(750L);
		} else if (id == MouseEvent.MOUSE_EXITED) {
			hideTip();
		}
		if (!getBoolean(component, "enabled", true)) {
			return;
		}
		String classname = getClass(component);
		if (popuptrigger) {// && (id == MouseEvent.MOUSE_RELEASED)) {
			if (insidepart == "proximity") {
				return;
			} else {
			    // we need to update the selection for left click too, it's too
			    // cumbersome to use the applet forcing a left-click before right-click
			    // just to delete a togglebutton.
			    Object tsel[];
			    int tlpc;
			    
			    tsel = getSelectedItems(getParent(component));
			    for (tlpc = 0; tlpc < tsel.length; tlpc++) {
				//setBoolean(tsel[tlpc], "selected", false);
				//invoke(tsel[tlpc], null, "selectchanged");
				changeCheck(tsel[tlpc],false);
			    }
			    //setFocus(component);
			    //setBoolean(component,"selected",true);
			    //invoke(component, null, "selectchanged");
			    changeCheck(component,false);
				Object popupmenu = get(component, "popupmenu");
				if (popupmenu != null) {
					popupPopup(popupmenu, x, y);
					mousepressed = popupmenu;
					return;
				}
			}
		}
		if (("list" == classname) || ("table" == classname)
			|| ("tree" == classname)) {
			// check if we need to change the column selection and invoke the
			// header its action
			Rectangle bounds = getRectangle(component, "bounds");
			Rectangle port = getRectangle(component, ":port");
			Object header = get(component, "header");
			boolean noScroll = false;
			if (header != null) {
				if (get(header, ":resizecomponent") != null) {
					noScroll = true;
					switch (id) {
					case MouseEvent.MOUSE_PRESSED:
						referencex = x;
						set(header, ":resizing", "true");
						break;
					case MouseEvent.MOUSE_DRAGGED:
						if (get(header, ":resizing") != null) {
							// resize the column, but limit its minimum size to
							// a width of 10
							Object column = get(header, ":resizecomponent");
							int newSize = getInteger(column, "width") + x
											- referencex;
							if (newSize > 10) {
								if (setInteger(column, "width", newSize, 80)) {
									update(column, "parent");
								}
								referencex = x;
							}
						}
						break;
					case MouseEvent.MOUSE_RELEASED:
						set(header, ":resizing", null);
						break;
					case MouseEvent.MOUSE_ENTERED:
						setCursor(Cursor
								.getPredefinedCursor(Cursor.E_RESIZE_CURSOR));
						break;
					}
				} else {
					set(header, ":resizing", null);
					setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
				}
				if (getBoolean(header, "sortable", false)) {
					if (pressedpart instanceof Object[]
						&& "column" == getClass(pressedpart)) {
						noScroll = true;
					}
					if (insidepart != null && insidepart instanceof Object[]
						&& "column" == getClass(insidepart)) {
						noScroll = true;
						if (id == MouseEvent.MOUSE_ENTERED
							|| id == MouseEvent.MOUSE_PRESSED
							|| id == MouseEvent.MOUSE_RELEASED) {
							if (id == MouseEvent.MOUSE_RELEASED
								&& part == insidepart
								&& mouseinside == component) {
								// set selected column its sort and selected
								// property
								// and set the sort of all other columns to none
								// (null)
								Object column = get(get(component, "header"),
										":comp");
								String sortKey = null;
								Object sort = null;
								while (column != null) {
									if (column == insidepart) {
										sort = get(column, "sort");
										sortKey = getString(column, "key");
										if (null == sort || "none" == sort
											|| "descent" == sort)
											sort = "ascent";
										else if ("ascent" == sort)
											sort = "descent";
										set(component, "key", sortKey);
										set(component, "sort", sort);
										invoke(column, null, "action");
									} else {
										sort = null;
									}
									set(column, "sort", sort);
									setBoolean(column, "selected",
											sort != null && sort != "none");
									column = get(column, ":next");
								}
								invoke(header, null, "action");
							}
							repaint(component, bounds.x, bounds.y, Math
									.abs(bounds.width), port.y);
						}
					} else if (id == MouseEvent.MOUSE_EXITED) {
						repaint(component, bounds.x, bounds.y, Math
								.abs(bounds.width), port.y);
					}
				}
			}
			if (!noScroll && !processScroll(x, y, id, component, part)) {
				if (((id == MouseEvent.MOUSE_PRESSED) || ((id == MouseEvent.MOUSE_DRAGGED)
															&& !shiftdown && !controldown))) {
					// Rectangle view = getRectangle(component, ":view");
					int my = y + port.y - referencey;
					for (Object item = get(component, ":comp"); item != null;) {
						Rectangle r = getRectangle(item, "bounds");
						if (my < r.y + r.height) {
							if (id == MouseEvent.MOUSE_DRAGGED) { // !!!
								scrollToVisible(component, r.x, r.y, 0,
										r.height);
							} else if ("tree" == classname) {
								int mx = x + port.x - referencex;
								if (mx < r.x) {
									if ((mx >= r.x - block)
										&& (get(item, ":comp") != null)) {
										boolean expanded = getBoolean(item,
												"expanded", true);
										setBoolean(item, "expanded", !expanded,
												true);
										selectItem(component, item, true);
										setLead(component, get(component,
												":lead"), item);
										setFocus(component);
										validate(component);
										invoke(component, item,
												expanded ? "collapse"
														: "expand"); // item
									}
									break;
								}
							}
							if ((id != MouseEvent.MOUSE_DRAGGED)
								|| !getBoolean(item, "selected", false)) {
								if (id != MouseEvent.MOUSE_DRAGGED) {
									if (setFocus(component)) {
										repaint(component, classname, item);
									} // ?
								}
								select(component, item, ("tree" == classname),
										shiftdown, controldown);
								if (clickcount == 2) {
									invoke(component, item, "perform");
								}
							}
							break;
						}
						item = getNextItem(component, item,
								("tree" == classname));
					}
				}
			}
		} else if ("bean" == classname) {
			Component bean = (Component) get(component, "bean");
			int modifiers = 0;
			if (shiftdown)
				modifiers |= KeyEvent.SHIFT_MASK;
			if (controldown)
				modifiers |= KeyEvent.CTRL_MASK;
			if (id == MouseEvent.MOUSE_PRESSED
				|| id == MouseEvent.MOUSE_RELEASED
				|| id == MouseEvent.MOUSE_EXITED
				|| id == MouseEvent.MOUSE_ENTERED
				|| id == MouseEvent.MOUSE_MOVED
				|| id == MouseEvent.MOUSE_DRAGGED) {
				Rectangle pbounds = getRectangle(getParent(component), "bounds");
				Rectangle bounds = getRectangle(component, "bounds");

				x -= pbounds.x + bounds.x;
				y -= pbounds.y + bounds.y;
				bean.dispatchEvent(new MouseEvent(this, id, System
						.currentTimeMillis(), modifiers, x, y, clickcount,
						popuptrigger));
				// ((Integer)get(component,":mousex")).intValue(),
				// ((Integer)get(component,":mousey")).intValue(),clickcount,popuptrigger));
			}
		} else if (("button" == classname) || ("checkbox" == classname)
					|| ("togglebutton" == classname)
					|| ("connector" == classname)) {
			int proximity = getInteger(component, "proximity", 0);
			if ((id == MouseEvent.MOUSE_ENTERED)
				|| (id == MouseEvent.MOUSE_EXITED)
				|| (id == MouseEvent.MOUSE_PRESSED)
				|| (id == MouseEvent.MOUSE_RELEASED)) {

				if (id == MouseEvent.MOUSE_PRESSED) {
					setFocus(component);
				}
				if (("button" == classname)
					&& ((mousepressed == null) || (mousepressed == component))
					&& ((id == MouseEvent.MOUSE_ENTERED) || (id == MouseEvent.MOUSE_EXITED))
					&& (get(component, "type") == "link")) {
					setCursor(Cursor
							.getPredefinedCursor((id == MouseEvent.MOUSE_ENTERED) ? Cursor.HAND_CURSOR
									: Cursor.DEFAULT_CURSOR));
					updateLinkStatus(component, (id == MouseEvent.MOUSE_ENTERED));
				} else if ((("button" == classname) || ("togglebutton" == classname))
							&& (tds = (ThinletDragSource) getDelegate(
									component, "dragsrc")) != null) {
					Rectangle bounds, cbounds;
					Point point;

					cbounds = getRectangle(content, "bounds");
					if (id == MouseEvent.MOUSE_PRESSED) {
						bounds = getRectangle(component, "bounds");
						point = getPoint(component, ":mouse");
						if ((point == null)
							|| (((point.x + 1) > proximity)
								&& ((point.y + 1) > proximity)
								&& (point.x < (bounds.width - proximity)) && (point.y < (bounds.height - proximity)))) {
							if ("button" != classname) {
								if (!shiftdown) {
									Object sel[];
									int lpc;
									
									// System.out.println("start changing");
									sel = getSelectedItems(getParent(component));
									for (lpc = 0; lpc < sel.length; lpc++) {
										changeCheck(sel[lpc], true);
									}
								}
								changeCheck(component, true);
							}
						} else {
							if (!getBoolean(component, "selected", false)) {
								Object parent = getParent(component);
								
								invoke(parent, null, "deselect");
								setBoolean(component, "selected", true);
								invoke(parent, null, "selectchanged");
							}
							
							Object sel[] = getSelectedItems(getParent(component));
							
							connectorList = new ArrayList();
							if (sel != null && sel.length > 0) {
								int lpc;
								
								for (lpc = 0; lpc < sel.length; lpc++) {
									if (getInteger(sel[lpc], "proximity", 0) > 0) {
										connectorList.add(sel[lpc]);
									}
								}
							}
							else {
								connectorList.add(component);
							}
							connx = dragx + bounds.x;
							conny = dragy + bounds.y;
							connector = component;
						}
					} else if (id == MouseEvent.MOUSE_RELEASED) {
						int modifiers = 0;

						if (shiftdown)
							modifiers |= KeyEvent.SHIFT_MASK;
						if (controldown)
							modifiers |= KeyEvent.CTRL_MASK;
						if (tdi.source != null) {
							Rectangle pbounds;

							pbounds = getRectangle(mouseinside, "bounds");
							bounds = getRectangle(tdi.source, "bounds");
							bounds.x = tdi.x - dragx;
							// System.out.println("drop bounds " + bounds + "; " + pbounds);
							if ((bounds.x + proximity) < 0) {
								bounds.x = -proximity;
							} else if ((mouseinside != null)
										&& ((bounds.x + bounds.width + proximity) > pbounds.width)) {
								bounds.x = pbounds.width - bounds.width
											+ proximity;
							}
							bounds.y = tdi.y - dragy;
							if ((bounds.y + proximity) < 0) {
								// System.out.println("too far " + bounds);
								bounds.y = -proximity;
							} else if ((mouseinside != null)
										&& ((bounds.y + bounds.height - proximity) > pbounds.height)) {
								// System.out.println("too far");
								bounds.y = pbounds.height - bounds.height
											+ proximity;
							}
							if (mouseinside != null) {
								Rectangle oldbounds;

								// System.out.println("dropped mi");
								// printcomp(mouseinside);
								// System.out.println("=========");
								if (tdd != null) {
									// System.out.println(" has tdd");
									if (tdd.prepareForDrop((Thinlet) this, tdi)) {
										tdd.performDrop((Thinlet) this, tdi);
										tdd.finishDrop((Thinlet) this, tdi);

										cbounds.width *= -1;
									}
									tdd = null;
									draggerParent = null;
								} else if (getClass(mouseinside) == "dropzone") {
									// System.out.println("add to zone!");
									add(mouseinside, tdi.source);

									Object sel[];
									int lpc;

									sel = getSelectedItems(mouseinside);
									for (lpc = 0; lpc < sel.length; lpc++) {
										setBoolean(sel[0], "selected", false);
									}

									cbounds.width *= -1;
									draggerParent = null;
								} else if ((oldbounds = getRectangle(
										tdi.source, ":oldbounds")) != null) {
									// System.out.println("**** RESTORING OLD "
									// + oldbounds);
									bounds.x = oldbounds.x;
									bounds.y = oldbounds.y;
									bounds.width = oldbounds.width;
									bounds.height = oldbounds.height;
								}
								// printcomp(mouseinside);

								if (invoke(mouseinside, null, "drop")) {
								} else if (draggerParent != null) {
									// System.out.println("ADDING IT BACK
									// !!!!!!");
									// printcomp(tdi.source);
									tds.dragAbort((Thinlet)this, tdi.source);
									add(draggerParent, tdi.source);
								}
							} else if (draggerParent != null) {
								// System.out.println("dropped dp");
								// printcomp(draggerParent);
								tds.dragAbort((Thinlet)this, tdi.source);
								add(draggerParent, tdi.source);
								invoke(draggerParent, null, "drop");
							}

							if (!getBoolean(tdi.source, "selected", false))
								changeCheck(tdi.source, false);
							tdi.source = null;
						} else if (connector != null) {
							// System.out.println("connect " +
							// getClass(mouseinside));
							if (!mousewasdragged) {
							} else {
								if ((mouseinside != null)
									&& (!connectorList.contains(mouseinside))
									&& (getParent(connector) == getParent(mouseinside))) {
									int lpc, size;
									
									connectee = mouseinside;
									size = connectorList.size();
									for (lpc = 0; lpc < size; lpc++) {
										connector = connectorList.get(lpc);
										invoke(mouseinside, null, "connect");
									}
								}
								connector = null;
								connectorList = null;
							}
						} else if (mouseinside == component) {
							// System.out.println("change check " + classname);
							if ("button" != classname) {
							} else {
								invoke(component, null, "action");
							}
						}
						if (cbounds.width < 0) {
							cbounds.width = Math.abs(cbounds.width);
							doLayout(content);
							findComponent(content, x, y);
						}
						repaint(dragBounds.x, dragBounds.y, dragBounds.width, dragBounds.height);
						repaint(getParent(component));
					}
				} else if ((id == MouseEvent.MOUSE_RELEASED)
							&& (mouseinside == component)) {
					if ("button" != classname) {
						Object parent = getParent(component);
						if (!shiftdown && getClass(parent) == "dropzone") {
							Object sel[];
							int lpc;
							
							// System.out.println("start changing");
							sel = getSelectedItems(parent);
							for (lpc = 0; lpc < sel.length; lpc++) {
								changeCheck(sel[lpc], true);
							}
						}
						changeCheck(component, true);
						// printcomp(component);
					} else {
						invoke(component, null, "action");
					}
				}
				repaint(component);
			} else if (id == MouseEvent.MOUSE_DRAGGED
						&& (mousebutton == MouseEvent.BUTTON1)
						&& (popupowner == null)
						&& (("button" == classname) || ("togglebutton" == classname))
						&& ((tdi.source != null) || (tds = (ThinletDragSource) getDelegate(
								component, "dragsrc")) != null)) {
				Rectangle bounds;
				if (connector != null) {
				} else if (tdi.source == null) {
					Object parent;

					bounds = getRectangle(component, "bounds");
					// System.out.println("starting drag");
					parent = getParent(component);
					if (getClass(parent) == "dropzone") {
						setRectangle(component, ":oldbounds", bounds.x,
								bounds.y, bounds.width, bounds.height);
						repaint(component,
								bounds.x - 5,
								bounds.y - 5,
								bounds.width + 10,
								bounds.height + 10);
						removeItemImpl(parent, component);
						// System.out.println("*** not cloning ***");
						tdi.source = tds.dragStart((Thinlet) this, component);
						draggerParent = parent;
					} else {
						tdi.source = tds.dragStart((Thinlet) this, component);
						draggerParent = null;
						// printcomp(component);
						// System.out.println("===");
						// printcomp(tdi.source);
					}
					// System.out.println("!! ** starting dx " + dragx);
					// System.out.println("starting dy " + dragy);
					// System.out.println("starting b " + bounds);
					// System.out.println("starting x " + x);
					// System.out.println("starting y " + y);
					if (tdi.source != null) {
						Dimension dim;
						Graphics ig;
						
						dragProximity = proximity;
						dim = getPreferredSize(tdi.source);
						setRectangle(tdi.source, "bounds", 0, 0, dim.width,
								dim.height);
						doLayout(tdi.source);
						// System.out.println("** tdi.source");
						// printcomp(tdi.source);
						bounds = getRectangle(tdi.source, "bounds");
						dragImage = this.createImage(bounds.width, bounds.height);
						ig = dragImage.getGraphics();
						ig.setFont(font);
						paint(ig, 0, 0,
								bounds.width,
								bounds.height,
								tdi.source, true);
						dragBounds.x = x - dragx;
						dragBounds.y = y - dragy;
						dragBounds.width = bounds.width;
						dragBounds.height = bounds.height;
						tdi.dropzone = parent;
						if ((tdd = (ThinletDropDestination) getDelegate(parent,
								"dropdst")) != null) {
							if (!tdd.draggingEntered((Thinlet) this, tdi))
								tdd = null;
						}
						findComponent(content, x, y);
					}
				} else {
					// System.out.println("repaint2");
					repaint(dragBounds.x, dragBounds.y, dragBounds.width,
							dragBounds.height);
					dragBounds.x = x - dragx;
					dragBounds.y = y - dragy;
					repaint(dragBounds.x, dragBounds.y, dragBounds.width,
							dragBounds.height);

					//System.out.println(" dz = " + tdi.dropzone + "; mi = "
					//					+ mouseinside);
					if (tdi.dropzone == mouseinside) {
						if (tdd != null)
							tdd.draggingUpdated((Thinlet) this, tdi);
					} else if (mouseinside != null) {
						if (tdd != null)
							tdd.draggingExited((Thinlet) this, tdi);
						tdi.dropzone = mouseinside;
						if ((tdd = (ThinletDropDestination) getDelegate(
								mouseinside, "dropdst")) != null) {
							if (!tdd.draggingEntered((Thinlet) this, tdi)) {
								tdd = null;
							}
							else {
								// System.out.println("set tdd");
							}
						}
					}
				}
			}
			if (connector != null) {
				int lpc, size;
				
				connectBounds.setBounds(connx - 5, conny - 5, 10, 10);
				connectBounds.add(lastconnx - 5, lastconny - 5);
				connectBounds.add(lastconnx + 5, lastconny + 5);
				size = connectorList.size();
				for (lpc = 0; lpc < size; lpc++) {
					Rectangle cbounds = getRectangle(connectorList.get(lpc), "bounds");
					
					connectBounds.add(cbounds);
				}
				repaint(getParent(connector), connectBounds.x, connectBounds.y, connectBounds.width, connectBounds.height);
			}
			if (tdi.source == null && proximity > 0) {
				if (id == MouseEvent.MOUSE_EXITED) {
					setPoint(component, ":mouse", -1, -1);
				}
				repaint(component);
			}
		} else if ("combobox" == classname) {
			boolean editable = getBoolean(component, "editable", true)
								|| getBoolean(component, "selectable");
			if (editable && (part == null)) { // textfield area
				Image icon = null;
				int left = ((id == MouseEvent.MOUSE_PRESSED) && ((icon = getIcon(
						component, "icon", null)) != null)) ? icon
						.getWidth(this) : 0;
				processField(x, y, clickcount, id, component, false, false,
						left);
			} else if (part != "icon") { // part = "down"
				if (!getBoolean(component, "pickable")) {
				} else if (((id == MouseEvent.MOUSE_ENTERED) || (id == MouseEvent.MOUSE_EXITED))
							&& (mousepressed == null)) {
					if (editable) {
						repaint(component, "combobox", part);
					} // hover the arrow button
					else {
						repaint(component);
					} // hover the whole combobox
				} else if (id == MouseEvent.MOUSE_PRESSED) {
					Object combolist = get(component, ":combolist");
					if (combolist == null) { // combolist is closed
						setFocus(component);
						repaint(component);
						popupCombo(component);
					} else { // combolist is visible
						closeCombo(component, combolist, null);
					}
				} else if (id == MouseEvent.MOUSE_RELEASED) {
					if (mouseinside != component) {
						Object combolist = get(component, ":combolist");
						closeCombo(
								component,
								combolist,
								((mouseinside == combolist) && (insidepart instanceof Object[])) ? insidepart
										: null);
					} else {
						repaint(component);
					}
				}
			}
		} else if (":combolist" == classname) {
			if (!processScroll(x, y, id, component, part)) {
				if ((id == MouseEvent.MOUSE_ENTERED) || (id == DRAG_ENTERED)) {
					if (part != null) { // + scroll if dragged
						setInside(component, part, false);
						setTimer(750L);
					}
				} else if (id == MouseEvent.MOUSE_RELEASED) {
					closeCombo(get(component, "combobox"), component, part);
				}
			}
		} else if (("textfield" == classname) || ("passwordfield" == classname)) {
			processField(x, y, clickcount, id, component, false,
					("passwordfield" == classname), 0);
		} else if ("textarea" == classname) {
			if (!processScroll(x, y, id, component, part)) {
				processField(x, y, clickcount, id, component, true, false, 0);
			}
		} else if ("panel" == classname || "dropzone" == classname) {
			processScroll(x, y, id, component, part);
			if (mousebutton == MouseEvent.BUTTON1 && classname == "dropzone" && part == null) {
				if (id == MouseEvent.MOUSE_PRESSED) {
					if (focusowner == component || setFocus(component)) {
						dragSelectInit.clear();
						if (shiftdown) {
							Object sel[] = getSelectedItems(component);
							
							dragSelectInit.addAll(Arrays.asList(sel));
						}
						else {
							invoke(component, null, "deselect");
						}
						dragSelect = 1;
						dragSelectStart.x = x;
						dragSelectStart.y = y;
						dragSelectCurrent.x = x;
						dragSelectCurrent.y = y;
						dragSelectLimit.x = x - dragx;
						dragSelectLimit.y = y - dragy;
					}
				}
				if (dragSelect != 0 && id == MouseEvent.MOUSE_DRAGGED) {
					boolean hits = false;

					dragSelect = 2;
					dragSelectCurrent.x = x;
					dragSelectCurrent.y = y;
					dragPaint = true;
					this.dragSelectInt.x = Math.min(this.dragSelectStart.x, this.dragSelectCurrent.x);
					this.dragSelectInt.y = Math.min(this.dragSelectStart.y, this.dragSelectCurrent.y);
					this.dragSelectInt.width = Math.max(this.dragSelectStart.x,
							this.dragSelectCurrent.x) - this.dragSelectInt.x;
					this.dragSelectInt.height = Math.max(this.dragSelectStart.y,
							this.dragSelectCurrent.y) - this.dragSelectInt.y;
					this.dragSelectInt.x -= dragSelectLimit.x;
					this.dragSelectInt.y -= dragSelectLimit.y;
					
					for (Object comp = get(component, ":comp"); comp != null; comp = get(comp, ":next")) {
						Rectangle bounds = getRectangle(comp, "bounds");
						int proximity = getInteger(comp, "proximity", 0);
						
						if (bounds != null && dragSelectInt.intersects(bounds.x + proximity,
								bounds.y + proximity,
								bounds.width - (proximity * 2),
								bounds.height - (proximity * 2))) {
							if (setBoolean(comp, "selected", !dragSelectInit.contains(comp), false)) {
								repaint(comp);
								hits = true;
							}
						}
						else if (setBoolean(comp, "selected", dragSelectInit.contains(comp), false)) {
							repaint(comp);
							hits = true;
						}
					}

					dragSelectRepaint.x = dragSelectCurrent.x;
					dragSelectRepaint.y = dragSelectCurrent.y;
					dragSelectRepaint.width = dragSelectRepaint.height = 1;
					dragSelectRepaint.add(dragSelectBounds);
					// System.out.println("rp " + dragSelectRepaint + "; " + dragSelectCurrent + "; " + dragSelectStart);
					repaint(dragSelectRepaint.x, dragSelectRepaint.y, dragSelectRepaint.width + 1, dragSelectRepaint.height + 1);
					if (hits)
						invoke(component, null, "selectchanged");
				}
				if (id == MouseEvent.MOUSE_RELEASED) {
					if (dragSelect == 2) {
						dragSelect = 0;
						dragPaint = true;
						repaint();
					}
				}
			}
		} else if ("desktop" == classname) {
			if (part == "modal") {
				if (id == MouseEvent.MOUSE_ENTERED) {
					// setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
				} else if (id == MouseEvent.MOUSE_EXITED) {
					setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
				}
			}
		} else if ("spinbox" == classname) {
			if (part == null) {
				processField(x, y, clickcount, id, component, false, false, 0);
			} else { // part = "up" || "down"
				if ((id == MouseEvent.MOUSE_ENTERED)
					|| (id == MouseEvent.MOUSE_EXITED)
					|| (id == MouseEvent.MOUSE_PRESSED)
					|| (id == MouseEvent.MOUSE_RELEASED)) {
					if (id == MouseEvent.MOUSE_PRESSED) {
						setFocus(component);
						if (processSpin(component, part)) {
							setTimer(375L);
						}
						// settext: start end selection, parse exception...
					} else {
						if (id == MouseEvent.MOUSE_RELEASED) {
							setTimer(0L);
						}
					}
					repaint(component, classname, part);
				}
			}
		} else if ("tabbedpane" == classname) {
			if ((id == MouseEvent.MOUSE_ENTERED)
				|| (id == MouseEvent.MOUSE_EXITED)) {
				if ((part != null)
					&& getBoolean(part, "enabled", true)
					&& (getInteger(component, "selected", 0) != getIndex(
							component, part))) {
					repaint(component, "tabbedpane", part);
				}
			} else if ((part != null) && (id == MouseEvent.MOUSE_PRESSED)
						&& getBoolean(part, "enabled", true)) {
				int selected = getInteger(component, "selected", 0);
				int current = getIndex(component, part);
				if (selected == current) {
					setFocus(component);
					repaint(component, "tabbedpane", part);
				} else {
					setInteger(component, "selected", current, 0);
					// Object tabcontent = getItem(component, current);
					// setFocus((tabcontent != null) ? tabcontent : component);
					setNextFocusable(component, false);
					checkOffset(component);
					repaint(component);
					invoke(component, part, "action");
				}
			}
		} else if ("slider" == classname) {
			if ((id == MouseEvent.MOUSE_PRESSED)
				|| (id == MouseEvent.MOUSE_DRAGGED)) {
				if (id == MouseEvent.MOUSE_PRESSED) {
					setReference(component, block / 2, block / 2);
					setFocus(component);
				}
				int minimum = getInteger(component, "minimum", 0);
				int maximum = getInteger(component, "maximum", 100);
				int value = getInteger(component, "value", 0);
				Rectangle bounds = getRectangle(component, "bounds");
				boolean horizontal = ("vertical" != get(component,
						"orientation"));
				int newvalue = minimum
								+ (horizontal ? (x - referencex)
										: (y - referencey))
								* (maximum - minimum)
								/ ((horizontal ? bounds.width : bounds.height) - block); // ...
																							// +0.5
				newvalue = Math.max(minimum, Math.min(newvalue, maximum));
				if (value != newvalue) {
					setInteger(component, "value", newvalue, 0);
					invoke(component, null, "action");
				}
				if ((value != newvalue) || (id == MouseEvent.MOUSE_PRESSED)) {
					repaint(component);
				}
			}
		} else if ("splitpane" == classname) {
			if (id == MouseEvent.MOUSE_PRESSED) {
				setReference(component, 2, 2);
			} else if (id == MouseEvent.MOUSE_DRAGGED) {
				int divider = getInteger(component, "divider", -1);
				boolean horizontal = ("vertical" != get(component,
						"orientation"));
				int moveto = horizontal ? (x - referencex) : (y - referencey);
				Rectangle bounds = getRectangle(component, "bounds");
				moveto = Math.max(0, Math.min(moveto, Math
						.abs(horizontal ? bounds.width : bounds.height) - 5));
				if (divider != moveto) {
					setInteger(component, "divider", moveto, -1);
					validate(component);
				}
			} else if ((id == MouseEvent.MOUSE_ENTERED)
						&& (mousepressed == null)) {
				boolean horizontal = ("vertical" != get(component,
						"orientation"));
				setCursor(Cursor
						.getPredefinedCursor(horizontal ? Cursor.E_RESIZE_CURSOR
								: Cursor.S_RESIZE_CURSOR));
			} else if (((id == MouseEvent.MOUSE_EXITED) && (mousepressed == null))
						|| ((id == MouseEvent.MOUSE_RELEASED) && (mouseinside != component))) {
				setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
			}
		} else if (("list" == classname) || ("table" == classname)
					|| ("tree" == classname)) {
			// XXX I think this is dead now.
			if (!processScroll(x, y, id, component, part)) {
				if (((id == MouseEvent.MOUSE_PRESSED) || ((id == MouseEvent.MOUSE_DRAGGED)
															&& !shiftdown && !controldown))) {
					// Rectangle view = getRectangle(component, ":view");
					Rectangle port = getRectangle(component, ":port");
					int my = y + port.y - referencey;
					for (Object item = get(component, ":comp"); item != null;) {
						setTimer(0L);
						Rectangle r = getRectangle(item, "bounds");
						if (my < r.y + r.height) {
							if (id == MouseEvent.MOUSE_DRAGGED) { // !!!
								scrollToVisible(component, r.x, r.y, 0,
										r.height);
							} else if ("tree" == classname) {
								int mx = x + port.x - referencex;
								if (mx < r.x) {
									if ((mx >= r.x - block)
										&& (get(item, ":comp") != null)) {
										boolean expanded = getBoolean(item,
												"expanded", true);
										setBoolean(item, "expanded", !expanded,
												true);
										selectItem(component, item, true);
										setLead(component, get(component,
												":lead"), item);
										setFocus(component);
										validate(component);
										invoke(component, item,
												expanded ? "collapse"
														: "expand"); // item
									}
									break;
								}
							}
							if ((id != MouseEvent.MOUSE_DRAGGED)
								|| !getBoolean(item, "selected", false)) {
								if (id != MouseEvent.MOUSE_DRAGGED) {
									if (setFocus(component)) {
										repaint(component, classname, item);
									} // ?
								}
								select(component, item, ("tree" == classname),
										shiftdown, controldown);
								if (clickcount == 2) {
									invoke(component, item, "perform");
								}
							}
							break;
						}
						item = getNextItem(component, item,
								("tree" == classname));
					}
				}
			}
		} else if ("menubar" == classname) {
			Object selected = get(component, "selected");
			if (((id == MouseEvent.MOUSE_ENTERED) || (id == MouseEvent.MOUSE_EXITED))
				&& (part != null)
				&& (selected == null)
				&& getBoolean(part, "enabled", true)) {
				repaint(component, classname, part);
			} else if ((part != null)
						&& ((selected == null) ? (id == MouseEvent.MOUSE_PRESSED)
								: ((id == MouseEvent.MOUSE_ENTERED) || (id == DRAG_ENTERED)))
						&& getBoolean(part, "enabled", true)) {
				// || ((id == MouseEvent.MOUSE_PRESSED) && (insidepart != part))
				set(component, "selected", part);
				popupMenu(component);
				repaint(component, classname, part);
			} else if ((id == MouseEvent.MOUSE_PRESSED) && (selected != null)) {
				closeup();
			} else if (id == MouseEvent.MOUSE_RELEASED) {
				if ((part != insidepart)
					&& ((insidepart == null) || ((insidepart instanceof Object[]) && (getClass(insidepart) != "menu")))) {
					if ((insidepart != null)
						&& getBoolean(insidepart, "enabled", true)) {
						if (getClass(insidepart) == "checkboxmenuitem") {
							changeCheck(insidepart, false);
						} else
							invoke(insidepart, null, "action");
					}
					closeup();
				}
			}
		} else if (":popup" == classname || "popupmenu" == classname) {
			if (part == null) {
				if (id == MouseEvent.MOUSE_RELEASED) {
					if ((insidepart == null)
						|| (getClass(insidepart) != "menu")) {
						if ((insidepart != null)
							&& getBoolean(insidepart, "enabled", true)) {
							if (getClass(insidepart) == "checkboxmenuitem") {
								changeCheck(insidepart, false);
							} else
								invoke(insidepart, null, "action");
						}
						closeup();
					}
				}
			}
			else if (part != "proximity") {
				if (((id == MouseEvent.MOUSE_ENTERED) || (id == DRAG_ENTERED))
					&& getBoolean(part, "enabled", true)) {
					set(component, "selected", part);
					popupMenu(component);
					repaint(component, classname, part);
				} else if (id == MouseEvent.MOUSE_RELEASED) {
					if ((insidepart == null)
						|| (getClass(insidepart) != "menu")) {
						if ((insidepart != null)
							&& getBoolean(insidepart, "enabled", true)) {
							if (getClass(insidepart) == "checkboxmenuitem") {
								changeCheck(insidepart, false);
							} else
								invoke(insidepart, null, "action");
						}
						closeup();
					}
				} else if (((id == MouseEvent.MOUSE_EXITED) || (id == DRAG_EXITED))
							&& getBoolean(part, "enabled", true)) {
					if (getClass(part) != "menu") {
						set(component, "selected", null);
					}
					repaint(component, classname, part);
				}
			}
		} else if ("dialog" == classname) {
			if (part == "header") {
				if (id == MouseEvent.MOUSE_PRESSED) {
					Rectangle bounds = getRectangle(component, "bounds");
					referencex = x - bounds.x;
					referencey = y - bounds.y;
					Object parent = getParent(component);
					if (get(parent, ":comp") != component) { // to front
						removeItemImpl(parent, component);
						insertItemImpl(parent, ":comp", component, 0);
						repaint(component); // to front always...
						setNextFocusable(component, false);
					}
				} else if (id == MouseEvent.MOUSE_DRAGGED) {
					Rectangle bounds = getRectangle(component, "bounds");
					Rectangle parents = getRectangle(getParent(component),
							"bounds");
					int mx = Math.max(0, Math.min(x - referencex,
							parents.width - bounds.width));
					int my = Math.max(0, Math.min(y - referencey,
							parents.height - bounds.height));
					if ((bounds.x != mx) || (bounds.y != my)) {
						// repaint the union of the previous and next bounds
						repaint(component, Math.min(bounds.x, mx), Math.min(
								bounds.y, my), bounds.width
												+ Math.abs(mx - bounds.x),
								bounds.height + Math.abs(my - bounds.y));
						bounds.x = mx;
						bounds.y = my;
					}
				} else if (id == MouseEvent.MOUSE_RELEASED) {
					Rectangle pbounds, bounds;
					String halign, valign;
					boolean xlock, ylock;

					pbounds = getRectangle(getParent(component), "bounds");
					bounds = getRectangle(component, "bounds");
					// System.out.println("P BOUNDS = " + pbounds);
					// System.out.println("NEW BOUNDS = " + bounds);
					if (bounds.x <= 8) {
						xlock = true;
						halign = "left";
					} else if ((pbounds.width - (bounds.x + bounds.width)) <= 8) {
						xlock = true;
						halign = "right";
					} else {
						xlock = false;
						halign = "center";
					}
					setBoolean(component, "xlock", xlock);
					setChoice(component, "halign", halign);
					if (bounds.y <= 8) {
						ylock = true;
						valign = "top";
					} else if ((pbounds.height - (bounds.y + bounds.height)) <= 8) {
						ylock = true;
						valign = "bottom";
					} else {
						ylock = false;
						valign = "center";
					}
					setBoolean(component, "ylock", ylock);
					setChoice(component, "valign", valign);
				}
			}
			// close dialog button
			else if (part == ":closebutton") {
				if (id == MouseEvent.MOUSE_ENTERED
					|| id == MouseEvent.MOUSE_EXITED
					|| id == MouseEvent.MOUSE_PRESSED
					|| id == MouseEvent.MOUSE_RELEASED) {
					if (id == MouseEvent.MOUSE_RELEASED
						&& mouseinside == component)
						invoke(component, null, "close");
					repaint(component);
				}
			} else if (!processScroll(x, y, id, component, part)
						&& (part != null)) {
				if (id == MouseEvent.MOUSE_PRESSED) {
					referencex = x;
					referencey = y;
				} else if (id == MouseEvent.MOUSE_DRAGGED) {
					repaint(component);
					Rectangle bounds = getRectangle(component, "bounds");
					Rectangle pbounds = getRectangle(getParent(component),
							"bounds");
					Rectangle minbounds = getRectangle(component, ":minbounds");
					int oldx = bounds.x, oldy = bounds.y;
					int right = bounds.x + bounds.width;
					int bottom = bounds.y + bounds.height;
					
					// System.out.println("h " + bounds.height + "; " + minbounds);
					minbounds.width = Math.min(pbounds.width, minbounds.width);
					minbounds.height = Math.min(pbounds.height, minbounds.height);
					
					if ((part == ":nw") || (part == ":n") || (part == ":ne")) {
						bounds.y += y - referencey;
						if (bounds.y < 0)
							bounds.y = 0;
						bounds.height = bottom - bounds.y;
						if (bounds.height < minbounds.height) {
							bounds.y = bottom - minbounds.height;
							bounds.height = minbounds.height;
						}
					}
					if ((part == ":ne") || (part == ":e") || (part == ":se")) {
						bounds.width += x - referencex;
						if (bounds.width < minbounds.width)
							bounds.width = minbounds.width;
					}
					if ((part == ":sw") || (part == ":s") || (part == ":se")) {
						bounds.height += y - referencey;
						if (bounds.height < minbounds.height)
							bounds.height = minbounds.height;
					}
					if ((part == ":nw") || (part == ":w") || (part == ":sw")) {
						bounds.x += x - referencex;
						if (bounds.x < 0)
							bounds.x = 0;
						bounds.width = right - bounds.x;
						if (bounds.width < minbounds.width) {
							bounds.x = right - minbounds.width;
							bounds.width = minbounds.width;
						}
					}
					
					// System.out.println("h1 " + bounds.height + "; " + minbounds);

					if (bounds.x + bounds.width > pbounds.width) {
						bounds.width = pbounds.width - bounds.x;
						if (bounds.width < minbounds.width) {
							bounds.width = minbounds.width;
							bounds.x = pbounds.width - bounds.width;
						}
					}
					if (bounds.y + bounds.height > pbounds.height) {
						bounds.height = pbounds.height - bounds.y;
						if (bounds.height < minbounds.height) {
							bounds.height = minbounds.height;
							bounds.y = pbounds.height - bounds.height;
						}
					}

					if ((part == ":nw") || (part == ":n") || (part == ":ne")) {
						referencey += bounds.y - oldy;
					}
					if ((part == ":ne") || (part == ":e") || (part == ":se")) {
						referencex += bounds.x + bounds.width - right;
					}
					if ((part == ":sw") || (part == ":s") || (part == ":se")) {
						referencey += bounds.y + bounds.height - bottom;
					}
					if ((part == ":nw") || (part == ":w") || (part == ":sw")) {
						referencex += bounds.x - oldx;
					}

					// System.out.println("h2 " + bounds.height + "; " + minbounds);
					
					doLayout(component);
					repaint(component);
				} else if (id == MouseEvent.MOUSE_ENTERED) {
					setCursor(Cursor
							.getPredefinedCursor((part == ":n") ? Cursor.N_RESIZE_CURSOR
									: (part == ":ne") ? Cursor.NE_RESIZE_CURSOR
											: (part == ":e") ? Cursor.E_RESIZE_CURSOR
													: (part == ":se") ? Cursor.SE_RESIZE_CURSOR
															: (part == ":s") ? Cursor.S_RESIZE_CURSOR
																	: (part == ":sw") ? Cursor.SW_RESIZE_CURSOR
																			: (part == ":w") ? Cursor.W_RESIZE_CURSOR
																					: Cursor.NW_RESIZE_CURSOR));
				} else if (id == MouseEvent.MOUSE_EXITED) {
					setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
				}
			}
		}

	}

	protected void updateLinkStatus(Object component, boolean entered) {
	}

	/**
	 * Calculate the given point in a component relative to the thinlet desktop
	 * and set as reference value
	 * 
	 * @param component
	 *            a widget
	 * @param x
	 *            reference point relative to the component left edge
	 * @param y
	 *            relative to the top edge
	 */
	void setReference(Object component, int x, int y) {
		referencex = x;
		referencey = y;
		for (; component != null; component = getParent(component)) {
			Rectangle bounds = getRectangle(component, "bounds");
			referencex += bounds.x;
			referencey += bounds.y;

			Rectangle port = getRectangle(component, ":port");
			if (port != null) { // content scrolled
				Rectangle view = getRectangle(component, ":view");
				referencex -= view.x - port.x;
				referencey -= view.y - port.y;
			}
		}
	}

	/**
	 * 
	 */
	void select(Object component, Object row, boolean recursive,
			boolean shiftdown, boolean controldown) {
		String selection = getString(component, "selection", "single");
		Object lead = null;
		if (shiftdown && (selection != "single")
			&& ((lead = get(component, ":lead")) != null)) {
			extend(component, lead, row, recursive);
		} else {
			if (controldown && (selection == "multiple")) {
				setBoolean(row, "selected",
						!getBoolean(row, "selected", false), false);
				repaint(component, null, row);
				invoke(component, row, "action");
				set(component, ":anchor", null);
			} else if (controldown && getBoolean(row, "selected", false)) {
				for (Object item = row; item != null; item = getNextItem(
						component, item, recursive)) {
					if (setBoolean(item, "selected", false, false)) {
						repaint(component, null, item);
					}
				}
				invoke(component, row, "action");
				set(component, ":anchor", null);
			} else {
				selectItem(component, row, recursive);
			}
		}
		setLead(component, (lead != null) ? lead : get(component, ":lead"), row);
	}

	/**
	 * Find the next item after the given
	 * 
	 * @param component
	 *            a list/tree/table widget
	 * @param item
	 *            the next item after this, or the first if null
	 * @param recursive
	 *            true if tree
	 * @return next (or first) item
	 */
	Object getNextItem(Object component, Object item, boolean recursive) {
		if (!recursive) {
			return get(item, ":next");
		}
		Object next = get(item, ":comp");
		if ((next == null) || !getBoolean(item, "expanded", true)) {
			while ((item != component) && ((next = get(item, ":next")) == null)) {
				item = getParent(item);
			}
		}
		return next;
	}

	/**
	 * 
	 */
	void processField(int x, int y, int clickcount, int id, Object component,
			boolean multiline, boolean hidden, int left) {
		if (id == MouseEvent.MOUSE_PRESSED) {
			// + middle=alt paste clipboard content
			setReference(component, 2 + left, 2);
			int mx = x - referencex;
			int my = 0;
			if (!multiline) {
				mx += getInteger(component, ":offset", 0);
			} else {
				Rectangle port = getRectangle(component, ":port");

				mx += port.x - 1;
				my = y - referencey + port.y - 1;
			}
			int caretstart = getCaretLocation(component, mx, my, multiline,
					hidden);
			int caretend = caretstart;
			if (clickcount > 1) {
				String text = getString(component, "text", "");
				while ((caretstart > 0)
						&& ((clickcount == 2) ? Character.isLetterOrDigit(text
								.charAt(caretstart - 1)) : (text
								.charAt(caretstart - 1) != '\n'))) {
					caretstart--;
				}
				while ((caretend < text.length())
						&& ((clickcount == 2) ? Character.isLetterOrDigit(text
								.charAt(caretend))
								: (text.charAt(caretend) != '\n'))) {
					caretend++;
				}
			}
			setInteger(component, "start", caretstart, 0);
			setInteger(component, "end", caretend, 0);
			if (getBoolean(component, "editable")
				&& ((MouseEvent) event).isAltDown()) {
				DataFlavor flavors[];
				Transferable trans;
				String insert;

				trans = getToolkit().getSystemSelection().getContents(null);
				flavors = trans.getTransferDataFlavors();
				try {
					StringBuffer sb = new StringBuffer();
					char bits[] = new char[1024];
					Reader reader;
					int rc;

					reader = flavors[0].getReaderForText(trans);
					while ((rc = reader.read(bits)) != -1) {
						sb.append(bits, 0, rc);
					}
					insert = sb.toString();
				} catch (Exception e) {
					e.printStackTrace();
					insert = clipboard;
				}
				if (insert != null) { // no text on system clipboard nor
										// internal clipboard text
					insert = filter(insert, multiline);
					changeField(component, getString(component, "text"),
							insert, caretstart, caretend, caretstart,
							caretstart);
				}
			}
			setFocus(component);
			validate(component); // caret check only
		} else if (id == MouseEvent.MOUSE_DRAGGED) {
			int mx = x - referencex;
			int my = 0;
			if (!multiline) {
				mx += getInteger(component, ":offset", 0);
			} else {
				Rectangle port = getRectangle(component, ":port");
				mx += port.x - 1;
				my = y - referencey + port.y - 1;
			}
			int dragcaret = getCaretLocation(component, mx, my, multiline,
					hidden);
			if (dragcaret != getInteger(component, "end", 0)) {
				setInteger(component, "end", dragcaret, 0);
				validate(component); // caret check only
			}
		} else if ((id == MouseEvent.MOUSE_ENTERED) && (mousepressed == null)) {
			setCursor(Cursor.getPredefinedCursor(Cursor.TEXT_CURSOR));
		} else if (((id == MouseEvent.MOUSE_EXITED) && (mousepressed == null))
					|| ((id == MouseEvent.MOUSE_RELEASED) && ((mouseinside != component) || (insidepart != null)))) {
			setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
		} else if ((id == MouseEvent.MOUSE_RELEASED)) {
			int start, end;
			String text;

			start = getInteger(component, "start");
			end = getInteger(component, "end");
			if (start != end) {
				text = getString(component, "text");
				try {
					getToolkit().getSystemSelection().setContents(
							new StringSelection(text.substring(Math.min(start,
									end), Math.max(start, end))), null);
				} catch (Exception exc) {
				}
			}
		}
	}

	/**
	 * 
	 */
	int getCaretLocation(Object component, int x, int y, boolean multiline,
			boolean hidden) {
		Font currentfont = (Font) get(component, "font");
		FontMetrics fm = getFontMetrics((currentfont != null) ? currentfont
				: font);
		char[] chars = multiline ? ((char[]) get(component, ":text"))
				: getString(component, "text", "").toCharArray(); // update it
		int linestart = 0;
		if (multiline) {
			int height = fm.getHeight(); // find the line start by y value
			for (int i = 0; (y >= height) && (i < chars.length); i++) {
				if ((chars[i] == '\n') || (chars[i] == 0)) { /* wrap */
					linestart = i + 1;
					y -= height;
				}
			}
		}
		for (int i = linestart; i < chars.length; i++) {
			if ((chars[i] == '\n') || (chars[i] == 0)) {
				return i;
			} /* wrap */
			if (chars[i] == '\b') {
				i++;
				continue;
			}
			int charwidth = fm.charWidth(hidden ? '*' : chars[i]);
			if (x <= (charwidth / 2)) {
				return i;
			}
			x -= charwidth;
		}
		return chars.length;
	}

	/**
	 * 
	 */
	boolean processScroll(int x, int y, int id, Object component, Object part) {
		if ((part == "up") || (part == "down") || (part == "left")
			|| (part == "right")) {
			if ((id == MouseEvent.MOUSE_ENTERED)
				|| (id == MouseEvent.MOUSE_EXITED)
				|| (id == MouseEvent.MOUSE_PRESSED)
				|| (id == MouseEvent.MOUSE_RELEASED)) {
				if (id == MouseEvent.MOUSE_PRESSED) {
					if (processScroll(component, part)) {
						setTimer(300L);
						return true;
					}
				} else {
					if (id == MouseEvent.MOUSE_RELEASED) {
						setTimer(0L);
					}
					repaint(component, null, part);
				}
			}
		} else if ((part == "uptrack") || (part == "downtrack")
					|| (part == "lefttrack") || (part == "righttrack")) {
			if (id == MouseEvent.MOUSE_PRESSED) {
				if (processScroll(component, part)) {
					setTimer(300L);
				}
			} else if (id == MouseEvent.MOUSE_RELEASED) {
				setTimer(0L);
			}
		} else if ((part == "vknob") || (part == "hknob")) {
			if (id == MouseEvent.MOUSE_PRESSED) {
				Rectangle port = getRectangle(component, ":port");
				Rectangle view = getRectangle(component, ":view");
				if (part == "hknob") {
					referencex = x - view.x * (port.width - 2 * block)
									/ view.width;
				} else {
					referencey = y - view.y * (port.height - 2 * block)
									/ view.height;
				}
			} else if (id == MouseEvent.MOUSE_DRAGGED) {
				Rectangle port = getRectangle(component, ":port");
				Rectangle view = getRectangle(component, ":view");
				if (part == "hknob") {
					int viewx = (x - referencex) * view.width
								/ (port.width - 2 * block);
					viewx = Math.max(0, Math
							.min(viewx, view.width - port.width));
					if (view.x != viewx) {
						view.x = viewx;
						repaint(component, null, "horizontal");
					}
				} else { // (part == "vknob")
					int viewy = (y - referencey) * view.height
								/ (port.height - 2 * block);
					viewy = Math.max(0, Math.min(viewy, view.height
														- port.height));
					if (view.y != viewy) {
						view.y = viewy;
						repaint(component, null, "vertical");
					}
				}
			}
		} else if (part == "corner") {
			part = "corner"; // compiler bug
		} else { // ?
			if (id == MouseEvent.MOUSE_PRESSED
				|| id == MouseEvent.MOUSE_DRAGGED) {
				Rectangle port = getRectangle(component, ":port");
				if (port != null) {
					setReference(component, port.x, port.y);
				}
			}
			return false;
		}
		return true;
	}

	/**
	 * 
	 */
	boolean processScroll(Object component, Object part) {
		Rectangle view = getRectangle(component, ":view");
		Rectangle port = ((part == "left") || (part == "up")) ? null
				: getRectangle(component, ":port");
		int dx = 0;
		int dy = 0;
		if (part == "left") {
			dx = -10;
		} else if (part == "lefttrack") {
			dx = -port.width;
		} else if (part == "right") {
			dx = 10;
		} else if (part == "righttrack") {
			dx = port.width;
		} else if (part == "up") {
			dy = -10;
		} else if (part == "uptrack") {
			dy = -port.height;
		} else if (part == "down") {
			dy = 10;
		} else if (part == "downtrack") {
			dy = port.height;
		}
		if (dx != 0) {
			dx = (dx < 0) ? Math.max(-view.x, dx) : Math.min(dx,
					view.width - port.width - view.x);
		} else if (dy != 0) {
			dy = (dy < 0) ? Math.max(-view.y, dy) : Math.min(dy,
					view.height - port.height - view.y);
		} else
			return false;
		if ((dx == 0) && (dy == 0)) {
			return false;
		}
		view.x += dx;
		view.y += dy;
		repaint(component, null, (dx != 0) ? "horizontal" : "vertical");
		return (((part == "left") || (part == "lefttrack")) && (view.x > 0))
				|| (((part == "right") || (part == "righttrack")) && (view.x < view.width
																				- port.width))
				|| (((part == "up") || (part == "uptrack")) && (view.y > 0))
				|| (((part == "down") || (part == "downtrack")) && (view.y < view.height
																				- port.height));
	}

	/**
	 * 
	 */
	boolean processSpin(Object component, Object part) {
		String text = getString(component, "text", "");
		try {
			int itext;

			if (text.length() == 0)
				text = getString(component, "empty", "");
			if (focusformatter != null) {
				try {
					itext = ((Number) focusformatter.getObjectValue(text))
							.intValue();
				} catch (Exception e) {
					itext = 0;
				}
			} else {
				itext = Integer.parseInt(text);
			}
			int step = getInteger(component, "step", 1);
			if ((part == "up") ? (itext + step <= getInteger(component,
					"maximum", Integer.MAX_VALUE))
					: (itext - step >= getInteger(component, "minimum",
							Integer.MIN_VALUE))) {
				String value = String.valueOf((part == "up") ? (itext + step)
						: (itext - step));
				if (focusformatter != null) {
					try {
						Object objvalue;

						objvalue = focusformatter.getObjectValue(value);
						set(component, "text", objvalue);
						String cacheKey = keyForCache("text");
						if (cacheKey != null)
							setReal(component, cacheKey, null);
						value = focusformatter.stringForObjectValue(
								(Thinlet) this, component, objvalue);
					} catch (Exception e) {
						handleException(component, e);
						return true;
					}
				} else {
					setString(component, "text", value, null);
				}
				setInteger(component, "start", value.length(), 0);
				setInteger(component, "end", 0, 0);
				repaint(component, "spinbox", "text");
				invoke(component, null, "action");
				return true;
			}
		} catch (NumberFormatException nfe) {
		}
		return false;
	}

	/**
	 * 
	 */
	boolean findComponent(Object component, int x, int y) {
		if (component == content) {
			mouseinside = insidepart = null;
			mousex = x;
			mousey = y;
		}
		if (!getBoolean(component, "visible", true)) {
			return false;
		}
		if (component == tdi.source) {
			return false;
		}
		Rectangle bounds = getRectangle(component, "bounds");
		if ((bounds == null) || !(bounds.contains(x, y))) {
			if (connector != null) {
				int prox = getInteger(component, "proximity", -1);
				
				if (prox != -1) {
					setPoint(component, ":mouse", -1, -1);
					repaint(component);
				}
			}
			return false;
		}
		{
			int prox = getInteger(component, "proximity", -1);

			if ((prox != -1)
				&& ((x < (bounds.x + prox))
					|| (x > (bounds.x + bounds.width - prox))
					|| (y < (bounds.y + prox)) || (y > (bounds.y
														+ bounds.height - prox)))) {
				insidepart = "proximity";
				if (tdi.source != null)
					return false;
				else
					proxinside = component;
			}
		}
		mouseinside = component;
		x -= bounds.x;
		y -= bounds.y;
		String classname = getClass(component);

		if (connector != null && getParent(connector) == component) {
			lastconnx = connx;
			lastconny = conny;
			connx = x;
			conny = y;
		}
		if (tdi.source == null) {
			if (mousepressed == null) {
				dragx = x;
				dragy = y;
				dragSelectLimit.width = bounds.width;
				dragSelectLimit.height = bounds.height;
			}
		} else {
			tdi.x = x;
			tdi.y = y;
		}
		if ("combobox" == classname) {
			if ((getBoolean(component, "editable", true) || getBoolean(
					component, "selectable", true))
				&& (x <= bounds.width - block)) {
				Image icon = getIcon(component, "icon", null);
				insidepart = ((icon != null) && (x <= 2 + icon.getWidth(this))) ? "icon"
						: null;
			} else {
				insidepart = "down";
			}
		} else if ("bean" == classname) {
			// set(component, ":mousex", new Integer(x));
			// set(component, ":mousey", new Integer(y));
		} else if ("button" == classname || "togglebutton" == classname) {
			if (getInteger(component, "proximity", -1) > 0) {
				setPoint(component, ":mouse", x, y);
				repaint(component);
			}
		} else if (":combolist" == classname) {
			if (!findScroll(component, x, y)) {
				y += getRectangle(component, ":view").y;
				for (Object choice = get(get(component, "combobox"), ":comp"); choice != null; choice = get(
						choice, ":next")) {
					Rectangle r = getRectangle(choice, "bounds");
					if ((r != null) && (y >= r.y) && (y < r.y + r.height)) {
						insidepart = choice;
						break;
					}
				}
			}
		} else if ("textarea" == classname) {
			findScroll(component, x, y);
		} else if ("tabbedpane" == classname) {
			int selected = getInteger(component, "selected", 0);
			int i = 0;
			for (Object tab = get(component, ":comp"); tab != null; tab = get(
					tab, ":next")) {
				Rectangle r = getRectangle(tab, "bounds");
				if (i == selected) {
					Object tabcontent = get(tab, ":comp");
					if ((tabcontent != null)
						&& findComponent(tabcontent, x - r.x, y - r.y)) {
						break;
					}
				}
				if (r.contains(x, y)) {
					insidepart = tab;
					break;
				}
				i++;
			}
		} else if (("panel" == classname) || ("desktop" == classname)
					|| ("dialog" == classname) || ("dropzone" == classname)) {
			if ("dialog" == classname) {
				int titleheight = getInteger(component, ":titleheight", 0);
				boolean resizable = getBoolean(component, "resizable", false);
				if (resizable && (x < 4)) {
					insidepart = (y < block) ? ":nw"
							: (y >= bounds.height - block) ? ":sw" : ":w";
				} else if (resizable && (y < 4)) {
					insidepart = (x < block) ? ":nw"
							: (x >= bounds.width - block) ? ":ne" : ":n";
				} else if (resizable && (x >= bounds.width - 4)) {
					insidepart = (y < block) ? ":ne"
							: (y >= bounds.height - block) ? ":se" : ":e";
				} else if (resizable && (y >= bounds.height - 4)) {
					insidepart = (x < block) ? ":sw"
							: (x >= bounds.width - block) ? ":se" : ":s";
				} else {
					if (y < 4 + titleheight) {
						insidepart = "header";
					}
				}
				// close dialog button
				if (get(component, "close") != null) {
					int buttonX = bounds.width - titleheight - 1;
					int buttonY = 3;
					if (x > buttonX && x < buttonX + titleheight - 2
						&& y > buttonY && y < buttonY + titleheight - 2)
						insidepart = ":closebutton";
				}
			}
			if ((insidepart == null) && !findScroll(component, x, y)) {
				Rectangle port = getRectangle(component, ":port");
				if (port != null) { // content scrolled
					Rectangle view = getRectangle(component, ":view");
					x += view.x - port.x;
					y += view.y - port.y;
				}
				proxinside = null;
				for (Object comp = get(component, ":comp"); comp != null; comp = get(
						comp, ":next")) {
					if (findComponent(comp, x, y)) {
						if (proxinside == null)
							break;
					}
					if (("desktop" == classname)
						&& getBoolean(comp, "modal", false)) {
						insidepart = "modal";
						break;
					} // && dialog
				}
			}
		} else if ("spinbox" == classname) {
			insidepart = (x <= bounds.width - block) ? null
					: ((y <= bounds.height / 2) ? "up" : "down");
		} else if ("splitpane" == classname) {
			Object comp1 = get(component, ":comp");
			if (comp1 != null) {
				if (!findComponent(comp1, x, y)) {
					Object comp2 = get(comp1, ":next");
					if (comp2 != null) {
						findComponent(comp2, x, y);
					}
				}
			}
		} else if ("list" == classname) {
			findScroll(component, x, y);
		} else if ("table" == classname) {
			// check if we are inside the header, then if we have an action on
			// this header or it is a
			// resizable one, do further checks to see if we need to change the
			// column drawing states or
			// even need to invoke the action or to resize the column. We do not
			// need to make this whole
			// block if we are resizing a column right now
			Object header = get(component, "header");
			if (header != null && get(header, ":resizing") == null) {
				boolean isResizable = getBoolean(header, "resizable");
				boolean isSortable = getBoolean(header, "sortable");
				if (isResizable || isSortable) {
					Rectangle view = getRectangle(component, ":view");
					Rectangle port = getRectangle(component, ":port");
					if ((0 < x) && (x < port.width) && (0 < y)
						&& (y < port.y - 1)) {
						int[] columnwidths = (int[]) get(component, ":widths");
						Object column = get(header, ":comp");
						int left = -view.x;
						for (int i = 0; i < columnwidths.length; i++) {
							if (i != 0) {
								column = get(column, ":next");
							}
							int width = (i == columnwidths.length - 1) ? (view.width
																			- left + 2)
									: columnwidths[i];
							if (isResizable
								&& ((x > left + width - 4 && x < left + width) || (i < columnwidths.length - 1
																					&& x >= left
																							+ width && x < left
																											+ width
																											+ 4))) {
								set(header, ":resizecomponent", column);
								break;
							} else {
								set(header, ":resizecomponent", null);
								if (isSortable && getCount(component) > 0
									&& left < x && x < left + width) {
									insidepart = column;
									break;
								}
							}
							left += width;
						}
					} else if (isResizable) {
						set(header, ":resizecomponent", null);
						set(header, ":resizing", null);
					}
				}
			}
			if (insidepart == null && get(header, ":resizecomponent") == null)
				findScroll(component, x, y);
		} else if ("tree" == classname) {
			findScroll(component, x, y);
		} else if ("menubar" == classname) {
			for (Object menu = get(component, ":comp"); menu != null; menu = get(
					menu, ":next")) {
				Rectangle r = getRectangle(menu, "bounds");
				if ((x >= r.x) && (x < r.x + r.width)) {
					insidepart = menu;
					break;
				}
			}
		} else if (":popup" == classname) {
			for (Object menu = get(get(component, "menu"), ":comp"); menu != null; menu = get(
					menu, ":next")) {
				if (!getBoolean(menu, "visible", true))
					continue;
				Rectangle r = getRectangle(menu, "bounds");
				if ((y >= r.y) && (y < r.y + r.height)) {
					insidepart = menu;
					break;
				}
			}
		}
		return true;
	}

	/**
	 * @param component
	 *            a scrollable widget
	 * @param x
	 *            point x location
	 * @param y
	 *            point y location
	 * @return true if the point (x, y) is inside scroll-control area
	 *         (scrollbars, corners, borders), false otherwise (vievport,
	 *         header, or no scrollpane)
	 */
	boolean findScroll(Object component, int x, int y) {
		Rectangle port = getRectangle(component, ":port");
		if ((port == null) || port.contains(x, y)) {
			return false;
		}
		Rectangle view = getRectangle(component, ":view");
		Rectangle horizontal = getRectangle(component, ":horizontal");
		Rectangle vertical = getRectangle(component, ":vertical");
		if ((horizontal != null) && horizontal.contains(x, y)) {
			findScroll(x - horizontal.x, horizontal.width, port.width, view.x,
					view.width, true);
		} else if ((vertical != null) && vertical.contains(x, y)) {
			findScroll(y - vertical.y, vertical.height, port.height, view.y,
					view.height, false);
		} else {
			insidepart = "corner";
		}
		return true;
	}

	/**
	 * @param p
	 *            x or y relative to the scrollbar begin
	 * @param size
	 *            scrollbar width or height
	 * @param portsize
	 *            viewport width or height
	 * @param viewp
	 *            view x or y
	 * @param viewsize
	 *            view width or height
	 * @param horizontal
	 *            if true horizontal, vertical otherwise
	 */
	void findScroll(int p, int size, int portsize, int viewp, int viewsize,
			boolean horizontal) {
		if (p < block) {
			insidepart = horizontal ? "left" : "up";
		} else if (p > size - block) {
			insidepart = horizontal ? "right" : "down";
		} else {
			int track = size - 2 * block;
			if (track < 10) {
				insidepart = "corner";
				return;
			} // too small
			int knob = Math.max(track * portsize / viewsize, 10);
			int decrease = viewp * (track - knob) / (viewsize - portsize);
			if (p < block + decrease) {
				insidepart = horizontal ? "lefttrack" : "uptrack";
			} else if (p < block + decrease + knob) {
				insidepart = horizontal ? "hknob" : "vknob";
			} else {
				insidepart = horizontal ? "righttrack" : "downtrack";
			}
		}
	}

	/**
	 * mnemonic (e.g. Alt-X): - check: label, button, checkbox, togglebutton,
	 * menubar menus, tabbedpane tabs - path: panel, desktop, dialog, splitpane
	 * components, tabbedpane selected component accelerator (e.g. Ctrl-Shift-X,
	 * F4): - check: menuitem, checkboxmenuitem - path: see above, and menubar,
	 * and menu items menubar F10: check menubar only button enter, escape:
	 * check button only
	 * 
	 * @param component
	 * @param parent
	 *            check upwards if true
	 * @param checked
	 *            this leaf is already checked
	 * @param mnemonic
	 * @return true if the char was consumed
	 */
	boolean checkMnemonic(Object component, boolean parent, Object checked,
			int keycode, int modifiers) {
		if ((component == null) || !getBoolean(component, "visible", true)
			|| !getBoolean(component, "enabled", true)) { // + enabled comp in
															// disabled parent
			return false;
		}
		String classname = getClass(component);
		if ("label" == classname) {
			if (hasMnemonic(component, keycode, modifiers)) {
				Object labelfor = get(component, "for");
				if (labelfor != null) {
					requestFocus(labelfor);
					return true;
				}
			}
		} else if ("button" == classname) {
			if (((modifiers == 0) && (((keycode == KeyEvent.VK_ENTER) && (get(
					component, "type") == "default")) || ((keycode == KeyEvent.VK_ESCAPE) && (get(
					component, "type") == "cancel"))))
				|| hasMnemonic(component, keycode, modifiers)) {
				invoke(component, null, "action");
				repaint(component);
				return true;
			}
		} else if (("checkbox" == classname) || ("togglebutton" == classname)) {
			if (hasMnemonic(component, keycode, modifiers)) {
				changeCheck(component, true);
				repaint(component);
				return true;
			}
		} else if ("menubar" == classname) {
			for (Object menu = get(component, ":comp"); menu != null; menu = get(
					menu, ":next")) {
				if (hasMnemonic(menu, keycode, modifiers)
					|| ((modifiers == 0) && (keycode == KeyEvent.VK_F10))) {
					closeup();
					set(component, "selected", menu);
					popupMenu(component);
					repaint(component, "menubar", menu);
					return true;
				}
			}
		} else if (("menuitem" == classname)
					|| ("checkboxmenuitem" == classname)) {
			if (hasAccelerator(component, keycode, modifiers)) {
				invoke(component, null, "action");
			}
		} else if ("tabbedpane" == classname) {
			int selected = getInteger(component, "selected", 0);
			int i = 0;
			for (Object tab = get(component, ":comp"); tab != null; tab = get(
					tab, ":next")) {
				if (hasMnemonic(tab, keycode, modifiers)) {
					if (selected != i) {
						setInteger(component, "selected", i, 0);
						repaint(component);
						invoke(component, getItem(component, i), "action");
					}
					return true;
				}
				i++;
			}
			Object comp = get(getItem(component, selected), ":comp");
			if ((comp != null) && (comp != checked)
				&& checkMnemonic(comp, false, null, keycode, modifiers)) {
				return true;
			}
		}
		// check subcomponents
		if (("panel" == classname) || ("desktop" == classname)
			|| ("dialog" == classname) || ("splitpane" == classname)
			|| ("menubar" == classname) || ("menu" == classname) || ("dropzone" == classname)) {
			for (Object comp = get(component, ":comp"); comp != null; comp = get(
					comp, ":next")) {
				if ((comp != checked)
					&& checkMnemonic(comp, false, null, keycode, modifiers)) {
					return true;
				}
			}
		}
		// check parent
		if (parent
			&& (("dialog" != classname) || !getBoolean(component, "modal",
					false))) {
			if (checkMnemonic(getParent(component), true,
					("tab" == classname) ? checked : component, keycode,
					modifiers)) {
				return true;
			}
		}
		return false;
	}

	/**
	 * Pop up the list of choices for the given combobox
	 * 
	 * @param combobox
	 * @return the created combolist
	 */
	Object popupCombo(Object combobox) {
		// combobox bounds relative to the root desktop
		int combox = 0, comboy = 0, combowidth = 0, comboheight = 0;
		for (Object comp = combobox; comp != content; comp = getParent(comp)) {
			Rectangle r = getRectangle(comp, "bounds");
			if (r.width < 0) {
				r.width = Math.abs(r.width);
				doLayout(comp);
			}
			combox += r.x;
			comboy += r.y;
			Rectangle view = getRectangle(comp, ":view");
			if (view != null) {
				combox -= view.x;
				comboy -= view.y;
				Rectangle port = getRectangle(comp, ":port");
				combox += port.x;
				comboy += port.y;
			}
			if (comp == combobox) {
				combowidth = r.width;
				comboheight = r.height;
			}
		}
		// :combolist -> combobox and combobox -> :combolist
		Object combolist = createImpl(":combolist");
		set(combolist, "combobox", combobox);
		set(combobox, ":combolist", combolist);
		// add :combolist to the root desktop and set the combobox as popupowner
		popupowner = combobox;
		insertItemImpl(content, ":comp", combolist, 0);
		// lay out choices verticaly and calculate max width and height sum
		int pw = 0;
		int ph = 0;
		for (Object item = get(combobox, ":comp"); item != null; item = get(
				item, ":next")) {
			Dimension d = getSize(item, 8, 4);
			setRectangle(item, "bounds", 0, ph, d.width, d.height);
			pw = Math.max(pw, d.width);
			ph += d.height;
		}
		// set :combolist bounds
		int listy = 0, listheight = 0;
		int bellow = getRectangle(content, "bounds").height - comboy
						- comboheight - 1;
		if ((ph + 2 > bellow) && (comboy - 1 > bellow)) { // popup above
															// combobox
			listy = Math.max(0, comboy - 1 - ph - 2);
			listheight = Math.min(comboy - 1, ph + 2);
		} else { // popup bellow combobox
			listy = comboy + comboheight + 1;
			listheight = Math.min(bellow, ph + 2);
		}
		setRectangle(combolist, "bounds", combox, listy, combowidth, listheight);
		layoutScroll(combolist, pw, ph, 0, 0, 0, 0, true, 0);
		repaint(combolist);
		// hover the selected item
		int selected = getInteger(combobox, "selected", -1);
		setInside(combolist, (selected != -1) ? getItem(combobox, selected)
				: null, true);
		return combolist;
	}

	/**
	 * @param component
	 *            menubar or :popup
	 * @return the created popupmenu
	 */
	Object popupMenu(Object component) {
		Object popup = get(component, ":popup"); // first :popup child
		Object selected = get(component, "selected"); // selected menu in of
														// the component
		if (popup != null) { // remove its current :popup
			if (get(popup, "menu") == selected) {
				return null;
			} // but the currect one
			set(popup, "selected", null);
			set(popup, "menu", null);
			repaint(popup);
			removeItemImpl(content, popup);
			set(popup, ":parent", null);
			set(component, ":popup", null);
			checkLocation(popup);
			popupMenu(popup); // remove recursively
		}
		// pop up the selected menu only
		if ((selected == null) || (getClass(selected) != "menu")) {
			return null;
		}
		// create the :popup, :popup.menu -> menu,
		// menubar|:popup.:popup -> :popup, menubar|:popup.selected -> menu
		popup = createImpl(":popup");
		set(popup, "menu", selected);
		set(component, ":popup", popup);
		insertItemImpl(content, ":comp", popup, 0);
		// calculates the bounds of the previous menubar/:popup relative to the
		// root desktop
		int menux = 0, menuy = 0, menuwidth = 0, menuheight = 0;
		for (Object comp = component; comp != content; comp = getParent(comp)) {
			Rectangle r = getRectangle(comp, "bounds");
			menux += r.x;
			menuy += r.y;
			Rectangle view = getRectangle(comp, ":view");
			if (view != null) {
				menux -= view.x;
				menuy -= view.y;
				Rectangle port = getRectangle(comp, ":port");
				menux += port.x;
				menuy += port.y;
			}
			if (comp == component) {
				menuwidth = r.width;
				menuheight = r.height;
			}
		}
		// set :popup bounds
		Rectangle menubounds = getRectangle(selected, "bounds");
		boolean menubar = ("menubar" == getClass(component));
		if (menubar) {
			popupowner = component;
		}
		popup(selected, popup, menubar ? (("bottom" != get(component,
				"placement")) ? 'D' : 'U') : 'R',
				menubar ? (menux + menubounds.x) : menux, menuy + menubounds.y,
				menubar ? menubounds.width : menuwidth, menubar ? menuheight
						: menubounds.height, menubar ? 1 : 3);
		return popup;
	}

	/**
	 * @param popupmenu
	 */
	void popupPopup(Object popupmenu, int x, int y) {
		// :popup.menu -> popupmenu, popupmenu.:popup -> :popup
		Object popup = createImpl(":popup");
		set(popup, "menu", popupmenu);
		set(popupmenu, ":popup", popup);
		// add :popup to the root desktop and set the combobox as popupowner
		popupowner = popupmenu;
		insertItemImpl(content, ":comp", popup, 0);
		// lay out
		popup(popupmenu, popup, 'D', x, y, 0, 0, 0);
		// invoke menushown listener
		invoke(popupmenu, null, "menushown"); // TODO before
		findComponent(content, x, y);
		repaint(content);
	}

	/**
	 * Lays out a popupmenu
	 * 
	 * @param menu
	 *            menubar's menu, menu's menu, or component's popupmenu
	 *            including items
	 * @param popup
	 *            created popupmenu
	 * @param direction
	 *            'U' for up, 'D' for down, and 'R' for right
	 * @param x
	 *            menu's x location relative to the desktop
	 * @param y
	 *            menu's y location
	 * @param width
	 *            menu's width, or zero for popupmenu
	 * @param height
	 *            menu's height
	 * @param offset
	 *            inner padding relative to the menu's bounds
	 */
	void popup(Object menu, Object popup, char direction, int x, int y,
			int width, int height, int offset) {
		int pw = 0;
		int ph = 0;
		for (Object item = get(menu, ":comp"); item != null; item = get(item,
				":next")) {
			if (!getBoolean(item, "visible", true))
				continue;
			String itemclass = getClass(item);
			Dimension d = (itemclass == "separator") ? new Dimension(1, 1)
					: getSize(item, 8, 4);
			if (itemclass == "checkboxmenuitem") {
				d.width = d.width + block + 3;
				d.height = Math.max(block, d.height);
			} else if (itemclass == "menu") {
				d.width += block;
			}
			String accelerator = getAccelerator(item); // add accelerator width
			if (accelerator != null) {
				d.width += 4 + getFontMetrics(font).stringWidth(accelerator); // TODO
																				// font,
																				// height
																				// and
																				// gap
			}
			setRectangle(item, "bounds", 1, 1 + ph, d.width, d.height);
			pw = Math.max(pw, d.width);
			ph += d.height;
		}
		pw += 2;
		ph += 2; // add border widths
		// set :popup bounds
		Rectangle desktop = getRectangle(content, "bounds");
		if (direction == 'R') {
			x += ((x + width - offset + pw > desktop.width) && (x >= pw
																		- offset)) ? (offset - pw)
					: (width - offset);
			if ((y + ph > desktop.height) && (ph <= y + height)) {
				y -= ph - height;
			}
		} else {
			boolean topspace = (y >= ph - offset); // sufficient space above
			boolean bottomspace = (desktop.height - y - height >= ph - offset);
			y += ((direction == 'U') ? (topspace || !bottomspace)
					: (!bottomspace && topspace)) ? (offset - ph)
					: (height - offset);
		}
		setRectangle(popup, "bounds", Math.max(0, Math.min(x, desktop.width
																- pw)), Math
				.max(0, Math.min(y, desktop.height - ph)), pw, ph);
		repaint(popup);
	}

	/**
	 * @param item
	 *            //TODO can be scrollbar string
	 */
	void closeCombo(Object combobox, Object combolist, Object item) {
		if ((item != null) && getBoolean(item, "enabled", true)) {
			String text = getString(item, "text", "");
			set(combobox, "text", text); // if editable
			putProperty(combobox, "i18n.text", null); // for I18N
			setInteger(combobox, "start", text.length(), 0);
			setInteger(combobox, "end", 0, 0);
			set(combobox, "icon", get(item, "icon"));
			validate(combobox);
			setInteger(combobox, "selected", getIndex(combobox, item), -1);
			invoke(combobox, item, "action");
		}
		set(combolist, "combobox", null);
		set(combobox, ":combolist", null);
		removeItemImpl(content, combolist);
		repaint(combolist);
		set(combolist, ":parent", null);
		popupowner = null;
		checkLocation(combolist);
	}

	/**
	 * 
	 */
	void closeup() {
		if (popupowner != null) {
			String classname = getClass(popupowner);
			if ("menubar" == classname) {
				set(popupowner, "selected", null);
				popupMenu(popupowner);
				repaint(popupowner); // , selected
			} else if ("combobox" == classname) {
				closeCombo(popupowner, get(popupowner, ":combolist"), null);
			} else { // "popupmenu"
				popupMenu(popupowner);
			}
			popupowner = null;
		}
	}

	/**
	 * 
	 */
	private synchronized void showTip() {
		String text = null;
		tooltipowner = null;
		String classname = getClass(mouseinside);
		if ((classname == "tabbedpane") || (classname == "menubar")
			|| (classname == ":popup")) {
			if (insidepart != null) {
				text = getString(insidepart, "tooltip", null);
			}
		} else if (classname == ":combolist") {
			if (insidepart instanceof Object[]) {
				text = getString(insidepart, "tooltip", null);
			}
		} else if (insidepart == "proximity") {
			text = getString(mouseinside, "proximitytooltip", null);
			tooltipowner = mouseinside;
		}
		// TODO list table tree
		if (text == null) {
			text = getString(mouseinside, "tooltip", null);
		} else if (tooltipowner == null) {
			tooltipowner = insidepart;
		}
		if (text != null) {
			Font ttFont = null;

			if (insidepart != "proximity")
				ttFont = (Font) get(mouseinside, "tooltipfont");

			if (ttFont == null)
				ttFont = this.font;
			FontMetrics fm = getFontMetrics(ttFont);
			int width = fm.stringWidth(text) + 4;
			int height = fm.getAscent() + fm.getDescent() + 4;
			if (tooltipowner == null) {
				tooltipowner = mouseinside;
			}
			Rectangle bounds = getRectangle(content, "bounds");
			int tx = Math.max(0, Math.min(mousex + 10, bounds.width - width));
			int ty = Math.max(0, Math.min(mousey + 10, bounds.height - height));
			setRectangle(tooltipowner, ":tooltipbounds", tx, ty, width, height);
			repaint(tx, ty, width, height);
		}
	}

	/**
	 * 
	 */
	private synchronized void hideTip() {
		setTimer(0);
		if (tooltipowner != null) {
			Rectangle bounds = getRectangle(tooltipowner, ":tooltipbounds");
			set(tooltipowner, ":tooltipbounds", null);
			tooltipowner = null;
			if (bounds != null)
				repaint(bounds.x, bounds.y, bounds.width, bounds.height);
		}
	}

	abstract void paint(Graphics g, int clipx, int clipy, int clipwidth,
			int clipheight, Object component, boolean enabled);

}
