package thinlet;

import java.awt.Image;
import java.awt.Rectangle;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.List;
import java.util.StringTokenizer;

public abstract class ThinletInvoke extends ThinletModel {
	transient Object connector;
	
	transient List connectorList;
	
	transient Rectangle connectBounds = new Rectangle();

	transient Object connectee;

	transient int connx, lastconnx;

	transient int conny, lastconny;

	class ThinletMethod {
		Object component;

		String value;

		Object root;

		Object handler;

		Method method;

		Object args[];

		public String toString() {
			return "ThinletMethod[component=" + this.component + "; value="
					+ this.value + "; root=" + this.root + "; handler="
					+ this.handler + "; method=" + this.method + "; args="
					+ this.args + "]";
		}
	}

	/**
	 * 
	 */
	public boolean invoke(Object component, Object part, String event) {
		Object method = get(component, event);
		if (method != null) {
			if (method instanceof ThinletMethod) {
				/* Regular method. */
				invokeImpl((ThinletMethod) method, part);
				return true;
			} else {
				/* A flow "method". */
				if ("flow" == getClass(method)) {
					/* Always start flows in an empty state. */
					if ("".equals(getString(method, "state", ""))) {
						invoke(component, method);
					} else {
						setString(method, "state", "");
					}
				} else {
					invoke(component, method);
				}
				return true;
			}
		}
		return false;
	}
	
	public boolean invoke(Object component, String event) {
		return this.invoke(component, null, event);
	}

	/**
	 * Invoke all items in a flow.
	 * 
	 * @param component
	 *            The component that invoked this flow.
	 * @param flow
	 *            The XML flow to execute.
	 * @return True if the caller should stop processing.
	 */
	public boolean invokeAll(Object component, Object flow) {
		boolean retval = false;
		Object items[];
		int lpc;

		items = getItems(flow);
		for (lpc = 0; (lpc < items.length) && !retval; lpc++) {
			retval = invoke(component, items[lpc]);
		}
		return retval;
	}

	/**
	 * The "flow" confirmation dialog object.
	 */
	transient Object confirmDialog;

	/**
	 * Find the parent "flow" for the given object.
	 * 
	 * @param obj
	 *            An XML executable object.
	 * @return The parent "flow" the given object.
	 */
	public Object findFlow(Object obj) {
		Object retval;

		for (retval = obj; "flow" != getClass(retval); retval = getParent(retval)) {
		}
		return retval;
	}

	/**
	 * Invoke an XML executable object.
	 * 
	 * @param component
	 *            The component that invoked this operation.
	 * @param flow
	 *            The XML object to execute.
	 */
	public boolean invoke(Object component, Object flow) {
		String classname = getClass(flow);
		boolean retval = false;

		if ("flow" == classname) {
			retval = invokeAll(component, flow);
		} else if ("action" == classname) {
			String state;

			if ((state = getString(flow, "state")) != null) {
				setString(findFlow(flow), "state", state);
			}
			invoke(flow, null, "method");
		} else if ("cond" == classname) {
			String fstate, cstate;

			fstate = getString(findFlow(flow), "state", "");
			cstate = getString(flow, "state", "");

			if ((fstate == cstate)
				|| ((fstate != null) && fstate.equals(cstate))) {
				Object method = get(flow, "test");
				boolean value = true;

				if (method != null) {
					Object rc;

					rc = invokeImpl((ThinletMethod) method, null);
					value = ((Boolean) rc).booleanValue();
				}
				if (value) {
					invokeAll(component, flow);
					retval = true;
				}
			}
		} else if ("confirm" == classname) {
			String title = getString(flow, "title", " ");
			String text = getString(flow, "text");
			String expl = getString(flow, "explanation");
			Object panel, comp, options[];
			int lpc;

			confirmDialog = createConfirmDialog(title, text, expl);

			panel = create("panel");
			setInteger(panel, "gap", 4);
			options = getItems(flow);
			for (lpc = 0; lpc < options.length; lpc++) {
				if (((lpc == 0) && ((options.length == 2) || (options.length == 1)))
					|| ((lpc == 1) && (options.length == 3))) {
					comp = create("label");
					setInteger(comp, "weightx", 2);
					add(panel, comp);
				}
				text = getString(options[lpc], "text");
				comp = create("button");
				setString(comp, "text", text);
				set(comp, "action", options[lpc]);
				switch (options.length) {
				case 1:
					setColor(comp, "background", c_default_button);
					setBoolean(comp, "usegradient", true);
					setChoice(comp, "type", "default");
					break;
				case 2:
					setChoice(comp, "type",
							(lpc == 0) ? "default" : "cancel");
					if (lpc == 0) {
						setColor(comp, "background", c_default_button);
						setBoolean(comp, "usegradient", true);
					}
					break;
				case 3:
					if (lpc != 0) {
						setChoice(comp, "type",
								(lpc == 1) ? "default" : "cancel");
						if (lpc == 1) {
							setColor(comp, "background", c_default_button);
							setBoolean(comp, "usegradient", true);
						}
					}
					break;
				}
				add(panel, comp);
			}
			add(confirmDialog, panel);

			add(confirmDialog);

			retval = true;
		} else if ("option" == classname) {
			remove(confirmDialog);
			retval = invokeAll(component, flow);
			confirmDialog = null;
		} else {
			throw new IllegalArgumentException(classname);
		}
		return retval;
	}
	
	public static int computeRowCount(String text, int cols) {
		int index, retval = 0;
		
		for (index = text.indexOf('\n'); index != -1; index = text.indexOf(
				'\n', index + 1)) {
			retval += 1;
		}

		retval += (text.length() + cols - 1) / cols;
		
		return retval;
	}

	/**
	 * Create the main bits of a confirmation dialog box.
	 * 
	 * @param title
	 *            The title of the dialog.
	 * @param text
	 *            The text message to put inside the dialog.
	 * @return The constructed dialog component.
	 */
	protected Object createConfirmDialog(String title, String text, String expl) {
		int cols = 32, rows = 0;
		Object comp, retval;

		rows = computeRowCount(text, cols);

		retval = create("dialog");
		setString(retval, "text", title);
		setBoolean(retval, "modal", true);
		setInteger(retval, "columns", 1);
		setInteger(retval, "gap", 4);
		setInteger(retval, "left", 4);
		setInteger(retval, "top", 4);
		setInteger(retval, "right", 4);
		setInteger(retval, "bottom", 4);
		comp = create("textarea");
		setString(comp, "text", text);
		setInteger(comp, "columns", cols);
		setInteger(comp, "rows", rows);
		setBoolean(comp, "editable", false);
		setBoolean(comp, "border", false);
		setBoolean(comp, "wrap", true);
		setFont(comp, "font", "bold");
		add(retval, comp);

		if (expl != null) {
			cols = 32;
			rows = computeRowCount(expl, cols);
			comp = create("textarea");
			setString(comp, "text", expl);
			setInteger(comp, "columns", cols);
			setInteger(comp, "rows", rows);
			setBoolean(comp, "editable", false);
			setBoolean(comp, "border", false);
			setBoolean(comp, "wrap", true);
			setFont(comp, "font", String.valueOf(font.getSize() - 2));
			add(retval, comp);
		}

		comp = create("separator");
		add(retval, comp);
		return retval;
	}

	/**
	 * 
	 */
	Object invokeImpl(ThinletMethod tm, Object part) {
		Object[] args = (tm.args.length > 2) ? new Object[tm.args.length / 3]
				: null;
		Object retval = null;

		if (args != null)
			for (int i = 0; i < args.length; i++) {
				Object target = tm.args[3 * i];
				if ("thinlet" == target) {
					args[i] = this;
				} else if ("constant" == target) {
					args[i] = tm.args[3 * i + 1];
				} else if ("connector" == target) {
					args[i] = connector;
				} else if ("connectee" == target) {
					args[i] = connectee;
				} else {
					if ("item" == target) {
						target = part;
					}
					Object parametername = tm.args[3 * i + 1];
					if (parametername == null) {
						args[i] = target;
						// args[i] = new Widget(this, target);
					} else {
						args[i] = (target != null) ? get(target, parametername)
								: null;
						if (args[i] == null) {
							args[i] = tm.args[3 * i + 2];
						}
					}
				}
			}
		try {
			retval = tm.method.invoke(tm.handler, args);
		} catch (InvocationTargetException ite) {
			handleException(tm.component, ite.getTargetException());
		} catch (Throwable throwable) {
			handleException(tm.component, throwable);
		}
		return retval;
	}

	/**
	 * Sets a new event handler method for a component
	 * 
	 * @param component
	 *            the target component
	 * @param key
	 *            the key name of the parameter (e.g. <i>action</i>)
	 * @param value
	 *            the method name and parameters (e.g. <i>foo(this, this.text,
	 *            mybutton, mybutton.enabled)</i> for <i>public void foo(Object
	 *            component, String text, Object mybutton, boolean enabled)</i>)
	 * @param root
	 *            the search starting component for name components in the
	 *            arguments
	 * @param handler
	 *            the target event handler object including the method
	 * @throws java.lang.IllegalArgumentException
	 */
	public void setMethod(Object component, String key, String value,
			Object root, Object handler) {
		key = (String) getDefinition(getClass(component), key, "method")[1];
		ThinletMethod method = getMethod(component, value, root, handler);
		set(component, key, method);
	}

	ThinletMethod getMethod(Object component, String value, Object root,
			Object handler) {
		ThinletMethod retval;

		retval = new ThinletMethod();
		retval.component = component;
		retval.value = value;
		retval.root = root;
		retval.handler = handler;
		getMethod(retval);
		return retval;
	}

	/**
	 * @return an object list including as follows: - handler object, - method, -
	 *         list of parameters including 3 values: - ("thinlet", null, null)
	 *         for the single thinlet component, - (target component, null,
	 *         null) for named widget as parameter, e.g. mybutton, - (target,
	 *         parameter name, default value) for a widget's given property,
	 *         e.g. mylabel.enabled, - ("item", null, null) for an item of the
	 *         target component as parameter, e.g. tree node, - ("item",
	 *         parameter name, default value) for the item's given property e.g.
	 *         list item's text.
	 */
	void getMethod(ThinletMethod tm) {
		StringTokenizer st = new StringTokenizer(tm.value, "(, \r\n\t)");
		String methodname = st.nextToken();
		int mdot;
		while ((mdot = methodname.indexOf('.')) != -1) {
			String fieldName = methodname.substring(0, mdot);
			methodname = methodname.substring(mdot + 1);

			tm.handler = tkv.getKeyValue(tm.handler, fieldName);
		}
		int n = st.countTokens();
		tm.args = new Object[3 * n];
		Class[] parametertypes = (n > 0) ? new Class[n] : null;
		for (int i = 0; i < n; i++) {
			String arg = st.nextToken();
			if ("thinlet".equals(arg)) {
				tm.args[3 * i] = "thinlet"; // the target component
				parametertypes[i] = Thinlet.class;
			} else if ("connector".equals(arg)) {
				tm.args[3 * i] = "connector";
				parametertypes[i] = Object.class;
			} else if ("connectee".equals(arg)) {
				tm.args[3 * i] = "connectee";
				parametertypes[i] = Object.class;
			} else if ((arg.length() > 1) && // constant string value
						(arg.charAt(0) == '\'')
						&& (arg.charAt(arg.length() - 1) == '\'')) {
				/* Its a constant string. */
				tm.args[3 * i] = "constant";
				tm.args[3 * i + 1] = arg.substring(1, arg.length() - 1).intern();
				parametertypes[i] = String.class;
			} else {
				int dot = arg.indexOf('.');
				String compname = (dot == -1) ? arg : arg.substring(0, dot);
				Object comp = null;
				String classname = null;
				if ("item".equals(compname)) {
					comp = "item";
					classname = getChildClass(getClass(tm.component));
				} else if ("this".equals(compname)) {
					comp = tm.component;
					classname = getClass(comp);
				} else if ((comp = find(tm.root, compname)) != null) {
					classname = getClass(comp);
				} else {
					try { // maybe constant number
						if (arg
								.regionMatches(true, arg.length() - 1, "F", 0,
										1)) { // float
							tm.args[3 * i + 1] = Float.valueOf(arg.substring(0,
									arg.length() - 1));
							parametertypes[i] = Float.TYPE;
						} else if (arg.regionMatches(true, arg.length() - 1,
								"L", 0, 1)) { // long
							tm.args[3 * i + 1] = Long.valueOf(arg.substring(0,
									arg.length() - 1));
							parametertypes[i] = Long.TYPE;
						} else if (dot != -1) { // double
							tm.args[3 * i + 1] = Double.valueOf(arg);
							parametertypes[i] = Double.TYPE;
						} else { // integer
							tm.args[3 * i + 1] = Integer.valueOf(arg);
							parametertypes[i] = Integer.TYPE;
						}
						tm.args[3 * i] = "constant";
						continue;
					} catch (NumberFormatException nfe) { // widget's name not
															// found nor
															// constant
						throw new IllegalArgumentException("unknown " + arg);
					}
				}
				tm.args[3 * i] = comp; // the target component
				if (dot == -1) {
					parametertypes[i] = Object.class; // Widget.class
				} else {
					Object[] definition = getDefinition(classname, arg
							.substring(dot + 1), null);
					tm.args[3 * i + 1] = definition[1]; // parameter name, e.g.
														// enabled
					tm.args[3 * i + 2] = definition[3]; // default value, e.g.
														// Boolean.TRUE
					Object fieldclass = definition[0];
					if ((fieldclass == "string") || (fieldclass == "choice")) {
						parametertypes[i] = String.class;
					} else if (fieldclass == "boolean") {
						parametertypes[i] = Boolean.TYPE;
					} else if (fieldclass == "integer") {
						parametertypes[i] = Integer.TYPE;
					} else if (fieldclass == "icon") {
						parametertypes[i] = Image.class;
					} else if (fieldclass == "object") {
						if (get(tm.component, "formatter") == null) {
							parametertypes[i] = String.class;
						} else {
							parametertypes[i] = Object.class;
						}
					} else
						throw new IllegalArgumentException((String) fieldclass);
				}
			}
		}
		try {
			tm.method = tm.handler.getClass().getMethod(methodname,
					parametertypes);
		} catch (Exception exc) {
			exc.printStackTrace();
			throw new IllegalArgumentException(tm.value + " "
												+ exc.getMessage());
		}
	}

	/**
	 * 
	 */
	abstract boolean requestFocus(Object component); // TODO: SJC: remove??
}
