package thinlet;

import java.awt.Component;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Image;
import java.awt.Rectangle;

public abstract class ThinletLayout extends ThinletPaintMgr {
	/**
	 * 
	 */
	protected void doLayout(Object component) {
		String classname = getClass(component);
		if (":combolist" == classname) {
			Object cb;

			cb = get(component, "combobox");
			closeCombo(cb, component, null);
			popupCombo(cb);
		} else if ("combobox" == classname) {
			int selected = getInteger(component, "selected", -1);

			if (getBoolean(component, "editable", true)) {
				Object combolist = get(component, ":combolist");
				Image icon = getIcon(component, "icon", null);
				String style;
				int dx = 0;

				style = getString(component, "style", "normal");
				if ("round" == style)
					dx = block;
				layoutField(component, block, false, (icon != null) ? icon
						.getWidth(this) + dx
						: dx);
				if (combolist != null) {
					if (selected == -1) {
						// System.out.println("dl closecombo");
						closeCombo(component, combolist, null);
					} else {
						Object choice = getItem(component, selected);
						Object lead = get(combolist, ":lead");

						/*
						 * Make sure the selection is synced with the display.
						 */
						if ((lead != null) && (choice != lead)) {
							setInside(combolist, choice, true);
						}
					}
				} else if (selected != -1) {
					Object choice = getItem(component, selected);
					String text = getString(choice, "text", "");
					
					if (!text.equals(getString(component, "text")) &&
							!text.equals(getString(component, "alt"))) {
						setInteger(component, "selected", -1);
					}
				}
				else {
					String text = getString(component, "text", "");
					String alt = getString(component, "alt", "");
					
					if (!"".equals(text)) {
						Object choice;
						
						selected = 0;
						for (choice = get(component, ":comp"); choice != null; choice = get(choice, ":next")) {
							String ctext = getString(choice, "text", "");
							
							if (ctext.equals(text) || ctext.equals(alt)) {
								break;
							}
							selected += 1;
						}
						
						if (choice != null) {
							setInteger(component, "selected", selected);
						}
					}
				}
			} // set editable -> validate (overwrite textfield repaint)
			else if (selected != -1) {
				Object choice = getItem(component, selected), icon;

				set(component, "text", get(choice, "text"));
				if ((icon = get(choice, "icon")) != null)
					set(component, "icon", icon);
			}
		} else if ("label" == classname) {
			Rectangle bounds = getRectangle(component, "bounds");

			if (bounds == null) {
				Dimension dim = getPreferredSize(component);

				setRectangle(component, "bounds", 0, 0, dim.width, dim.height);
			}
		} else if ("button" == classname || "togglebutton" == classname) {
			int proximity = getInteger(component, "proximity", 0);
			Rectangle bounds;

			if (proximity > 0) {
				Dimension dim;

				bounds = getRectangle(component, "bounds");
				dim = getPreferredSize(component);
				// System.out.println("prox " + getString(component, "name"));
				// System.out.println("prox " + bounds + " " + dim);
				if (bounds == null) {
					setRectangle(component, "bounds", 0, 0, dim.width,
							dim.height);
				} else {
					if (dim.width != bounds.width) {
						int diff = dim.width - bounds.width;

						bounds.x -= diff / 2;
						bounds.width += diff;
					}
					if (dim.height != bounds.height) {
						int diff = dim.height - bounds.height;

						bounds.y -= diff / 2;
						bounds.height += diff;
					}
					// System.out.println("new " + bounds);
				}
			} else if ((get(component, "proximity") != null)
						&& (bounds = getRectangle(component, "bounds")) != null) {
				Dimension dim;

				dim = getPreferredSize(component);
				if (dim.width != bounds.width) {
					int diff = bounds.width - dim.width;

					bounds.x += diff / 2;
					bounds.width -= diff;
				}
				if (dim.height != bounds.height) {
					int diff = bounds.height - dim.height;

					bounds.y += diff / 2;
					bounds.height -= diff;
				}
			}
		} else if (("textfield" == classname) || ("passwordfield" == classname)) {
			layoutField(component, 0, ("passwordfield" == classname), 0);
		} else if ("textarea" == classname) {
			String text = getString(component, "text", "");
			int start = getInteger(component, "start", 0);
			if (start > text.length()) {
				setInteger(component, "start", start = text.length(), 0);
			}
			int end = getInteger(component, "end", 0);
			if (end > text.length()) {
				setInteger(component, "end", end = text.length(), 0);
			}

			boolean wrap = getBoolean(component, "wrap", false);
			char[] chars = null;
			if (wrap) {
				Rectangle bounds = getRectangle(component, "bounds");
				chars = getChars(component, text, true, bounds.width - 4,
						bounds.height);
				if (chars == null) { // need scrollbars
					chars = getChars(component, text, true, bounds.width
															- block - 4, 0);
				}
			} else {
				chars = getChars(component, text, false, 0, 0);
			}

			Font currentfont = (Font) get(component, "font");
			FontMetrics fm = getFontMetrics((currentfont != null) ? currentfont
					: font);
			int width = 0, height = 0;
			int caretx = 0;
			int carety = 0;
			for (int i = 0, j = 0; j <= chars.length; j++) {
				if ((j == chars.length) || (chars[j] == '\n')) {
					width = Math.max(width,
							asciiCharsWidth(fm, chars, i, j - i));
					if ((end >= i) && (end <= j)) {
						caretx = asciiCharsWidth(fm, chars, i, end - i);
						carety = height;
					}
					height += fm.getHeight();
					i = j + 1;
				}
			}
			layoutScroll(component, width + 2, height - fm.getLeading() + 2, 0,
					0, 0, 0, getBoolean(component, "border", true), 0);
			scrollToVisible(component, caretx, carety, 2, fm.getAscent()
															+ fm.getDescent()
															+ 2); // ?
		} else if ("tabbedpane" == classname) {
			// tabbedpane (not selected) tab padding are 1, 3, 1, and 3 pt
			Rectangle bounds = getRectangle(component, "bounds");
			String placement = getString(component, "placement", "top");
			boolean horizontal = ((placement == "top") || (placement == "bottom"));
			boolean stacked = (placement == "stacked");

			// draw up tabs in row/column
			int tabd = 0;
			Rectangle first = null; // x/y location of tab left/top
			int tabsize = 0; // max height/width of tabs
			for (Object tab = get(component, ":comp"); tab != null; tab = get(
					tab, ":next")) {
				if ((tabd == 0)
					&& ((first = getRectangle(tab, "bounds")) != null)) {
					tabd = horizontal ? first.x : first.y; // restore previous
															// offset
				}
				Dimension d = getSize(tab, stacked ? 8 : horizontal ? 12 : 9,
						stacked ? 3 : horizontal ? 5 : 8);
				setRectangle(tab, "bounds", horizontal ? tabd : 0,
						horizontal ? 0 : tabd,
						stacked ? bounds.width : d.width, d.height);
				if (stacked) {
					tabd += d.height;
				} else {
					tabd += (horizontal ? d.width : d.height) - 3;
					tabsize = Math
							.max(tabsize, horizontal ? d.height : d.width);
				}
			}

			// match tab height/width, set tab content size
			int cx = (placement == "left") ? (tabsize + 1) : 2;
			int cy = (placement == "top") ? (tabsize + 1) : 2;
			int cwidth = bounds.width
							- ((horizontal || stacked) ? 4 : (tabsize + 3));
			int cheight = bounds.height
							- (stacked ? (tabd + 3)
									: (horizontal ? (tabsize + 3) : 4));
			for (Object tab = get(component, ":comp"); tab != null; tab = get(
					tab, ":next")) {
				Rectangle r = getRectangle(tab, "bounds");
				if (!stacked) {
					if (horizontal) {
						if (placement == "bottom") {
							r.y = bounds.height - tabsize;
						}
						r.height = tabsize;
					} else {
						if (placement == "right") {
							r.x = bounds.width - tabsize;
						}
						r.width = tabsize;
					}
				}

				Object comp = get(tab, ":comp"); // relative to the tab
													// location
				if ((comp != null) && getBoolean(comp, "visible", true)) {
					setRectangle(comp, "bounds", cx - r.x,
							stacked ? (r.height + 1) : (cy - r.y), cwidth,
							cheight);
					doLayout(comp);
				}
			}
			checkOffset(component);
		} else if (("panel" == classname) || (classname == "dialog") || (classname == "dropzone")) {
			int gap = getInteger(component, "gap", 0);
			int[][] grid = getGrid(component);
			int top = 0;
			int left = 0;
			int contentwidth = 0;
			int contentheight = 0;
			if (classname == "dropzone") {
				/*
				 * Components in a dropzone have absolute coordinates, so we
				 * find the farthest one on the right and bottom to determine
				 * the contentwidth/contentheight.
				 */
				for (Object comp = get(component, ":comp"); comp != null; comp = get(
						comp, ":next")) {
					if (getBoolean(comp, "visible", true)) {
						Rectangle bounds = getRectangle(comp, "bounds");

						if (bounds != null) {
							contentwidth = Math.max(contentwidth,
									bounds.x + bounds.width);
							contentheight = Math.max(contentheight,
									bounds.y + bounds.height);
						}
					}
				}
				contentwidth += getInteger(component, "right", 0);
				contentheight += getInteger(component, "bottom", 0);
			} else if (grid != null) { // has subcomponents
				top = getInteger(component, "top", 0);
				left = getInteger(component, "left", 0);
				int bottom = getInteger(component, "bottom", 0);
				int right = getInteger(component, "right", 0);
				// sums the preferred size of cell widths and heights, gaps
				contentwidth = left
								+ getSum(grid[0], 0, grid[0].length, gap, false)
								+ right;
				contentheight = top
								+ getSum(grid[1], 0, grid[1].length, gap, false)
								+ bottom;
			}

			int titleheight = getSize(component, 0, 0).height; // title text
																// and icon
			setInteger(component, ":titleheight", titleheight, 0);
			boolean scrollable = getBoolean(component, "scrollable", false);
			boolean border = (("panel" == classname) || ("dropzone" == classname))
								&& getBoolean(component, "border", false);
			int iborder = (border ? 1 : 0);
			if (scrollable) { // set scrollpane areas
				if ("panel" == classname || "dropzone" == classname) {
					int head = titleheight / 2;
					int headgap = (titleheight > 0) ? (titleheight - head - iborder)
							: 0;
					scrollable = layoutScroll(component, contentwidth,
							contentheight, head, 0, 0, 0, border, headgap);
				} else { // dialog
					scrollable = layoutScroll(component, contentwidth,
							contentheight, 3 + titleheight, 3, 3, 3, true, 0);
				}
			}
			if (!scrollable) { // clear scrollpane bounds //+
				set(component, ":view", null);
				set(component, ":port", null);
			}

			if (grid != null) {
				int areax = 0;
				int areay = 0;
				int areawidth = 0;
				int areaheight = 0;
				if (scrollable) {
					// components are relative to the viewport
					Rectangle view = getRectangle(component, ":view");
					areawidth = view.width;
					areaheight = view.height;
				} else { // scrollpane isn't required
					// components are relative to top/left corner
					Rectangle bounds = getRectangle(component, "bounds");
					areawidth = bounds.width;
					areaheight = bounds.height;
					if ("panel" == classname || "dropzone" == classname) {
						areax = iborder;
						areay = Math.max(iborder, titleheight);
						areawidth -= 2 * iborder;
						areaheight -= areay + iborder;
					} else { // dialog
						areax = 4;
						areay = 4 + titleheight;
						areawidth -= 8;
						areaheight -= areay + 4;
						// System.out.println("dsizing" + bounds);
					}
				}

				for (int i = 0; i < 2; i++) { // i=0: horizontal, i=1:
												// vertical
					// remaining space
					int d = ((i == 0) ? (areawidth - contentwidth)
							: (areaheight - contentheight));
					if (d != 0) { // + > 0
						int w = getSum(grid[2 + i], 0, grid[2 + i].length, 0,
								false);
						if (w > 0) {
							for (int j = 0; j < grid[i].length; j++) {
								if (grid[2 + i][j] != 0) {
									grid[i][j] += d * grid[2 + i][j] / w;
								}
							}
						}
					}
				}

				Object comp = get(component, ":comp");
				for (int i = 0; comp != null; comp = get(comp, ":next")) {
					if (!getBoolean(comp, "visible", true)) {
						continue;
					}
					if (getBoolean(comp, "overlay", false)) {
						doLayout(comp);
						continue;
					}
					int ix = areax + left
								+ getSum(grid[0], 0, grid[4][i], gap, true);
					int iy = areay + top
								+ getSum(grid[1], 0, grid[5][i], gap, true);
					int iwidth = getSum(grid[0], grid[4][i], grid[6][i], gap,
							false);
					int iheight = getSum(grid[1], grid[5][i], grid[7][i], gap,
							false);
					String halign = getString(comp, "halign", "fill");
					String valign = getString(comp, "valign", "fill");
					if ((halign != "fill") || (valign != "fill")) {
						Dimension d = getPreferredSize(comp);
						if (halign != "fill") {
							int dw = Math.max(0, iwidth - d.width);
							if (halign == "center") {
								ix += dw / 2;
							} else if (halign == "right") {
								ix += dw;
							}
							iwidth -= dw;
						}
						if (valign != "fill") {
							int dh = Math.max(0, iheight - d.height);
							if (valign == "center") {
								iy += dh / 2;
							} else if (valign == "bottom") {
								iy += dh;
							}
							iheight -= dh;
						}
					}
					// System.out.println("newb " + classname + "; " + ix + "; " + iy + "; " + iwidth + "; " + iheight);
					setRectangle(comp, "bounds", ix, iy, iwidth, iheight);
					doLayout(comp);
					i++;
				}
			} else if (classname == "dropzone") {
				for (Object comp = get(component, ":comp"); comp != null; comp = get(
						comp, ":next")) {
					doLayout(comp);
				}
			}
		} else if ("desktop" == classname) {
			Rectangle bounds = getRectangle(component, "bounds");

			for (Object comp = get(component, ":comp"); comp != null; comp = get(
					comp, ":next")) {
				String iclass = getClass(comp);
				if (iclass == "dialog") {
					if (!getBoolean(comp, "visible", true))
						continue;
					
					Rectangle sbounds;
					boolean resizable = getBoolean(comp, "resizable", false);
					boolean scrollable = getBoolean(comp, "scrollable", false);
					sbounds = getRectangle(comp, "bounds");

					// System.out.println("sizing " + comp + "; " + sbounds);
					
					Dimension d = getPreferredSize(comp);
					int x, y, width, height;
					String halign, valign;

					halign = getString(comp, "halign", "center");
					valign = getString(comp, "valign", "center");

					if (sbounds != null && scrollable)
						width = sbounds.width;
					else if (sbounds != null && resizable)
						width = Math.min(Math.abs(bounds.width), Math.max(d.width, Math.abs(sbounds.width)));
					else
						width = d.width;
					if (sbounds != null && scrollable)
						height = sbounds.height;
					else if (sbounds != null && resizable)
						height = Math.min(bounds.height, Math.max(d.height, sbounds.height));
					else
						height = d.height;

					if (sbounds != null && !getBoolean(comp, "xlock"))
						x = sbounds.x;
					else if ("left".equals(halign))
						x = 0;
					else if ("center".equals(halign) || "fill".equals(halign))
						x = (bounds.width - width) / 2;
					else if ("right".equals(halign))
						x = bounds.width - width;
					else
						throw new InternalError();

					if (sbounds != null && !getBoolean(comp, "ylock"))
						y = sbounds.y;
					else if ("top".equals(valign))
						y = 0;
					else if ("center".equals(valign) || "fill".equals(valign))
						y = (bounds.height - height) / 2;
					else if ("bottom".equals(valign))
						y = bounds.height - height;
					else
						throw new InternalError();

					setRectangle(comp, "bounds", x, y, width, height);
					// System.out.println("dbounds = " + getRectangle(comp, "bounds"));
				} else if ((iclass != ":combolist") && (iclass != ":popup") && bounds != null) {
					setRectangle(comp, "bounds", 0, 0, bounds.width,
							bounds.height);
				}
				doLayout(comp);
			}
		} else if ("spinbox" == classname) {
			layoutField(component, block, false, 0);
		} else if ("splitpane" == classname) {
			Rectangle bounds = getRectangle(component, "bounds");
			boolean horizontal = ("vertical" != get(component, "orientation"));
			int divider = getInteger(component, "divider", -1);
			int maxdiv = Math.max(0,
					(horizontal ? bounds.width : bounds.height) - 5);

			Object comp1 = get(component, ":comp");
			boolean visible1 = (comp1 != null)
								&& getBoolean(comp1, "visible", true);
			if (divider == -1) {
				int d1 = 0;
				if (visible1) {
					Dimension d = getPreferredSize(comp1);
					d1 = horizontal ? d.width : d.height;
				}
				divider = Math.min(d1, maxdiv);
				setInteger(component, "divider", divider, -1);
			} else if (divider > maxdiv) {
				setInteger(component, "divider", divider = maxdiv, -1);
			}

			if (visible1) {
				setRectangle(comp1, "bounds", 0, 0, horizontal ? divider
						: bounds.width, horizontal ? bounds.height : divider);
				doLayout(comp1);
			}
			Object comp2 = (comp1 != null) ? get(comp1, ":next") : null;
			if ((comp2 != null) && getBoolean(comp2, "visible", true)) {
				setRectangle(comp2, "bounds", horizontal ? (divider + 5) : 0,
						horizontal ? 0 : (divider + 5),
						horizontal ? (bounds.width - 5 - divider)
								: bounds.width, horizontal ? bounds.height
								: (bounds.height - 5 - divider));
				doLayout(comp2);
			}
		} else if (("list" == classname) || ("table" == classname)
					|| ("tree" == classname)) {
			int line = getBoolean(component, "line", true) ? 1 : 0;
			int width = 0;
			int columnheight = 0;
			if ("table" == classname) {
				Object header = get(component, "header");
				int[] columnwidths = null;
				Object[] columneditors = null;
				if (header != null) {
					columnwidths = new int[getCount(header)];
					columneditors = new Object[columnwidths.length];
					Object column = get(header, ":comp");
					for (int i = 0; i < columnwidths.length; i++) {
						if (i != 0) {
							column = get(column, ":next");
						}
						if (!getBoolean(column, "visible", true))
							continue;
						columnwidths[i] = getInteger(column, "width", 80);
						width += columnwidths[i];
						Dimension d = getSize(column, 2, 2);
						columnheight = Math.max(columnheight, d.height);
						if (getBoolean(column, "editable")) {
							columneditors[i] = get(column, ":editor");
						}
					}
				}
				set(component, ":widths", columnwidths);
				set(component, ":editors", columneditors);
			}
			int y = 0;
			int level = 0;
			for (Object item = get(component, ":comp"); item != null;) {
				int x = 0;
				int iwidth = 0;
				int iheight = 0;
				if ("table" == classname) {
					iwidth = width;
					for (Object cell = get(item, ":comp"); cell != null; cell = get(
							cell, ":next")) {
						iheight = Math.max(iheight, getSizeHeight(cell, 2));
					}
				} else {
					if ("tree" == classname) {
						x = (level + 1) * block;
					}
					Dimension d = getSize(item, 6, 2);
					iwidth = d.width;
					iheight = d.height;
					width = Math.max(width, x + d.width);
				}
				setRectangle(item, "bounds", x, y, iwidth, iheight);
				y += iheight + line;
				if ("tree" == classname) {
					Object next = get(item, ":comp");
					if ((next != null) && getBoolean(item, "expanded", true)) {
						level++;
					} else {
						while (((next = get(item, ":next")) == null)
								&& (level > 0)) {
							item = getParent(item);
							level--;
						}
					}
					item = next;
				} else {
					item = get(item, ":next");
				}
			}
			layoutScroll(component, width, y - line, columnheight, 0, 0, 0,
					true, 0);
		} else if ("menubar" == classname) {
			Rectangle bounds = getRectangle(component, "bounds");
			int x = 0;
			for (Object menu = get(component, ":comp"); menu != null; menu = get(
					menu, ":next")) {
				Dimension d = getSize(menu, 8, 4);
				setRectangle(menu, "bounds", x, 0, d.width, bounds.height);
				x += d.width;
			}
		} else if ("bean" == classname) {
			Rectangle r = getRectangle(component, "bounds");
			((Component) get(component, "bean")).setBounds(r);
		} else if ("connector" == classname) {
			Rectangle lbounds, rbounds;
			Object parent, left, right;
			Dimension dim;
			int lx, ly, rx, ry, x, y;

			parent = getParent(component);
			dim = getPreferredSize(component);
			left = find(parent, getString(component, "left"));
			right = find(parent, getString(component, "right"));
			if (left != null && right != null) {
				int prox;

				prox = getInteger(left, "proximity");
				lbounds = getRectangle(left, "bounds");
				lx = lbounds.x + lbounds.width / 2;
				ly = lbounds.y + lbounds.height / 2;
				rbounds = getRectangle(right, "bounds");
				rx = rbounds.x + rbounds.width / 2;
				ry = rbounds.y + rbounds.height / 2;

				double angle = Math.toDegrees(Math.atan2(ly - ry, lx - rx));

				if ((-30.0 <= angle) && (angle <= 30.0)) {
					lx = lbounds.x + prox;
					rx = rbounds.x + rbounds.width - prox;
				} else if ((30.0 <= angle) && (angle <= 150.0)) {
					ly = lbounds.y + prox;
					ry = rbounds.y + rbounds.height - prox;
				} else if (((150.0 <= angle) && (angle <= 180.0))
							|| ((-150.0 >= angle) && (angle >= -180.0))) {
					lx = lbounds.x + lbounds.width - prox;
					rx = rbounds.x + prox;
				} else {
					ly = lbounds.y + lbounds.height - prox;
					ry = rbounds.y + prox;
				}

				x = Math.min(lx, rx) + Math.abs(lx - rx) / 2;
				y = Math.min(ly, ry) + Math.abs(ly - ry) / 2;

				x -= dim.width / 2;
				y -= dim.height / 2;

				setRectangle(component, "bounds", x, y, dim.width, dim.height);
				setRectangle(component, "cbounds", lx, ly, rx - lx + 1, ry - ly + 1);
				set(component, ":angles", new Object[] {
						new int[] { 5, (dim.height - 1) / 2, (dim.height - 1) / 2, 5 },
						new int[] { (dim.height - 1) / 2, 5, (dim.height - 5 - 1), (dim.height - 1) / 2 },
						new int[] { dim.width - (dim.height - 1) / 2, dim.width - 5,
								dim.width - (dim.height - 1) / 2,
								dim.width - (dim.height - 1) / 2 },
						new int[] { 5, (dim.height - 1) / 2, (dim.height - 5 - 1), 5 } });
			} else if (getRectangle(component, "cbounds") != null) {
				setReal(component, "cbounds", null);
			}
		} else if (instance(classname, "prim")) {
			String forName;

			if ((forName = getString(component, "for")) != null) {
				Object forComp;

				if ((forComp = get(component, ":for")) == null) {
					if ((forComp = find(this.content, forName)) != null) {
						set(component, ":for", forComp);
					}
				}
				if (forComp != null) {
					Rectangle forBounds;

					if ((forBounds = getRectangle(forComp, "bounds")) != null) {
						int xoffset, yoffset, fwidth, fheight, x, y, width, height;
						String placement;

						placement = getString(component, "placement");
						if (placement == null)
							placement = "s";
						xoffset = getInteger(component, "xoffset");
						yoffset = getInteger(component, "yoffset");
						x = forBounds.x;
						y = forBounds.y;
						width = fwidth = forBounds.width;
						height = fheight = forBounds.height;
						done: while ((forComp = getParent(forComp)) != null) {
							Object parent = component;

							while ((parent = getParent(parent)) != null) {
								if (parent == forComp)
									break done;
							}
							if ((forBounds = getRectangle(forComp, "bounds")) != null) {
								x += forBounds.x;
								y += forBounds.y;
							} else {
								break;
							}
						}
						if ("note" == classname) {
							Dimension dim;

							dim = getSize(component, 0, 0);
							width = dim.width;
							height = dim.height;
						} else {
							width += Math.abs(xoffset) * 2;
							height += Math.abs(yoffset) * 2;
						}
						switch (placement.charAt(0)) {
						case 'q':
							x -= xoffset + width;
							y -= yoffset + height;
							break;
						case 'w':
							x -= Math.abs((width - fwidth) / 2);
							y -= yoffset + height;
							break;
						case 'e':
							x += xoffset + fwidth;
							y -= yoffset + height;
							break;

						case 'a':
							x -= xoffset + width;
							y += yoffset;
							break;
						case 's':
							x -= (width - fwidth) / 2;
							y += yoffset;
							break;
						case 'd':
							x += xoffset + fwidth;
							y += yoffset;
							break;

						case 'z':
							x -= xoffset + width;
							y += yoffset + fheight;
							break;
						case 'x':
							x -= (width - fwidth) / 2;
							y += yoffset + fheight;
							break;
						case 'c':
							x += xoffset + fwidth;
							y += yoffset + fheight;
							break;
						}
						if ("note" == classname) {
							switch (placement.charAt(0)) {
							case 'x':
								y -= yoffset;
								height += yoffset;
								break;
							}
						}
						setRectangle(component, "bounds", x, y, width, height);
					}
				}
			}
		} else {
			// System.out.println("** no layout for " + classname + " " +
			// getClass(getParent(component)));
		}
	}

	/**
	 * Scroll tabs to make the selected one visible
	 * 
	 * @param component
	 *            a tabbedpane
	 */
	void checkOffset(Object component) {
		String placement = getString(component, "placement", "top");
		int selected = getInteger(component, "selected", 0);
		int i = 0;
		if (placement == "stacked") {
			int dy = 0;
			for (Object tab = get(component, ":comp"); tab != null; tab = get(
					tab, ":next")) {
				Rectangle r = getRectangle(tab, "bounds");
				r.y = dy;
				dy += r.height;
				if (i == selected) {
					dy += getRectangle(get(tab, ":comp"), "bounds").height + 2;
				}
				i++;
			}
			checkLocation(component);
			return;
		}
		boolean horizontal = ((placement == "top") || (placement == "bottom"));
		Rectangle bounds = getRectangle(component, "bounds");
		int panesize = horizontal ? bounds.width : bounds.height;
		int first = 0;
		int last = 0;
		int d = 0;
		for (Object tab = get(component, ":comp"); tab != null; tab = get(tab,
				":next")) {
			Rectangle r = getRectangle(tab, "bounds");
			if (i == 0) {
				first = (horizontal ? r.x : r.y);
			}
			last = (horizontal ? (r.x + r.width) : (r.y + r.height));
			if (i == selected) {
				int ifrom = (horizontal ? r.x : r.y) - 6;
				int ito = (horizontal ? (r.x + r.width) : (r.y + r.height)) + 6;
				if (ifrom < 0) {
					d = -ifrom;
				} else if (ito > panesize) {
					d = panesize - ito;
				}
			}
			i++;
		}
		d = Math.min(-first, Math.max(d, panesize - last));
		if (d != 0) {
			for (Object tab = get(component, ":comp"); tab != null; tab = get(
					tab, ":next")) {
				Rectangle r = getRectangle(tab, "bounds");
				if (horizontal) {
					r.x += d;
				} else {
					r.y += d;
				}
				Object comp = get(tab, ":comp"); // relative to the tab
													// location
				if ((comp != null) && getBoolean(comp, "visible", true)
					&& !getBoolean(comp, "overlay", false)) {
					Rectangle rc = getRectangle(comp, "bounds");
					if (horizontal) {
						rc.x -= d;
					} else {
						rc.y -= d;
					}
				}
			}
			checkLocation(component);
		}
	}

	/**
	 * 
	 */
	char[] getChars(Object component, String text, boolean wrap, int width,
			int height) {
		char[] chars = (char[]) get(component, ":text");
		if ((chars == null) || (chars.length != text.length())) {
			chars = text.toCharArray();
			set(component, ":text", chars);
		} else
			text.getChars(0, chars.length, chars, 0);

		if (wrap) {
			Font currentfont = (Font) get(component, "font");
			FontMetrics fm = getFontMetrics((currentfont != null) ? currentfont
					: font);
			int lines = (height - 4 + fm.getLeading()) / fm.getHeight();
			boolean prevletter = false;
			int n = chars.length;
			int linecount = 0;
			for (int i = 0, j = -1, k = 0; k <= n; k++) { // j is the last
															// space index
															// (before k)
				if (((k == n) || (chars[k] == '\n') || (chars[k] == ' '))
					&& (j > i) && (fm.charsWidth(chars, i, k - i) > width)) {
					chars[j] = '\n';
					k--; // draw line to the begin of the current word (+
							// spaces) if it is out of width
				} else if ((k == n) || (chars[k] == '\n')) { // draw line to
																// the text/line
																// end
					j = k;
					prevletter = false;
				} else {
					if ((chars[k] == ' ') && (prevletter || (j > i))) {
						j = k;
					} // keep spaces starting the line
					prevletter = (chars[k] != ' ');
					continue;
				}
				linecount++;
				if ((lines != 0) && (linecount == lines)) {
					return null;
				}
				i = j + 1;
			}
		}
		return chars;
	}

	/**
	 * 
	 */
	void layoutField(Object component, int dw, boolean hidden, int left) {
		int width = getRectangle(component, "bounds").width - left - dw;
		String text = getString(component, "text", "");
		int start = getInteger(component, "start", 0);
		if (start > text.length()) {
			setInteger(component, "start", start = text.length(), 0);
		}
		int end = getInteger(component, "end", 0);
		if (end > text.length()) {
			setInteger(component, "end", end = text.length(), 0);
		}
		int offset = getInteger(component, ":offset", 0);
		int off = offset;
		Font currentfont = (Font) get(component, "font");
		FontMetrics fm = getFontMetrics((currentfont != null) ? currentfont
				: font);
		int textwidth = hidden ? (fm.charWidth('*') * text.length()) : fm
				.stringWidth(text);
		int caret = hidden ? (fm.charWidth('*') * end) : fm.stringWidth(text
				.substring(0, end));
		if (textwidth <= width - 4) { // text fits inside the available space
			Object align = get(component, "alignment");
			if (align == null) {
				off = 0;
			} // left alignment
			else {
				off = textwidth - width + 4; // right alignment
				if (align == "center") {
					off /= 2;
				} // center alignment
			}
		} else { // text is scrollable
			if (off > caret) {
				off = caret;
			} else if (off < caret - width + 4) {
				off = caret - width + 4;
			}
			off = Math.max(0, Math.min(off, textwidth - width + 4));
		}
		if (off != offset) {
			setInteger(component, ":offset", off, 0);
		}
	}

	/**
	 * Set viewport (:port) bounds excluding borders, view position and content
	 * size (:view), horizontal (:horizontal), and vertical (:vertical)
	 * scrollbar bounds
	 * 
	 * @param component
	 *            scrollable widget
	 * @param contentwidth
	 *            preferred component width
	 * @param contentheight
	 *            preferred component height
	 * @param top
	 *            top inset (e.g. table header, dialog title, half of panel
	 *            title)
	 * @param left
	 *            left inset (e.g. dialog border)
	 * @param bottom
	 *            bottom inset (e.g. dialog border)
	 * @param right
	 *            right inset (e.g. dialog border)
	 * @param topgap
	 *            (lower half of panel title)
	 * @return true if scrollpane is required, otherwise false
	 * 
	 * list: 0, 0, 0, 0, true, 0 | table: header, ... | dialog: header, 3, 3, 3,
	 * true, 0 title-border panel: header / 2, 0, 0, 0, true, head
	 */
	boolean layoutScroll(Object component, int contentwidth, int contentheight,
			int top, int left, int bottom, int right, boolean border, int topgap) {
		Rectangle bounds = getRectangle(component, "bounds");
		int iborder = border ? 1 : 0;
		int iscroll = block + 1 - iborder;
		int portwidth = bounds.width - left - right - 2 * iborder; // available
																	// horizontal
																	// space
		int portheight = bounds.height - top - topgap - bottom - 2 * iborder; // vertical
																				// space
		boolean hneed = contentwidth > portwidth; // horizontal scrollbar
													// required
		boolean vneed = contentheight > portheight - (hneed ? iscroll : 0); // vertical
																			// scrollbar
																			// needed
		if (vneed) {
			portwidth -= iscroll;
		} // subtract by vertical scrollbar width
		hneed = hneed || (vneed && (contentwidth > portwidth));
		if (hneed) {
			portheight -= iscroll;
		} // subtract by horizontal scrollbar height

		setRectangle(component, ":port", left + iborder,
				top + iborder + topgap, portwidth, portheight);
		if (hneed) {
			setRectangle(component, ":horizontal", left, bounds.height - bottom
															- block - 1,
					bounds.width - left - right - (vneed ? block : 0),
					block + 1);
		} else {
			set(component, ":horizontal", null);
		}
		if (vneed) {
			setRectangle(component, ":vertical", bounds.width - right - block
													- 1, top, block + 1,
					bounds.height - top - bottom - (hneed ? block : 0));
		} else {
			set(component, ":vertical", null);
		}

		contentwidth = Math.max(contentwidth, portwidth);
		contentheight = Math.max(contentheight, portheight);
		int viewx = 0, viewy = 0;
		Rectangle view = getRectangle(component, ":view");
		if (view != null) { // check the previous location
			viewx = Math.max(0, Math.min(view.x, contentwidth - portwidth));
			viewy = Math.max(0, Math.min(view.y, contentheight - portheight));
		}
		setRectangle(component, ":view", viewx, viewy, contentwidth,
				contentheight);
		return vneed || hneed;
	}

	/**
	 * Gets the preferred size of the root component
	 * 
	 * @return a dimension object indicating the root component's preferred size
	 */
	public Dimension getPreferredSize() {
		return getPreferredSize(content);
	}

	/**
	 * 
	 * @throws java.lang.IllegalArgumentException
	 */
	Dimension getPreferredSize(Object component) {
		int width = getInteger(component, "width", 0);
		int height = getInteger(component, "height", 0);
		String classname = getClass(component);
		if ((width > 0) && (height > 0)) {
			if (classname == "dialog"
				&& getBoolean(component, "resizable", false)) {
				if (get(component, ":minbounds") == null ||
						!getBoolean(component, "scrollable", false)) {
					setRectangle(component, ":minbounds", 0, 0, width, height);
				}
			}
			return new Dimension(width, height);
		}
		if ("label" == classname) {
			return getSize(component, 0, 0);
		}
		if (("button" == classname) || ("togglebutton" == classname)) {
			boolean link = ("button" == classname)
							&& (get(component, "type") == "link");
			boolean round = ("togglebutton" == classname) && (get(component, "style") == "round");
			int proximity = getInteger(component, "proximity", 0);
			int dw = (round ? 8 : 0);
			return getSize(component, link ? 0 : 12 + (proximity * 2), link ? 0
					: 6 + (proximity * 2));
		}
		if ("checkbox" == classname) {
			Dimension d = getSize(component, 0, 0);
			d.width = d.width
						+ block
						+ ((getString(component, "text", "").length() > 0) ? 3
								: 0);
			d.height = Math.max(block, d.height);
			return d;
		}
		if ("combobox" == classname) {
			if (getBoolean(component, "editable", true)
				|| getBoolean(component, "selectable", true)) {
				Dimension size = getFieldSize(component);
				Image icon = getIcon(component, "icon", null);
				String style;
				int dw = 0;

				style = getString(component, "style", "normal");
				if ("round" == style)
					dw = block;
				if (icon != null) {
					size.width += icon.getWidth(this);
					size.height = Math.max(size.height,
							icon.getHeight(this) + 2);
				}
				size.width += block + dw;
				return size;
			} else {
				// maximum size of current values and choices including 2-2-2-2
				// insets
				Dimension size = getSize(component, 4, 4);
				for (Object item = get(component, ":comp"); item != null; item = get(
						item, ":next")) {
					Dimension d = getSize(item, 4, 4);
					size.width = Math.max(d.width, size.width);
					size.height = Math.max(d.height, size.height);
				}
				size.width += block;
				if (size.height == 4) { // no content nor items, set text height
					Font customfont = (Font) get(component, "font");
					FontMetrics fm = getFontMetrics((customfont != null) ? customfont
							: font);
					size.height = fm.getAscent() + fm.getDescent() + 4;
				}
				return size;
			}
		}
		if (("textfield" == classname) || ("passwordfield" == classname)) {
			return getFieldSize(component);
		}
		if ("textarea" == classname) {
			int columns = getInteger(component, "columns", 0);
			int rows = getInteger(component, "rows", 0); // 'e' -> 'm' ?
			Font currentfont = (Font) get(component, "font");
			FontMetrics fm = getFontMetrics((currentfont != null) ? currentfont
					: font);
			if (getBoolean(component, "editable")) {
				return new Dimension(
						((columns > 0) ? (columns * fm.charWidth('e') + 2) : 76)
								+ 2 + block,
						((rows > 0) ? (rows * fm.getHeight() - fm.getLeading() + 2)
								: 76)
								+ 2 + block);
			} else {
				return new Dimension(
						((columns > 0) ? (columns * fm.charWidth('e') + 2) : 76) + 2,
						((rows > 0) ? (rows * fm.getHeight() - fm.getLeading() + 2)
								: 76) + 2);
			}
		}
		if ("tabbedpane" == classname) {
			String placement = getString(component, "placement", "top");
			boolean horizontal = ((placement != "left") && (placement != "right"));
			int tabsize = 0; // max tab height (for horizontal),
			// max tabwidth (for vertical), or sum of tab heights for stacked
			int contentwidth = 0;
			int contentheight = 0; // max content size
			for (Object tab = get(component, ":comp"); tab != null; tab = get(
					tab, ":next")) {
				Dimension d = getSize(tab, 0, 0);
				if (placement == "stacked") {
					tabsize += d.height + 3;
				} else {
					tabsize = Math.max(tabsize, horizontal ? d.height + 5
							: d.width + 9);
				}

				Object comp = get(tab, ":comp");
				if ((comp != null) && getBoolean(comp, "visible", true)) {
					Dimension dc = getPreferredSize(comp);
					contentwidth = Math.max(contentwidth, dc.width);
					contentheight = Math.max(contentheight, dc.height);
				}
			}
			return new Dimension(contentwidth
									+ (horizontal ? 4 : (tabsize + 3)),
					contentheight + (horizontal ? (tabsize + 3) : 4));
		}
		if (("panel" == classname) || (classname == "dialog") || (classname == "dropzone")) {
			// title text and icon height
			Dimension size = getSize(component, 0, 0);
			// add border size
			if (classname == "dialog") {
				size.width = 8;
				size.height += 8; // title width neglected
			} else if (getBoolean(component, "border", false)) { // bordered
																	// panel
				size.width = 2;
				size.height += (size.height > 0) ? 1 : 2; // title includes
															// line
			} else {
				size.width = 0;
			} // title width is clipped
			// add paddings
			size.width += getInteger(component, "left", 0)
							+ getInteger(component, "right", 0);
			size.height += getInteger(component, "top", 0)
							+ getInteger(component, "bottom", 0);
			// add content preferred size
			int gap = getInteger(component, "gap", 0);
			int[][] grid = getGrid(component);
			if (grid != null) { // has components
				size.width += getSum(grid[0], 0, grid[0].length, gap, false);
				size.height += getSum(grid[1], 0, grid[1].length, gap, false);
			}
			if (getBoolean(component, "scrollable", false)) {
				size.width += block;
				size.height += block;
			}
			if (classname == "dialog"
				&& getBoolean(component, "resizable", false)) {
				if (get(component, ":minbounds") == null ||
						!getBoolean(component, "scrollable")) {
					setRectangle(component, ":minbounds", 0, 0, size.width,
							size.height);
				}
			}

			return size;
		} else if ("desktop" == classname) {
			Dimension size = new Dimension();
			for (Object comp = get(component, ":comp"); comp != null; comp = get(
					comp, ":next")) {
				String iclass = getClass(comp);
				if ((iclass != "dialog") && (iclass != ":popup")
					&& (iclass != ":combolist")) {
					Dimension d = getPreferredSize(comp);
					size.width = Math.max(d.width, size.width);
					size.height = Math.max(d.height, size.height);
				}
			}
			return size;
		}
		if ("spinbox" == classname) {
			Dimension size = getFieldSize(component);
			size.width += block;
			return size;
		}
		if ("progressbar" == classname) {
			boolean horizontal = ("vertical" != get(component, "orientation"));
			return new Dimension(horizontal ? 76 : 6, horizontal ? 6 : 76);
		}
		if ("slider" == classname) {
			boolean horizontal = ("vertical" != get(component, "orientation"));
			return new Dimension(horizontal ? 76 : 10, horizontal ? 10 : 76);
		}
		if ("splitpane" == classname) {
			boolean horizontal = ("vertical" != get(component, "orientation"));
			Object comp1 = get(component, ":comp");
			Dimension size = ((comp1 == null) || !getBoolean(comp1, "visible",
					true)) ? new Dimension() : getPreferredSize(comp1);
			Object comp2 = get(comp1, ":next");
			if ((comp2 != null) && getBoolean(comp2, "visible", true)) {
				Dimension d = getPreferredSize(comp2);
				size.width = horizontal ? (size.width + d.width) : Math.max(
						size.width, d.width);
				size.height = horizontal ? Math.max(size.height, d.height)
						: (size.height + d.height);
			}
			if (horizontal) {
				size.width += 5;
			} else {
				size.height += 5;
			}
			return size;
		}
		if (("list" == classname) || ("table" == classname)
			|| ("tree" == classname)) {
			return new Dimension(76 + 2 + block, 76 + 2 + block);
		}
		if ("separator" == classname) {
			return new Dimension(1, 1);
		}
		if ("menubar" == classname) {
			Dimension size = new Dimension(0, 0);
			for (Object menu = get(component, ":comp"); menu != null; menu = get(
					menu, ":next")) {
				Dimension d = getSize(menu, 8, 4);
				size.width += d.width;
				size.height = Math.max(size.height, d.height);
			}
			return size;
		}
		if ("bean" == classname) {
			return ((Component) get(component, "bean")).getPreferredSize();
		}
		if ("connector" == classname) {
			Dimension size = getSize(component, 0, 6);

			size.width += size.height;
			size.width += 10;
			size.height += 10;
			return size;
		}
		if (instance(classname, "prim")) {
			return new Dimension(1, 1);
		}
		throw new IllegalArgumentException(classname);
	}

	/**
	 * @param component
	 *            a container
	 * @return null for zero visible subcomponent, otherwise an array contains
	 *         the following lists:
	 *         <ul>
	 *         <li>columnwidths, preferred width of grid columns</li>
	 *         <li>rowheights, preferred heights of grid rows</li>
	 *         <li>columnweights, grid column-width weights</li>
	 *         <li>rowweights, grid row-height weights</li>
	 *         <li>gridx, horizontal location of the subcomponents</li>
	 *         <li>gridy, vertical locations</li>
	 *         <li>gridwidth, column spans</li>
	 *         <li>gridheight, row spans</li>
	 *         </ul>
	 */
	int[][] getGrid(Object component) {
		if (getClass(component) == "dropzone") {
			/*
			 * System.out.println("ignoring
			 * dropzone");
			 */
			return null;
		}
		int count = 0; // count of the visible subcomponents
		for (Object comp = get(component, ":comp"); comp != null; comp = get(
				comp, ":next")) {
			if (getBoolean(comp, "visible", true)
				&& !getBoolean(comp, "overlay", false)) {
				count++;
			}
		}
		if (count == 0) {
			return null;
		} // zero subcomponent
		int columns = getInteger(component, "columns", 0);
		int icols = (columns != 0) ? columns : count;
		int irows = (columns != 0) ? ((count + columns - 1) / columns) : 1;
		int[][] grid = { new int[icols], new int[irows], // columnwidths,
															// rowheights
				new int[icols], new int[irows], // columnweights, rowweights
				new int[count], new int[count], // gridx, gridy
				new int[count], new int[count] }; // gridwidth, gridheight
		int[] columnheight = new int[icols];
		int[][] cache = null; // preferredwidth, height, columnweight,
								// rowweight

		int i = 0;
		int x = 0;
		int y = 0;
		int nextsize = 0;
		for (Object comp = get(component, ":comp"); comp != null; comp = get(
				comp, ":next")) {
			if (!getBoolean(comp, "visible", true)) {
				continue;
			}
			if (getBoolean(comp, "overlay", false)) {
				continue;
			}
			int colspan = ((columns != 0) && (columns < count)) ? Math.min(
					getInteger(comp, "colspan", 1), columns) : 1;
			int rowspan = (columns != 1) ? getInteger(comp, "rowspan", 1) : 1;

			for (int j = 0; j < colspan; j++) {
				if ((columns != 0) && (x + colspan > columns)) {
					x = 0;
					y++;
					j = -1;
				} else if (columnheight[x + j] > y) {
					x += (j + 1);
					j = -1;
				}
			}
			if (y + rowspan > grid[1].length) {
				int[] rowheights = new int[y + rowspan];
				System.arraycopy(grid[1], 0, rowheights, 0, grid[1].length);
				grid[1] = rowheights;
				int[] rowweights = new int[y + rowspan];
				System.arraycopy(grid[3], 0, rowweights, 0, grid[3].length);
				grid[3] = rowweights;
			}
			for (int j = 0; j < colspan; j++) {
				columnheight[x + j] = y + rowspan;
			}

			int weightx = getInteger(comp, "weightx", 0);
			int weighty = getInteger(comp, "weighty", 0);
			Dimension d = getPreferredSize(comp);

			if (colspan == 1) {
				grid[0][x] = Math.max(grid[0][x], d.width); // columnwidths
				grid[2][x] = Math.max(grid[2][x], weightx); // columnweights
			} else {
				if (cache == null) {
					cache = new int[4][count];
				}
				cache[0][i] = d.width;
				cache[2][i] = weightx;
				if ((nextsize == 0) || (colspan < nextsize)) {
					nextsize = colspan;
				}
			}
			if (rowspan == 1) {
				grid[1][y] = Math.max(grid[1][y], d.height); // rowheights
				grid[3][y] = Math.max(grid[3][y], weighty); // rowweights
			} else {
				if (cache == null) {
					cache = new int[4][count];
				}
				cache[1][i] = d.height;
				cache[3][i] = weighty;
				if ((nextsize == 0) || (rowspan < nextsize)) {
					nextsize = rowspan;
				}
			}
			grid[4][i] = x; // gridx
			grid[5][i] = y; // gridy
			grid[6][i] = colspan; // gridwidth
			grid[7][i] = rowspan; // gridheight

			x += colspan;
			i++;
		}

		while (nextsize != 0) {
			int size = nextsize;
			nextsize = 0;
			for (int j = 0; j < 2; j++) { // horizontal, vertical
				for (int k = 0; k < count; k++) {
					if (grid[6 + j][k] == size) { // gridwidth, gridheight
						int gridpoint = grid[4 + j][k]; // gridx, gridy

						int weightdiff = cache[2 + j][k];
						for (int m = 0; (weightdiff > 0) && (m < size); m++) {
							weightdiff -= grid[2 + j][gridpoint + m];
						}
						if (weightdiff > 0) {
							int weightsum = cache[2 + j][k] - weightdiff;
							for (int m = 0; (weightsum > 0) && (m < size); m++) {
								int weight = grid[2 + j][gridpoint + m];
								if (weight > 0) {
									int weightinc = weight * weightdiff
													/ weightsum;
									grid[2 + j][gridpoint + m] += weightinc;
									weightdiff -= weightinc;
									weightsum -= weightinc;
								}
							}
							grid[2 + j][gridpoint + size - 1] += weightdiff;
						}

						int sizediff = cache[j][k];
						int weightsum = 0;
						for (int m = 0; (sizediff > 0) && (m < size); m++) {
							sizediff -= grid[j][gridpoint + m];
							weightsum += grid[2 + j][gridpoint + m];
						}
						if (sizediff > 0) {
							for (int m = 0; (weightsum > 0) && (m < size); m++) {
								int weight = grid[2 + j][gridpoint + m];
								if (weight > 0) {
									int sizeinc = weight * sizediff / weightsum;
									grid[j][gridpoint + m] += sizeinc;
									sizediff -= sizeinc;
									weightsum -= weight;
								}
							}
							grid[j][gridpoint + size - 1] += sizediff;
						}
					} else if ((grid[6 + j][k] > size)
								&& ((nextsize == 0) || (grid[6 + j][k] < nextsize))) {
						nextsize = grid[6 + j][k];
					}
				}
			}
		}
		return grid;
	}

	/**
	 * 
	 */
	int getSum(int[] values, int from, int length, int gap, boolean last) {
		if (length <= 0) {
			return 0;
		}
		int value = 0;
		for (int i = 0; i < length; i++) {
			value += values[from + i];
		}
		return value + (length - (last ? 0 : 1)) * gap;
	}

	/**
	 * 
	 */
	Dimension getFieldSize(Object component) {
		int columns = getInteger(component, "columns", 0);
		Font currentfont = (Font) get(component, "font");
		FontMetrics fm = getFontMetrics((currentfont != null) ? currentfont
				: font);
		return new Dimension(((columns > 0) ? (columns * fm.charWidth('e'))
				: 76) + 4, fm.getAscent() + fm.getDescent() + 4); // fm.stringWidth(text)
	}

	int getSizeHeight(Object component, int dy) {
		Font customfont = (Font) get(component, "font");
		FontMetrics fm = getFontMetrics((customfont != null) ? customfont
				: font);
		Image icon;
		int retval;

		icon = getIcon(component, "icon", null);
		retval = Math.max(fm.getAscent() + fm.getDescent(), icon != null ? icon
				.getHeight(this) : 0)
					+ dy;
		return retval;
	}

	/**
	 * @param component
	 *            a widget including the text and icon parameters
	 * @param dx
	 *            increase width by this value
	 * @param dy
	 *            increase height by this value
	 * @return size of the text and the image (plus a gap) including the given
	 *         offsets
	 */
	Dimension getSize(Object component, int dx, int dy) {
		String text = getString(component, "text", null);
		int tw = 0;
		int th = 0;
		if (text != null) {
			Font customfont = (Font) get(component, "font");
			FontMetrics fm = getFontMetrics((customfont != null) ? customfont
					: font);
			tw = fm.stringWidth(text);
			th = fm.getAscent() + fm.getDescent();
		}
		Image icon = getIcon(component, "icon", null);
		int iw = 0;
		int ih = 0;
		if (icon != null) {
			iw = icon.getWidth(this);
			ih = icon.getHeight(this);
			if (text != null) {
				iw += 2;
			}
		}
		return new Dimension(tw + iw + dx, Math.max(th, ih) + dy);
	}

	int asciiCharsWidth(FontMetrics fm, char arr[], int start, int length) {
		int lpc, retval = 0;

		for (lpc = start; lpc < (start + length); lpc++) {
			if (arr[lpc] == '\b') {
				int diff;

				diff = (lpc - start - 1);
				retval += fm.charsWidth(arr, start, diff);
				start += diff + 2;
				length -= diff + 2;
				lpc += 1;
			}
		}
		retval += fm.charsWidth(arr, start, length);

		return retval;
	}

	abstract void checkLocation(Object component);

	abstract Object popupCombo(Object combobox);

	abstract void setInside(Object component, Object part, boolean scroll);

	abstract void closeCombo(Object combobox, Object combolist, Object item);
}
