package thinlet;

import java.awt.Color;
import java.awt.Component;
import java.awt.Composite;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Image;
import java.awt.Point;
import java.awt.Polygon;
import java.awt.Rectangle;
import java.awt.RenderingHints;
import java.awt.geom.CubicCurve2D;
import java.awt.image.ImageObserver;
import java.awt.image.MemoryImageSource;
import java.lang.reflect.Method;
import java.util.Hashtable;

public class ThinletPaint extends ThinletEvent {
	protected static Method renderinghint;

	protected static Object[] TXT_AA, G_AA, G_AA_OFF;

	transient final Hashtable imageToVisible = new Hashtable();

	transient final Rectangle clipBounds = new Rectangle();
	transient final Rectangle paintBounds = new Rectangle();

	static int evm;
	static { // evm
		// EVM has larger fillRect, fillOval, and drawImage(part), others are
		// correct
		// contributed by Ibsen Ramos-Bonilla and AK
		try {
			if ((System.getProperty("java.vendor").indexOf("Insignia") != -1)
				&& System.getProperty("os.name").indexOf("Windows CE") == -1) {
				evm = -1;
			}
		} catch (Exception exc) { /* never */
		}
		try {
			Class hintsclass = Class.forName("java.awt.RenderingHints");
			TXT_AA = new Object[] {
					hintsclass.getField("KEY_TEXT_ANTIALIASING").get(null),
					hintsclass.getField("VALUE_TEXT_ANTIALIAS_ON").get(null) };
			G_AA = new Object[] {
					hintsclass.getField("KEY_ANTIALIASING").get(null),
					hintsclass.getField("VALUE_ANTIALIAS_ON").get(null) };
			G_AA_OFF = new Object[] {
					hintsclass.getField("KEY_ANTIALIASING").get(null),
					hintsclass.getField("VALUE_ANTIALIAS_OFF").get(null) };
		} catch (Exception exc) {
		}
	} // evm

	/**
	 * Invokes the paint method
	 */
	public void update(Graphics g) {
		paint(g);
	}

	public boolean imageUpdate(Image img, int infoflags, int x, int y,
			int width, int height) {
		boolean retval = true;

		if ((infoflags & ImageObserver.FRAMEBITS) != 0) {
			Object comp;

			if ((comp = imageToVisible.get(img)) == null)
				retval = false;
			else if (isVisible(comp) && getIcon(comp, "icon") == img)
				repaint(comp);
			else
				imageToVisible.remove(img);
		} else if (false) {
			retval = super.imageUpdate(img, infoflags, x, y, width, height)
						&& retval;
		}
		return retval;
	}
	
	protected void antialiasToggle(Graphics g, boolean on) {
		if (TXT_AA != null) {
			try {
				renderinghint.invoke(g, on ? G_AA : G_AA_OFF);
			} catch(Exception exc) {
			}
		}
	}

	public void paint(Graphics g) {
		if (TXT_AA != null) { // set antialiasing
			try {
				if (renderinghint == null) {
					Class cl = Class.forName("java.awt.Graphics2D");
					
					renderinghint = cl.getMethod(
							"setRenderingHint",
							new Class[] { TXT_AA[0].getClass().getSuperclass(),
									Object.class });
				}
				renderinghint.invoke(g, TXT_AA); // text rendering with
													// antialiasing
			} catch (Exception exc) { /* never */
				TXT_AA = null;
			}
		}

		synchronized(this) {
			for (int lpc = 0; lpc < relayoutQueueCount; lpc++) {
				if (false) {
					System.out.println("relayout " + lpc + " of "
							+ relayoutQueueCount);
				}
				doLayout(relayoutQueue[lpc]);
				repaint(relayoutQueue[lpc]);
			}
		}
		
		relayoutQueueCount = 0;
		if (repaintQueueCount == 0) {
			// System.out.println("full");
			paintBounds.setRect(this.getBounds());
			paintInternal(g);
		} else {
			if (false && repaintQueueCount > 1)
				System.out.println("rp len = " + repaintQueueCount);
			else if (false)
				System.out.println("clip " + repaintQueue[0]);
			for (int lpc = 0; lpc < repaintQueueCount; lpc++) {
				if (lpc == 0)
					this.paintBounds.setRect(repaintQueue[lpc]);
				else
					this.paintBounds.add(repaintQueue[lpc]);
				g.setClip(repaintQueue[lpc]);
				if (false && repaintQueueCount > 1)
					System.out.println("clip " + repaintQueue[lpc]);
				paintInternal(g);
			}
			if (false && repaintQueueCount > 1)
				System.out.println("   len = " + repaintQueueCount);
			repaintQueueCount = 0;
			if (dragPaint) {
				g.setClip(null);
				g.setColor(Color.BLACK);
				if (this.dragSelect != 0) {
					this.dragSelectBounds.x = Math.min(this.dragSelectStart.x, this.dragSelectCurrent.x);
					this.dragSelectBounds.y = Math.min(this.dragSelectStart.y, this.dragSelectCurrent.y);
					this.dragSelectBounds.width = Math.max(this.dragSelectStart.x,
							this.dragSelectCurrent.x) - this.dragSelectBounds.x;
					this.dragSelectBounds.height = Math.max(this.dragSelectStart.y,
							this.dragSelectCurrent.y) - this.dragSelectBounds.y;

					int nx, ny;
					
					nx = Math.max(this.dragSelectBounds.x, this.dragSelectLimit.x);
					ny = Math.max(this.dragSelectBounds.y, this.dragSelectLimit.y);
					this.dragSelectBounds.width = Math.min(this.dragSelectBounds.x + this.dragSelectBounds.width,
							this.dragSelectLimit.x + this.dragSelectLimit.width) - nx;
					this.dragSelectBounds.height = Math.min(this.dragSelectBounds.y + this.dragSelectBounds.height,
							this.dragSelectLimit.y + this.dragSelectLimit.height) - ny;
					
					this.dragSelectBounds.x = nx;
					this.dragSelectBounds.y = ny;

					g.drawRect(this.dragSelectBounds.x,
							this.dragSelectBounds.y,
							this.dragSelectBounds.width,
							this.dragSelectBounds.height);
				}
				else {
					this.dragSelectBounds.x = 0;
					this.dragSelectBounds.y = 0;
					this.dragSelectBounds.width = 0;
					this.dragSelectBounds.height = 0;
				}
				g.setPaintMode();
				dragPaint = false;
			}
		}
	}
	
	protected void addGradient(Color c1, Color c2) {
		if (hgradients.containsKey(c1))
			return;
		
		int[][] pix = new int[2][block * block];
		int r1 = c1.getRed();
		int r2 = c2.getRed();
		int g1 = c1.getGreen();
		int g2 = c2.getGreen();
		int b1 = c1.getBlue();
		int b2 = c2.getBlue();
		for (int i = 0; i < block; i++) {
			int cr = r1 - (r1 - r2) * i / block;
			int cg = g1 - (g1 - g2) * i / block;
			int cb = b1 - (b1 - b2) * i / block;
			int color = (255 << 24) | (cr << 16) | (cg << 8) | cb;
			for (int j = 0; j < block; j++) {
				pix[0][i * block + j] = color;
				pix[1][j * block + i] = color;
			}
		}
		hgradients.put(c1, createImage(new MemoryImageSource(block, block, pix[0],
				0, block)));
		vgradients.put(c1, createImage(new MemoryImageSource(block, block, pix[1],
				0, block)));
	}

	/**
	 * Paints the components inside the graphics clip area
	 */
	protected void paintInternal(Graphics g) {
		g.setFont(font);
		if (hgradient == null) {
			this.addGradient(c_bg, c_press);
			hgradient = (Image)this.hgradients.get(c_bg);
			vgradient = (Image)this.vgradients.get(c_bg);
		}
		Rectangle clip = g.getClipBounds(this.clipBounds);
		if (ThinletBase.SHOW_REPAINT) {
			g.setColor(Color.orange);
			g.fillRect(0, 0, getSize().width, getSize().height);
			long time = System.currentTimeMillis();
			if (true)
				System.out.println("paintclip " + clip);
			try {
				Thread.sleep(125);
			} catch (Throwable th) {
			}
			// dg.setClip(r.x, r.y, r.width, r.height);
		}
		paint(g, clip.x, clip.y, clip.width, clip.height, content, isEnabled());
		if (tdi.source != null) {
			Graphics2D g2d;
			Composite oc;
			if (true) {

				g2d = (Graphics2D) g;
				oc = g2d.getComposite();
				g2d.setComposite(dragComposite);
			}
			if (true) {
				g.setClip(dragBounds.x + dragProximity,
						dragBounds.y + dragProximity,
						dragBounds.width - dragProximity * 2,
						dragBounds.height - dragProximity * 2);
				g.drawImage(this.dragImage, dragBounds.x, dragBounds.y, this);
			} else {
			g.drawImage(this.dragImage,
					dragBounds.x + dragProximity, dragBounds.y + dragProximity,
					dragBounds.x + dragBounds.width - dragProximity,
					dragBounds.y + dragBounds.height - dragProximity,
					dragProximity, dragProximity,
					dragBounds.width - dragProximity, dragBounds.height - dragProximity,
					this);
			}
			// paint(g, clip.x, clip.y, clip.width, clip.height, dragger, true);
			if (true) {
				g2d.setComposite(oc);
			}
		} else if (connector != null && false) {
			int lpc, size;
			
			antialiasToggle(g, true);
			size = connectorList.size();
			for (lpc = 0; lpc < size; lpc++) {
				Rectangle bounds = getRectangle(connectorList.get(lpc), "bounds");

				g.setColor(Color.black);
				g.drawLine(connx + bounds.x + (bounds.width / 2),
						conny + bounds.y + (bounds.height / 2),
						mousex, mousey);
				g.drawOval(mousex - 4, mousey - 4, 8, 8);
				g.drawOval(mousex - 3, mousey - 3, 6, 6);
			}
			antialiasToggle(g, false);
		}
		// System.out.println(System.currentTimeMillis() - time);
		// g.setClip(0, 0, getSize().width, getSize().height);
		// g.setColor(Color.red); g.drawRect(clip.x, clip.y, clip.width - 1,
		// clip.height - 1);
	}

	/**
	 * @param clipx
	 *            the cliping rectangle is relative to the component's parent
	 *            location similar to the component's bounds rectangle
	 * @param clipy
	 * @param clipwidth
	 * @param clipheight
	 * @throws java.lang.IllegalArgumentException
	 */
	void paint(Graphics g, int clipx, int clipy, int clipwidth, int clipheight,
			Object component, boolean enabled) {
		if (!getBoolean(component, "visible", true)) {
			return;
		}
		Rectangle bounds = getRectangle(component, "bounds");
		if (bounds == null) {
			return;
		}
		// negative component width indicates invalid component layout
		while (bounds.width < 0) {
			if (false)
				System.out.println("doing layout! " + getClass(component));
			bounds.width = Math.abs(bounds.width);
			doLayout(component);
		}
		// return if the component was out of the cliping rectangle
		if (!bounds.intersects(clipx, clipy, clipwidth, clipheight)) {
			return;
		}
		// set the clip rectangle relative to the component location
		clipx -= bounds.x;
		clipy -= bounds.y;
		g.translate(bounds.x, bounds.y);
		// g.setClip(0, 0, bounds.width, bounds.height);
		String classname = getClass(component);
		boolean pressed = (mousepressed == component);
		boolean inside = (mouseinside == component)
							&& ((mousepressed == null) || pressed);
		boolean focus = focusinside && (focusowner == component);
		enabled = getBoolean(component, "enabled", true); // enabled &&

		if ("label" == classname) {
			paint(component, 0, 0, bounds.width, bounds.height, g, clipx,
					clipy, clipwidth, clipheight, false, false, false, false,
					0, 0, 0, 0, false, enabled ? 'e' : 'd', "left", true, false);
		} else if (("button" == classname) || ("togglebutton" == classname)
					|| ("connector" == classname)) {
			boolean toggled = (("togglebutton" == classname) || ("connector" == classname))
								&& getBoolean(component, "selected", false);
			boolean link = ("button" == classname)
							&& (get(component, "type") == "link");
			if (link) {
				paint(component, 0, 0, bounds.width, bounds.height, g, clipx,
						clipy, clipwidth, clipheight, false, false, false,
						false, 0, 0, 0, 0, focus, enabled ? (pressed ? 'e'
								: 'l') : 'd', "center", true,
						enabled && (inside != pressed));
			} else { // disabled toggled
				int proximity = getInteger(component, "proximity", 0);
				Object parent = getParent(component);

				if (getClass(parent) == "dropzone") {
					focus = false;

					if (toggled) {
						int rrx, rry, rrw, rrh;
						
						// System.out.println("cx " + clipx + "; " + clipy + "; " + bounds);
						rrx = proximity - 4;
						rry = proximity - 4;
						rrw = bounds.width - (proximity * 2) + 8;
						rrh = bounds.height - (proximity * 2) + 8;
						if (classname == "connector") {
							rrx += 5; rry += 5;
							rrw -= 10; rrh -= 10;
						}
						g.setColor(c_select);
						g.fillRoundRect(rrx, rry, rrw, rrh, 12, 12);
						toggled = false;
					}
				}

				char mode = enabled ? ((inside != pressed) ? 'h'
						: ((pressed || toggled) ? 'p' : 'g')) : 'd';

				if ((tdi.source == null) && (popupowner == null)
					&& ((mouseinside == component) || (connector != null))
					&& ((mousepressed == null) || (connector != null))
					&& (proximity > 0)) {
					Point point;

					if (((point = getPoint(component, ":mouse")) != null)
						&& (connector != component) && (point.x != -1)
						&& (point.y != -1)) {
						int x, y;

						x = point.x;
						y = point.y;
						if (x < 4) {
							x = proximity / 2;
							switch (point.x) {
							case 0:
								break;
							case 1:
								x -= (x / 4);
								break;
							case 2:
								x -= (x / 3);
								break;
							case 3:
								x -= (x / 2);
								break;
							}
						} else if (x >= (bounds.width - 4)) {
							x = bounds.width - proximity / 2;
							switch (bounds.width - point.x) {
							case 1:
								break;
							case 2:
								x += ((proximity / 2) / 4);
								break;
							case 3:
								x += ((proximity / 2) / 3);
								break;
							case 4:
								x += ((proximity / 2) / 2);
								break;
							}
						}
						if (y < 4) {
							y = proximity / 2;
							switch (point.y) {
							case 0:
								break;
							case 1:
								y -= (y / 4);
								break;
							case 2:
								y -= (y / 3);
								break;
							case 3:
								y -= (y / 2);
								break;
							}
						} else if (y > (bounds.height - 4)) {
							y = bounds.height - proximity / 2;
							switch (bounds.height - point.y) {
							case 1:
								break;
							case 2:
								y += ((proximity / 2) / 4);
								break;
							case 3:
								y += ((proximity / 2) / 3);
								break;
							case 4:
								y += ((proximity / 2) / 2);
								break;
							}
						}
						antialiasToggle(g, true);
						g.setColor(Color.black);
						g.drawLine(bounds.width / 2, bounds.height / 2, x, y);
						g.drawOval(x - 4, y - 4, 8, 8);
						g.drawOval(x - 3, y - 3, 6, 6);
						antialiasToggle(g, false);
					}
				}

				int apadding = ("connector" == classname) ? (bounds.height / 2)
						: 0;
				int xpadding = ("connector" == classname) ? 0 : 0;
				int cspace = ("connector" == classname) ? 5 : 0;
				int dx = 0;
				boolean rounded = false;
				
				rounded = ("round" == getString(component, "style", "normal"));
					
				if (rounded) {
					dx = 6;
				}

				if (rounded) {
					int rh = (bounds.height - (proximity * 2));
					Graphics2D g2d = (Graphics2D)g;
					CubicCurve2D cc = new CubicCurve2D.Double(proximity + 6, proximity,
							proximity, proximity + rh / 4,
							proximity, proximity + rh - rh / 4 - 1,
							proximity + 6, proximity + rh - 1);
					CubicCurve2D cc2 = new CubicCurve2D.Double(bounds.width - 6 - proximity, proximity,
							bounds.width - proximity, proximity + rh / 4,
							bounds.width - proximity, proximity + rh - rh / 4 - 1,
							bounds.width - 6 - proximity, proximity + rh - 1);
					Rectangle oc;
					Color bg;
					
					antialiasToggle(g, true);
					bg = getColor(component, "background", c_bg);
					switch (mode) {
					case 'h':
						bg = bg.brighter();
						break;
					case 'p':
						bg = bg.darker();
						break;
					default:
						break;
					}
					
					oc = g.getClipBounds();
					g2d.clip(cc);
					if (g2d != null && (mode != 'h') && (mode != 'p')) {
						fill(g, proximity, proximity + 1, 6, rh - 2, true, bg);
					}
					else {
						g.setColor(bg);
						g2d.fill(cc);
					}
					g.setClip(oc);
					g.setColor(c_border);
					g2d.draw(cc);
					g2d.clip(cc2);
					if (g2d != null && (mode != 'h') && (mode != 'p')) {
						fill(g, bounds.width - 6 - proximity, proximity + 1, 8, rh - 2, true, bg);
					}
					else {
						g.setColor(bg);
						g2d.fill(cc2);
					}
					g.setClip(oc);
					g.setColor(c_border);
					g2d.draw(cc2);
					antialiasToggle(g, false);
					xpadding = 0;
				}
				
				paint(component,
						proximity + apadding + dx, proximity + cspace,
						bounds.width - (proximity * 2) - (("connector" == classname) ? bounds.height - 2: 0) - (dx * 2) + (dx > 0 ? 1 : 0),
						bounds.height - (proximity * 2) - (cspace * 2), g,
						clipx, clipy, clipwidth, clipheight,
						true, ("connector" != classname && !rounded),
						true, ("connector" != classname && !rounded), 2, xpadding, 2,
						xpadding, focus, mode, "center", true, false);

				if ("connector" == classname) {
					int lax[], lay[], rax[], ray[];
					Object angles[];
					Graphics2D g2d = null;
					Color bg;

					if (g instanceof Graphics2D) {
						g2d = (Graphics2D)g;
					}
					bg = getColor(component, "background", c_bg);
					switch (mode) {
					case 'h':
						if (bg == c_bg)
							bg = c_hover;
						else
							bg = bg.brighter();
						break;
					case 'p':
						if (bg == c_bg)
							bg = c_press;
						else
							bg = bg.darker();
						break;
					default:
						break;
					}
					/* XXX stupid */
					angles = (Object[]) get(component, ":angles");
					lax = (int[]) angles[0];
					lay = (int[]) angles[1];
					rax = (int[]) angles[2];
					ray = (int[]) angles[3];
					g.setColor(bg);
					if (g2d != null && (mode != 'h') && (mode != 'p')) {
						Polygon p = new Polygon(lax, lay, 4);
						Rectangle clip;
						
						clip = g.getClipBounds();
						g2d.clip(p);
						fill(g, 1, lay[1] + 1, lax[1], lay[2] - lay[1] - 1, true, bg);
						g.setClip(clip);
						p = new Polygon(rax, ray, 4);
						g2d.clip(p);
						fill(g, rax[0] - 1, ray[0] + 1, rax[1], ray[2] - ray[0] - 1, true, bg);
						g.setClip(clip);
					}
					else {
						g.fillPolygon(lax, lay, 4);
						g.fillPolygon(rax, ray, 4);
					}
					g.setColor(c_border);
					antialiasToggle(g, true);
					g.drawLine(lax[0], lay[0], lax[1], lay[1]);
					g.drawLine(lax[2], lay[2], lax[3], lay[3]);
					g.drawLine(rax[0], ray[0], rax[1], ray[1]);
					g.drawLine(rax[1], ray[1], rax[2], ray[2]);
					antialiasToggle(g, false);
				}

				if (connector == component ||
						(connectorList != null && connectorList.contains(component))) {
					int x, y;

					antialiasToggle(g, true);
					x = (bounds.width / 2);
					y = (bounds.height / 2);
					g.setColor(Color.black);
					g.drawOval(x - 4, y - 4, 8, 8);
					g.setColor(Color.magenta);
					g.fillOval(x - 3, y - 3, 6, 6);
					antialiasToggle(g, false);
				}
				// (enabled && ("button" == classname) && get(component, "type")
				// == "default")...
			}
		} else if ("checkbox" == classname) {
			boolean selected = getBoolean(component, "selected", false);
			String style = getString(component, "style", "normal");
			String group = getString(component, "group", null);
			Color border = enabled ? c_border : c_disable;
			Color foreground = enabled ? ((inside != pressed) ? c_hover
					: (pressed ? c_press : c_ctrl)) : c_bg;
			int dy = (bounds.height - block + 2) / 2;
			char mode;

			if (("arrow" == style) || ("blank" == style))
				mode = enabled ? ((inside != pressed) ? 'h' : 'e') : 'd';
			else
				mode = enabled ? 'e' : 'd';
			paint(component, 0, 0, bounds.width, bounds.height, g, clipx,
					clipy, clipwidth, clipheight, false, false, false, false,
					0, block + 3, 0, 0, false, mode, "left", true, false);

			if (group == null) {
				if ("normal" == style) {
					paintRect(g, 1, dy + 1, block - 2, block - 2, border,
							foreground, true, true, true, true, true);
				} else if ("arrow" == style) {
					// dy = 1;
					g.setColor(enabled ? c_text : c_disable);
					paintArrow(g, 1 + (selected ? 1 : 0), dy + 1, block - 2,
							block - 2, selected ? 'S' : 'E', 5);
				}
			} else {
				g.setColor((foreground != c_ctrl) ? foreground : c_bg);
				g.fillOval(1, dy + 1, block - 3 + evm, block - 3 + evm);
				g.setColor(border);
				g.drawOval(1, dy + 1, block - 3, block - 3);
			}
			if (focus) {
				g.setColor(c_focus);
				if (group == null) {
					if ("normal" == style) {
						g.drawRect(3, dy + 3, block - 7, block - 7);
					} else {
						g.drawRect(3, dy + 2, block - 7, block - 6);
					}
				} else {
					g.drawOval(3, dy + 3, block - 7, block - 7);
				}
			}
			if ((!selected && inside && pressed)
				|| (selected && (!inside || !pressed))) {
				antialiasToggle(g, true);
				g.setColor(enabled ? c_text : c_disable);
				if (group == null) {
					if ("normal" == style) {
						g.fillRect(3, dy + block - 9, 2 + evm, 6 + evm);
						g.drawLine(3, dy + block - 4, block - 4, dy + 3);
						g.drawLine(4, dy + block - 4, block - 4, dy + 4);
					}
				} else {
					g.fillOval(5, dy + 5, block - 10 + evm, block - 10 + evm);
					g.drawOval(4, dy + 4, block - 9, block - 9);
				}
				antialiasToggle(g, false);
			}
		} else if ("combobox" == classname) {
			if (getBoolean(component, "editable", true)
				|| getBoolean(component, "selectable", true)) {
				Image icon = getIcon(component, "icon", null);
				boolean rounded;
				String style;
				int dx = 0;

				style = getString(component, "style", "normal");
				rounded = ("round" == style);
				if (rounded)
					dx = block + 1;
				int left = (icon != null) ? icon.getWidth(this) + dx : dx;
				paintField(g, clipx, clipy, clipwidth, clipheight, component,
						bounds.width - block, bounds.height, focus, enabled,
						false, left, rounded);
				if (rounded)
					paintRoundedEdge(component, g, block, bounds.height);
				if (icon != null) {
					g.drawImage(icon, 2,
							(bounds.height - icon.getHeight(this)) / 2, this);
				}
				enabled = enabled && getBoolean(component, "pickable", true);
				paintArrow(g, bounds.width - block, 0, block, bounds.height,
						'S', enabled, inside, pressed, "down", true, false,
						true, true, true, 4);
			} else {
				paint(component, 0, 0, bounds.width, bounds.height, g, clipx,
						clipy, clipwidth, clipheight, true, true, true, true,
						1, 1, 1, 1 + block, focus,
						enabled ? ((inside != pressed) ? 'h' : (pressed ? 'p'
								: 'g')) : 'd', "left", false, false);
				g.setColor(enabled ? c_text : c_disable);
				paintArrow(g, bounds.width - block, 0, block, bounds.height,
						'S', 4);
			}
		} else if (":combolist" == classname) {
			paintScroll(component, classname, pressed, inside, focus, enabled,
					g, clipx, clipy, clipwidth, clipheight);
		} else if (("textfield" == classname) || ("passwordfield" == classname)) {
			paintField(g, clipx, clipy, clipwidth, clipheight, component,
					bounds.width, bounds.height, focus, enabled,
					("passwordfield" == classname), 0, false);
		} else if ("textarea" == classname) {
			paintScroll(component, classname, pressed, inside, focus, enabled,
					g, clipx, clipy, clipwidth, clipheight);
		} else if ("tabbedpane" == classname) {
			int i = 0;
			Object selectedtab = null;
			int selected = getInteger(component, "selected", 0);
			String placement = getString(component, "placement", "top");
			boolean horizontal = ((placement == "top") || (placement == "bottom"));
			boolean stacked = (placement == "stacked");
			int bx = stacked ? 0 : horizontal ? 2 : 1, by = stacked ? 0
					: horizontal ? 1 : 2, bw = 2 * bx, bh = 2 * by;
			// paint tabs except the selected one
			g.clipRect(0, 0, bounds.width, bounds.height); // +clip
			for (Object tab = get(component, ":comp"); tab != null; tab = get(
					tab, ":next")) {
				Rectangle r = getRectangle(tab, "bounds");
				if (selected != i) {
					boolean hover = inside && (mousepressed == null)
									&& (insidepart == tab);
					boolean tabenabled = enabled
											&& getBoolean(tab, "enabled", true);
					paint(tab, r.x + bx, r.y + by, r.width - bw, r.height - bh,
							g, clipx, clipy, clipwidth, clipheight,
							(placement != "bottom"), (placement != "right"),
							!stacked && (placement != "top"),
							(placement != "left"), 1, 3, 1, 3, false,
							tabenabled ? (hover ? 'h' : 'g') : 'd', "left",
							true, false);

				} else {
					selectedtab = tab;
					// paint tabbedpane border
					paint(
							tab,
							(placement == "left") ? r.width - 1 : 0,
							stacked ? (r.y + r.height - 1)
									: (placement == "top") ? r.height - 1 : 0,
							(horizontal || stacked) ? bounds.width
									: (bounds.width - r.width + 1),
							stacked ? (bounds.height - r.y - r.height + 1)
									: horizontal ? (bounds.height - r.height + 1)
											: bounds.height, g, true, true,
							true, true, enabled ? 'e' : 'd');
					Object comp = get(selectedtab, ":comp");
					if ((comp != null) && getBoolean(comp, "visible", true)) {
						clipx -= r.x;
						clipy -= r.y;
						g.translate(r.x, r.y); // relative to tab
						paint(g, clipx, clipy, clipwidth, clipheight, comp,
								enabled);
						clipx += r.x;
						clipy += r.y;
						g.translate(-r.x, -r.y);
					}
				}
				i++;
			}

			// paint selected tab and its content
			if (selectedtab != null) {
				Rectangle r = getRectangle(selectedtab, "bounds");
				// paint selected tab
				int ph = stacked ? 3 : (horizontal ? 5 : 4);
				int pv = stacked ? 1 : (horizontal ? 2 : 3);
				paint(selectedtab, r.x, r.y, r.width, r.height, g, clipx,
						clipy, clipwidth, clipheight, (placement != "bottom"),
						(placement != "right"), !stacked
												&& (placement != "top"),
						(placement != "left"), pv, ph, pv, ph, focus,
						enabled ? 'b' : 'i', "left", true, false);
			}
			g.setClip(clipx, clipy, clipwidth, clipheight); // +clip
		} else if (("panel" == classname) || ("dialog" == classname) || ("dropzone" == classname)) {
			int titleheight = getInteger(component, ":titleheight", 0);
			if ("dialog" == classname) {
				paint(component, 0, 0, bounds.width, 3 + titleheight, g, clipx,
						clipy, clipwidth, clipheight, true, true, false, true,
						1, 2, 1, 2, false, 'g', "left", false, false);
				int controlx = bounds.width - titleheight - 1;
				if (get(component, "close") != null) {
					paint(component, g, controlx, 3, titleheight - 2,
							titleheight - 2, 'c',
							pressedpart == ":closebutton" ? 'p'
									: insidepart == ":closebutton" ? 'h' : 'g');
					controlx -= titleheight;
				}
				if (getBoolean(component, "maximizable", false)) {
					paint(component, g, controlx, 3, titleheight - 2,
							titleheight - 2, 'm', 'g');
					controlx -= titleheight;
				}
				if (getBoolean(component, "iconifiable", false)) {
					paint(component, g, controlx, 3, titleheight - 2,
							titleheight - 2, 'i', 'g');
				}
				paintRect(g, 0, 3 + titleheight, bounds.width, bounds.height
																- 3
																- titleheight,
						c_border, c_press, false, true, true, true, true); // lower
																			// part
																			// excluding
																			// titlebar
				// drawing a visible decorator in the lower right corner of the
				// dialog if it is resizable
				if (getBoolean(component, "resizable", false)) {
					g.setColor(c_resizeborder);
					g.fillRect(bounds.width - 3, bounds.height - 16, 3, 16);
					g.fillRect(bounds.width - 16, bounds.height - 3, 16, 3);
				}
				paint(component, // content area
						3, 3 + titleheight, bounds.width - 6, bounds.height - 6
																- titleheight,
						g, true, true, true, true, 'b');
			} else { // panel
				boolean border = getBoolean(component, "border", false);
				int radius = getInteger(component, "radius", -1);
				
				if (radius > -1) {
					Color bg = getColor(component, "background", c_bg);
					
					if (!enabled)
						bg = bg.darker();
					antialiasToggle(g, true);
					g.setColor(bg);
					g.fillRoundRect(0, titleheight / 2, bounds.width, bounds.height, radius, radius);
					if (border) {
						Color inner = bg.darker();

						g.setColor(inner);
						g.drawRoundRect(1, titleheight / 2 + 1, bounds.width - 2, bounds.height - 2, radius - 2, radius - 2);
						g.setColor(inner.darker());
						g.drawRoundRect(0, titleheight / 2, bounds.width, bounds.height, radius, radius);
					}
					antialiasToggle(g, false);
				}
				else {
					paint(component, 0, titleheight / 2, bounds.width,
							bounds.height - (titleheight / 2), g, border, border,
							border, border, enabled ? 'e' : 'd');
				}
				
				paint(component, 0, 0, bounds.width, titleheight, // panel
																	// title
						g, clipx, clipy, clipwidth, clipheight, false, false,
						false, false, 0, 3, 0, 3, false, enabled ? 'x' : 'd',
						"left", false, false);
			}

			if (get(component, ":port") != null) {
				paintScroll(component, classname, pressed, inside, focus,
						enabled, g, clipx, clipy, clipwidth, clipheight);
			} else {
				for (Object comp = get(component, ":comp"); comp != null; comp = get(
						comp, ":next")) {
					if ("connector" == getClass(comp)) {
						Rectangle cbounds;

						cbounds = getRectangle(comp, "cbounds");
						/*
						 * XXX Can't use cbounds.intesects because the width and
						 * height can be negative, which is how we tell which
						 * direction the line should go.
						 */
						if ((cbounds != null)) {
							int x1, x2, y1, y2;

							x1 = cbounds.x;
							y1 = cbounds.y;
							x2 = cbounds.x + cbounds.width - 1;
							y2 = cbounds.y + cbounds.height - 1;
							
							antialiasToggle(g, true);
							g.setColor(Color.black);
							g.drawLine(x1, y1, x2, y2);
							g.fillOval(x1 - 4, y1 - 4, 9, 9);
							g.fillOval(x2 - 4, y2 - 4, 9, 9);
							antialiasToggle(g, false);
						}
					}
				}
				for (Object comp = get(component, ":comp"); comp != null; comp = get(
						comp, ":next")) {
					paint(g, clipx, clipy, clipwidth, clipheight, comp, enabled);
				}
				if (connectorList != null && getParent(connector) == component) {
					int lpc, size;

					antialiasToggle(g, true);
					size = connectorList.size();
					for (lpc = 0; lpc < size; lpc++) {
						Rectangle cbounds = getRectangle(connectorList.get(lpc), "bounds");

						g.setColor(Color.black);
						g.drawLine(cbounds.x + (cbounds.width / 2),
								cbounds.y + (cbounds.height / 2),
								connx, conny);
						g.drawOval(connx - 4, conny - 4, 8, 8);
						g.drawOval(connx - 3, conny - 3, 6, 6);
					}
					antialiasToggle(g, false);
				}
			}

		} else if ("desktop" == classname) {
			paintRect(g, 0, 0, bounds.width, bounds.height, c_border, c_bg,
					false, false, false, false, true);
			paintReverse(g, clipx, clipy, clipwidth, clipheight, get(component,
					":comp"), enabled);
			// g.setColor(Color.red); if (clip != null) g.drawRect(clipx, clipy,
			// clipwidth, clipheight);
			if ((tooltipowner != null) && (component == content)) {
				Rectangle r = getRectangle(tooltipowner, ":tooltipbounds");
				paintRect(g, r.x, r.y, r.width, r.height, c_border, c_tooltip,
						true, true, true, true, true);
				String text = null;

				if (insidepart == "proximity") {
					text = getString(tooltipowner, "proximitytooltip", null);
				}
				if (text == null) {
					text = getString(tooltipowner, "tooltip", null);
				}

				if (text != null) {
					Font ttFont = null;

					if (insidepart != "proximity")
						ttFont = (Font) get(mouseinside, "tooltipfont");

					if (ttFont == null)
						ttFont = this.font;
					g.setFont(ttFont);
					g.setColor(c_text);
					g.drawString(text, r.x + 2, r.y
												+ g.getFontMetrics()
														.getAscent() + 2); // +nullpointerexception
				}
			}
		} else if ("spinbox" == classname) {
			paintField(g, clipx, clipy, clipwidth, clipheight, component,
					bounds.width - block, bounds.height, focus, enabled, false,
					0, false);
			paintArrow(g, bounds.width - block, 0, block, bounds.height / 2,
					'N', enabled, inside, pressed, "up", true, false, false,
					true, true, 4);
			paintArrow(g, bounds.width - block, bounds.height / 2, block,
					bounds.height - (bounds.height / 2), 'S', enabled, inside,
					pressed, "down", true, false, true, true, true, 4);
		} else if ("progressbar" == classname) {
			int minimum = getInteger(component, "minimum", 0);
			int maximum = getInteger(component, "maximum", 100);
			int value = getInteger(component, "value", 0);
			boolean horizontal = ("vertical" != get(component, "orientation"));
			int length;
			if (value < minimum)
				value = minimum;
			if (maximum != minimum) {
				length = (value - minimum)
							* ((horizontal ? bounds.width : bounds.height) - 1)
							/ (maximum - minimum);
			} else {
				length = horizontal ? bounds.width : bounds.height;
			}
			paintRect(g, 0, 0, horizontal ? length : bounds.width,
					horizontal ? bounds.height : length, enabled ? c_border
							: c_disable, c_select, true, true, horizontal,
					!horizontal, true);
			paintRect(g, horizontal ? length : 0, horizontal ? 0 : length,
					horizontal ? (bounds.width - length) : bounds.width,
					horizontal ? bounds.height : (bounds.height - length),
					enabled ? c_border : c_disable, c_bg, true, true, true,
					true, true);
		} else if ("slider" == classname) {
			int minimum = getInteger(component, "minimum", 0);
			int maximum = getInteger(component, "maximum", 100);
			int value = getInteger(component, "value", minimum);
			boolean horizontal = ("vertical" != get(component, "orientation"));
			int length = (value - minimum)
							* ((horizontal ? bounds.width : bounds.height) - block)
							/ (maximum - minimum);
			paintRect(g, horizontal ? 0 : 3, horizontal ? 3 : 0,
					horizontal ? length : (bounds.width - 6),
					horizontal ? (bounds.height - 6) : length,
					enabled ? c_border : c_disable, c_bg, true, true,
					horizontal, !horizontal, true);
			paintRect(g, horizontal ? length : 0, horizontal ? 0 : length,
					horizontal ? block : bounds.width,
					horizontal ? bounds.height : block, enabled ? c_border
							: c_disable, enabled ? c_ctrl : c_bg, true, true,
					true, true, true);
			if (focus) {
				g.setColor(c_focus);
				g.drawRect(horizontal ? (length + 2) : 2, horizontal ? 2
						: (length + 2),
						(horizontal ? block : bounds.width) - 5,
						(horizontal ? bounds.height : block) - 5);
				// g.drawRect(length + 1, 1, block - 3, bounds.height - 3);
			}
			paintRect(g, horizontal ? (block + length) : 3, horizontal ? 3
					: (block + length), bounds.width
										- (horizontal ? (block + length) : 6),
					bounds.height - (horizontal ? 6 : (block + length)),
					enabled ? c_border : c_disable, c_bg, horizontal,
					!horizontal, true, true, true);
		} else if ("splitpane" == classname) {
			boolean horizontal = ("vertical" != get(component, "orientation"));
			int divider = getInteger(component, "divider", -1);
			paintRect(g, horizontal ? divider : 0, horizontal ? 0 : divider,
					horizontal ? 5 : bounds.width, horizontal ? bounds.height
							: 5, c_border, c_bg, false, false, false, false,
					true);
			g.setColor(enabled ? (focus ? c_focus : c_border) : c_disable);
			int xy = horizontal ? bounds.height : bounds.width;
			int xy1 = Math.max(0, xy / 2 - 12);
			int xy2 = Math.min(xy / 2 + 12, xy - 1);
			for (int i = divider + 1; i < divider + 4; i += 2) {
				if (horizontal) {
					g.drawLine(i, xy1, i, xy2);
				} else {
					g.drawLine(xy1, i, xy2, i);
				}
			}
			Object comp1 = get(component, ":comp");
			if (comp1 != null) {
				paint(g, clipx, clipy, clipwidth, clipheight, comp1, enabled);
				Object comp2 = get(comp1, ":next");
				if (comp2 != null) {
					paint(g, clipx, clipy, clipwidth, clipheight, comp2,
							enabled);
				}
			}
		} else if (("list" == classname) || ("table" == classname)
					|| ("tree" == classname)) {
			paintScroll(component, classname, pressed, inside, focus, enabled,
					g, clipx, clipy, clipwidth, clipheight);
		} else if ("separator" == classname) {
			g.setColor(enabled ? c_border : c_disable);
			g.fillRect(0, 0, bounds.width + evm, bounds.height + evm);
		} else if ("menubar" == classname) {
			Object selected = get(component, "selected");
			boolean bottom = ("bottom" == get(component, "placement"));
			int lastx = 0;
			for (Object menu = get(component, ":comp"); menu != null; menu = get(
					menu, ":next")) {
				Rectangle mb = getRectangle(menu, "bounds");
				if (clipx + clipwidth <= mb.x) {
					break;
				}
				if (clipx >= mb.x + mb.width) {
					continue;
				}
				boolean menuenabled = enabled
										&& getBoolean(menu, "enabled", true);
				boolean armed = (selected == menu);
				boolean hoover = (selected == null) && (insidepart == menu);
				paint(menu, mb.x, 0, mb.width, bounds.height, g, clipx, clipy,
						clipwidth, clipheight, // TODO disabled
						bottom || armed, armed, !bottom || armed, armed, 1, 3,
						1, 3, false, enabled ? (menuenabled ? (armed ? 's'
								: (hoover ? 'h' : 'g')) : 'r') : 'd', "left",
						true, false);
				lastx = mb.x + mb.width;
			}
			paintRect(g, lastx, 0, bounds.width - lastx, bounds.height,
					enabled ? c_border : c_disable, enabled ? c_ctrl : c_bg,
					bottom, false, !bottom, false, true);
		} else if (":popup" == classname) {
			paintRect(g, 0, 0, bounds.width, bounds.height, c_border, c_textbg,
					true, true, true, true, true);
			Object selected = get(component, "selected");
			for (Object menu = get(get(component, "menu"), ":comp"); menu != null; menu = get(
					menu, ":next")) {
				if (!getBoolean(menu, "visible", true))
					continue;
				Rectangle r = getRectangle(menu, "bounds");
				if (clipy + clipheight <= r.y) {
					break;
				}
				if (clipy >= r.y + r.height) {
					continue;
				}
				String itemclass = getClass(menu);
				if (itemclass == "separator") {
					g.setColor(c_border);
					g
							.fillRect(r.x, r.y, bounds.width - 2 + evm,
									r.height + evm);
				} else {
					boolean armed = (selected == menu);
					boolean menuenabled = getBoolean(menu, "enabled", true);
					paint(
							menu,
							r.x,
							r.y,
							bounds.width - 2,
							r.height,
							g,
							clipx,
							clipy,
							clipwidth,
							clipheight,
							false,
							false,
							false,
							false,
							2,
							(itemclass == "checkboxmenuitem") ? (block + 7) : 4,
							2, 4, false, menuenabled ? (armed ? 's' : 't')
									: 'd', "left", true, false);
					if (itemclass == "checkboxmenuitem") {
						boolean checked = getBoolean(menu, "selected", false);
						String group = getString(menu, "group", null);
						g.translate(r.x + 4, r.y + 2);
						g.setColor(menuenabled ? c_border : c_disable);
						if (group == null) {
							g.drawRect(1, 1, block - 3, block - 3);
						} else {
							g.drawOval(1, 1, block - 3, block - 3);
						}
						if (checked) {
							g.setColor(menuenabled ? c_text : c_disable);
							if (group == null) {
								g.fillRect(3, block - 9, 2 + evm, 6 + evm);
								g.drawLine(3, block - 4, block - 4, 3);
								g.drawLine(4, block - 4, block - 4, 4);
							} else {
								g.fillOval(5, 5, block - 10 + evm, block - 10
																	+ evm);
								g.drawOval(4, 4, block - 9, block - 9);
							}
						}
						g.translate(-r.x - 4, -r.y - 2);
					}
					if (itemclass == "menu") {
						paintArrow(g, r.x + bounds.width - block, r.y, block,
								r.height, 'E', 4);
					} else {
						String accelerator = getAccelerator(menu);
						if (accelerator != null) { // TODO
							g.drawString(accelerator,
									bounds.width
											- 4
											- getFontMetrics(font).stringWidth(
													accelerator), r.y + 2 + 10);
						}
					}
				}
			}
		} else if ("bean" == classname) {
			g.clipRect(0, 0, bounds.width, bounds.height);
			((Component) get(component, "bean")).paint(g);
			g.setClip(clipx, clipy, clipwidth, clipheight);
		} else if ("connector" == classname) {
			paint(component, 0, 0, bounds.width, bounds.height, g, clipx,
					clipy, clipwidth, clipheight, true, true, true, true, 2, 5,
					2, 5, focus, 'g', "center", true, false);
		} else if (instance(classname, "prim")) {
			boolean border = getBoolean(component, "border");
			String placement = getString(component, "placement");
			Color fg;

			paint(component, 0, 0, bounds.width, bounds.height, g, border,
					border, border, border, enabled ? 'e' : 'd');
			if ((fg = (Color) get(component, "foreground")) == null)
				fg = Color.black;
			g.setColor(fg);
			if ("oval" == classname) {
				if (getBoolean(component, "fill"))
					g.fillOval(0, 0, bounds.width, bounds.height);
				else
					g.drawOval(0, 0, bounds.width, bounds.height);
			} else if ("rect" == classname) {
				if (getBoolean(component, "fill"))
					g.fillRect(0, 0, bounds.width, bounds.height);
				else
					g.drawRect(0, 0, bounds.width, bounds.height);
			} else if ("note" == classname) {
				int x = 0, y = 0, width = bounds.width, height = bounds.height, yoffset;

				yoffset = getInteger(component, "yoffset");
				switch (placement.charAt(0)) {
				case 'x':
					y += yoffset;
					height -= yoffset;
					break;
				}
				paint(component, x, y, width, height, g, clipx, clipy,
						clipwidth, clipheight, border, border, border, border,
						0, 0, 0, 0, false, 'e', "center", false, false);
				switch (placement.charAt(0)) {
				case 'x':
					// System.out.println("y " + y + " " + bounds.height);
					g.drawLine(x + (bounds.width / 2), 0,
							x + (bounds.width / 2), y);
					break;
				}
			}
		} else {
			// printcomp(component);
			throw new IllegalArgumentException(classname);
		}
		g.translate(-bounds.x, -bounds.y);
		clipx += bounds.x;
		clipy += bounds.y;
	}

	/**
	 * 
	 */
	void paintReverse(Graphics g, int clipx, int clipy, int clipwidth,
			int clipheight, Object component, boolean enabled) {
		if (component != null) {
			Rectangle bounds = getRectangle(component, "bounds");

			if (bounds != null &&
					((clipx < bounds.x)
							|| (clipx + clipwidth > bounds.x + bounds.width)
							|| (clipy < bounds.y)
							|| (clipy + clipheight > bounds.y + bounds.height))) {
				paintReverse(g, clipx, clipy, clipwidth, clipheight, get(
						component, ":next"), enabled);
				if (getBoolean(component, "modal", false)) {
					g.setColor(new Color(0, 0, 0, 0xff / 3));
					g.fillRect(clipx, clipy, clipwidth, clipheight);
				}
			}
			paint(g, clipx, clipy, clipwidth, clipheight, component, enabled);
		}
	}

	void paintRoundedEdge(Object component, Graphics g, int width, int height) {
		Color bg;
		
		antialiasToggle(g, true);
		bg = getColor(component, "background", c_bg);
		g.setColor(bg);
		g.fillRect(0, 0, width, height);
		g.setColor(c_textbg);
		g.fillArc(width / 2 + 2, 0, width - 1, height - 1, 90, 180);
		g.setColor(c_border);
		g.drawArc(width / 2 + 1, 0, width, height - 1, 90, 180);
		antialiasToggle(g, false);
	}

	/**
	 * 
	 */
	void paintField(Graphics g, int clipx, int clipy, int clipwidth,
			int clipheight, Object component, int width, int height,
			boolean focus, boolean enabled, boolean hidden, int left,
			boolean rounded) {
		boolean editable = getBoolean(component, "editable", true);
		Color tf_bg = getColor(component, "background", c_bg);
		Color tf_border = getBoolean(component, "border") ? (enabled ? c_border
				: c_disable) : tf_bg;
		paintRect(g, 0, 0, width, height, tf_border, editable ? c_textbg
				: tf_bg, true, !rounded, true, !rounded, true);
		g.clipRect(1 + left, 1, width - left - 2, height - 2);

		String text = getString(component, "text", "");
		int offset = getInteger(component, ":offset", 0);
		Font currentfont = (Font) get(component, "font");
		if (currentfont != null) {
			g.setFont(currentfont);
		}
		FontMetrics fm = g.getFontMetrics();

		int caret = 0;
		if (focus) {
			int start = getInteger(component, "start", 0);
			int end = getInteger(component, "end", 0);
			caret = hidden ? (fm.charWidth('*') * end) : fm.stringWidth(text
					.substring(0, end));
			if (start != end) {
				int is = hidden ? (fm.charWidth('*') * start) : fm
						.stringWidth(text.substring(0, start));
				g.setColor(c_select);
				g.fillRect(2 + left - offset + Math.min(is, caret), 1, Math
						.abs(caret - is)
																		+ evm,
						height - 2 + evm);
			}
		}

		if (focus) {
			g.setColor(c_caret);
			g.fillRect(1 + left - offset + caret, 2, 1 + evm, height - 4 + evm);
		}

		g.setColor(enabled ? c_text : c_disable);
		int fx = 2 + left - offset;
		int fy = (height + fm.getAscent() - fm.getDescent()) / 2;
		if (hidden) {
			int fh = fm.charWidth('*');
			for (int i = text.length(); i > 0; i--) {
				g.drawString("*", fx, fy);
				fx += fh;
			}
		} else if ((focusowner != component)
					&& ((text == ThinletController.MULTIPLE_SELECTION)
						|| (text == ThinletController.NO_SELECTION) || (text == ThinletController.NOT_APPLICABLE))) {
			g.setColor(c_empty);
			if (text == ThinletController.MULTIPLE_SELECTION) {
				if (getClass(component) == "spinbox")
					text = "?";
				else
					text = "<Multiple Selection>";
			}
			else if (text == ThinletController.NO_SELECTION) {
				if (getClass(component) == "spinbox")
					text = "-";
				else
					text = "<No Selection>";
			}
			else
				text = "<Not Applicable>";
			g.drawString(text, fx, fy);
		} else if ((focusowner != component) && ((text == null) || text.length() == 0)) {
			g.setColor(c_empty);
			g.drawString(getString(component, "empty", ""), fx, fy);
		} else {
			boolean underline[] = (boolean[]) get(component, "underline");
			g.drawString(text, fx, fy);

			if (underline != null) {
				int ux = fx, lpc, length;

				length = text.length();
				if (false)
					System.out.println(text);
				for (lpc = 0; (lpc < length) && (lpc < underline.length); lpc++) {
					int cw = fm.charWidth(text.charAt(lpc));

					if (underline[lpc]) {
						if (false)
							System.out.print("_");
						g.drawLine(ux + 1, fy + 1, ux + cw - 2, fy + 1);
					} else {
						if (false)
							System.out.print(" ");
					}
					ux += cw;
				}
				if (false)
					System.out.println();
			}
		}
		if (currentfont != null) {
			g.setFont(font);
		}
		g.setClip(clipx, clipy, clipwidth, clipheight);
	}

	/**
	 * @param component
	 *            scrollable widget
	 * @param classname
	 * @param pressed
	 * @param inside
	 * @param focus
	 * @param enabled
	 * @param g
	 *            grahics context
	 * @param clipx
	 *            current cliping x location relative to the component
	 * @param clipy
	 *            y location of the cliping area relative to the component
	 * @param clipwidth
	 *            width of the cliping area
	 * @param clipheight
	 *            height of the cliping area
	 * @param header
	 *            column height
	 * @param topborder
	 *            bordered on the top if true
	 * @param border
	 *            define left, bottom, and right border if true
	 */
	void paintScroll(Object component, String classname, boolean pressed,
			boolean inside, boolean focus, boolean enabled, Graphics g,
			int clipx, int clipy, int clipwidth, int clipheight) {
		Rectangle port = getRectangle(component, ":port");
		Rectangle horizontal = getRectangle(component, ":horizontal");
		Rectangle vertical = getRectangle(component, ":vertical");
		Rectangle view = getRectangle(component, ":view");

		if (port == null)
			return;

		if (horizontal != null) { // paint horizontal scrollbar
			int x = horizontal.x;
			int y = horizontal.y;
			int width = horizontal.width;
			int height = horizontal.height;
			paintArrow(g, x, y, block, height, 'W', enabled, inside, pressed,
					"left", true, true, true, false, true, 4);
			paintArrow(g, x + width - block, y, block, height, 'E', enabled,
					inside, pressed, "right", true, false, true, true, true, 4);

			int track = width - (2 * block);
			if (track < 10) {
				paintRect(g, x + block, y, track, height, enabled ? c_border
						: c_disable, c_bg, true, true, true, true, true);
			} else {
				int knob = Math.max(track * port.width / view.width, 10);
				int decrease = view.x * (track - knob)
								/ (view.width - port.width);
				paintRect(g, x + block, y, decrease, height, enabled ? c_border
						: c_disable, c_bg, false, true, true, false, true);
				paintRect(g, x + block + decrease, y, knob, height,
						enabled ? c_border : c_disable,
						enabled ? c_ctrl : c_bg, true, true, true, true, true);
				int n = Math.min(5, (knob - 4) / 3);
				g.setColor(enabled ? c_border : c_disable);
				int cx = (x + block + decrease) + (knob + 2 - n * 3) / 2;
				for (int i = 0; i < n; i++) {
					g.drawLine(cx + i * 3, y + 3, cx + i * 3, y + height - 5);
				}
				int increase = track - decrease - knob;
				paintRect(g, x + block + decrease + knob, y, increase, height,
						enabled ? c_border : c_disable, c_bg, false, false,
						true, true, true);
			}
		}

		if (vertical != null) { // paint vertical scrollbar
			int x = vertical.x;
			int y = vertical.y;
			int width = vertical.width;
			int height = vertical.height;
			paintArrow(g, x, y, width, block, 'N', enabled, inside, pressed,
					"up", true, true, false, true, false, 4);
			paintArrow(g, x, y + height - block, width, block, 'S', enabled,
					inside, pressed, "down", false, true, true, true, false, 4);

			int track = height - (2 * block);
			if (track < 10) {
				paintRect(g, x, y + block, width, track, enabled ? c_border
						: c_disable, c_bg, true, true, true, true, false);
			} else {
				int knob = Math.max(track * port.height / view.height, 10);
				int decrease = view.y * (track - knob)
								/ (view.height - port.height);
				paintRect(g, x, y + block, width, decrease, enabled ? c_border
						: c_disable, c_bg, true, false, false, true, false);
				paintRect(g, x, y + block + decrease, width, knob,
						enabled ? c_border : c_disable,
						enabled ? c_ctrl : c_bg, true, true, true, true, false);
				int n = Math.min(5, (knob - 4) / 3);
				g.setColor(enabled ? c_border : c_disable);
				int cy = (y + block + decrease) + (knob + 2 - n * 3) / 2;
				for (int i = 0; i < n; i++) {
					g.drawLine(x + 3, cy + i * 3, x + width - 5, cy + i * 3);
				}
				int increase = track - decrease - knob;
				paintRect(g, x, y + block + decrease + knob, width, increase,
						enabled ? c_border : c_disable, c_bg, false, false,
						true, true, false);
			}
		}

		boolean hneed = (horizontal != null);
		boolean vneed = (vertical != null);
		if (("panel" != classname)
			&& ("dialog" != classname) && ("dropzone" != classname)
			&& (("textarea" != classname) || getBoolean(component, "border",
					true))) {
			paintRect(g, port.x - 1, port.y - 1, port.width + (vneed ? 1 : 2),
					port.height + (hneed ? 1 : 2), enabled ? c_border
							: c_disable, c_textbg, true, true, !hneed, !vneed,
					true);
			if ("table" == classname) {
				Object header = get(component, "header");
				if (header != null) {
					int[] columnwidths = (int[]) get(component, ":widths");
					Object column = get(header, ":comp");
					int x = 0;
					Object prevcolumn = null;
					int pi = 0;
					g.clipRect(0, 0, port.width + 2, port.y); // not 2 and
																// decrease clip
																// area...
					for (int i = 0; i < columnwidths.length; i++) {

						if (prevcolumn != null && columnwidths[i] > 0) {
							int width = columnwidths[pi];
							// pressed (p), hovered (h), selected (b), default
							// (g) or disabled (d)
							paint(
									prevcolumn,
									x - view.x,
									0,
									width,
									port.y - 1,
									g,
									clipx,
									clipy,
									clipwidth,
									clipheight,
									true,
									true,
									false,
									false,
									1,
									1,
									0,
									0,
									false,
									enabled ? (((pressedpart == prevcolumn) && (insidepart == prevcolumn)) ? 'p'
											: (insidepart == prevcolumn ? 'h'
													: (getBoolean(prevcolumn,
															"selected") ? 'b'
															: 'g')))
											: 'd', "left", false, false);

							Object sort = get(prevcolumn, "sort"); // "none",
																	// "ascent",
																	// "descent"
							if (sort != null && sort != "none") {
								paintArrow(g, x - view.x + width - block, 0,
										block, port.y,
										(sort == "descent") ? 'S' : 'N', 4);
							}
							x += width;
							prevcolumn = null;
						}

						if (columnwidths[i] > 0) {
							prevcolumn = column;
							pi = i;
						}
						column = get(column, ":next");
					}
					if (prevcolumn != null) {
						int width = (view.width - x + 2);
						// pressed (p), hovered (h), selected (b), default (g)
						// or disabled (d)
						paint(
								prevcolumn,
								x - view.x,
								0,
								width,
								port.y - 1,
								g,
								clipx,
								clipy,
								clipwidth,
								clipheight,
								true,
								true,
								false,
								true,
								1,
								1,
								0,
								0,
								false,
								enabled ? (((pressedpart == prevcolumn) && (insidepart == prevcolumn)) ? 'p'
										: (insidepart == prevcolumn ? 'h'
												: (getBoolean(prevcolumn,
														"selected") ? 'b' : 'g')))
										: 'd', "left", false, false);

						Object sort = get(prevcolumn, "sort"); // "none",
																// "ascent",
																// "descent"
						if (sort != null && sort != "none") {
							paintArrow(g, x - view.x + width - block, 0, block,
									port.y, (sort == "descent") ? 'S' : 'N', 4);
						}
					}
					g.setClip(clipx, clipy, clipwidth, clipheight);
				}
			}
		}
		int x1 = Math.max(clipx, port.x);
		int x2 = Math.min(clipx + clipwidth, port.x + port.width);
		int y1 = Math.max(clipy, port.y);
		int y2 = Math.min(clipy + clipheight, port.y + port.height);
		if ((x2 > x1) && (y2 > y1)) {
			g.clipRect(x1, y1, x2 - x1, y2 - y1);
			g.translate(port.x - view.x, port.y - view.y);

			paint(component, classname, focus, enabled, g,
					view.x - port.x + x1, view.y - port.y + y1, x2 - x1,
					y2 - y1, port.width, view.width);

			g.translate(view.x - port.x, view.y - port.y);
			g.setClip(clipx, clipy, clipwidth, clipheight);
		}
	}

	/**
	 * Paint scrollable content
	 * 
	 * @param component
	 *            a panel
	 */
	void paint(Object component, String classname, boolean focus,
			boolean enabled, Graphics g, int clipx, int clipy, int clipwidth,
			int clipheight, int portwidth, int viewwidth) {
		if ("textarea" == classname) {
			char[] chars = (char[]) get(component, ":text");
			int start = focus ? getInteger(component, "start", 0) : 0;
			int end = focus ? getInteger(component, "end", 0) : 0;
			int is = Math.min(start, end);
			int ie = Math.max(start, end);
			Font customfont = (Font) get(component, "font");
			if (customfont != null) {
				g.setFont(customfont);
			}
			FontMetrics fm = g.getFontMetrics();
			int fontascent = fm.getAscent();
			int fontheight = fm.getHeight();
			boolean editable, didHighlight = false;
			Color fg = enabled ? getColor(component, "foreground", c_text) : c_disable;
			int ascent = 1;
			int xoff = 1, coff = 0;

			editable = getBoolean(component, "editable", true);
			for (int i = 0, j = 0; j <= chars.length; j++) {
				if ((j == chars.length) || (chars[j] == '\n')
					|| (chars[j] == '\b')) {
					if (clipy + clipheight <= ascent) {
						break;
					} // the next lines are bellow paint rectangle
					if (clipy < ascent + fontheight) { // this line is not
														// above painting area
						if (focus && !didHighlight && (is != ie) && (ie >= i)
							&& (is <= j)) {
							int xs = (is < i) ? -1
									: ((j < chars.length) && (chars[j] != '\b') && (is > j) ? (viewwidth - 1)
											: asciiCharsWidth(fm, chars, i, is
																			- i));
							int xe = ((j != -1) && (ie > j) && (chars[j] != '\b')) ? (viewwidth - 1)
									: asciiCharsWidth(fm, chars, i, ie - i);
							g.setColor(c_select);
							g.fillRect(1 + xs, ascent, xe - xs + evm,
									fontheight + evm);
							didHighlight = true;
						}
						g.setColor(fg);
						g.drawChars(chars, i,
								j
										- i
										- (j != chars.length
											&& chars[j] == '\b' ? 1 : 0), xoff,
								ascent + fontascent);
						if (focus) {
							if ((end >= i) && (end <= j) && editable) {
								int caret = coff
											+ fm.charsWidth(chars, i, end - i);
								g.setColor(c_caret);
								g.fillRect(caret, ascent, 1 + evm, fontheight
																	+ evm);
							} else if ((j < chars.length) && (chars[j] == '\b')) {
								coff += fm.charsWidth(chars, i, j - i - 1);
							}
						}
					}
					if (j < chars.length) {
						switch (chars[j]) {
						case '\b':
							xoff += fm.charsWidth(chars, i, j - i - 1);
							if (j < chars.length) {
								g.drawChars(chars, j - 1, 1, xoff,
										ascent + fontascent + 1);
							}
							break;
						default:
							didHighlight = false;
							ascent += fontheight;
							xoff = 1;
							coff = 0;
							break;
						}
					} else {
						ascent += fontheight;
						xoff = 1;
					}
					i = j + 1;
				}
			}
			if (customfont != null) {
				g.setFont(font);
			} // restore the default font
		} else if (":combolist" == classname) {
			Object lead = get(component, ":lead");
			for (Object choice = get(get(component, "combobox"), ":comp"); choice != null; choice = get(
					choice, ":next")) {
				Rectangle r = getRectangle(choice, "bounds");
				if (r == null) {
					continue;
				}
				if (clipy + clipheight <= r.y) {
					break;
				}
				if (clipy >= r.y + r.height) {
					continue;
				}
				paint(
						choice,
						r.x,
						r.y,
						portwidth,
						r.height,
						g,
						clipx,
						clipy,
						clipwidth,
						clipheight,
						false,
						false,
						false,
						false,
						2,
						4,
						2,
						4,
						false,
						getBoolean(choice, "enabled", true) ? ((lead == choice) ? 's'
								: 't')
								: 'd', "left", false, false);
			}
		} else if (("panel" == classname) || ("dialog" == classname) || ("dropzone" == classname)) {
			for (Object comp = get(component, ":comp"); comp != null; comp = get(
					comp, ":next")) {
				if ("connector" == getClass(comp)) {
					Rectangle lbounds, rbounds;
					Object left, right;

					left = find(getParent(comp), getString(comp, "left"));
					right = find(getParent(comp), getString(comp, "right"));
					lbounds = getRectangle(left, "bounds");
					rbounds = getRectangle(right, "bounds");
					if ((lbounds != null) && (rbounds != null)) {
						g.setColor(Color.black);
						g.drawLine(lbounds.x + (lbounds.width / 2),
								lbounds.y + (lbounds.height / 2),
								rbounds.x + (rbounds.width / 2),
								rbounds.y + (rbounds.height / 2));
					}
				}
			}
			for (Object comp = get(component, ":comp"); comp != null; comp = get(
					comp, ":next")) {
				paint(g, clipx, clipy, clipwidth, clipheight, comp, enabled);
			}
		} else { // if (("list" == classname) || ("table" == classname) ||
					// ("tree" == classname))
			Object lead = get(component, ":lead");
			int[] columnwidths = ("table" == classname) ? ((int[]) get(
					component, ":widths")) : null;
			boolean line = getBoolean(component, "line", true);
			int iline = line ? 1 : 0;
			boolean angle = ("tree" == classname)
							&& getBoolean(component, "angle", false);
			Object item = get(component, ":comp");
			if (item == null) {
				String empty = getString(component, "empty");

				if (empty != null) {
					FontMetrics fm;

					fm = g.getFontMetrics();
					g.setColor(c_empty);
					g.drawString(empty, 2, 2 + fm.getAscent());
				}
			}
			for (Object next = null; item != null; item = next) {
				if (focus && (lead == null)) {
					set(component, ":lead", lead = item); // draw first item
															// focused when lead
															// is null
				}
				Rectangle r = getRectangle(item, "bounds");
				if (r == null) {
					break;
				}
				if (clipy + clipheight <= r.y) {
					break;
				} // clip rectangle is above
				boolean subnode = false;
				boolean expanded = false;
				if ("tree" != classname) {
					next = get(item, ":next");
				} else {
					subnode = (next = get(item, ":comp")) != null;
					expanded = subnode && getBoolean(item, "expanded", true);
					if (!expanded) {
						for (Object node = item; (node != component)
													&& ((next = get(node,
															":next")) == null); node = getParent(node))
							;
					}
				}
				if (clipy >= r.y + r.height + iline) {
					if (angle) {
						Object nodebelow = get(item, ":next");
						if (nodebelow != null) { // and the next node is
													// bellow clipy
							g.setColor(c_bg);
							int x = r.x - block / 2;
							g.drawLine(x, r.y, x, getRectangle(nodebelow,
									"bounds").y);
						}
					}
					continue; // clip rectangle is bellow
				}

				boolean selected = getBoolean(item, "selected", false);
				boolean focused = focus && (lead == item);
				paintRect(g, ("tree" != classname) ? 0 : r.x, r.y,
						("tree" != classname) ? viewwidth : r.width, r.height,
						c_focus, selected ? c_select : c_textbg, focused,
						focused, focused, focused, true);
				if (line) {
					g.setColor(c_bg);
					g.drawLine(0, r.y + r.height, viewwidth, r.y + r.height);
				}
				if ("table" != classname) { // list or tree
					boolean itemenabled = enabled
											&& getBoolean(item, "enabled", true);
					paint(item, r.x, r.y, viewwidth, r.height, g, clipx, clipy,
							clipwidth, clipheight, false, false, false, false,
							1, 3, 1, 3, false, itemenabled ? 'e' : 'd', "left",
							false, false);
					if ("tree" == classname) {
						int x = r.x - block / 2;
						int y = r.y + (r.height - 1) / 2;
						if (angle) {
							g.setColor(c_bg);
							g.drawLine(x, r.y, x, y);
							g.drawLine(x, y, r.x - 1, y);
							Object nodebelow = get(item, ":next");
							if (nodebelow != null) {
								g.drawLine(x, y, x, getRectangle(nodebelow,
										"bounds").y);
							}
						}
						if (subnode) {
							paintRect(g, x - 4, y - 4, 9, 9,
									itemenabled ? c_border : c_disable,
									itemenabled ? c_ctrl : c_bg, true, true,
									true, true, true);
							g.setColor(itemenabled ? c_text : c_disable);
							g.drawLine(x - 2, y, x + 2, y);
							if (!expanded) {
								g.drawLine(x, y - 2, x, y + 2);
							}
						}
					}
				} else { // table
					Object prevcell = null;
					int i = 0, pi = 0, x = 0;
					for (Object cell = get(item, ":comp"); cell != null; cell = get(
							cell, ":next")) {
						if (clipx + clipwidth <= x) {
							break;
						}

						if (prevcell != null && columnwidths[i] > 0) {
							// column width is defined by header calculated in
							// layout, otherwise is 80
							int iwidth = 80;
							if ((columnwidths != null)
								&& (columnwidths.length > pi)) {
								iwidth = columnwidths[pi];
							}
							if (iwidth > 0 && clipx < x + iwidth) {
								boolean cellenabled = enabled
														&& getBoolean(prevcell,
																"enabled", true);
								paint(prevcell, r.x + x, r.y, iwidth,
										r.height - 1, g, clipx, clipy,
										clipwidth, clipheight, false, false,
										false, false, 1, 1, 1, 1, false,
										cellenabled ? (selected ? 's' : 'e') : 'd', "left", false,
										false);
							}
							x += iwidth;
						}

						if (columnwidths[i] > 0) {
							prevcell = cell;
							pi = i;
						}
						i++;
					}
					if (prevcell != null) {
						int iwidth = viewwidth - x;
						boolean cellenabled = enabled
												&& getBoolean(prevcell,
														"enabled", true);
						paint(prevcell, r.x + x, r.y, iwidth, r.height - 1, g,
								clipx, clipy, clipwidth, clipheight, false,
								false, false, false, 1, 1, 1, 1, false,
								cellenabled ? (selected ? 's' : 'e') : 'd', "left", false, false);
					}
				}
			}
		}
	}

	/**
	 * 
	 */
	void paintRect(Graphics g, int x, int y, int width, int height,
			Color border, Color bg, boolean top, boolean left, boolean bottom,
			boolean right, boolean horizontal) {
		if ((width <= 0) || (height <= 0))
			return;
		g.setColor(border);
		if (top) {
			g.drawLine(x + width - 1, y, x, y);
			y++;
			height--;
			if (height <= 0)
				return;
		}
		if (left) {
			g.drawLine(x, y, x, y + height - 1);
			x++;
			width--;
			if (width <= 0)
				return;
		}
		if (bottom) {
			g.drawLine(x, y + height - 1, x + width - 1, y + height - 1);
			height--;
			if (height <= 0)
				return;
		}
		if (right) {
			g.drawLine(x + width - 1, y + height - 1, x + width - 1, y);
			width--;
			if (width <= 0)
				return;
		}

		if (bg == c_ctrl) {
			fill(g, x, y, width, height, horizontal, c_bg);
		} else {
			g.setColor(bg);
			g.fillRect(x, y, width + evm, height + evm);
		}
	}

	/**
	 * Fill the given rectangle with gradient
	 */
	void fill(Graphics g, int x, int y, int width, int height,
			boolean horizontal, Color bg) {
		if (bg == null)
			bg = c_bg;
		Image hg = (Image)this.hgradients.get(bg);
		if (hg == null) {
			this.addGradient(bg, bg.darker());
			hg = (Image)this.hgradients.get(bg);
		}
		Image vg = (Image)this.vgradients.get(bg);
		if (horizontal) {
			if (height > block) {
				g.setColor(bg);
				g.fillRect(x, y, width + evm, height - block + evm);
			}
			for (int i = 0; i < width; i += block) {
				g.drawImage(hg, x + i,
						(height > block) ? (y + height - block) : y,
						x + Math.min(i + block, width) + evm, y + height + evm,
						0, 0, Math.min(block, width - i) + evm, Math.min(block,
								height)
																+ evm, null);
			}
		} else {
			if (width > block) {
				g.setColor(bg);
				g.fillRect(x, y, width - block + evm, height + evm);
			}
			for (int i = 0; i < height; i += block) {
				g.drawImage(vg, (width > block) ? (x + width - block)
						: x, y + i, x + width + evm, y
														+ Math.min(i + block,
																height) + evm,
						0, 0, Math.min(block, width) + evm, Math.min(block,
								height - i)
															+ evm, null);
			}
		}
	}

	/**
	 * 
	 */
	void paintArrow(Graphics g, int x, int y, int width, int height, char dir,
			boolean enabled, boolean inside, boolean pressed, String part,
			boolean top, boolean left, boolean bottom, boolean right,
			boolean horizontal, int size) {
		inside = inside && (insidepart == part);
		pressed = pressed && (pressedpart == part);
		paintRect(g, x, y, width, height, enabled ? c_border : c_disable,
				enabled ? ((inside != pressed) ? c_hover : (pressed ? c_press
						: c_ctrl)) : c_bg, top, left, bottom, right, horizontal);
		g.setColor(enabled ? c_text : c_disable);
		paintArrow(g, x + (left ? 1 : 0), y + (top ? 1 : 0),
				width - (left ? 1 : 0) - (right ? 1 : 0), height
															- (top ? 1 : 0)
															- (bottom ? 1 : 0),
				dir, size);
	}

	/**
	 * 
	 */
	void paintArrow(Graphics g, int x, int y, int width, int height, char dir,
			int size) {
		int cx = x + width / 2 - 2;
		int cy = y + height / 2 - 2;
		for (int i = 0; i < size; i++) {
			if (dir == 'N') { // north
				g.drawLine(cx + 1 - i, cy + i, cx + 1/* 2 */+ i, cy + i);
			} else if (dir == 'W') { // west
				g.drawLine(cx + i, cy + 1 - i, cx + i, cy + 1/* 2 */+ i);
			} else if (dir == 'S') { // south
				g.drawLine(cx + 1 - i, cy + 4 - i, cx + 1/* 2 */+ i, cy + 4 - i);
			} else { // east
				g.drawLine(cx + 4 - i, cy + 1 - i, cx + 4 - i, cy + 1/* 2 */+ i);
			}
		}
	}

	/**
	 * Paint component's borders and background
	 */
	void paint(Object component, int x, int y, int width, int height,
			Graphics g, boolean top, boolean left, boolean bottom,
			boolean right, char mode) {
		if ((width <= 0) || (height <= 0)) {
			return;
		}

		if (top || left || bottom || right) { // draw border
			g.setColor(((mode != 'd') && (mode != 'i')) ? c_border : c_disable);
			if (top) {
				g.drawLine(x + width - 1, y, x, y);
				y++;
				height--;
				if (height <= 0) {
					return;
				}
			}
			if (left) {
				g.drawLine(x, y, x, y + height - 1);
				x++;
				width--;
				if (width <= 0) {
					return;
				}
			}
			if (bottom) {
				g.drawLine(x, y + height - 1, x + width - 1, y + height - 1);
				height--;
				if (height <= 0) {
					return;
				}
			}
			if (right) {
				g.drawLine(x + width - 1, y + height - 1, x + width - 1, y);
				width--;
				if (width <= 0) {
					return;
				}
			}
		}
		
		if (getClass(component) == "connector") {
			x--; width++; // XXX Hack for connectors.
		}

		Color background = (Color) get(component, "background");
		switch (mode) {
		case 'e':
		case 'l':
		case 'd':
		case 'g':
		case 'r':
			break;
		case 'b':
		case 'i':
		case 'x':
			if (background == null) {
				background = c_bg;
			}
			break;
		case 'h':
			background = (background != null) ? background.brighter() : c_hover;
			break;
		case 'p':
			background = (background != null) ? background.darker() : c_press;
			break;
		case 't':
			if (background == null) {
				background = c_textbg;
			}
			break;
		case 's':
			background = c_select;
			break;
		default:
			throw new IllegalArgumentException();
		}
		if (((mode == 'g') || (mode == 'r')) && ((background == null) || getBoolean(component, "usegradient", false))) {
			fill(g, x, y, width, height, true, background);
		} else if (background != null) {
			g.setColor(background);
			if (mode != 'x') {
				g.fillRect(x, y, width + evm, height + evm);
			}
		}
	}

	/**
	 * Paints the dialog close, maxmize and iconify buttons
	 */
	void paint(Object component, Graphics g, int x, int y, int width,
			int height, char type, char mode) {
		paint(component, x, y, width, height, g, true, true, true, true, mode);
		g.setColor(Color.black);
		switch (type) {
		case 'c': // closable dialog button
			g.drawLine(x + 3, y + 4, x + width - 5, y + height - 4);
			g.drawLine(x + 3, y + 3, x + width - 4, y + height - 4);
			g.drawLine(x + 4, y + 3, x + width - 4, y + height - 5);
			g.drawLine(x + width - 5, y + 3, x + 3, y + height - 5);
			g.drawLine(x + width - 4, y + 3, x + 3, y + height - 4);
			g.drawLine(x + width - 4, y + 4, x + 4, y + height - 4);
			break;
		case 'm': // maximizable dialog button
			g.drawRect(x + 3, y + 3, width - 7, height - 7);
			g.drawLine(x + 4, y + 4, x + width - 5, y + 4);
			break;
		case 'i': // iconifiable dialog button
			g.fillRect(x + 3, y + height - 5, width - 6, 2);
			break;
		}
	}

	/**
	 * Paint component icon and text (using default or custom font)
	 * 
	 * @param mnemonic
	 *            find mnemonic index and underline text
	 */
	void paint(Object component, int x, int y, int width, int height,
			Graphics g, int clipx, int clipy, int clipwidth, int clipheight,
			boolean top, boolean left, boolean bottom, boolean right,
			int toppadding, int leftpadding, int bottompadding,
			int rightpadding, boolean focus, char mode, String alignment,
			boolean mnemonic, boolean underline) {
		paint(component, x, y, width, height, g, top, left, bottom, right, mode);
		if (top) {
			y++;
			height--;
		}
		if (left) {
			x++;
			width--;
		}
		if (bottom) {
			height--;
		}
		if (right) {
			width--;
		}
		if ((width <= 0) || (height <= 0)) {
			return;
		}

		if (focus) {
			g.setColor(c_focus);
			g.drawRect(x + 1, y + 1, width - 3, height - 3);
		}

		String text = getString(component, "text", null);
		
		if (text == ThinletController.MULTIPLE_SELECTION)
			text = "?";
		else if (text == ThinletController.NO_SELECTION)
			text = "-";
		
		Image icon = getIcon(component, "icon", null);
		if ((text == null) && (icon == null)) {
			return;
		}

		x += leftpadding;
		y += toppadding;
		width -= leftpadding + rightpadding;
		height -= toppadding + bottompadding;

		alignment = getString(component, "alignment", alignment);
		Font customfont = (text != null) ? (Font) get(component, "font") : null;
		if (customfont != null) {
			g.setFont(customfont);
		}

		FontMetrics fm = null;
		int tw = 0, th = 0;
		int ta = 0;
		if (text != null) {
			fm = g.getFontMetrics();
			tw = fm.stringWidth(text);
			ta = fm.getAscent();
			th = fm.getDescent() + ta;
		}
		int iw = 0, ih = 0;
		if (icon != null) {
			iw = icon.getWidth(this);
			ih = icon.getHeight(this);
			if (text != null) {
				iw += 2;
			}
		}

		boolean clipped = (tw + iw > width) || (th > height) || (ih > height);
		int cx = x;
		if ("center" == alignment) {
			cx += (width - tw - iw) / 2;
		} else if ("right" == alignment) {
			cx += width - tw - iw;
		}

		if (clipped) {
			g.clipRect(x, y, width, height);
		}
		if (mode == 'x') {
			g.drawLine(cx, y + height / 2, cx + iw + tw, y + height / 2);
		}
		if (icon != null) {
			imageToVisible.put(icon, component);
			g.drawImage(icon, cx, y + (height - ih) / 2, this);
			cx += iw;
		}
		if (text != null) {
			Color foreground = (Color) get(component, "foreground");
			if (foreground == null) {
				foreground = (mode == 'l') ? Color.blue
						: (((mode != 'd') && (mode != 'r')) ? c_text
								: c_disable);
			}
			g.setColor(foreground);
			int ty = y + (height - th) / 2 + ta;

			if ("fill" == alignment) {
				String splittext[];
				int tab = -1;

				if (((splittext = (String[]) get(component, ":splittext")) != null)
					|| ((tab = text.indexOf('\t')) != -1)) {
					String first, second = null;

					/* XXX optimize me. */
					if (splittext != null && splittext[0] == text) {
						first = splittext[1];
						second = splittext[2];
					} else if (tw < width) {
						tab = text.indexOf('\t');
						first = text.substring(0, tab);
						second = text.substring(tab + 1);
						set(component, ":splittext", new String[] { text,
								first, second });
					} else {
						first = text;
					}
					text = first;
					if (second != null) {
						g.drawString(second,
								x + width - fm.stringWidth(second), ty);
					}
				}
			}
			g.drawString(text, cx, ty);
			if (mnemonic) {
				int imnemonic = getInteger(component, "mnemonic", -1);
				if ((imnemonic != -1) && (imnemonic < text.length())) {
					int mx = cx + fm.stringWidth(text.substring(0, imnemonic));
					g
							.drawLine(mx, ty + 1,
									mx + fm.charWidth(text.charAt(imnemonic)),
									ty + 1);
				}
			}
			if (underline) { // for link button
				g.drawLine(cx, ty + 1, cx + tw, ty + 1);
			}
		}
		if (clipped) {
			g.setClip(clipx, clipy, clipwidth, clipheight);
		}

		if (customfont != null) {
			g.setFont(font);
		} // restore the default font
	}

}
