package thinlet;

import java.awt.Rectangle;

public abstract class ThinletPaintMgr extends ThinletFocus {
	protected final Object relayoutQueue[] = new Object[16];

	protected int relayoutQueueCount = 0;

	protected final Rectangle repaintQueue[] = new Rectangle[16];
	{
		int lpc;

		for (lpc = 0; lpc < repaintQueue.length; lpc++) {
			repaintQueue[lpc] = new Rectangle();
		}
	}

	protected int repaintQueueCount = 0;
	
	protected void printQueue() {
		int lpc;
		
		for (lpc = 0; lpc < repaintQueueCount; lpc++) {
			System.out.println("q " + repaintQueue[lpc]);
		}
	}

	protected synchronized void relayout(Object component) {
		if (component == null) {
			throw new IllegalArgumentException("comonent is null");
		}
		if (relayoutQueue == null) {
		} else if (relayoutQueueCount == relayoutQueue.length) {
			relayoutQueueCount = 1;
			relayoutQueue[0] = content;
		} else if (isVisible(component)) {
			relayoutQueue[relayoutQueueCount] = component;
			relayoutQueueCount += 1;
		}
	}

	public synchronized void repaint(long tm, int x, int y, int width,
			int height) {
		if (false) {
			System.out.println("repaint " + x + "; " + y + "; " + width + "; " + height);
			printCaller(new Throwable().getStackTrace());
		}
		// System.out.println("rep " + x + "; " + y + "; " + width + "; " + height);
		// printQueue();
		if (repaintQueue == null) {
		} else if (repaintQueueCount == repaintQueue.length) {
			if (false)
				System.out.println("overflow!");
			repaintQueueCount = 1;
			repaintQueue[0].setBounds(0, 0, getSize().width, getSize().height);
		} else {
			boolean contained = false;
			Rectangle rectangle;
			int lpc;

			if (x == 0) {
				// new Throwable().printStackTrace();
			}
			rectangle = repaintQueue[repaintQueueCount];
			rectangle.setBounds(x, y, width, height);
			// System.out.println("new " + rectangle);
			for (lpc = 0; (lpc < repaintQueueCount) && !contained; lpc++) {
				Rectangle qrect;

				qrect = repaintQueue[lpc];
				contained = qrect.contains(rectangle);
				if (!contained
					&& (rectangle.intersects(qrect)
						|| ((qrect.x + qrect.width + 1) == rectangle.x)
						|| ((qrect.y + qrect.height + 1) == rectangle.y)
						|| ((rectangle.x + rectangle.width + 1) == qrect.x)
						|| ((rectangle.y + rectangle.height + 1) == qrect.y))) {
					if (false)
						System.out.println("merge " + rectangle + " " + qrect);
					qrect.add(rectangle);
					contained = true;
					if (false)
						System.out.println("merge2 " + rectangle + " " + qrect);
				}
			}
			if (!contained) {
				// System.out.println("!contained " + rectangle);
				repaintQueueCount += 1;
			}
		}
		// printQueue();
		super.repaint(tm, x, y, width, height);
	}

	/**
	 * Repaint the given component's area later
	 * 
	 * @param component
	 *            a visible widget inside thinlet desktop
	 */
	public void repaint(Object component) {
		// System.out.println(" repc " + getClass(component) +"; " + getString(component, "name"));
		Rectangle bounds = getRectangle(component, "bounds");
		if (bounds != null) {
			repaint(component, bounds.x, bounds.y, Math.abs(bounds.width),
					bounds.height);
		}
	}

	/**
	 * Layout and paint the given component later
	 * 
	 * @param component
	 */
	protected void validate(Object component) {
		repaint(component);
		Rectangle bounds = getRectangle(component, "bounds");
		if (bounds != null) {
			bounds.width = -1 * Math.abs(bounds.width);
		}
	}

	/**
	 * Repaint the given component's area later
	 * 
	 * @param component
	 * @param x
	 * @param y
	 * @param width
	 * @param height
	 */
	void repaint(Object component, int x, int y, int width, int height) {
		while ((component = getParent(component)) != null) {
			Rectangle bounds = getRectangle(component, "bounds");
			if (bounds == null) {
				continue;
			}
			x += bounds.x;
			y += bounds.y;
			Rectangle view = getRectangle(component, ":view");
			if (view != null) {
				Rectangle port = getRectangle(component, ":port");
				x += -view.x + port.x;
				y += -view.y + port.y; // + clip :port
			}
		}
		repaint(x, y, width, height);
	}

	/**
	 * 
	 */
	protected void update(Object component, Object mode) {
		if ("reeval" == mode) {
			invoke(getParent(component), component);
			return;
		}
		if ("parent" == mode) {
			component = getParent(component);
			// System.out.println("inv " + component + " mode " + mode);
			Rectangle bounds = getRectangle(component, "bounds");
			if (bounds != null) {
				if ("dialog" == getClass(component)) {
					if (!getBoolean(component, "resizable", false))
						component = getParent(component);
					relayout(component);
					return;
				}
				bounds.width = -1 * Math.abs(bounds.width);
			}
			mode = "validate";
		}
		if ("gparent" == mode) {
			while (component != null && component != this.content) {
				component = getParent(component);
			}
			if (component != null) {
				Rectangle bounds = getRectangle(component, "bounds");
				
				if (bounds != null) {
					bounds.width = -1 * Math.abs(bounds.width);
					relayout(component);
				}
			}
			repaint();
			
			return;
		}
		// System.out.println("update " + getClass(component) + " " + mode);
		boolean firstpaint = true;
		int x = 0;
		int y = 0;
		int width = 0;
		int height = 0;
		while (component != null) {
			if (!getBoolean(component, "visible", true)) { /*
															 * System.out.println("not
															 * vis");
															 */
				break;
			}
			if ("paint" == mode) {// || (firstpaint && (component == content))
				Rectangle bounds = getRectangle(component, "bounds");
				if (bounds == null) {
					return;
				}
				// System.out.println("oldbounds " + x + " " + y + " " + width +
				// " " + height + "; " + bounds + "; " + getClass(component));
				if (firstpaint) {
					// new Throwable().printStackTrace();
					x = bounds.x;
					y = bounds.y;
					width = Math.abs(bounds.width);
					height = bounds.height;
					firstpaint = false;
				} else {
					x += bounds.x;
					y += bounds.y;
				}
				// System.out.println("newbounds " + getClass(component) + " " +
				// x + " " + y + " " + width + " " + height);
				if (component == content) {
					if (false) {
						System.out.println("repaint " + x + " " + y + " "
											+ width + " " + height + " "
											+ System.currentTimeMillis());
					}
					repaint(20, x, y, width, height);
				}
			}
			Object parent = getParent(component);
			String classname = getClass(parent);
			// System.out.println("cname = " + classname + " " + mode);
			if ("combobox" == classname) {
				parent = get(parent, ":combolist");
			} else if ("menu" == classname) {
				parent = get(parent, ":popup");
			} else if ("tab" == classname) {
				Rectangle r = getRectangle(parent, "bounds");
				Object tab = parent;
				if (r != null) {
					x += r.x;
					y += r.y;
				}
				parent = getParent(parent);
				if (getSelectedItem(parent) != tab) {
					if (false)
						System.out.println("Skip hidden tab");
					return;
				}
				// classname = getClass(parent);
			} else if (("paint" == mode) && ("tabbedpane" == classname)) {
				if (getItem(parent, getInteger(parent, "selected", 0)) != component) {
					if (false)
						System.out.println("break");
					break;
				}
			}
			if (("layout" == mode)
				|| (("validate" == mode) && (("list" == classname)
												|| ("table" == classname)
												|| ("tree" == classname)
												|| ("dialog" == classname)
												|| ("dropzone" == classname)
												|| ("panel" == classname) || (parent == content)))) {
				// System.out.println("inv " + parent + " mode " + mode + " cl " + classname);
				Rectangle bounds = getRectangle(parent, "bounds");
				if (bounds == null) {
					return;
				}
				if ("dialog" == classname) {
					relayout(parent);
					return;
				}
				bounds.width = -1 * Math.abs(bounds.width);
				if ("layout" == mode) {
					mode = "paint";
					continue;
				}
				mode = "paint";
			}
			component = parent;
		}
	}

	/**
	 * 
	 */
	protected void repaint(Object component, Object classname, Object part) {
		// System.out.println("repaint " + getClass(component) + " " + part);
		Rectangle b = getRectangle(component, "bounds");
		if (classname == "combobox") { // combobox down arrow
			repaint(component, b.x + b.width - block, b.y, block, b.height); // icon?+
		} else if (classname == "spinbox") {
			if (part == "text") { // spinbox textfield content
				repaint(component, b.x, b.y, b.width - block, b.height);
			} else { // spinbox increase or decrease button
				repaint(component, b.x + b.width - block, (part == "up") ? b.y
						: (b.y + b.height - b.height / 2), block, b.height / 2);
			}
		}
		// else if (classname == "dialog") {}
		// int titleheight = getInteger(component, ":titleheight", 0);
		// else if (classname == "splitpane") {}
		else if ((classname == "tabbedpane") || // tab
					(classname == "menubar") || (classname == ":popup")) { // menuitem
			Rectangle r = getRectangle(part, "bounds");
			repaint(component, b.x + r.x, b.y + r.y,
					(classname == ":popup") ? b.width : r.width, r.height);
		}
		// classname: ":combolist" "textarea" "list" "table" "tree"
		else if ((part == "left") || (part == "right")) { // horizontal
															// scrollbar button
			Rectangle r = getRectangle(component, ":horizontal");
			if (r != null)
				repaint(component, b.x
									+ ((part == "left") ? r.x
											: (r.x + r.width - block)), b.y
																		+ r.y,
						block, r.height);
		} else if ((part == "up") || (part == "down")) { // vertical
															// scrollbar button
			Rectangle r = getRectangle(component, ":vertical");
			if (r != null)
				repaint(
						component,
						b.x + r.x,
						b.y + ((part == "up") ? r.y : (r.y + r.height - block)),
						r.width, block);
		} else if ((part == "text") || (part == "horizontal")
					|| (part == "vertical")) {
			Rectangle port = getRectangle(component, ":port"); // textarea or
																// content
			repaint(component, b.x + port.x, b.y + port.y, port.width,
					port.height);
			if (part == "horizontal") {
				Rectangle r = getRectangle(component, ":horizontal");
				repaint(component, b.x + r.x, b.y + r.y, r.width, r.height);
				repaint(component, b.x + r.x, b.y, r.width, port.y); // paint
																		// header
																		// too
			} else if (part == "vertical") {
				Rectangle r = getRectangle(component, ":vertical");
				repaint(component, b.x + r.x, b.y + r.y, r.width, r.height);
			}
		} else { // repaint the whole line of its subcomponent
			Rectangle port = getRectangle(component, ":port");
			Rectangle view = getRectangle(component, ":view");
			Rectangle r = getRectangle(part, "bounds");
			if ((r != null) && (r.y + r.height >= view.y)
				&& (r.y <= view.y + port.height)) {
				repaint(component, b.x + port.x, b.y + port.y - view.y + r.y,
						port.width, r.height);
				// ? need cut item rectangle above/bellow viewport
			}
		}
	}

	/**
	 * 
	 */
	void scrollToVisible(Object component, int x, int y, int width, int height) {
		Rectangle view = getRectangle(component, ":view");
		Rectangle port = getRectangle(component, ":port");
		int vx = Math.max(x + width - port.width, Math.min(view.x, x));
		int vy = Math.max(y + height - port.height, Math.min(view.y, y));
		if ((view.x != vx) || (view.y != vy)) {
			repaint(component); // horizontal | vertical
			view.x = vx;
			view.y = vy;
		}
	}
}
