/*
 * Created on Nov 21, 2005
 *
 * To change the template for this generated file go to
 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
 */

import java.io.File;
import java.io.FileReader;
import java.io.FilenameFilter;
import java.io.IOException;
import java.util.Hashtable;
import java.util.List;
import java.util.Vector;

import net.emulab.netlab.client.ExperimentParameters;
import net.emulab.netlab.client.emulab.EmulabExperimentProxy;
import net.emulab.netlab.client.emulab.EmulabUserProxy;
import net.emulab.netlab.client.emulab.EmulabXmlRpcException;
import net.emulab.ns.DBState;
import net.emulab.ns.NSTopology;
import net.emulab.util.EnvironmentVariables;
import net.emulab.util.FileDiff;
import net.emulab.util.ServerProperties;
import net.emulab.xmlrpc.AutoXmlRpcClient;
import thinlet.ObjectFactory;
import thinlet.ObjectPool;
import thinlet.ThinletKeyValue;

public class DBStateTest {
	
	private static final String EID = "netlabclienttestexp";
	private static final String FWEID = "ncfwtestexp";
	
	private static final String BLANK_NSFILE = ""
		+ "set ns [new Simulator]\n"
		+ "source tb_compat.tcl\n"
		+ "set node [$ns node]\n"
		+ "$ns run\n";
	
	private static String NSFILTER;
	
	private static String readFile(File file) throws IOException {
		StringBuffer retval = new StringBuffer();
		char buf[] = new char[1024];
		FileReader fr;
		int rc;
		
		fr = new FileReader(file);
		while ((rc = fr.read(buf)) != -1) {
			retval.append(buf, 0, rc);
		}
		
		return retval.toString();
	}
	
	private static void printDiffs(String fn, List diffs) {
		int lpc, size;
		
		size = diffs.size();
		for (lpc = 0; lpc < size; lpc++) {
			FileDiff fd = (FileDiff)diffs.get(lpc);
			
			if (!"".equals(fd.getType())) {
				String line = fd.getLine().trim();
				String otype = "+".equals(fd.getType()) ? "-" : "+";
				
				if (!"".equals(line) && !line.startsWith("#") &&
						FileDiff.indexOf(diffs, otype, fd.getLine()) == -1)
					System.out.println(fn + ": " + fd.getType() + " " + fd.getLine());
			}
		}
	}
	
	private static void createExperiment(EmulabExperimentProxy eep, String pid, String eid) throws Throwable {
		try {
			eep.state(pid, eid);
		}
		catch (EmulabXmlRpcException e) {
			ExperimentParameters ep = new ExperimentParameters();
			
			ep.setProject(pid);
			ep.setName(eid);
			ep.setDescription("Experiment used by the netlab client test suite");
			ep.setNSFileString(BLANK_NSFILE);
			ep.setNoSwapIn(true);
			ep.setWait(true);
			eep.startexp(ep);
		}
	}
	
	public static void main(String args[]) throws Throwable {
		ThinletKeyValue tkv = new ThinletKeyValue();
		ObjectPool op = new ObjectPool();
		
		tkv.setKeyValue(op, "applet", null);
		tkv.setKeyValue(op, "environ", new EnvironmentVariables());
		tkv.setKeyValue(op, "server", new ServerProperties(tkv, op));
		tkv.setKeyValue(op, "xmlrpcClientFactory",
				new ObjectFactory(tkv, op, AutoXmlRpcClient.class));
		
		EmulabUserProxy user = new EmulabUserProxy(tkv, op);
		EmulabExperimentProxy eep = new EmulabExperimentProxy(tkv, op);
		String pid, nsfiles[];
		Hashtable ht;
		File srcdir;
		int lpc;
		
		srcdir = new File(tkv.getKeyValueAsString(op, "environ.srcdir"));
		ht = user.membership(EmulabUserProxy.PERMISSION_CREATE_EXPERIMENT);
		pid = (String)ht.keys().nextElement();
		
		createExperiment(eep, pid, EID);
		createExperiment(eep, pid, FWEID);
		
		try {
			NSFILTER = tkv.getKeyValueAsString(op, "environ.NSFILES");
		}
		catch (IllegalArgumentException e) {
		}
		
		nsfiles = srcdir.list(new FilenameFilter() {
			public boolean accept(File dir, String name) {
				return (name.endsWith(".ns") &&
						(NSFILTER == null || NSFILTER.indexOf(name) != -1));
			}
		});
		
		for (lpc = 0; lpc < nsfiles.length; lpc++) {
			File file = new File(srcdir, nsfiles[lpc]);
			String nsfile = readFile(file);
			NSTopology nst = new NSTopology();
			DBState dbs = new DBState();
			String eid;
			
			if (nsfile.indexOf("new Firewall") != -1) {
				eid = FWEID;
			}
			else {
				eid = EID;
			}
			
			eep.modify(pid, eid, nsfile, true);
			ht = eep.virtual_topology(pid, eid, new Vector());
			dbs.fromVirtualTopology(tkv, nst, ht, nsfile);
			
			printDiffs(nsfiles[lpc], dbs.getDiffs());
		}
		
		System.out.println("done");
	}
}

/* Expected Output:
done
*/
