
import java.util.Arrays;

import java.text.DecimalFormat;

import java.awt.Font;
import java.awt.Color;

import thinlet.ThinletAPI;
import thinlet.NumberFormatter;
import thinlet.ThinletBase;

public class ThinletAPITest extends ThinletAPI
{
	private static final NumberFormatter NUMBER_FORMATTER =
		new NumberFormatter(new DecimalFormat("#,##0"));
	
	public ThinletAPITest()
	{
	}
	
	public void run()
	{
	}
	
	protected Object[] getDefinition(Object classname,
			String key,
			String type)
	{
		if ("foreground".equals(key)) {
			return new Object[] { "color", "foreground", "paint", null };
		}
		else if ("name".equals(key)) {
			return new Object[] { "string", "name", null, null };
		}
		else if ("text".equals(key)) {
			return new Object[] { "object", "text", "layout", "" };
		}
		else if ("empty".equals(key)) {
			return new Object[] { "object", "empty", "layout", "" };
		}
		else if ("formatter".equals(key)) {
			return new Object[] { "delegate", "formatter" };
		}
		else if ("valign".equals(key)) {
			return new Object[] {
					"choice", "valign", "layout", new String[] {
							"center", "top", "bottom",
					}
			};
		}
		else if ("font".equals(key)) {
			return new Object[] { "font", "font", "layout", null };
		}
		throw new IllegalArgumentException("unknown key " + key);
	}
	
	public Object create(String classname)
	{
		return ThinletBase.createImpl(classname);
	}
	
	public boolean set(Object component,
			Object key,
			Object value)
	{
		return super.set(component, key, value);
	}
	
	public static Object get(Object component, Object key)
	{
		return ThinletAPI.get(component, key);
	}
	
	public int getIndex(Object component, Object value)
	{
		return super.getIndex(component, value);
	}
	
	public void insertItem(Object parent, Object component, int index)
	{
		super.insertItemImpl(parent, ":comp", component, index);
	}
	
	public int getSelectedIndex(Object component)
	{
		return -1;
	}
	
	public Object getSelectedItem(Object component)
	{
		return null;
	}
	
	public Object[] getSelectedItems(Object component)
	{
		return null;
	}
	
	public void setLead(Object component, Object oldlead, Object lead)
	{
	}
	
	public boolean isVisible(Object component)
	{
		return true;
	}
	
	public String keyForCache(Object key)
	{
		if (key.equals("text"))
			return ":tcache";
		else if (key.equals("empty"))
			return ":ecache";
		return null;
	}
	
	public void update(Object component, Object mode)
	{
		System.out.println("update: " + getClass(component) + " " + mode);
	}
	
	public void doLayout(Object component)
	{
	}
	
	public void repaint(Object component)
	{
	}
	
	public void repaint(Object component, Object classname, Object part)
	{
	}
	
	public void validate(Object component)
	{
	}
	
	public static void main(String args[]) throws Throwable
	{
		ThinletAPITest tat = new ThinletAPITest();
		
		Object obj = tat.create("foo");
		Color color;
		Font font;
		
		if (!tat.getString(obj, "text", "bar").equals("bar"))
			System.out.println("foo.text != bar?");
		tat.setString(obj, "text", "baz");
		if (!tat.getString(obj, "text", "bar").equals("baz"))
			System.out.println("foo.text != baz?");
		
		tat.setString(obj, "text", "", "");
		if (ThinletBase.get(obj, "text") != null)
			System.out.println("setting default string value did not clear?");
		
		tat.setObject(obj, "text", new Integer(4096));
		System.out.println(tat.getString(obj, "text"));
		
		tat.setDelegate(obj, "formatter", NUMBER_FORMATTER);
		System.out.println(tat.getString(obj, "text"));
		
		if (ThinletBase.get(obj, ":tcache") == null)
			System.out.println("no :tcache for formatted object?");
		
		tat.setObject(obj, "empty", new Integer(8192));
		System.out.println(tat.getString(obj, "empty"));
		
		if (ThinletBase.get(obj, ":ecache") == null)
			System.out.println("no :ecache for formatted object?");
		
		
		/* set/get choice */
		System.out.println(tat.getChoice(obj, "valign"));
		
		try
		{
			tat.setChoice(obj, "valign", "badbad");
			System.out.println("valign has 'badbad' choice?");
		}
		catch(IllegalArgumentException e)
		{
		}
		
		tat.setChoice(obj, "valign", "top");
		System.out.println(tat.getChoice(obj, "valign"));
		
		tat.setChoice(obj, "valign", null);
		System.out.println(tat.getChoice(obj, "valign"));
		
		tat.setFont(obj, "font", "bold");
		font = tat.getFont(obj, "font");
		if( !font.isBold() )
			System.out.println("Font is not bold?");
		
		tat.setFont(obj, "font", "italic");
		font = tat.getFont(obj, "font");
		if( !font.isItalic() )
			System.out.println("Font is not italic?");
		
		tat.setFont(obj, "font", "bold italic");
		font = tat.getFont(obj, "font");
		if( !font.isBold() || !font.isItalic() )
			System.out.println("Font is not bold/italic?");
		
		tat.setFont(obj, "font", "24");
		font = tat.getFont(obj, "font");
		if( font.getSize() != 24 )
			System.out.println("Font is not 24 pt?");
		
		tat.setFont(obj, "font", "24 bold italic");
		font = tat.getFont(obj, "font");
		if( !font.isBold() || !font.isItalic() || font.getSize() != 24 )
			System.out.println("Font is not 24 pt bold/italic?");
		
		/* set/get color */
		try
		{
			tat.setColor(obj, "foreground", "");
			System.out.println("Able to set foreground to empty string?");
		}
		catch(IllegalArgumentException e)
		{
			System.out.println(e.getMessage());
		}
		
		tat.setColor(obj, "foreground", "#ff00ff");
		color = tat.getColor(obj, "foreground");
		System.out.println("#" + Integer.toHexString(color.getRGB()));
		
		tat.setColor(obj, "foreground", "0xaabbcc");
		color = tat.getColor(obj, "foreground");
		System.out.println("0x" + Integer.toHexString(color.getRGB()));
		
		tat.setColor(obj, "foreground", "red");
		color = tat.getColor(obj, "foreground");
		System.out.println("equals red? " + (Color.red == color));
		
		tat.setColor(obj, "foreground", "1 2 3");
		color = tat.getColor(obj, "foreground");
		System.out.println("#" + Integer.toHexString(color.getRGB()));
		
		tat.setColor(obj, "foreground", "4,5,6");
		color = tat.getColor(obj, "foreground");
		System.out.println("#" + Integer.toHexString(color.getRGB()));
		
	}

	protected void addGradient(Color c1, Color c2) {
		// TODO Auto-generated method stub
		
	}

	protected boolean setNextFocusable(Object comp, boolean b) {
		// TODO Auto-generated method stub
		return false;
	}
}

/* Expected Output:
update: foo layout
update: foo layout
4096
4,096
update: foo layout
8,192
center
update: foo layout
top
update: foo layout
center
update: foo layout
update: foo layout
update: foo layout
update: foo layout
update: foo layout
Color value for 'foreground' needs to be in hex format (e.g. #efef66, 0xefef66), a color name (e.g. red), or a triple of ints representing the rgb value (e.g. '256,128,256'): 
update: foo paint
#ffff00ff
update: foo paint
0xffaabbcc
update: foo paint
equals red? true
update: foo paint
#ff010203
update: foo paint
#ff040506
*/
